#include "ExpMapInterface.h"
#include "..\ExpMap\MeshObject.h"
#include "..\..\SSSLib\SSSFunc.h"
#include <iostream>

using namespace std;
using namespace SSSLib;

void ExpMapExp::LoadAndAppendSamples(const char* iFileName,
									   float iDiskSize,
									   SSSLib::Vec3f iCtr,
									   float iRadius,
									   std::vector<Point>& oPt)
{
	MeshObject mesh;
	Wml::Vector3f ctr(iCtr.val);
	mesh.ReadMeshOBJ(iFileName, ctr, 1/iRadius);
	mesh.generateASP(iDiskSize, 10, string("pd"), 0);

	oPt.resize(mesh.aspoints.size());
	for (int i=0; i<(int)oPt.size(); i++)
	{
		oPt[i].mPst = mesh.aspoints[i].point;
		oPt[i].mNormal = mesh.aspoints[i].normal;
	}
}

void ExpMapExp::FilterSamples(Mesh& iMesh,
		const std::vector<Point>& iSamples,
		std::vector<Point>& oSamples)
{
	oSamples.clear();
	for (int i=0; i<(int)iSamples.size(); i++)
	{
		//i=272;
		if (i%100==0)
			cout<<"Filtering "<<i<<endl;
		Vec3f v = iSamples[i].mNormal.Norm();
		//assert(v.Length()>0);
		if (v.Length()<=0)
			continue;
		Vec3f u;
		do
		{
			u = Vec3f(randf<float>(), randf<float>(), randf<float>());
			if (u.Length()==0)
				continue;
			u = u.Norm();
			if (fabs(dot<float>(u, v))<0.05)
				continue;
			u = cross3(u, v).Norm();
			break;
		}
		while (1);
		vec uvec(u[0], u[1], u[2]);
		vec vvec(v[0], v[1], v[2]);
		vec pvec(iSamples[i].mPst[0], iSamples[i].mPst[1], iSamples[i].mPst[2]);
		if (!iMesh.InOutCheck(pvec, uvec, vvec))
			oSamples.push_back(iSamples[i]);
	}
}

void ExpMapExp::WriteSamples(const char* iFileName,
		const std::vector<Point>& iSamples)
{
	ofstream outFile(iFileName, ios_base::binary);
	int sampleNum = iSamples.size();
	outFile.write((const char*)&sampleNum, sizeof(int));
	outFile.write((const char*)&iSamples[0], sizeof(Point)*iSamples.size());
}

void ExpMapExp::WriteSamplesObj(const char* iFileName,
		const std::vector<Point>& iSamples)
{
	ofstream outFile(iFileName);
	for (int i=0; i<(int)iSamples.size(); i++)
		outFile<<"v "<<iSamples[i].mPst[0]
				<<" "<<iSamples[i].mPst[1]
				<<" "<<iSamples[i].mPst[2]
				<<endl;


}

void ExpMapExp::WriteSamplesPly(const char* iFileName,
		const std::vector<Point>& iSamples)
{
	ofstream outFile(iFileName);
	outFile<<"ply\n"
		<<"format ascii 1.0\n"
		<<"element vertex "<<iSamples.size()<<endl
		<<"property float x\n"
		<<"property float y\n"
		<<"property float z\n"
		<<"property float nx\n"
		<<"property float ny\n"
		<<"property float nz\n"
		<<"element face 0\n"
		<<"property list uchar int vertex_indices\n"
		<<"end_header\n";
	for (int i=0; i<(int)iSamples.size(); i++)
	{
		Vec3f p = iSamples[i].mPst;
		Vec3f n = iSamples[i].mNormal;
		n = n.Norm();
		outFile<<p[0]<<" "<<p[1]<<" "<<p[2]<<" "
			<<n[0]<<" "<<n[1]<<" "<<n[2]<<endl;
	}
}

void ExpMapExp::ReadSamples(const char* iFileName,
		std::vector<Point>& oSamples)
{
	ifstream inFile(iFileName, ios_base::binary);
	int sampleNum;
	inFile.read((char*)&sampleNum, 4);
	oSamples.resize(sampleNum);
	inFile.read((char*)&oSamples[0], sizeof(Point)*sampleNum);
}