//============================================================================
//
// This file is part of the 'Modeling with High Level Attribute' project.
//
// This software is covered by the following BSD license, except for portions
// derived from other works which are covered by their respective licenses.
// For full licensing information including reproduction of these external
// licenses, see the file LICENSE.txt provided in the documentation.
//
// Copyright (c) 2013, UMass-Amherst
//
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// * Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
//
// * Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
//
// * Neither the name of the copyright holders nor the names of contributors
// to this software may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
//============================================================================


#include "helper.h"
#include "Mesh.h"
#include "myXForm.h"

int main(int argc, char *argv[]) {
#if defined(_MSC_VER) && defined(_DEBUG)
	_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

	srand( (unsigned)time( NULL ) );

	///////////////// INITIALIZATION /////////////////////
	MeshFeatureExporterOptions options;
	options.correct_mesh_mode = false; 
	clock_t start = clock();	

	// initialize options
	strcpy(options.inputPath, "./");
	options.numPointSamples = POINT_SAMPLES;

	// rewrite options based on command options
	if (argc != 2 && argc != 3) 
	{
		std::cout << "Usage: MeshFeatureExporter <execution_mode> input_path" << std::endl;
		std::cout << "Please specify execution mode: -pr (correct mesh mode)" << std::endl;
		return 0;
	}

	if (strcmp(argv[1], "-pr") == 0) 
	{
		options.correct_mesh_mode = true;
	} 
	else
	{
		std::cout << "Invalid option " << argv[1] << std::endl;
		std::cout << "Usage: MeshFeatureExporter <execution_mode> input_path" << std::endl;
		std::cout << "Please specify execution mode: -pr (correct mesh mode)" << std::endl;
		return 0;
	}

	if ( argc == 3 )
	{
		strcpy(options.inputPath, argv[2]);
		strcat(options.inputPath, "/");
	}

	char newdir[FILE_STRING_SIZE];
	strcpy(newdir, options.inputPath);
	strcat(newdir, "out/");
	mkdir(newdir);

	////////////////////////////////////////////////////
	///////// Mesh Correction Mode  //////////////////////
	////////////////////////////////////////////////////

	if (options.correct_mesh_mode == true) 
	{    
		char meshFileListName[FILE_STRING_SIZE];
		strcpy(meshFileListName, options.inputPath);
		strcat(meshFileListName, MESHFILELIST);
		ifstream meshFileListFile(meshFileListName);
		if (!meshFileListFile.good()) 
		{
			std::cerr << "fatal error reading mesh file list " << meshFileListName << std::endl;
			return -1;
		}

		while( meshFileListFile.good() ) 
		{
			char inputFileName[FILE_STRING_SIZE];
			meshFileListFile.getline(inputFileName, FILE_STRING_SIZE);
			if ( isemptychar(inputFileName) || !meshFileListFile.good() )
				break;
			strcpy(options.currentObjFilename, options.inputPath);
			strcat(options.currentObjFilename, inputFileName);

			Mesh m(options.currentObjFilename);
			m.write( string(newdir) + inputFileName );
		}
		meshFileListFile.close();
	} // end of mesh_preparer_mode

}