#ifndef __DEF_MYXFORM_H
#define __DEF_MYXFORM_H

#include "helper.h"

class myXForm
{
public:
	vector< vec >						rotation_matrix; // each row has a vector
	vec											translation;

	myXForm()
	{
		rotation_matrix.push_back( vec(1.0, 0.0, 0.0) );
		rotation_matrix.push_back( vec(0.0, 1.0, 0.0) );
		rotation_matrix.push_back( vec(0.0, 0.0, 1.0) );
		translation = vec(0.0, 0.0, 0.0);
	}

	myXForm(float i)
	{
		rotation_matrix.push_back( vec(i, i, i) );
		rotation_matrix.push_back( vec(i, i, i) );
		rotation_matrix.push_back( vec(i, i, i) );
		translation = vec(i, i, i);
	}

	myXForm(float u1, float u2, float u3, float v1, float v2, float v3, float w1, float w2, float w3, float t1, float t2, float t3)
	{
		rotation_matrix.push_back( vec(u1, u2, u3) );
		rotation_matrix.push_back( vec(v1, v2, v3) );
		rotation_matrix.push_back( vec(w1, w2, w3) );
		rotation_matrix[0] = normalize( rotation_matrix[0] );
		rotation_matrix[1] = normalize( rotation_matrix[1] );
		rotation_matrix[2] = normalize( rotation_matrix[2] );
		translation = vec(t1, t2, t3);
	}

	void init_from_matlab_style(float u1, float v1, float w1, 
		                          float u2, float v2, float w2, 
												      float u3, float v3, float w3,
												      float t1 = 0.0, float t2 = 0.0, float t3 = 0.0)
	{
		rotation_matrix.clear();
		rotation_matrix.push_back( vec(u1, u2, u3) );
		rotation_matrix.push_back( vec(v1, v2, v3) );
		rotation_matrix.push_back( vec(w1, w2, w3) );
		rotation_matrix[0] = normalize( rotation_matrix[0] );
		rotation_matrix[1] = normalize( rotation_matrix[1] );
		rotation_matrix[2] = normalize( rotation_matrix[2] );
		translation = vec(t1, t2, t3);
	}

	//const float operator [] (size_t i) const
	//{ 
	//	return rotation_matrix[i]; 
	//}

	//float &operator [] (size_t i) 
	//{ 
	//	return rotation_matrix[i]; 
	//}


	void orthogonalize()
	{
		myXForm Q(0);
		myXForm R(0);

		// for each vector
		for (size_t j = 0; j < 3; j++)
		{
			vec v = rotation_matrix[j];
			for (size_t i = 0; i < j; i++)
			{
				R.rotation_matrix[j][i] = Q.rotation_matrix[i] DOT rotation_matrix[j];
				v = v - R.rotation_matrix[j][i] * Q.rotation_matrix[i];
			}
			R.rotation_matrix[j][j] = len(v);
			Q.rotation_matrix[j] = v / R.rotation_matrix[j][j];
		}

		rotation_matrix = Q.rotation_matrix;
	}


	void inv()
	{
		myXForm Q;
		for (size_t i = 0; i < 3; i++)
			for (size_t j = 0; j < 3; j++)
				Q.rotation_matrix[i][j] = rotation_matrix[j][i];

		rotation_matrix = Q.rotation_matrix;

		vec T;
		for (size_t i = 0; i < 3; i++)
			T[i] = rotation_matrix[i] DOT translation;

		translation = -T;
	}

};



//inline myXForm op(const vec& v1, const vec& v2)
//{
//	myXForm Q;
//	Q.rotation_matrix[0][0] = v1[0] * v2[0];
//	Q.rotation_matrix[0][1] = v1[0] * v2[1];
//	Q.rotation_matrix[0][2] = v1[0] * v2[2];
//	Q.rotation_matrix[1][0] = v1[1] * v2[0];
//	Q.rotation_matrix[1][1] = v1[1] * v2[1];
//	Q.rotation_matrix[1][2] = v1[1] * v2[2];
//	Q.rotation_matrix[2][0] = v1[2] * v2[0];
//	Q.rotation_matrix[2][1] = v1[2] * v2[1];
//	Q.rotation_matrix[2][2] = v1[2] * v2[2];
//	return Q;
//}

static inline myXForm operator + (const myXForm &xf1, const myXForm &xf2)
{
	myXForm newxf;
	for (size_t i = 0; i < 3; i++)
		for (size_t j = 0; j < 3; j++)
			newxf.rotation_matrix[i][j] = xf1.rotation_matrix[i][j] + xf2.rotation_matrix[i][j];

	newxf.translation = (xf1.translation + xf2.translation) * 0.5f;
	newxf.orthogonalize();
	return newxf;
}



static inline point operator * (const myXForm &xf, const point &p)
{
	bool isinf = false;
	for (size_t i = 0; i < 3; i++)
		for (size_t j = 0; j < 3; j++)
			isinf = isinf | xf.rotation_matrix[i][j] == FLT_MAX | _isnan( xf.rotation_matrix[i][j] );
	for (size_t i = 0; i < 3; i++)
		isinf = isinf | xf.translation[i] == FLT_MAX | _isnan( xf.translation[i] );
	if (isinf)
		return point(FLT_MAX, FLT_MAX, FLT_MAX);

	point newp;
	for (size_t i = 0; i < 3; i++)
		newp[i] = xf.rotation_matrix[i] DOT p;

	return newp + xf.translation;
}



static inline myXForm operator * (const myXForm &xf1, const myXForm &xf2)
{
	myXForm newxf;
	newxf.rotation_matrix[0][0] = xf1.rotation_matrix[0][0]*xf2.rotation_matrix[0][0] + xf1.rotation_matrix[0][1]*xf2.rotation_matrix[1][0] + xf1.rotation_matrix[0][2]*xf2.rotation_matrix[2][0];
	newxf.rotation_matrix[0][1] = xf1.rotation_matrix[0][0]*xf2.rotation_matrix[0][1] + xf1.rotation_matrix[0][1]*xf2.rotation_matrix[1][1] + xf1.rotation_matrix[0][2]*xf2.rotation_matrix[2][1];
	newxf.rotation_matrix[0][2] = xf1.rotation_matrix[0][0]*xf2.rotation_matrix[0][2] + xf1.rotation_matrix[0][1]*xf2.rotation_matrix[1][2] + xf1.rotation_matrix[0][2]*xf2.rotation_matrix[2][2];
	newxf.rotation_matrix[1][0] = xf1.rotation_matrix[1][0]*xf2.rotation_matrix[0][0] + xf1.rotation_matrix[1][1]*xf2.rotation_matrix[1][0] + xf1.rotation_matrix[1][2]*xf2.rotation_matrix[2][0];
	newxf.rotation_matrix[1][1] = xf1.rotation_matrix[1][0]*xf2.rotation_matrix[0][1] + xf1.rotation_matrix[1][1]*xf2.rotation_matrix[1][1] + xf1.rotation_matrix[1][2]*xf2.rotation_matrix[2][1];
	newxf.rotation_matrix[1][2] = xf1.rotation_matrix[1][0]*xf2.rotation_matrix[0][2] + xf1.rotation_matrix[1][1]*xf2.rotation_matrix[1][2] + xf1.rotation_matrix[1][2]*xf2.rotation_matrix[2][2];
	newxf.rotation_matrix[2][0] = xf1.rotation_matrix[2][0]*xf2.rotation_matrix[0][0] + xf1.rotation_matrix[2][1]*xf2.rotation_matrix[1][0] + xf1.rotation_matrix[2][2]*xf2.rotation_matrix[2][0];
	newxf.rotation_matrix[2][1] = xf1.rotation_matrix[2][0]*xf2.rotation_matrix[0][1] + xf1.rotation_matrix[2][1]*xf2.rotation_matrix[1][1] + xf1.rotation_matrix[2][2]*xf2.rotation_matrix[2][1];
	newxf.rotation_matrix[2][2] = xf1.rotation_matrix[2][0]*xf2.rotation_matrix[0][2] + xf1.rotation_matrix[2][1]*xf2.rotation_matrix[1][2] + xf1.rotation_matrix[2][2]*xf2.rotation_matrix[2][2];

	newxf.translation = xf1 * xf2.translation;
	return newxf;
}

static inline std::istream &operator >> (std::istream &is, myXForm& m)
{
	for (size_t i = 0; i < 3; i++)
	{
		for (size_t j = 0; j < 3; j++)
			is >> m.rotation_matrix[i][j];
		is >> m.translation[i];
	}

	m.rotation_matrix[0] = normalize( m.rotation_matrix[0] );
	m.rotation_matrix[1] = normalize( m.rotation_matrix[1] );
	m.rotation_matrix[2] = normalize( m.rotation_matrix[2] );

	return is;
}

static inline std::ostream &operator << (std::ostream &os, const myXForm &m)
{
	for (size_t i = 0; i < 3; i++)
	{
		for (size_t j = 0; j < 3; j++)
			os << m.rotation_matrix[i][j] << " ";
		os << m.translation[i] << std::endl;
	}

	return os;
}



inline point euler_angles( const myXForm& xft, const myXForm& xfs )
{
	myXForm invxfs = xfs;
	invxfs.inv();
	myXForm r = xft * invxfs;
	float thetaX = 0;
	float thetaY = asin( r.rotation_matrix[0][2] );
	float thetaZ = 0;
	if ( thetaY < M_PI_2 )
	{
		if ( thetaY > -M_PI_2 )
		{
			thetaX = atan2( -r.rotation_matrix[1][2], r.rotation_matrix[2][2] );
			thetaZ = atan2( -r.rotation_matrix[0][1], r.rotation_matrix[0][0] );
		} 
		else
		{
			thetaX = -atan2( r.rotation_matrix[1][0], r.rotation_matrix[1][1] );
			thetaZ = 0.0f;
		}
	}
	else
	{
		thetaX = atan2( r.rotation_matrix[1][0], r.rotation_matrix[1][1] );
		thetaZ = 0.0f;
	}
	return point( thetaX, thetaY, thetaZ );
}




	// testing
	//myXForm	 A(0.5, -0.5, 0,  
	//	         0.2, -0.2, 0.1, 
	//					 0.1,  0.1, 1,
	//					 1,    2,   4);	
	//myXForm	 B(1, 3, 0.5,  
	//	         2, -3, 1, 
	//					 0, 1, 2,
	//					 1,    .5,   3);	
	//XForm<float> AA(0.707107,   0,  0.707107, 0,
	//	              -0.707107,  0,  0.707107, 0,
	//								    0,      1,       0,   0,
	//								    1,      2,       4,   1 );

	//A.orthogonalize();
	//std::cout << A * A << std::endl;
	//std::cout << AA * AA << std::endl;

	//B.orthogonalize();
	//std::cout << A << std::endl;
	//std::cout << B << std::endl;
	//orthogonalize(AA);	
	//std::cout << A * point(0.4, 0.2, 0.9) << std::endl;
	//std::cout << A * point(0, 0, 1) << std::endl;
	//std::cout << AA * point(0.4, 0.2, 0.9) << std::endl;
	//std::cout << AA * point(0, 0, 1) << std::endl;
	//std::cout << AAA * point(0.4, 0.2, 0.9) << std::endl;
	//std::cout << AAA * point(0, 0, 1) << std::endl;
	//std::cout << AAA << std::endl;
	//std::cout << AA << std::endl;
	//std::cout << A << std::endl;
	//std::cout << B << std::endl;
	//std::cout << A * A << std::endl;
	//std::cout << A * B << std::endl;
	//std::cout << A * myXForm() << std::endl;
	//std::cout << A * myXForm(1) << std::endl;
	//A.inv();
	//std::cout << A << std::endl;
	//system("pause");

;


#endif




