/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <opencv\cxcore.hpp>
#include "SSSFunc.h"

namespace SSSLib
{

	template <typename _TVal>
	_TVal clampFetch(const cv::Mat& iMat, int x, int y)
	{
		return iMat.at<_TVal>(clamp(y, 0, iMat.rows-1), clamp(x, 0, iMat.cols-1));
	}

	template <typename _TVal>
	_TVal linearFilterFetch(const cv::Mat& iMat, float iX, float iY)
	{
		int x = max(int(iX-0.5f), 0);
		int xp = min(int(iX+0.5f), iMat.cols-1);
		int y = max(int(iY-0.5f), 0);
		int yp = min(int(iY+0.5f), iMat.rows-1);
		float xint = max(iX - x - 0.5f, 0.0f);
		float yint = max(iY - y - 0.5f, 0.0f);
#define _FETCH(xx,yy) clampFetch<_TVal>(iMat, xx, yy)
		return interpolate(
			interpolate( _FETCH(x, y), _FETCH(xp, y), xint ),
			interpolate( _FETCH(x, yp), _FETCH(xp, yp), xint ),
			yint
			);
#undef _FETCH
	}

	template <typename _TVal>
	_TVal linearFilterFetch_Norm(const cv::Mat& iMat, float iX, float iY) // coordinates are normalized to [0..1]
	{
		return linearFilterFetch<_TVal>(iMat, iX * iMat.cols, iY * iMat.rows);
	}

	inline float linearFilterFetchImg(const cv::Mat& iMat, float iX, float iY, int chn)
	{
		int x = Max(int(iX-0.5f), 0);
		int xp = Min(int(iX+0.5f), iMat.cols-1);
		int y = Max(int(iY-0.5f), 0);
		int yp = Min(int(iY+0.5f), iMat.rows-1);
		float xint = Max(iX - x - 0.5f, 0.0f);
		float yint = Max(iY - y - 0.5f, 0.0f);
#define _FETCH(xx,yy) clampFetch<cv::Vec3b>(iMat, xx, yy)
		return interpolate(
			interpolate( (float)_FETCH(x, y)[chn], (float)_FETCH(xp, y)[chn], xint ),
			interpolate( (float)_FETCH(x, yp)[chn], (float)_FETCH(xp, yp)[chn], xint ),
			yint
			);
#undef _FETCH
	}

	inline float linearFilterFetchImg_Norm(const cv::Mat& iMat, float iX, float iY, int chn)
	{
		return linearFilterFetchImg(iMat, iX * iMat.cols, iY * iMat.rows, chn);
	}

}