/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <cxcore.h>
#include <cv.h>
#include <highgui.h>
#include <iostream>


class SCVImg
{
public:
	SCVImg(){pImg = NULL;}
	SCVImg(const SCVImg& iImg)
		: pImg(NULL)
	{*this=iImg;}
	SCVImg(const IplImage* iImg){pImg = cvCloneImage(iImg);}
	SCVImg(int width, int height, int nchannels, int depth = IPL_DEPTH_8U)
	{
		pImg = cvCreateImage(cvSize(width, height),
			depth,
			nchannels);
	}
	void ReleaseData(){if (pImg) cvReleaseImage(&pImg);}
	~SCVImg(){ReleaseData();}
	SCVImg& TakeOver(IplImage* iImg){pImg = iImg; return *this;}
	IplImage* GetImg(){return pImg;}
	const IplImage* GetImg() const{return pImg;}
	int GetWidth() const{return pImg->width;}
	int GetHeight() const{return pImg->height;}
	int GetRowCount() const{return pImg->height;}
	int GetColCount() const{return pImg->width;}
	int GetChan() const{return pImg->nChannels;}
	int GetnChannels() const{return pImg->nChannels;}
	CvRect GetROI() const{return cvGetImageROI(pImg);}
	void SetROI(CvRect iRoi){cvSetImageROI(pImg, iRoi);}
	unsigned char& operator()(int w, int h, int ch = 0)
	{
		return (unsigned char&)
		(pImg->imageData[h*pImg->widthStep+w*GetnChannels()+ch]);
	}
	const unsigned char& operator()(int w, int h, int ch = 0) const
	{
		return (unsigned char&)
		(pImg->imageData[h*pImg->widthStep+w*GetnChannels()+ch]);
	}
	unsigned char GetEle(int row, int col, int ch = 0) const
	{
		return pImg->imageData[row*pImg->widthStep+col*GetnChannels()+ch];
	}
	void operator=(const SCVImg& iImg)
	{
		if (pImg)
			cvReleaseImage(&pImg);
		if (iImg.GetImg())
		pImg = cvCloneImage(iImg.GetImg());
	}
	SCVImg CvtColor(int flag)
	{
		SCVImg outImg;
		switch (flag)
		{
		case CV_RGB2GRAY:
		case CV_BGR2GRAY:
		case CV_BGRA2GRAY:
		case CV_RGBA2GRAY:
		case CV_BGR5652GRAY:
		case CV_BGR5552GRAY:
			outImg.TakeOver(cvCreateImage(
				cvSize(GetImg()->width,GetImg()->height),
				GetImg()->depth,1));
			break;
		case CV_BGR2YCrCb:
		case CV_RGB2YCrCb:
		case CV_YCrCb2BGR:
		case CV_YCrCb2RGB:
		case CV_BGR2HSV:
		case CV_RGB2HSV:
			outImg = SCVImg(GetWidth(), GetHeight(), 3, GetImg()->depth);

		default:
			std::cout<<"Convert Format Not Supported Yet!\n";
		}
		cvCvtColor(GetImg(), outImg.GetImg(), flag);
		return outImg;
	}
	SCVImg SplitChannel(int iChn)
	{
		SCVImg outImg(GetWidth(), GetHeight(), 1);
		switch (iChn)
		{
		case 0:
			cvSplit(GetImg(), outImg.GetImg(), NULL, NULL, NULL);
			break;
		case 1:
			cvSplit(GetImg(), NULL, outImg.GetImg(), NULL, NULL);
			break;
		case 2:
			cvSplit(GetImg(), NULL, NULL, outImg.GetImg(), NULL);
			break;
		case 3:
			cvSplit(GetImg(), NULL, NULL, NULL, outImg.GetImg());
			break;
		}
		return outImg;
	}

protected:
	IplImage* pImg;
};