/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "SSSSmallVec.h"
#include "SSSFunc.h"
#include <math.h>
#include <iostream>

namespace SSSLib
{
	inline Vec3f HSV2RGBf(Vec3f iHSV) // all in [0, 1)
	{
		float c = iHSV[1]*iHSV[2];
		float hp = iHSV[0]*6;
		//if (!(hp>=0 && hp<6))
		//	std::cout<<"!";
		float x = c*(1-fabs(fmod(hp, 2.0f)-1));
		Vec3f rgb1;
		switch (int(hp))
		{
		case 0:
			rgb1=Vec3f(c, x, 0);
			break;
		case 1:
			rgb1=Vec3f(x, c, 0);
			break;
		case 2:
			rgb1=Vec3f(0, c, x);
			break;
		case 3:
			rgb1=Vec3f(0, x, c);
			break;
		case 4:
			rgb1=Vec3f(x, 0, c);
			break;
		case 5:
			rgb1=Vec3f(c, 0, x);
			break;
		}
		float m = iHSV[2]-c;
		return rgb1+m;
		//return Vec3uc(clamp<int>((int)((rgb1[0]+m)*255), 0, 255),
		//				clamp<int>((int)((rgb1[1]+m)*255), 0, 255),
		//				clamp<int>((int)((rgb1[2]+m)*255), 0, 255));
	}

	inline Vec3uc HSV2RGB(Vec3f iHSV)
	{
		Vec3f rgbf = HSV2RGBf(iHSV);
		return Vec3uc(clamp<int>((int)(rgbf[0]*255), 0, 255),
						clamp<int>((int)(rgbf[1]*255), 0, 255),
						clamp<int>((int)(rgbf[2]*255), 0, 255));
	}

	inline Vec3f Value2HSVf(float iVal)
	{
		Vec3f hsv(iVal, .8f, 1.0f);
		return hsv;
	}

	inline Vec3uc Value2RGBuc(float iVal)
	{
		return HSV2RGB(Value2HSVf(iVal));
	}

	inline std::vector<Vec3uc> RandColor(int iSize)
	{
		std::vector<Vec3uc> retVec(iSize);
		float hue = randf<float>();
		float gr = 0.618033988749895f;
		for (int i=0; i<iSize; i++)
		{
			Vec3f hsv;
			hue = fmod(hue+gr, 1.0f);
			//hue = randf<float>();
			retVec[i] = HSV2RGB(Vec3f(hue, 0.5f, 0.95f));
		}
		return retVec;
	}
	inline Vec3uc Value2RGBJetUC(double iVal)
	{
		double v = iVal;
		int b = v < 0.125 ?
			int((v + 0.125) * 4 * 255) :
				v < 0.375 ? 255 : v<0.625 ? int((0.625 - v) * 4 * 255) : 0;
		int g = v < 0.125 ? 0 : v < 0.375 ? int((v - 0.125) * 4 * 255) : v < 0.625 ? 255 : v<0.875 ? int((0.875 - v) * 4 * 255) : 0;
		int r = v < 0.375 ? 0 : v < 0.625 ? int((v - 0.375) * 4 * 255) : v < 0.875 ? 255 : int((1.125 - v) * 4 * 255);

		return Vec3uc(r, g, b);
	}
	inline Vec3d Value2RGBJet(double iVal)
	{
		double v = iVal;
		double b = v < 0.125 ?
			((v + 0.125) * 4 ) :
				v < 0.375 ? 1.0 : v<0.625 ? ((0.625 - v) * 4 ) : 0;
		double g = v < 0.125 ? 0 : v < 0.375 ? ((v - 0.125) * 4 ) : v < 0.625 ? 1.0 : v<0.875 ? ((0.875 - v) * 4 ) : 0;
		double r = v < 0.375 ? 0 : v < 0.625 ? ((v - 0.375) * 4 ) : v < 0.875 ? 1.0 : ((1.125 - v) * 4 );

		return Vec3d(r, g, b);
	}
}