#pragma once

#include "kdtree.h"
#include <vector>
#include "../SSSSmallVec.h"
#include "kdtree.h"
#include <memory>

namespace SSSLib
{
	template <class _TData>
	class KDTree3f
	{
	public:
		KDTree3f()
			: mpKDTree(kd_create(3), kd_free)
			, mData(0)
			, mPst(0)
		{}
		~KDTree3f(){}

		std::vector<_TData> mData;
		std::vector<SSSLib::Vec3f> mPst;
		std::shared_ptr<kdtree> mpKDTree;

		void BuildTree(int iN, const SSSLib::Vec3f* iPst, const _TData* iData)
		{
			mpKDTree.reset(kd_create(3), kd_free);
			mData.resize(iN);
			mPst.resize(iN);
			for (int i=0; i<iN; i++)
			{
				mData[i] = iData[i];
				mPst[i] = iPst[i];
				kd_insert3f(mpKDTree.get(),
					iPst[i][0],
					iPst[i][1],
					iPst[i][2],
					(void*)((_TData*)&mData[i]));
			}
		}

		void InitTree()
		{
			mpKDTree.reset(kd_create(3), kd_free);
			for (int i=0; i<(int)mPst.size(); i++)
			{
				kd_insert3f(mpKDTree.get(),
					mPst[i][0],
					mPst[i][1],
					mPst[i][2],
					(void*)((_TData*)&mData[i]));
			}
		}

		void InsertPt(SSSLib::Vec3f iPst, const _TData& iData)
		{
			mData.push_back(iData);
			mPst.push_back(iPst);
			//kd_insert3f(mpKDTree.get(), iPst[0], iPst[1], iPst[2], (void*)((_TData*)&mData[mData.size()-1]));
		}

		void QueryClosestPt(SSSLib::Vec3f iVec, _TData* oData, SSSLib::Vec3f* oVec)
		{
			std::shared_ptr<kdres> kdpoint(
								kd_nearest3f(mpKDTree.get(), iVec[0], iVec[1], iVec[2]),
								kd_res_free);
			void* dataPtr;
			if (oVec)
				kd_res_item3f(kdpoint.get(), oVec->val, oVec->val+1, oVec->val+2);
			dataPtr = kd_res_item_data(kdpoint.get());
			if (oData)
				*oData = *((_TData*)dataPtr);
		}
	};

	template <class _TData>
	class KDTree3d
	{
	public:
		KDTree3d()
			: mpKDTree(kd_create(3), kd_free)
			, mData(0)
			, mPst(0)
		{}
		~KDTree3d(){}

		std::vector<_TData> mData;
		std::vector<SSSLib::Vec3d> mPst;
		std::shared_ptr<kdtree> mpKDTree;

		void BuildTree(int iN, const SSSLib::Vec3d* iPst, const _TData* iData)
		{
			mpKDTree.reset(kd_create(3), kd_free);
			mData.resize(iN);
			mPst.resize(iN);
			for (int i=0; i<iN; i++)
			{
				mData[i] = iData[i];
				mPst[i] = iPst[i];
				kd_insert3(mpKDTree.get(),
					iPst[i][0],
					iPst[i][1],
					iPst[i][2],
					(void*)((_TData*)&mData[i]));
			}
		}

		void InitTree()
		{
			mpKDTree.reset(kd_create(3), kd_free);
			for (int i=0; i<(int)mPst.size(); i++)
			{
				kd_insert3(mpKDTree.get(),
					mPst[i][0],
					mPst[i][1],
					mPst[i][2],
					(void*)((_TData*)&mData[i]));
			}
		}

		void InsertPt(SSSLib::Vec3d iPst, const _TData& iData)
		{
			mData.push_back(iData);
			mPst.push_back(iPst);
			//kd_insert3f(mpKDTree.get(), iPst[0], iPst[1], iPst[2], (void*)((_TData*)&mData[mData.size()-1]));
		}

		void QueryClosestPt(SSSLib::Vec3d iVec, _TData* oData, SSSLib::Vec3d* oVec)
		{
			std::shared_ptr<kdres> kdpoint(
								kd_nearest3(mpKDTree.get(), iVec[0], iVec[1], iVec[2]),
								kd_res_free);
			void* dataPtr;
			if (oVec)
				kd_res_item3(kdpoint.get(), oVec->val, oVec->val+1, oVec->val+2);
			dataPtr = kd_res_item_data(kdpoint.get());
			if (oData)
				*oData = *((_TData*)dataPtr);
		}
	};

}