/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once
#include "OGLCam.h"
#include "OGLFrame.h"
#include "../SSSFunc.h"



namespace SSSLib
{

class OGLCamModel
	: public OGLCam
        , public OGLScene::OGLMouse
	, public OGLScene::OGLMotion
	, public OGLScene::OGLMouseWheel
	, public OGLScene::OGLReshape
{
public:
	OGLCamModel(void);
	~OGLCamModel(void);
	Vec3f ModelCenter;
	void EnableCam(OGLScene* ipScene = OGLScene::spCurrentScene)
	{
                OGLScene::OGLMouse::Enabled = true;
		OGLScene::OGLMotion::Enabled = true;
		OGLScene::OGLMouseWheel::Enabled = true;
		OGLScene::OGLReshape::Enabled = true;
                CamMouseID = ipScene->AddMouse(this);
		CamMotionID = ipScene->AddMotion(this);
		CamMouseWheelID = ipScene->AddMouseWheel(this);
	}
	void DisableCam()
	{
                OGLScene::OGLMouse::Enabled = false;
		OGLScene::OGLMotion::Enabled = false;
		OGLScene::OGLMouseWheel::Enabled = false;
		OGLScene::OGLReshape::Enabled = false;
		//OGLScene::DropMotionObj(CamMotionID);
		//OGLScene::DropMouseWheelObj(CamMouseWheelID);
	}
        virtual void MouseFunc(int button, int state, int x, int y)
        {
#ifdef __linux
            if (button==3 || button==4)
            {
                if (state==GLUT_UP)
                    return;
                int down=(button==3);
                MouseWheelFunc(button, down, x, y);
            }
#endif
        }
	virtual void MotionFunc(int x, int y)
	{
		OGLCommonScene* instance = (OGLCommonScene*)OGLScene::GetCurrentScene();
		if (instance->mMLDown)
		{
			Rot[0] = Rot[0]-instance->mDeltaY;
			Rot[0] = clamp(Rot[0], -89.99f, 89.99f);
			//while (Rot[0]<0) Rot[0]+=360;
			//while (Rot[0]>360) Rot[0]-=360;
			Rot[1] = Rot[1]-instance->mDeltaX;
			while (Rot[1]<0) Rot[1]+=360;
			while (Rot[1]>360) Rot[1]-=360;
			float crotx = -Rot[0]/180*float(_M_PI);
			float croty = -Rot[1]/180*float(_M_PI);
			CenterPst();
			//float len = (Posi-ModelCenter).Length();
			//Vec3f tmp(0,sin(crotx),cos(crotx));
			//Posi = Vec3f(
			//	-tmp[2]*sin(croty)*len,
			//	tmp[1]*len,
			//	tmp[2]*cos(croty)*len)+ModelCenter;
		}
	}
	virtual void CenterPst()
	{
		float crotx = -Rot[0]/180*float(_M_PI);
		float croty = -Rot[1]/180*float(_M_PI);
		float len = (Posi-ModelCenter).Length();
		Vec3f tmp(0,sin(crotx),cos(crotx));
		Posi = Vec3f(
			-tmp[2]*sin(croty)*len,
			tmp[1]*len,
			tmp[2]*cos(croty)*len)+ModelCenter;
	}
	virtual void MouseWheelFunc(int wheel, int direction, int x, int y)
	{
		float scale = 1;
		if (direction==1)
			scale = 0.9f;
		else
			scale = 1.1f;
		Posi = ModelCenter+(Posi-ModelCenter)*scale;
	}
	virtual void ReshapeFunc(int x, int y)
	{
		Aspect = float(x)/y;
	}
protected:
	int CamMotionID;
	int CamMouseWheelID;
        int CamMouseID;
};

}