/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "OGLTextureDisplay.h"
#include "../SSSFunc.h"
#include <vector>
#include <opencv/cxcore.h>
#include <opencv/highgui.h>

namespace SSSLib
{

template <typename TVal, int TIInternalFormat, int TIFormat, int TIType, int TIChannels>
class OGLDTexture2D: public OGLTexture2D
{
public:
	std::vector<TVal> mData;
	void CreateTexture(int iWidth, int iHeight, TVal* iData)
	{
		mData.resize(iWidth*iHeight*TIChannels);
		if (iData)
			memcpy((TVal*)(&mData[0]), iData, sizeof(TVal)*mData.size());
		OGLTexture2D::CreateTexture(iWidth, iHeight, TIInternalFormat, TIFormat, TIType, iData);
	}
	void UpdateFromMem()
	{
		Bind(false);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, mWidth, mHeight, TIFormat, TIType, (TVal*)(&mData[0]));
		OGLCheckError();
	}
	void UpdateFromFBO(GLenum iBufferName = GL_COLOR_ATTACHMENT0)
	{
		ReadCurrentBuffer(iBufferName, (TVal*)(&mData[0]));
	}
	TVal& At(int x, int y, int chn){return mData[(y*mWidth+x)*TIChannels+chn];}
	TVal& At(int x, int y, int chn) const {return mData[(y*mWidth+x)*TIChannels+chn];}
	TVal& operator()(int x, int y, int chn){return At(x,y,chn);}
	TVal& operator()(int x, int y, int chn) const {return At(x,y,chn);}

	void SaveToFile(const char* iFileName, float iScale = 1, float iAdd = 0)
	{
		cv::Mat img(mHeight, mWidth, CV_8UC3);
		for (int y=0; y<mHeight; y++)
		for (int x=0; x<mWidth; x++)
		for (int ch=0; ch<Min(3, TIChannels); ch++)
		{
			if (At(x, y, ch)<0)
				x=x;
			img.at<cv::Vec3b>(y, x)[3-ch] = (unsigned char)clamp<float>(At(x, y, ch)*iScale+iAdd, 0, 255);
		}
		cv::imwrite(iFileName, img);
	}
};

namespace DTex2D
{
	typedef OGLDTexture2D<float, GL_R32F, GL_RED, GL_FLOAT, 1> Tex1ChFl;
	typedef OGLDTexture2D<float, GL_RG32F, GL_RG, GL_FLOAT, 2> Tex2ChFl;
	typedef OGLDTexture2D<float, GL_RGB32F, GL_RGB, GL_FLOAT, 3> Tex3ChFl;
	typedef OGLDTexture2D<float, GL_RGBA32F, GL_RGBA, GL_FLOAT, 4> Tex4ChFl;
	typedef OGLDTexture2D<unsigned char, 1, GL_RED, GL_UNSIGNED_BYTE, 1> Tex1ChUB;
	typedef OGLDTexture2D<unsigned char, 3, GL_RGB, GL_UNSIGNED_BYTE, 3> Tex3ChUB;
	typedef OGLDTexture2D<unsigned char, 3, GL_BGR, GL_UNSIGNED_BYTE, 3> Tex3ChUBBGR;
	typedef OGLDTexture2D<unsigned char, 4, GL_RGBA, GL_UNSIGNED_BYTE, 4> Tex4ChUB;
}

}
