/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "OGLFrac.h"
#include "OGLSimpleObj.h"
#include <GL/glew.h>
#include <GL/freeglut.h>
#include "../SSSUtil.h"


using namespace SSSLib;
using namespace OGLFancy2D;

OGLFrac::OGLFrac(void)
{
}

OGLFrac::~OGLFrac(void)
{
}

void OGLFrac::Init(int iWidth, int iHeight)
{
	for (int i=0; i<2; i++)
		mTex[i].CreateTexture(iWidth, iHeight, NULL);

	for (int i=0; i<2; i++)
	{
		mTex[i].SetFilter(GL_TEXTURE_WRAP_S, GL_CLAMP);
		mTex[i].SetFilter(GL_TEXTURE_WRAP_T, GL_CLAMP);
		mTex[i].SetFilter(GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		mTex[i].SetFilter(GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}

	mDrawTex.CreateTexture(iWidth, iHeight, NULL);
	mDrawTex.SetFilter(GL_TEXTURE_WRAP_S, GL_REPEAT);
	mDrawTex.SetFilter(GL_TEXTURE_WRAP_T, GL_REPEAT);
	mDrawTex.SetFilter(GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	mDrawTex.SetFilter(GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	mUpdateShader.Create(NULL, "FracUpdate.fs");

	OGLCheckError();
	mCurLoop = 0;
}

void OGLFrac::UpdatePass()
{
	GLboolean rBlend;
	GLboolean rDepth;
	glGetBooleanv(GL_BLEND, &rBlend);
	glGetBooleanv(GL_DEPTH_TEST, &rDepth);

	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);

	mCurLoop = !mCurLoop;
	int cur = mCurLoop;
	int pre = !mCurLoop;

	mTex[cur].BindFrameBuffer(GL_COLOR_ATTACHMENT0_EXT);
	mDrawTex.BindFrameBuffer(GL_COLOR_ATTACHMENT1_EXT);
	OGLTexture2D::SetDrawBuffer(2);
	OGLCheckError();
	glClear(GL_COLOR_BUFFER_BIT);
	mUpdateShader.Use();
	mUpdateShader.Uniform1f("TimeCount", (float)_GET_TICK_COUNT());
	mUpdateShader.BindTexture("PrevWav", mTex[pre], 0);
	OGLCheckError();
	OGLRenderRectTex(1, -1, -1, 1);
	OGLCheckError();
	//mTex[cur].UpdateFromFBO();
	//glFramebufferRenderbuffer(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_RENDERBUFFER, 0);
	//glFramebufferRenderbuffer(GL_DRAW_FRAMEBUFFER, GL_COLOR_ATTACHMENT1, GL_RENDERBUFFER, 0);
	OGLTexture2D::SetDrawBuffer(1);
	OGLCheckError();
	mTex[cur].UnbindFrameBuffer();
	mUpdateShader.Unuse();
	glDisable(GL_TEXTURE_2D);

	if (rBlend)
		glEnable(GL_BLEND);
	if (rDepth)
		glEnable(GL_DEPTH_TEST);
}
