/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <GL/glew.h>
#include "../SSSSmallVec.h"
#include <math.h>
#include <opencv/cxcore.h>
#include <opencv/highgui.h>

namespace SSSLib
{
	template <typename _TVal>
	SmallVec<_TVal, 2> Turn(SmallVec<_TVal, 2> iVec, _TVal radius)
	{
		_TVal cosr = cos(radius);
		_TVal sinr = sin(radius);
		return SmallVec<_TVal, 2>(iVec[0]*cosr - iVec[1]*sinr, iVec[1]*cosr + iVec[0]*sinr);
	}

	inline cv::Mat ReadFB(int iW, int iH)
	{
		cv::Mat retMat(iH, iW, CV_8UC3);
		glReadPixels(0, 0, iW, iH, GL_BGR, GL_UNSIGNED_BYTE, &retMat.at<cv::Vec3b>(0, 0));
		cv::flip(retMat, retMat, 0);
		return retMat;
	}

	inline void PrintScreen(int iW, int iH, const char* iFileName)
	{
		cv::imwrite(iFileName, ReadFB(iW, iH));
	}
}
