/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <GL/glew.h>
#include <GL/freeglut.h>
#include "OGLShader.h"
#include "OGLSProgram.h"

namespace SSSLib
{
class OGLSEasyProgram
{
public:
	OGLSEasyProgram(void)
		: pVertex(NULL)
		, pFragment(NULL)
		, pProgram(NULL)
	{}
	~OGLSEasyProgram(void)
	{
		Release();
	}
	OGLSEasyProgram(const char* iVertex, const char* iFragment)
		: pVertex(NULL)
		, pFragment(NULL)
		, pProgram(NULL)
	{
		Create(iVertex, iFragment);
	}
	void Attach()
	{
		if (pVertex)
			pProgram->AttachShader(pVertex->GetID());
		if (pFragment)
			pProgram->AttachShader(pFragment->GetID());
		pProgram->Link();
	}
	void Create(const char* iVertex, const char* iFragment)
	{
		Release();
		pProgram = new OGLSProgram();
		if (iVertex)
		{
			pVertex = new OGLShader(iVertex, GL_VERTEX_SHADER);
			pProgram->AttachShader(pVertex->GetID());
		}
		if (iFragment)
		{
			pFragment = new OGLShader(iFragment, GL_FRAGMENT_SHADER);
			pProgram->AttachShader(pFragment->GetID());
		}
		pProgram->Link();
	}
	void Use(){pProgram->Use();}
	static void Unuse(){glUseProgram(0);}
	void Release()
	{
#define _DELETE(param) if (param) {delete param; param = NULL;}
		_DELETE(pVertex);
		_DELETE(pFragment);
		_DELETE(pProgram);
#undef _DELETE
	}
	GLuint GetVertexID(){return pVertex->GetID();}
	GLuint GetFragmentID(){return pFragment->GetID();}
	GLuint GetUniformLocation(const GLchar* iName){return pProgram->GetUniformLocation(iName);}
	void Uniform1i(const GLchar* iName, GLint iValue){pProgram->Uniform1i(iName, iValue);}
	void Uniform1f(const GLchar* iName, GLfloat iValue){pProgram->Uniform1f(iName, iValue);}
	void BindTexture(const GLchar* iName, OGLTexture2D& iTex, int iTexNum)
	{
		pProgram->BindTexture(iName, iTex, iTexNum);
	}
protected:
	OGLShader* pVertex;
	OGLShader* pFragment;
	OGLSProgram* pProgram;
};
}