/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <GL/glew.h>
#include <GL/freeglut.h>
#include "OGLShader.h"
#include "OGLTexture.h"

namespace SSSLib
{

class OGLSProgram
{
public:
	OGLSProgram(void){ID = glCreateProgram();}
	~OGLSProgram(void){Release();}
	GLuint GetID() const{return ID;}
	void Reset(){Release(); ID = glCreateProgram();}
	void AttachShader(GLuint iShader){glAttachShader( ID, iShader);}
	void AttachShader(const OGLShader& iShader){AttachShader(iShader.GetID());}
	void DetachShader(GLuint iShader){glDetachShader( ID, iShader);}
	void DetachShader(const OGLShader& iShader){DetachShader(iShader.GetID());}
	void Link(){glLinkProgram(ID);}
	void Use(){glUseProgram(ID);}
	static void Unuse(){glUseProgram(0);}
	GLuint GetUniformLocation(const GLchar* iName){return glGetUniformLocation(ID, iName);}
	void BindTexture(const GLchar* iName, OGLTexture2D& iTex, int iTexNum)
	{
		iTex.Bind(true, iTexNum);
		Uniform1i(iName, iTexNum);
	}
	void Uniform1i(const GLchar* iName, GLint iValue)
	{
		glUniform1i(GetUniformLocation(iName), iValue);
	}
	void Uniform1f(const GLchar* iName, GLfloat iValue)
	{
		glUniform1f(GetUniformLocation(iName), iValue);
	}
protected:
	void Release(){glDeleteProgram(ID);}
	GLuint ID;
};
}