/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <GL/freeglut.h>
#include <map>
#include <vector>
#include "../SSSFunc.h"
#include <list>
#include <iostream>

namespace SSSLib
{
	inline int OGLCheckError()
	{
		int err = glGetError();
		if (err>0)
			std::cout<<"Error Encountered:"<<err<<std::endl;
		return err;
	}
	class OGLScene
	{
	public:
		OGLScene(){}
		~OGLScene(){}

		int GetWindowWidth(){return glutGet(GLUT_WINDOW_WIDTH);}
		int GetWindowHeight(){return glutGet(GLUT_WINDOW_HEIGHT);}
		static int sWindowID;

		static bool InitWindow(int iWindowWidth,
			int iWindowHeight,
			const char* iWindowName = "default");

		static void Bind();
		static void EnableIdleRender()
		{
			glutIdleFunc(MultiPassRender);
		}
		static void DisableIdleRender()
		{
			glutIdleFunc(sIdleFunc);
		}

		typedef void (*DropFunc)(int iNum);

		inline void UseScene(){spCurrentScene = this;};
		static OGLScene* spCurrentScene;
		static OGLScene* GetCurrentScene(){return spCurrentScene;}

	// macro definition _NEW_FUNC_TYPE
#define _NEW_FUNC_TYPE(func,parameter1,parameter2) \
public: class OGL##func\
	{\
		friend class OGLScene;\
	public:\
		virtual void func##Func parameter1{}\
		bool Enabled;\
		OGL##func(bool iEnabled = true)\
		{\
			mpAttachScene = NULL;\
			Enabled = iEnabled;\
		}\
		~OGL##func()\
		{\
			Drop();\
		}\
		virtual void Drop(){if (mpAttachScene) mpAttachScene->Drop##func(mDropNum); mpAttachScene = NULL;}\
		OGLScene* mpAttachScene;\
		int mDropNum;\
	}; \
protected: struct OGL##func##Node\
	{\
		OGL##func* pFuncObj;\
		void (*pFunc) parameter1;\
		bool ObjOrFunc;\
		bool StaticFuncEnabled;\
		OGL##func##Node(){StaticFuncEnabled = true;};\
		OGL##func##Node(OGL##func* ipFuncObj, OGLScene* ipScene, int iDropNum)\
		{\
			pFuncObj = ipFuncObj;\
			ipFuncObj->mpAttachScene = ipScene;\
			ipFuncObj->mDropNum = iDropNum;\
			ObjOrFunc=0;\
		}\
		OGL##func##Node(OGL##func* ipFuncObj){pFuncObj = ipFuncObj; ObjOrFunc=0; };\
		OGL##func##Node(void (*ipFunc) parameter1){pFunc = ipFunc; ObjOrFunc=1; StaticFuncEnabled = true;};\
	};\
protected: std::map<int, OGL##func##Node> m##func##Tree; \
protected: int m##func##MaxID;\
public: void Add##func(int iNum, void (*iFunc) parameter1){m##func##Tree[iNum]=OGL##func##Node(iFunc); m##func##MaxID = Max(m##func##MaxID, iNum);}\
public: int Add##func(void (*iFunc) parameter1)\
	{\
		while (m##func##Tree.count(m##func##MaxID)>0)\
			m##func##MaxID++;\
		m##func##Tree[m##func##MaxID]=OGL##func##Node(iFunc);\
		return m##func##MaxID;\
	}\
	public: void Add##func(int iNum, OGL##func* iFunc){m##func##Tree[iNum]=OGL##func##Node(iFunc, this, iNum); m##func##MaxID = Max(m##func##MaxID, iNum);}\
public: int Add##func(OGL##func* iFunc)\
	{\
		while (m##func##Tree.count(m##func##MaxID)>0)\
			m##func##MaxID++;\
		m##func##Tree[m##func##MaxID]=OGL##func##Node(iFunc, this, m##func##MaxID);\
		return m##func##MaxID;\
	}\
public: void Enable##func(int iNum){SetEnable##func(iNum, true);}\
public: void Disable##func(int iNum){SetEnable##func(iNum, false);}\
public: void SetEnable##func(int iNum, bool iEnable)\
	{\
		OGL##func##Node& node = m##func##Tree[iNum];\
		if (node.ObjOrFunc==0) node.pFuncObj->Enabled = iEnable;\
		else node.StaticFuncEnabled = iEnable;\
	}\
public: void Drop##func(int iNum){if (m##func##Tree.count(iNum)>0) m##func##Tree.erase(iNum);}\
protected: static void s##func##Func parameter1{ \
	if (!spCurrentScene) return;\
	for (std::map<int, OGL##func##Node>::iterator ite = spCurrentScene->m##func##Tree.begin();\
		ite != spCurrentScene->m##func##Tree.end();ite++)\
		if (ite->second.ObjOrFunc==0 && ite->second.pFuncObj->Enabled) ite->second.pFuncObj->func##Func parameter2;\
			else if (ite->second.ObjOrFunc==1 && ite->second.StaticFuncEnabled) ite->second.pFunc parameter2;	\
	}
	// ending _NEW_FUNC_TYPE

	// defining callback functions
		_NEW_FUNC_TYPE(Display,(void),());
		_NEW_FUNC_TYPE(Idle,(void),());
		_NEW_FUNC_TYPE(ComputeBeforeDisplay,(void),());
		_NEW_FUNC_TYPE(Reshape,(int x, int y),(x,y));
		_NEW_FUNC_TYPE(Keyboard,(unsigned char key, int x, int y),(key,x,y));
		_NEW_FUNC_TYPE(KeyboardUp,(unsigned char key, int x, int y),(key,x,y));
		_NEW_FUNC_TYPE(Special,(int key, int x, int y),(key,x,y));
		_NEW_FUNC_TYPE(SpecialUp,(int key, int x, int y),(key,x,y));
		_NEW_FUNC_TYPE(Mouse,(int button, int state, int x, int y),(button,state,x,y));
		_NEW_FUNC_TYPE(Motion,(int x, int y),(x,y));
		_NEW_FUNC_TYPE(MouseWheel,(int wheel, int direction, int x, int y),(wheel,direction,x,y));

		static void ComputeAndRender()
		{
			sComputeBeforeDisplayFunc();
			sDisplayFunc();
			glFlush();
			glutSwapBuffers();
		}

		static void MultiPassRender()
		{
			glClearColor(0.0, 0.0, 0.0, 0.0);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			//ComputeAndRender();

			using namespace std;
			if (!spCurrentScene)
				return;
			for (spCurrentScene->mStage=0; spCurrentScene->mStage<4; spCurrentScene->mStage++)
			{
				TStage& curPass = spCurrentScene->GetPass((TStageName)spCurrentScene->mStage);
				for (spCurrentScene->mPass=0; spCurrentScene->mPass<(int)curPass.size(); spCurrentScene->mPass++)
				for (list<OGLRenderNode>::iterator ite = curPass[spCurrentScene->mPass].begin();
					ite != curPass[spCurrentScene->mPass].end();
					ite++)
					if (ite->ObjOrFunc==0 && ite->pFuncObj->Enabled)
						ite->pFuncObj->RenderFunc((TStageName)spCurrentScene->mStage, spCurrentScene->mPass);
					else if (ite->ObjOrFunc==1 && ite->StaticFuncEnabled)
						ite->pFunc((TStageName)spCurrentScene->mStage, spCurrentScene->mPass);
			}

			//glFlush();
			//glutSwapBuffers();

		}


	// Multi-pass Rendering
	public:
		enum TStageName
		{
			CALCULATE_PASS = 0, PRE_PASS, RENDER_PASS, EFFECT_PASS
		};

		_NEW_FUNC_TYPE(Render, (TStageName stage, int pass), (stage, pass));
	#undef _NEW_FUNC_TYPE

	public:
		class OGLRenderExt: public OGLRender
		{
		public:
			virtual void Drop()
			{
				OGLRender::Drop();
				if (mpList)
				{
					mpList->erase(mPstIte);
					mpList = NULL;
				}
			}
			std::list<OGLRenderNode>::iterator mPstIte;
			std::list<OGLRenderNode> * mpList;
		};

		typedef std::vector<std::list<OGLRenderNode> > TStage;

	protected:

		TStage mCalPass;
		TStage mPrePass;
		TStage mRenderPass;
		TStage mAfterPass;

	public:


		int GetPass(){return mPass;}
		TStageName GetStage(){return (TStageName)mStage;}

		int mStage;
		int mPass;

	protected:
		TStage& GetPass(TStageName iName)
		{
			switch (iName)
			{
			case CALCULATE_PASS:
				return mCalPass;	
			case PRE_PASS:
				return mPrePass;
			case RENDER_PASS:
				return mRenderPass;
			case EFFECT_PASS:
				return mAfterPass;
			default:
				return mRenderPass;
			}
		}

	public:

		void AddRenderExt(OGLRenderExt* iFunc, TStageName iStage = RENDER_PASS, int iPass = 1)
		{
			//glClearColor(0.0, 0.0, 0.0, 0.0);
			//glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			
			using namespace std;
			TStage& curPass = GetPass(iStage);
			if (iPass>=(int)curPass.size())
				curPass.resize(iPass+1);
			list<OGLRenderNode>& curList = curPass[iPass];
			curList.push_front(OGLRenderNode(iFunc));
			iFunc->mPstIte = curList.begin();
			iFunc->mpList = &curList;

			//glFlush();
			//glutSwapBuffers();

		}
		void AddRenderExt(void  (*iFunc)(TStageName stage, int pass), TStageName iStage = RENDER_PASS, int iPass = 1)
		{
			using namespace std;
			TStage& curPass = GetPass(iStage);
			if (iPass>=(int)curPass.size())
				curPass.resize(iPass+1);
			list<OGLRenderNode>& curList = curPass[iPass];
			curList.push_front(OGLRenderNode(iFunc));
		}

		static void OGLClean(TStageName stage, int pass)
		{
			glClearColor(0.0, 0.0, 0.0, 0.0);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		}

	};
}
