/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "OGLShader.h"
#include <iostream>
#include <list>
#include <string>
#include <fstream>

using namespace SSSLib;
using namespace std;

namespace
{
	
void printShaderInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten  = 0;
    char *infoLog;

	glGetShaderiv(obj, GL_INFO_LOG_LENGTH,&infologLength);

    if (infologLength > 0)
    {
        infoLog = (char *)malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
		printf("%s\n",infoLog);
        free(infoLog);
    }
}

}

void OGLShader::LoadShader(const char* iFileName, GLenum iShaderType)
{
	Release();
	FileContent fc;
	LoadFile(iFileName, fc);
	ID = glCreateShader(iShaderType);
	glShaderSource(ID, fc.Content.size(), fc.DataPtr, NULL);
	int err = glGetError();
	if (err>0)
	{
		cout<<"Shader Error, Code "<<err<<" File: "<<iFileName;
		return;
	}
	glCompileShader(ID);
	err = glGetError();
	if (err>0)
		cout<<"Shader Error, Code "<<err<<" File: "<<iFileName;
	else
		IsCreated = true;
	printShaderInfoLog(ID);
}

void OGLShader::LoadFile(const char* iFileName, FileContent& oData)
{
	oData = FileContent();
	ifstream file(iFileName);
	if (!file.is_open())
	{
		cout<<iFileName<<" File Open Error!\n";
		return;
	}
	char buffer[3000];
	while (!file.eof())
	{
		file.getline(buffer,3000);
		oData.Content.push_back(string(buffer)+"\n");
	}
	oData.DataPtr = new const char*[oData.Content.size()];
	oData.LineLength.resize(oData.Content.size());
	int i=0;
	for (list<string>::iterator ite = oData.Content.begin();
		ite != oData.Content.end();
		ite ++)
	{
		oData.DataPtr[i] = ite->c_str();
		oData.LineLength[i] = ite->length();
		i++;
	}
}