/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once
#include <GL/glew.h>
#include <GL/freeglut.h>
#include <string>
#include <list>
#include <vector>

namespace SSSLib
{
class OGLShader
{
public:
	OGLShader(void)
		: ID(-1)
		, IsCreated(false)
	{}
	OGLShader(const char* iFileName, GLenum iShaderType)
	{
		IsCreated = false;
		LoadShader(iFileName, iShaderType);
	}
	~OGLShader(void)
	{
		Release();
	}

	struct FileContent
	{
		FileContent(){DataPtr = NULL;}
		~FileContent()
		{
			Release();
		}
		void Release()
		{
			Content.clear();
			delete [] DataPtr;
		}
		FileContent& operator=(const FileContent& iFC)
		{
			Release();
			Content = iFC.Content;
			DataPtr = new const char*[Content.size()];
			int i=0;
			for (std::list<std::string>::iterator ite = Content.begin();
				ite!=Content.end();
				ite++)
			{
				DataPtr[i] = ite->c_str();
				i++;
			}
			return *this;
		}
		const char** DataPtr;
		std::vector<int> LineLength;
		std::list<std::string> Content;
	};
	void LoadShader(const char* iFileName, GLenum iShaderType);
	void LoadFile(const char* iFileName, FileContent& oData);
	void Release(){if (IsCreated) {glDeleteShader(ID); IsCreated = false;}}
	GLuint GetID() const{return ID;}
	bool GetIsCreated() const{return IsCreated;}
protected:
	GLuint ID;
	bool IsCreated;
};
}