/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <GL/glew.h>
#include <GL/freeglut.h>
#include "../SSSSmallVec.h"

namespace SSSLib
{

class OGLTexture2D
{
public:
	static GLuint sCommonFBO;
	static bool sIsFBOInit;
	static GLuint GetFBO();
	OGLTexture2D(void);
	~OGLTexture2D(void);
	void SetFilter(GLenum iName, GLint iParam)
	{
		glBindTexture(GL_TEXTURE_2D, ID);
		glTexParameteri(GL_TEXTURE_2D, iName, iParam);
	}
	void CreateTexture(GLsizei iWidth,
						GLsizei iHeight,
						GLint iChans,
						GLenum iFormat,
						GLenum iType,
						void *iData);
	void CreateTexture4ChFl(GLsizei iWidth,
					GLsizei iHeight,
					float* iData)
	{
		CreateTexture(iWidth,
			iHeight,
			GL_RGBA,
			GL_RGBA,
			GL_FLOAT,
			iData);
	}
	void CreateTexture1ChFl(GLsizei iWidth,
					GLsizei iHeight,
					float* iData)
	{
		CreateTexture(iWidth,
			iHeight,
			GL_RED,
			GL_RED,
			GL_FLOAT,
			iData);
	}
	void CreateTexture1ChI(GLsizei iWidth,
					GLsizei iHeight,
					float* iData)
	{
		CreateTexture(iWidth,
			iHeight,
			GL_RED,
			GL_RED,
			GL_INT,
			iData);
	}
	void CreateTexture3ChUB(GLsizei iWidth,
					GLsizei iHeight,
					unsigned char* iData)
	{
		CreateTexture(iWidth,
			iHeight,
			GL_RGB,
			GL_RGB,
			GL_UNSIGNED_BYTE,
			iData);
	}
	void CreateTexture4ChUB(GLsizei iWidth,
					GLsizei iHeight,
					unsigned char* iData)
	{
		CreateTexture(iWidth,
			iHeight,
			GL_RGBA,
			GL_RGBA,
			GL_UNSIGNED_BYTE,
			iData);
	}
	void CreateTexture3ChUBBGR(GLsizei iWidth,
					GLsizei iHeight,
					unsigned char* iData)
	{
		CreateTexture(iWidth,
			iHeight,
			GL_RGB,
			GL_BGR,
			GL_UNSIGNED_BYTE,
			iData);
	}
	void ManageTexture(GLuint iID){ID = iID; IsCreated = true;}
	void UnmanageTexture(){IsCreated = false;}
	void LoadFromFile(const char* iFileName, SSSLib::Vec3uc* ipColorKey = NULL);
	void Bind(bool iIsMultiTexture, int iTextureUnit = -1);
	static void Unbind(int iTextureUnit = -1);
	GLuint GetID() const{return ID;}
	virtual void Draw(float iTop, float iLeft, float iButton, float iRight)
	{
		Bind(false);
		glBegin(GL_QUADS);
			glTexCoord2f(1.0f, 0.0f);		glVertex2f(iTop, iLeft);
			glTexCoord2f(0.0f, 0.0f);		glVertex2f(iButton, iLeft);
			glTexCoord2f(0.0f, 1.0f);		glVertex2f(iButton, iRight);
			glTexCoord2f(1.0f, 1.0f);		glVertex2f(iTop, iRight);
		glEnd();
	}
	virtual void Draw(Vec2f* iPt)
	{
		Bind(false);
		glBegin(GL_QUADS);
			glTexCoord2f(1.0f, 0.0f);		glVertex2fv(iPt[0].val);
			glTexCoord2f(0.0f, 0.0f);		glVertex2fv(iPt[1].val);
			glTexCoord2f(0.0f, 1.0f);		glVertex2fv(iPt[2].val);
			glTexCoord2f(1.0f, 1.0f);		glVertex2fv(iPt[3].val);
		glEnd();
	}
	//static void BindFrameBuffer(){glBindFramebuffer(GL_FRAMEBUFFER, GetFBO());}
	void BindFrameBuffer(GLenum iBufferName = GL_COLOR_ATTACHMENT0_EXT);
	static void SetDrawBuffer(int iNum, GLenum* ipNames = NULL);
	static void UnbindFrameBuffer()
	{
		glBindFramebuffer(GL_FRAMEBUFFER, 0);
	}
	static void ReadCurrentBuffer(GLenum iBufferName, GLenum iFormat, GLenum iType, int iWidth, int iHeight, GLvoid* oData);
	void ReadCurrentBuffer(GLenum iBufferName, GLvoid* oData){ReadCurrentBuffer(iBufferName, mFormat, mType, mWidth, mHeight, oData);}
	GLenum mFormat;
	GLenum mType;
	int mWidth;
	int mHeight;
	GLuint ID;
	bool IsCreated;
};

}