/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "OGLWater2D.h"
#include "OGLSimpleObj.h"
#include <GL/glew.h>
#include <GL/freeglut.h>

using namespace std;
using namespace SSSLib;
using namespace OGLFancy2D;

OGLWater2D::OGLWater2D(void)
: mCurLoop(0)
{

}

OGLWater2D::~OGLWater2D(void)
{
}

void OGLWater2D::Init(int iWidth, int iHeight)
{
	OGLCheckError();

	for (int i=0; i<2; i++)
		mWaveTex[i].CreateTexture(iWidth, iHeight, NULL);

	OGLCheckError();

	mInitShader.Create(NULL, "WaterInit.fs");
	mUpdateShader.Create(NULL, "WaterUpdate.fs");
	mRenderShader.Create(NULL, "WaterRender.fs");

	OGLCheckError();

	mUpdateShader.Use();
	mUpdateShader.Uniform1f("RevWidth", 1.0f/iWidth);
	mUpdateShader.Uniform1f("RevHeight", 1.0f/iHeight);

	OGLCheckError();

	mInitShader.Use();
	mInitShader.Uniform1f("RevWidth", 1.0f/iWidth);
	mInitShader.Uniform1f("RevHeight", 1.0f/iHeight);

	OGLCheckError();

	mRenderShader.Use();
	mRenderShader.Uniform1f("RevWidth", 1.0f/iWidth);
	mRenderShader.Uniform1f("RevHeight", 1.0f/iHeight);

	OGLCheckError();

	mCurLoop=0;
	
	GLboolean rBlend;
	GLboolean rDepth;
	glGetBooleanv(GL_BLEND, &rBlend);
	glGetBooleanv(GL_DEPTH_TEST, &rDepth);

	OGLCheckError();
	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);

	for (int i=0; i<2; i++)
	{
		mWaveTex[i].SetFilter(GL_TEXTURE_WRAP_S, GL_REPEAT);
		mWaveTex[i].SetFilter(GL_TEXTURE_WRAP_T, GL_REPEAT);
		mWaveTex[i].SetFilter(GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		mWaveTex[i].SetFilter(GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	OGLCheckError();

	mWaveTex[0].BindFrameBuffer();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	mInitShader.Use();

//	mRenderShader.Use();
//	mRenderShader.BindTexture("Wav", mWaveTex[1], 0);

	mWaveTex[0].Draw(1, -1, -1, 1);
	mWaveTex[0].UpdateFromFBO();
	mWaveTex[0].UnbindFrameBuffer();
	OGLCheckError();

	//mWaveTex[0].SaveToFile("test.png", 128, 0);
	mInitShader.Unuse();
	mCurLoop = 0;

	if (rBlend)
		glEnable(GL_BLEND);
	if (rDepth)
		glEnable(GL_DEPTH_TEST);

}

void OGLWater2D::RenderFunc(SSSLib::OGLScene::TStageName stage, int pass)
{
	if (stage == mUpdateStage && pass == mUpdatePass)
		UpdatePass();
	if (stage == mRenderStage && pass == mRenderPass)
		RenderPass();
}

void OGLWater2D::UpdatePass()
{
	if (mIsUpdated)
		mWaveTex[mCurLoop].UpdateFromMem();
	mCurLoop = !mCurLoop;
	int cur = mCurLoop;
	int pre = !mCurLoop;
	GLboolean rBlend;
	GLboolean rDepth;
	glGetBooleanv(GL_BLEND, &rBlend);
	glGetBooleanv(GL_DEPTH_TEST, &rDepth);


	mWaveTex[cur].BindFrameBuffer();
	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	//glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT);
	mUpdateShader.Use();
	mUpdateShader.BindTexture("PrevWav", mWaveTex[pre], 0);
	mWaveTex[pre].Draw(1, -1, -1, 1);
	mWaveTex[cur].UpdateFromFBO();
	
	if (rBlend)
		glEnable(GL_BLEND);

	if (rDepth)
		glEnable(GL_DEPTH_TEST);

	mWaveTex[cur].UnbindFrameBuffer();
	mUpdateShader.Unuse();
	glDisable(GL_TEXTURE_2D);
//	mWaveTex[cur].SaveToFile("test.png", 255, 0);
	mIsUpdated = false;

}

void OGLWater2D::RenderPass()
{
	mRenderShader.Use();
	mRenderShader.BindTexture("Wav", mWaveTex[mCurLoop], 0);
	if (mpButtonTex)
		mRenderShader.BindTexture("BottonImg", *mpButtonTex, 1);
	//mWaveTex[mCurLoop].Draw(1, -1, -1, 1);
	OGLRenderRectTex(1, -1, -1, 1);
	OGLTexture2D::Unbind(1);
	OGLTexture2D::Unbind(0);
	mRenderShader.Unuse();
}

void OGLWater2D::RainDrop(float iDelta, float iWeight, SSSLib::Vec2i iPst)
{
	int width = mWaveTex[mCurLoop].mWidth;
	int height = mWaveTex[mCurLoop].mHeight;
	int len = int( (iDelta-0.8f)/0.3f+1.0f+0.5f);
	vector<float> coe(len);
	for (int i=0; i<len; i++)
		coe[i] = 1.0f/iDelta/sqrt(2.0f*(float)Pi)*exp(-sqr(i)/2.0f/sqr(iDelta));
	for (int dx = -len+1; dx < len-1; dx++)
	for (int dy = -len+1; dy < len-1; dy++)
	{
		mWaveTex[mCurLoop].At(warp(iPst[0] + dx, width), warp(iPst[1] + dy, height), 2) -= coe[abs(dx)] * coe[abs(dy)]*iWeight;
	}
	mIsUpdated = true;
}