/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "SSSMesh.h"
#include "../SSSFunc.h"
#include <fstream>
#include <iostream>
#include <string>

#ifndef _NO_GL
#include <GL/freeglut.h>
#endif

#include <map>
#include <set>
#include <algorithm>

using namespace SSSLib;
using namespace std;

void SSSMesh::ComputeBB(Vec3d& oLow, Vec3d& oHigh) const
{
	oLow = mVtx[0];
	oHigh = mVtx[0];
	for (int i=0; i<(int)mVtx.size(); i++)
	for (int c=0; c<3; c++)
	{
		oLow[c] = Min(oLow[c], mVtx[i][c]);
		oHigh[c] = Max(oHigh[c], mVtx[i][c]);
	}
}

bool SSSMesh::LoadObj(const char* iFileName, bool ibWithColor)
{
	ifstream inFile(iFileName);
	if (!inFile.is_open())
		return false;
	string command;
	mVtx.clear();
	mFace.clear();
	mNormal.clear();
	bool isNormal = false;
	char buffer[100];
	while (!inFile.eof())
	{
		//int c = inFile.peek();
		//if (c=='\n' || c==EOF)
		//{
		//	inFile.getline(buffer, 100);
		//	continue;
		//}
		inFile>>command;

		if (command == "v")
		{
			Vec3d vtx;
			if (ibWithColor)
			{
				Vec3d clr;
				inFile>>vtx[0]>>vtx[1]>>vtx[2]>>clr[0]>>clr[1]>>clr[2];
				mColor.push_back(clr);
			}
			else
				inFile>>vtx[0]>>vtx[1]>>vtx[2];
			mVtx.push_back(vtx);
		}
		else if (command == "vn")
		{
			isNormal = true;
			Vec3d vtx;
			inFile>>vtx[0]>>vtx[1]>>vtx[2];
			mNormal.push_back(vtx);
		}
		else if (command == "f")
		{
			Vec3i face;
			Vec3i normalIdx;
			char c;
			if (!isNormal)
				inFile>>face[0]>>face[1]>>face[2];
			else
				inFile>>face[0]>>c>>c>>normalIdx[0]
					>>face[1]>>c>>c>>normalIdx[1]
					>>face[2]>>c>>c>>normalIdx[2];
			mFace.push_back(face-1);
		}
		else if (!inFile.eof())
			inFile.getline(buffer, 100);
		command = "";
	}
	return true;
}

void SSSMesh::WriteObj(const char* iFileName, const char* iMtrFileName, const char* iMtrName)
{
	ofstream outFile(iFileName);
	bool isNormal;
	if (mNormal.size()>0)
		isNormal = true;
	else
		isNormal = false;

	if (iMtrFileName)
	{
		outFile<<"mtllib "<<iMtrFileName<<endl;
			//<<"usemtl "<<iMtrName<<endl;
	}


	for (int v=0; v<(int)mVtx.size(); v++)
	{
		if (isNormal)
			outFile<<"vn"
					<<" "<<mNormal[v][0]
					<<" "<<mNormal[v][1]
					<<" "<<mNormal[v][2]
					<<endl;

		if (mTexCoord.size()>0)
			outFile<<"vt"
					<<" "<<mTexCoord[v][0]
					<<" "<<mTexCoord[v][1]
					<<endl;


		outFile<<"v"
				<<" "<<mVtx[v][0]
				<<" "<<mVtx[v][1]
				<<" "<<mVtx[v][2];

		if (mColor.size()>0)
			outFile<<" "<<mColor[v][0]
					<<" "<<mColor[v][1]
					<<" "<<mColor[v][2];

		outFile<<endl;
	}
	if (iMtrName)
		outFile<<"usemtl "<<iMtrName<<endl;
	for (int f=0; f<(int)mFace.size(); f++)
	{
		outFile<<"f";
		for (int c=0; c<3; c++)
		{
			int v = mFace[f][c]+1;
			outFile<<" "<<v;
			if (mTexCoord.size()>0)
				outFile<<"/"<<v;
			else if (isNormal)
				outFile<<"/";
			if (isNormal)
				outFile<<"/"<<v;
		}
		outFile<<endl;
		//if (isNormal)
		//	outFile<<"f"
		//			<<" "<<mFace[f][0]+1<<"//"<<mFace[f][0]+1
		//			<<" "<<mFace[f][1]+1<<"//"<<mFace[f][1]+1
		//			<<" "<<mFace[f][2]+1<<"//"<<mFace[f][2]+1
		//			<<endl;
		//else
		//	outFile<<"f"
		//			<<" "<<mFace[f][0]+1
		//			<<" "<<mFace[f][1]+1
		//			<<" "<<mFace[f][2]+1
		//			<<endl;
	}
	for (int f=0; f<(int)mQuad.size(); f++)
	{
		outFile<<"f";
		for (int c=0; c<4; c++)
		{
			int v = mQuad[f][c]+1;
			outFile<<" "<<v;
			if (mTexCoord.size()>0)
				outFile<<"/"<<v;
			else if (isNormal)
				outFile<<"/";
			if (isNormal)
				outFile<<"/"<<v;
		}
		outFile<<endl;
	}

}

void SSSMesh::WriteStl(const char* iFileName)
{
	ofstream outFile(iFileName);
	outFile<<"solid stl_for_print"<<endl;
	for (int f=0; f<(int)mFace.size(); f++)
	{
		Vec3i face = mFace[f];
		Vec3d vtx[] = {mVtx[face[0]], mVtx[face[1]], mVtx[face[2]]};
		Vec3d norm = cross3(vtx[1]-vtx[0], vtx[2]-vtx[1]);
		norm = norm.Norm();
		outFile<<"facet normal "<<norm[0]<<" "<<norm[1]<<" "<<norm[2]<<endl
			<<"\touter loop"<<endl;
		for (int v=0; v<3; v++)
			outFile<<"\t\tvertex "<<vtx[v][0]<<" "<<vtx[v][1]<<" "<<vtx[v][2]<<endl;
		outFile<<"\tendloop"<<endl
			<<"endfacet"<<endl;
	}

	for (int h=0; h<(int)mQuad.size(); h++)
	{
		Vec4i quad = mQuad[h];
		for (int i=0; i<2; i++)
		{
			Vec3i face;
			if (i==0)
				face = Vec3i(quad[0], quad[1], quad[3]);
			else
				face = Vec3i(quad[2], quad[3], quad[1]);
			Vec3d vtx[] = {mVtx[face[0]], mVtx[face[1]], mVtx[face[2]]};
			Vec3d norm = cross3(vtx[1]-vtx[0], vtx[2]-vtx[1]);
			norm = norm.Norm();
			outFile<<"facet normal "<<norm[0]<<" "<<norm[1]<<" "<<norm[2]<<endl
				<<"\touter loop"<<endl;
			for (int v=0; v<3; v++)
				outFile<<"\t\tvertex "<<vtx[v][0]<<" "<<vtx[v][1]<<" "<<vtx[v][2]<<endl;
			outFile<<"\tendloop"<<endl
				<<"endfacet"<<endl;
		}
	}

}

void SSSMesh::LoadDisteneMesh(const char* iFileName)
{
	ifstream inFile(iFileName);
	char buffer[100];
	while (!inFile.eof())
	{
		string name;
		inFile>>name;
		if (name[0]=='#')
			inFile.getline(buffer, 100);
		else if (name=="Vertices")
		{
			int vtxNum;
			inFile>>vtxNum;
			mVtx.resize(vtxNum);
			for (int i=0; i<vtxNum; i++)
			{
				int tmp;
				for (int c=0; c<3; c++)
					inFile>>mVtx[i][c];
				inFile>>tmp;
			}
		}
		else if (name=="Quadrilaterals")
		{
			int quadNum;
			inFile>>quadNum;
			mQuad.resize(quadNum);
			for (int i=0; i<quadNum; i++)
			{
				int tmp;
				for (int c=0; c<4; c++)
					inFile>>mQuad[i][c];
				mQuad[i] = mQuad[i]-1;
				inFile>>tmp;

			}
		}
	}
}

void SSSMesh::WriteDisteneMesh(const char* iFileName)
{
	ofstream outFile(iFileName);
	outFile<<"MeshVersionFormatted 1\n"
			<<"Dimension\n3\n"
			<<"Vertices\n"<<mVtx.size()<<endl;

	for (int i=0; i<(int)mVtx.size(); i++)
	{
		for (int c=0; c<3; c++)
			outFile<<mVtx[i][c]<<" ";
		outFile<<"0\n";
	}

	outFile<<"Triangles\n"
			<<mFace.size()<<endl;
	for (int i=0; i<(int)mFace.size(); i++)
	{
		for (int v=0; v<3; v++)
			outFile<<mFace[i][v]+1<<" ";
		outFile<<"0\n";
	}

	outFile<<"End\n";
}

void SSSMesh::Render()
{
    
#ifndef _NO_GL
	glBegin(GL_TRIANGLES);
		for (int f=0; f<(int)mFace.size(); f++)
		{
			for (int i=0; i<3; i++)
			{
				int v = mFace[f][i];
				if (mNormal.size()>0)
					glNormal3dv(mNormal[v].val);
				if (mTexCoord.size()>0)
					glTexCoord3dv(mTexCoord[v].val);
				if (mColor.size()>0)
					glColor3dv(mColor[v].val);
				glVertex3dv(mVtx[v].val);
			}
		}
	glEnd();
#endif
}

double SSSMesh::DisPt2Plane(Vec3d iPt, int iFaceID, Vec3d* oProjPt)
{
	Vec3d pt[3];
	for (int i=0; i<3; i++)
		pt[i] = mVtx[mFace[iFaceID][i]];

	Vec3d normal = cross3(pt[2]-pt[0], pt[1]-pt[0]);
	normal = normal.Norm();
	double dis = dot<double>(iPt, normal)-dot<double>(pt[0], normal);
	if (oProjPt)
		*oProjPt = iPt-dis*normal;
	return fabs(dis);
}

double SSSMesh::DisPt2Triangle(Vec3d iPt, int iFaceID, Vec3d* oProjPt)
{
	Vec3d projPt;
	double dis = DisPt2Plane(iPt, iFaceID, &projPt);

	Vec3d pt[3];
	for (int i=0; i<3; i++)
		pt[i] = mVtx[mFace[iFaceID][i]];

	double area[3];
	for (int i=0; i<3; i++)
		area[i] = cross2<double>(projPt-pt[i], pt[(i+1)%3]-pt[i]);

	if (fabs(area[0]+area[1]+area[2])==fabs(area[0])+fabs(area[1])+fabs(area[2]))
	{
		if (oProjPt)
			*oProjPt = projPt;
		return dis;
	}

	double minDis = -1;
	Vec3d returnPt;
	for (int i=0; i<3; i++)
	{
		double newDis = (pt[i]-projPt).Length();
		if (i==0 || newDis<minDis)
		{
			minDis = newDis;
			returnPt = pt[i];
		}
	}

	for (int i=0; i<3; i++)
	{
		Vec3d& pt0 = pt[i];
		Vec3d& pt1 = pt[(i+1)%3];
		Vec3d lineTan = (pt1-pt0).Norm();
		Vec3d linePt = dot<double>(lineTan, projPt)*lineTan+pt0;
		double newDis = (projPt-linePt).Length();
		if (newDis<minDis && dot<double>(projPt-pt0, pt1-pt0)>=0 && dot<double>(projPt-pt1, pt0-pt1)>=0)
		{
			minDis = newDis;
			returnPt = linePt;
		}
	}

	if (oProjPt)
		*oProjPt = returnPt;
	return minDis;
}

void SSSMesh::QueryPoint(Vec3d iPt, int& oFaceID, Vec3d& oWeight)
{
	Vec3d projPt;
	double dis;
	//for (int f=0; f<(int)mFace.size(); f++)
	//{
	//	Vec3d newPt;
	//	double newDis = DisPt2Triangle(iPt, f, &newPt);
	//	if (f==0 || newDis<dis)
	//	{
	//		dis = newDis;
	//		projPt = newPt;
	//		oFaceID = f;
	//	}
	//}
	oFaceID = QueryClosestFace(iPt);
	dis = DisPt2Triangle(iPt, oFaceID, &projPt);
	Vec3d pts[3] = {mVtx[mFace[oFaceID][0]], mVtx[mFace[oFaceID][1]], mVtx[mFace[oFaceID][2]]};

	for (int i=0; i<3; i++)
	{
		Vec3d pt0 = pts[(i+1)%3];
		Vec3d pt1 = pts[(i+2)%3];
		oWeight[i] = cross2<double>(projPt-pt0, projPt-pt1);
	}

	double normVal = cross2<double>(pts[0]-pts[1], pts[0]-pts[2]);
	if (normVal!=0)
		oWeight /= normVal;
}

Vec3d SSSMesh::FaceNorm(int iFaceID)
{
	Vec3d pts[3] = {mVtx[mFace[iFaceID][0]], mVtx[mFace[iFaceID][1]], mVtx[mFace[iFaceID][2]]};
	return cross3(pts[1]-pts[0], pts[2]-pts[0]).Norm();
}

std::vector<SSSLib::Vec3d> SSSMesh::SampleFaces(int iFace)
{
	vector<Vec3d> retVec;
	Vec3d pts[3] = {mVtx[mFace[iFace][0]], mVtx[mFace[iFace][1]], mVtx[mFace[iFace][2]]};
	for (int i=0; i<3; i++)
		retVec.push_back(pts[i]);
	Vec3d norm = FaceNorm(iFace);
	Vec3d normPlace = norm*dot<double>(pts[0], norm);
	Vec3d longEdge;
	Vec3d otherEdges[2];
	Vec3d longPt;
	{
		int longID;
		double longLength;
		for (int i=0; i<3; i++)
		{
			double edgeLength = (pts[(i+1)%3]-pts[i]).Length();
			if (i==0 || edgeLength>longLength)
			{
				longID = i;
				longLength = edgeLength;
			}
		}
		longEdge = pts[(longID+1)%3]-pts[longID];
		for (int i=0; i<2; i++)
			otherEdges[i] = pts[(longID+i+2)%3]-pts[(longID+i+1)%3];
		longPt = pts[longID];
	}
	Vec3d t0 = longEdge.Norm();
	Vec3d t1 = cross3(norm, t0).Norm();
	double low0;
	double low1;
	double high0;
	double high1;
	for (int i=0; i<3; i++)
	{
		double val0 = dot<double>(pts[i], t0);
		double val1 = dot<double>(pts[i], t1);
		
		if (i==0 || val0<low0)
			low0 = val0;
		if (i==0 || val0>high0)
			high0 = val0;

		if (i==0 || val1<low1)
			low1 = val1;
		if (i==0 || val1>high1)
			high1 = val1;
	}

	for (double u=low0; u<=high0; u+=mFacePtSampleDis)
	for (double v=low1; v<=high1; v+=mFacePtSampleDis)
	{

		Vec3d newPt = normPlace+u*t0+v*t1;
		//check if the point is on the triangle
		if (dot<double>(cross3(otherEdges[0], newPt-longPt), cross3(newPt-longPt, otherEdges[1]))>0)
			retVec.push_back(newPt);
	}
	return retVec;
}

void SSSMesh::CreateKdTree(double iSampleDistance)
{
	mFacePtSampleDis = iSampleDistance;
	//mTreeDataList.resize(mFace.size());
	vector<Vec3d> samples;
	vector<int> data;
	for (int f=0; f<(int)mFace.size(); f++)
	{
		if (f%100==0)
			cout<<"Creating KdTree Face # "<<f<<endl;
		vector<Vec3d> faceSamples = SampleFaces(f);
		for (int i=0; i<(int)faceSamples.size(); i++)
		{
			samples.push_back(faceSamples[i]);
			data.push_back(f);
		}
	}
	mKDTree.BuildTree(samples.size(), &samples[0], &data[0]);
	
	/*mpKDTree.reset(kd_create(3), kd_free);
	for (int f=0; f<(int)mFace.size(); f++)
	{
		if (f%100==0)
			cout<<"Creating KdTree Face # "<<f<<endl;
		mTreeDataList[f] = f;
		vector<Vec3d> faceSamples = SampleFaces(f);
		for (int i=0; i<(int)faceSamples.size(); i++)
			kd_insert3f(mpKDTree.get(),
				faceSamples[i][0],
				faceSamples[i][1],
				faceSamples[i][2],
				(void*)((int*)&mTreeDataList[f]));
	}*/
}

int SSSMesh::QueryClosestFace(SSSLib::Vec3d iPt)
{
	int retVal;
	mKDTree.QueryClosestPt(iPt, &retVal, NULL);
	return retVal;
	/*
	std::shared_ptr<kdres> kdpoint(
								kd_nearest3f(mpKDTree.get(), iPt[0], iPt[1], iPt[2]),
								kd_res_free);
	return *((int*)kd_res_item_data(kdpoint.get()));*/
}

bool SSSMesh::UnifyOrientation()
{
	bool isError = false;
	int flipCount = 0;
	set<Vec2i> stillEdge;
	vector<int> faceFlipping(mFace.size(), 0);
	Vec3i face = mFace[0];
	for (int i=0; i<3; i++)
	{
		stillEdge.insert(Vec2i(face[i], face[(i+1)%3]));
	}
	faceFlipping[0] = 1;
	set<int> restFaceSet;
	for (int f=1; f<(int)mFace.size(); f++)
		restFaceSet.insert(f);
	int lastFaceSize = restFaceSet.size()+1;
	while (restFaceSet.size()!=lastFaceSize)
	{
		lastFaceSize = restFaceSet.size();
		for (set<int>::iterator ite = restFaceSet.begin();
			ite != restFaceSet.end();
			)
		{
			set<int>::iterator nextIte = ite;
			nextIte++;
			face = mFace[*ite];
			int isInsert = 0; // 0: no insertion; 1: insert edge; -1: insert reversed edge; -2: error
			for (int i=0; i<3 && isInsert==0; i++)
			{
				Vec2i edge(face[i], face[(i+1)%3]);
				Vec2i reEdge(edge[1], edge[0]);
				if (stillEdge.count(edge) && !stillEdge.count(reEdge))
				{
					faceFlipping[*ite] = -1;
					flipCount++;
					isInsert = -1;
				}
				else if (!stillEdge.count(edge) && stillEdge.count(reEdge))
				{
					faceFlipping[*ite] = 1;
					isInsert = 1;
				}
				else if (stillEdge.count(edge) && stillEdge.count(reEdge))
					isInsert = -2;
			}

			if (isInsert==-1 || isInsert==1 || isInsert==-2)
			{
				for (int i=0; i<3; i++)
				{
					Vec2i edge;
					if (isInsert!=-1)
						edge = Vec2i(face[i], face[(i+1)%3]);
					else
						edge = Vec2i(face[(i+1)%3], face[i]);
					stillEdge.insert(edge);
				}
				restFaceSet.erase(*ite);
				if (isInsert==-2)
					isError = true;
			}

			ite = nextIte;
		}
	}

	int flip = 1;
	if (flipCount>(int)mFace.size()/2)
		flip=-1;
	for (int f=0; f<(int)mFace.size(); f++)
	{
		if (faceFlipping[f]*flip<0)
		{
			Vec3i face = mFace[f];
			mFace[f][0] = face[2];
			mFace[f][2] = face[0];
		}
	}
	if (restFaceSet.size()>0)
	{
		cout<<"Non-oriented faces "<<restFaceSet.size()<<endl;
		isError = true;
	}
	return isError;
}

void SSSMesh::RemoveDuplicatedFaces()
{
	set<Vec3i> faceSet;
	vector<int> markErase;
	for (int i=0; i<(int)mFace.size(); i++)
	{
		Vec3i face = mFace[i];
		if (face[1]<face[2] && face[1]<face[0])
			face = Vec3i(face[1], face[2], face[0]);
		else if (face[2]<face[1] && face[2]<face[0])
			face = Vec3i(face[2], face[0], face[1]);
		if (faceSet.count(face))
			markErase.push_back(i);
		else
			faceSet.insert(face);
	}
	for (int i=(int)markErase.size()-1; i>=0; i--)
		mFace.erase(mFace.begin()+markErase[i]);
}

void SSSMesh::ComputeVtxAreaWeight()
{
	mVtxArea.clear();
	mVtxArea.resize(mVtx.size(), 0);
	for (int f=0; f<(int)mFace.size(); f++)
	{
		Vec3i face = mFace[f];
		Vec3d pts[3] = {mVtx[face[0]], mVtx[face[1]], mVtx[face[2]]};

		double area = cross3(pts[1]-pts[0], pts[2]-pts[0]).Length()/2;

		double angleSum = 0;

		for (int i=0; i<3; i++)
		{
			Vec3d edges[2] = {pts[(i+1)%3]-pts[i], pts[(i+2)%3]-pts[i]};
			double angle = acos(dot<double>(edges[0], edges[1])
				/edges[0].Length()/edges[1].Length());
			angleSum += angle;

			mVtxArea[face[i]] += area*angle;

		}

		assert(fabs(angleSum-Pi)<1e-5);
	}

	for (int q=0; q<(int)mQuad.size(); q++)
	{
		Vec4i quad = mQuad[q];
		double faceArea = 0;
		for (int i=0; i<2; i++)
		{
			Vec3d pts[3];
			Vec3i face;
			for (int v=0; v<3; v++)
			{
				if (i==0)
					face[v] = quad[v];
				else
					face[v] = quad[3-v];
				pts[v] = mVtx[face[v]];
			}

			faceArea += cross3(pts[1]-pts[0], pts[2]-pts[0]).Length()/2;

			//for (int i=0; i<3; i++)
			//	mVtxArea[face[i]] += area/3;
		}

		double w[4];
		double wsum = 0;

		for (int v=0; v<4; v++)
		{
			Vec3d edge0 = mVtx[quad[(v+1)%4]]-mVtx[quad[v]];
			Vec3d edge1 = mVtx[quad[(v+3)%4]]-mVtx[quad[v]];

			w[v] = acos(dot<double>(edge0, edge1)/edge0.Length()/edge1.Length());

			wsum += w[v];
		}

		for (int v=0; v<4; v++)
			mVtxArea[quad[v]] += w[v]/wsum;

		//assert(fabs(wsum-Pi*2)<1e-5);
	}
}

double SSSMesh::ComputeSignedVol() const
{
	double retVal = 0;

	for (int t=0; t<(int)mFace.size(); t++)
	{
		const Vec3i& f = mFace[t];
		retVal += TripleProduct<double>(mVtx[f[0]], mVtx[f[1]], mVtx[f[2]]);
	}

	for (int q=0; q<(int)mQuad.size(); q++)
	{
		const Vec4i& quad = mQuad[q];
		retVal += TripleProduct<double>(mVtx[quad[0]], mVtx[quad[1]], mVtx[quad[2]]);
		retVal += TripleProduct<double>(mVtx[quad[0]], mVtx[quad[2]], mVtx[quad[3]]);
	}

	return retVal/6;
}
