/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <vector>
#include "../SSSSmallVec.h"
#include "../SSSSmallMat.h"
#include "kdtree.h"
#include <memory>
#include "KDTree_wrap.h"
#include <map>
#include <set>

namespace SSSLib
{

typedef std::vector<Vec3d> SSSVtxSet;
typedef std::vector<Vec3i> SSSFaceSet;
typedef std::vector<Vec4i> SSSQuadFaceSet;

class SSSMesh
{
public:
	SSSMesh(void){}
	~SSSMesh(void){}

	SSSVtxSet mVtx;
	SSSFaceSet mFace;
	SSSQuadFaceSet mQuad;
	SSSVtxSet mNormal;
	SSSVtxSet mTexCoord;
	SSSVtxSet mColor;
	std::vector<SSSLib::Vec2i> mEdges;
	std::vector<std::set<int> > mVtxConnect;
	std::vector<bool> mbPlanarVtx;
	void ComputeBB(Vec3d& oLow, Vec3d& oHigh) const;

	std::vector<double> mVtxArea;
	void ComputeVtxAreaWeight();

	bool LoadObj(const char* iFileName, bool ibWithColor = false);
	void WriteObj(const char* iFileName, const char* iMtrFileName = NULL, const char* iMtrName = NULL);
	void WriteStl(const char* iFileName);
	void LoadDisteneMesh(const char* iFileName);
	void WriteDisteneMesh(const char* iFileName);
	void Render();

	double DisPt2Plane(Vec3d iPt, int iFaceID, Vec3d* oProjPt = NULL);
	double DisPt2Triangle(Vec3d iPt, int iFaceID, Vec3d* oProjPt = NULL);
	void QueryPoint(Vec3d iPt, int& oFaceID, Vec3d& oWeight);
	Vec3d FaceNorm(int iFaceID);

	KDTree3d<int> mKDTree;
	//std::shared_ptr<kdtree> mpKDTree;
	double mFacePtSampleDis;
	std::vector<SSSLib::Vec3d> SampleFaces(int iFace);
	void CreateKdTree(double iSampleDistance);
	int QueryClosestFace(SSSLib::Vec3d iPt);
	//std::vector<int> mTreeDataList;

	// only work with connected mesh
	bool UnifyOrientation();
	void RemoveDuplicatedFaces();

	void ComputeEdges(double iNormTolerance);
	void VtxPlanarTest(double iNormTolerance);
	void VtxPlanarRelax();
	void VtxPlanarRelaxProc(double iNormTolerance = 5e-2)
	{
		BuildVtxConnect();
		VtxPlanarTest(iNormTolerance);
		ComputeEdges(iNormTolerance);
		EdgePlanarRelax();
		VtxPlanarRelax();
	}
	void EdgePlanarRelax();
	void BuildVtxConnect();

	void ComputeQuadEdges(std::vector<Vec2i>& oEdges);

	void Scale(Vec3d iScale)
	{
		for (int i=0; i<(int)mVtx.size(); i++)
		for (int c=0; c<3; c++)
		{
			mVtx[i][c] *= iScale[c];
		}
	}

	void Translate(Vec3d iTrans)
	{
		for (int i=0; i<(int)mVtx.size(); i++)
		{
			mVtx[i] += iTrans;
		}
	}


	void Rotate(Vec3d iAxis, double iAngle)
	{
		Mat3x3d rotateMat;
		double sinTheta = sin(iAngle);
		double cosTheta = cos(iAngle);
	
		for (int i=0; i<3; i++)
		{
			rotateMat(i, i) = cosTheta+(iAxis[i]*iAxis[i])*(1-cosTheta);
			for (int j=1; j<3; j++)
				rotateMat(i, (i+j)%3) = iAxis[i]*iAxis[(i+j)%3]*(1-cosTheta)
										+iAxis[(i+3-j)%3]*sinTheta*(j*2-3);
		}

		for (int i=0; i<(int)mVtx.size(); i++)
		{
			mVtx[i] = rotateMat*mVtx[i];
		}
	}

	double ComputeSignedVol() const;
};

}

