/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "SSSMeshCarve.h"
#include <map>
#include <iostream>

using namespace SSSLib;

// std has some conflict with carve library
//using namespace std;
using namespace SSSLib;

// due to a bug in vs2010 x64 optimization
void Shape3D::Carve2S3(const std::vector<CFace>& iCFace,
					Shape3DMesh& oS3Mesh)
{
	using namespace carve::geom;

	std::map<const CVtx3*, int> vtxMap;
	oS3Mesh = Shape3DMesh();
	for (int i=0; i<(int)iCFace.size(); i++)
	{
		std::vector<int> s3faces;
		const CFace& face = iCFace[i];
		std::vector<carve::triangulate::tri_idx> result;
		std::vector<const CVtx3*> vtx;
		for (int c=0; c<(int)face.nVertices(); c++)
			vtx.push_back(face.vertex(c));
		carve::triangulate::triangulate<carve::poly::p2_adapt_project<3>, const CVtx3*>
			(carve::poly::p2_adapt_project<3>(face.project),
			vtx,
			result);
		carve::triangulate::improve<carve::poly::p2_adapt_project<3>, const CVtx3*>
			(carve::poly::p2_adapt_project<3>(face.project),
			vtx,
			result);
		for (int c=0; c<(int)face.nVertices(); c++)
		{
			if (vtxMap.count(face.vertex(c))==0)
			{
				vtxMap[face.vertex(c)] = (int)oS3Mesh.mVtx.size();
				oS3Mesh.mVtx.push_back(face.vertex(c)->v);
			}
			s3faces.push_back(vtxMap[face.vertex(c)]);
		}

		// simple triangulation
		//int left = 0;
		//int right = face.nVertices()-1;
		//std::vector<carve::geom2d::P2> vtx2d = face.projectedVertices();
		//while (right>left+1)
		//{
		//	Vec2f pl = vtx2d[left];
		//	Vec2f pr = vtx2d[right];
		//	Vec2f plp = vtx2d[left+1];
		//	if (cross2<_T_FLOAT>(plp-pl, pr-pl)<0)
		//	{
		//		oS3Mesh.mFace.push_back(Vec3i(s3faces[left], s3faces[left+1], s3faces[right]));
		//		left++;
		//	}
		//	else
		//	{
		//		oS3Mesh.mFace.push_back(Vec3i(s3faces[left], s3faces[right-1], s3faces[right]));
		//		right--;
		//	}
		//}

		
		int numFaces = (int)s3faces.size();
		//for (int f=1; f<numFaces-1; f++)
		//	oS3Mesh.mFace.push_back(Vec3i(s3faces[0], s3faces[f], s3faces[f+1]));
		for (int f=0; f<result.size(); f++)
			oS3Mesh.mFace.push_back(Vec3i(s3faces[result[f].a],
										s3faces[result[f].b],
										s3faces[result[f].c]));
		
	}
}

void Shape3D::S32Carve(const Shape3DMesh& iS3Mesh,
		std::vector<CVtx3>& oCVtx,
		std::vector<CFace>& oCFace)
{
	using namespace std;
	oCVtx.resize(iS3Mesh.mVtx.size());
	oCFace.resize(iS3Mesh.mFace.size());
	for (int v=0; v<(int)iS3Mesh.mVtx.size(); v++)
		for (int c=0; c<3; c++)
			oCVtx[v].v[c] = iS3Mesh.mVtx[v][c];

	for (int f=0; f<(int)iS3Mesh.mFace.size(); f++)
	{
		vector<const CVtx3*> face;
		for (int c=0; c<3; c++)
		{
			int vtxID = iS3Mesh.mFace[f][c];
			face.push_back(&oCVtx[vtxID]);
		}
		oCFace[f] = CFace(face);
	}
}

void Shape3D::S3Intersect(const Shape3DMesh& iS3Mesh0,
							const Shape3DMesh& iS3Mesh1,
							Shape3DMesh& oS3Mesh)
{
	using namespace std;
	carve::poly::Polyhedron* opoly = S3Intersect(iS3Mesh0, iS3Mesh1);
	Carve2S3(opoly->faces, oS3Mesh);
	cout<<"Seg Result "<<oS3Mesh.mVtx.size()<<endl;
	delete opoly;
}

carve::poly::Polyhedron* Shape3D::S3Intersect(const Shape3DMesh& iS3Mesh0,
							const Shape3DMesh& iS3Mesh1)
{
	using namespace std;
	vector<CVtx3> vtx0;
	vector<CVtx3> vtx1;
	vector<CFace> face0;
	vector<CFace> face1;

	S32Carve(iS3Mesh0, vtx0, face0);
	S32Carve(iS3Mesh1, vtx1, face1);

	carve::poly::Polyhedron p0(face0);
	carve::poly::Polyhedron p1(face1);

	carve::csg::V2Set sharedEdges;

	return carve::csg::CSG().compute(&p0, &p1, carve::csg::CSG::INTERSECTION, &sharedEdges, carve::csg::CSG::CLASSIFY_EDGE);
}

_T_FLOAT Shape3D::CarveVolume(const std::vector<CFace> &iFaces)
{
    _T_FLOAT vol = 0.0;
	const CVtx3* origin = iFaces[0].vertex(0);

    for (size_t f = 0; f < iFaces.size(); ++f) {
        const CFace *face = &iFaces[f];
		const CVtx3 *vtx = face->vertex(0);
		
		for (int i=1; i<(int)face->nVertices()-1; i++)
			vol += carve::geom3d::tetrahedronVolume(vtx->v, face->vertex(i)->v, face->vertex(i+1)->v, origin->v);

		//vector<const CEdge*>::const_iterator e1 = face->ebegin();
		//
		//for (vector<const CEdge*>::const_iterator e2 = e1+1; e2+1!=face->eend(); e2++)
		//{
		//	vol += carve::geom3d::tetrahedronVolume((*e1)->v1->v, (*e2)->v1->v, (*e2)->v2->v, origin->v);
  //      }
    }
    return fabs(vol);
}
//#pragma optimize("",on)
