/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "SSSMesh.h"
#include <fstream>
#include <iostream>
#include <string>
#include <map>
#include <set>
#include <algorithm>
#include <Eigen/Sparse>
#include <Eigen/SparseLU>
#include "SSSMeshUtil.h"

using namespace SSSLib;
using namespace std;

void SSSLib::SSSMesh::BuildVtxConnect()
{
	mVtxConnect.clear();
	mVtxConnect.resize(mVtx.size());

	SSSLib::BuildVtxConnect<Vec3i, 3>(mVtx.size(),
								mFace,
								mVtxConnect);
	
	vector<Vec2i> quadEdges;
	for (int i=0; i<(int)mQuad.size(); i++)
	for (int v=0; v<4; v++)
		quadEdges.push_back(Vec2i(mQuad[i][v], mQuad[i][(v+1)%4]));
	
	SSSLib::BuildVtxConnect<Vec2i, 2>(mVtx.size(), quadEdges, mVtxConnect);

	//mVtxConnect.clear();
	//mVtxConnect.resize(mVtx.size());

	//for (int f=0; f<(int)mFace.size(); f++)
	//{
	//	Vec3i face = mFace[f];
	//	for (int v1=0; v1<3; v1++)
	//	for (int v2=0; v2<3; v2++)
	//	if (v1!=v2)
	//	{
	//		mVtxConnect[face[v1]].insert(face[v2]);
	//		mVtxConnect[face[v2]].insert(face[v1]);
	//	}
	//}
}

void SSSMesh::ComputeEdges(double iNormTolerance)
{
	mEdges.clear();
	map<Vec2i, Vec3d> edgeNorm;
	set<Vec2i> edgeSet;
	for (int f=0; f<(int)mFace.size(); f++)
	{
		Vec3d faceNorm = FaceNorm(f);
		Vec3i face = mFace[f];
		for (int e=0; e<3; e++)
		{
			Vec2i edge(face[e], face[(e+1)%3]);
			sort(edge.val, edge.val+2);
			if (!edgeNorm.count(edge))
				edgeNorm[edge] = faceNorm;
			else if (!edgeSet.count(edge))
			{
				if ((edgeNorm[edge]-faceNorm).Length()>iNormTolerance)
				{
					mEdges.push_back(edge);
					edgeSet.insert(edge);
				}
			}
		}
	}
}

void SSSMesh::VtxPlanarTest(double iNormTolerance)
{
	mbPlanarVtx.clear();
	mbPlanarVtx.resize(mVtx.size(), true);
	SSSVtxSet vtxNorm(mVtx.size(), Vec3d(0, 0, 0));
	SSSVtxSet faceNorm(mFace.size(), Vec3d(0, 0, 0));
	for (int f=0; f<(int)mFace.size(); f++)
	{
		faceNorm[f] = FaceNorm(f);
		Vec3i face = mFace[f];
		for (int v=0; v<3; v++)
			vtxNorm[face[v]] += faceNorm[f];
	}
	for (int q=0; q<(int)mQuad.size(); q++)
	{
		Vec4i quad = mQuad[q];
		for (int v=0; v<4; v++)
		{
			Vec3d quadn = cross3(mVtx[quad[(v+1)%4]]-mVtx[quad[v]],
								mVtx[quad[(v+3)%4]]-mVtx[quad[v]]).Norm();
			vtxNorm[quad[v]] += quadn;
		}
	}

	for (int v=0; v<(int)vtxNorm.size(); v++)
	if (mVtxConnect[v].size()>0)
	{
		vtxNorm[v] = vtxNorm[v].Norm();
	}
	else
		mbPlanarVtx[v] = false;

	for (int f=0; f<(int)mFace.size(); f++)
	{
		Vec3i face = mFace[f];
		for (int v=0; v<3; v++)
			if (mbPlanarVtx[face[v]] && dot<double>(faceNorm[f], vtxNorm[face[v]])<1-iNormTolerance)
				mbPlanarVtx[face[v]] = false;
	}

	for (int q=0; q<(int)mQuad.size(); q++)
	{
		Vec4i quad = mQuad[q];
		for (int v=0; v<4; v++)
		if (mbPlanarVtx[quad[v]])
		{
			Vec3d quadn = cross3(mVtx[quad[(v+1)%4]]-mVtx[quad[v]],
								mVtx[quad[(v+3)%4]]-mVtx[quad[v]]).Norm();
			if (dot<double>(vtxNorm[quad[v]], quadn)<1-iNormTolerance)
				mbPlanarVtx[quad[v]] = false;
		}
	}
}

void SSSMesh::EdgePlanarRelax()
{
	vector<set<int> > edgeConnect(mVtx.size());
	for (int i=0; i<(int)mEdges.size(); i++)
	{
		Vec2i e = mEdges[i];
		edgeConnect[e[0]].insert(e[1]);
		edgeConnect[e[1]].insert(e[0]);
	}
	vector<bool> bNonFixed(mVtx.size());
	for (int i=0; i<(int)bNonFixed.size(); i++)
	if (edgeConnect[i].size()==2)
		bNonFixed[i] = true;
	else
		bNonFixed[i] = false;
	ZeroLapRelax(edgeConnect, bNonFixed, mVtx);
}

void SSSMesh::VtxPlanarRelax()
{
	ZeroLapRelax(mVtxConnect, mbPlanarVtx, mVtx);
	//typedef Eigen::Triplet<double> T;
	//std::vector<T> tripleList;
	//Eigen::MatrixXd b = Eigen::MatrixXd::Zero(mVtx.size()*3, 1);

	//for (int v=0; v<(int)mVtx.size(); v++)
	//if (mbPlanarVtx[v])
	//{
	//	for (int c=0; c<3; c++)
	//		tripleList.push_back(T(v*3+c, v*3+c, 1.0));
	//	int n = mVtxConnect[v].size();
	//	assert(n>0);
	//	double scale = -1.0/n;
	//	for (set<int>::iterator ite = mVtxConnect[v].begin();
	//		 ite != mVtxConnect[v].end();
	//		 ite++)
	//	for (int c=0; c<3; c++)
	//		 tripleList.push_back(T(v*3+c, *ite*3+c, scale));
	//}
	//else
	//{
	//	for (int c=0; c<3; c++)
	//	{
	//		tripleList.push_back(T(v*3+c, v*3+c, 1.0));
	//		b(v*3+c, 0) = (double)mVtx[v][c];
	//	}
	//}

	//Eigen::SparseMatrix<double> A(mVtx.size()*3, mVtx.size()*3);
	//A.setFromTriplets(tripleList.begin(), tripleList.end());

	//Eigen::SparseLU<Eigen::SparseMatrix<double>, Eigen::COLAMDOrdering<int> > solver;
	//solver.analyzePattern(A);

	//if (solver.lastErrorMessage().length()>0)
	//	cout<<solver.lastErrorMessage();
	//solver.factorize(A);
	//if (solver.lastErrorMessage().length()>0)
	//	cout<<solver.lastErrorMessage();
	//Eigen::MatrixXd x = solver.solve(b);

	//for (int v=0; v<(int)mVtx.size(); v++)
	//for (int c=0; c<3; c++)
	//	mVtx[v][c] = (double)x(v*3+c, 0);
}

void SSSMesh::ComputeQuadEdges(std::vector<Vec2i>& oEdges)
{
	for (int i=0; i<(int)mQuad.size(); i++)
	{
		Vec4i quad = mQuad[i];
		for (int j=0; j<4; j++)
			oEdges.push_back(Vec2i(quad[j], quad[(j+1)%4]));
	}
}
