/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "SSSMeshUtil.h"
#include <Eigen/Sparse>
#include <Eigen/SparseLU>
#include "SSSVolMesh.h"

using namespace std;
using namespace SSSLib;

void SSSLib::ZeroLapRelax(const std::vector<std::set<int> >& iConnect,
							const std::vector<bool>& ibNonFixed,
							std::vector<Vec3d>& ioVtx)
{
	typedef Eigen::Triplet<double> T;
	std::vector<T> tripleList;
	Eigen::MatrixXd b = Eigen::MatrixXd::Zero(ioVtx.size()*3, 1);

	for (int v=0; v<(int)ioVtx.size(); v++)
	if (ibNonFixed[v])
	{
		for (int c=0; c<3; c++)
			tripleList.push_back(T(v*3+c, v*3+c, 1.0));
		int n = iConnect[v].size();
		assert(n>0);
		double scale = -1.0/n;
		for (set<int>::iterator ite = iConnect[v].begin();
			 ite != iConnect[v].end();
			 ite++)
		for (int c=0; c<3; c++)
			 tripleList.push_back(T(v*3+c, *ite*3+c, scale));
	}
	else
	{
		for (int c=0; c<3; c++)
		{
			tripleList.push_back(T(v*3+c, v*3+c, 1.0));
			b(v*3+c, 0) = (double)ioVtx[v][c];
		}
	}

	Eigen::SparseMatrix<double> A(ioVtx.size()*3, ioVtx.size()*3);
	A.setFromTriplets(tripleList.begin(), tripleList.end());

	Eigen::SparseLU<Eigen::SparseMatrix<double>, Eigen::COLAMDOrdering<int> > solver;
	solver.analyzePattern(A);

	if (solver.lastErrorMessage().length()>0)
		cout<<solver.lastErrorMessage();
	solver.factorize(A);
	if (solver.lastErrorMessage().length()>0)
		cout<<solver.lastErrorMessage();
	Eigen::MatrixXd x = solver.solve(b);

	for (int v=0; v<(int)ioVtx.size(); v++)
	for (int c=0; c<3; c++)
		ioVtx[v][c] = x(v*3+c, 0);
}

int SSSLib::CountTet(SSSVolMesh& imesh, set<int>* oNegVtx, double iVolBound)
{
	int pos = 0;
	int neg = 0;
	int zero = 0;
	for (int i=0; i<(int)imesh.mTets.size(); i++)
	{
		double vol = imesh.TetSignedVol(i);
		if (vol<iVolBound)
		{
			cout<<i<<"\t"<<vol<<endl;
			neg++;
			//if (gWriteMask.size()>0)
			//{
			//	for (int j=0; j<4; j++)
			//		gWriteMask[imesh.mTets[i][j]] = false;
			//}
			if (oNegVtx)
				for (int j=0; j<4; j++)
					oNegVtx->insert(imesh.mTets[i][j]);
		}
		else
			pos++;
	}

	cout<<"Positive Tet: "<<pos<<endl
		<<"Negative Tet: "<<neg<<endl
		<<"Zero Tet: "<<zero<<endl;
	return neg;
}

void SSSLib::Hex2TriangleFaces(SmallVec<int, 8> iHex, Vec3i* oFaces)
{
	int validPoints[] = {0, 3, 5, 6};

	for (int i=0; i<4; i++)
	{
		int p0 = validPoints[i];
		Vec3i f;
		f[0] = iHex[p0];
		for (int j=0; j<3; j++)
		{
			for (int k=0; k<2; k++)
			{
				int newp = p0^(1<<((j+k)%3));
				f[1+k] = iHex[newp];
			}
			oFaces[i*3+j] = f;
		}
	}
};
