/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <vector>
#include "../SSSSmallVec.h"
#include <set>

namespace SSSLib
{
	class SSSVolMesh;

	template <class _TFace, int _FaceSize>
	void BuildVtxConnect(int iNumVtx,
								const std::vector<_TFace>& iFaces,
								std::vector<std::set<int> >& oVtxConnect)
	{

		for (int f=0; f<(int)iFaces.size(); f++)
		{
			_TFace face = iFaces[f];
			for (int v1=0; v1<_FaceSize; v1++)
			for (int v2=0; v2<_FaceSize; v2++)
			if (v1!=v2)
			{
				oVtxConnect[face[v1]].insert(face[v2]);
				oVtxConnect[face[v2]].insert(face[v1]);
			}
		}
	}

	extern void ZeroLapRelax(const std::vector<std::set<int> >& iConnect,
							const std::vector<bool>& ibNonFixed,
							std::vector<Vec3d>& ioVtx);

	extern int CountTet(SSSVolMesh& imesh, std::set<int>* oNegVtx, double iVolBound = 0.0f);

	inline Vec3d TriangleCross(Vec3d v0, Vec3d v1, Vec3d v2)
	{return cross3(v1-0, v2-v0);}

	inline Vec3d TriangleNorm(Vec3d v0, Vec3d v1, Vec3d v2)
	{return TriangleCross(v0, v1, v2).Norm();}

	extern void Hex2TriangleFaces(SmallVec<int, 8> iHex, Vec3i* oFaces);
}
