/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "SSSVolMesh.h"
#include <iomanip>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <algorithm>
#include <string.h>
//#include <ctype.h>

using namespace std;
using namespace SSSLib;

SSSVolMesh::SSSVolMesh(void)
{
}

SSSVolMesh::~SSSVolMesh(void)
{
}

void SSSVolMesh::LoadVolMesh(const char* iFileName)
{
	char nameBuffer[10000];
	sprintf(nameBuffer, "%s.node", iFileName);
	LoadNodes(nameBuffer);
	sprintf(nameBuffer, "%s.ele", iFileName);
	LoadTets(nameBuffer);
	sprintf(nameBuffer, "%s.face", iFileName);
	LoadFaces(nameBuffer);
}

void SSSVolMesh::LoadNodes(const char* iFileName)
{
	mbLoadZeroBase = false;
	LoadArray<Vec3d, 3>(iFileName, mVtx);
	//ifstream inFile(iFileName);
	//stringstream lineStream;
	//string line;
	//int count;
	//std::getline(inFile, line);
	//lineStream<<line;
	//lineStream>>count;

	//mVtx.resize(count);

	//for (int i=0; i<count; i++)
	//{
	//	lineStream.clear();
	//	std::getline(inFile, line);
	//	lineStream<<line;
	//	int id;
	//	Vec3d pst;
	//	lineStream>>id>>pst[0]>>pst[1]>>pst[2];
	//	mVtx[id-1] = pst;
	//}
}

void SSSVolMesh::LoadFaces(const char* iFileName)
{
	mSurfaceMesh.mVtx = mVtx;
	mSurfaceMesh.mNormal.clear();
	mSurfaceMesh.mTexCoord.clear();
	mSurfaceMesh.mFace.clear();

	LoadArray<Vec3i, 3>(iFileName, mSurfaceMesh.mFace);

	mbIntVtx.clear();
	mbIntVtx.resize(mVtx.size(), true);

	for (int i=0; i<(int)mSurfaceMesh.mFace.size(); i++)
	for (int v=0; v<3; v++)
	{
		if (!mbLoadZeroBase)
			mSurfaceMesh.mFace[i][v]--;
		mbIntVtx[mSurfaceMesh.mFace[i][v]] = false;
	}

	//ifstream inFile(iFileName);
	//stringstream lineStream;
	//string line;
	//int count;
	//std::getline(inFile, line);
	//lineStream<<line;
	//lineStream>>count;

	//mSurfaceMesh.mFace.resize(count);
	//for (int i=0; i<count; i++)
	//{
	//	lineStream.clear();
	//	std::getline(inFile, line);
	//	lineStream<<line;
	//	int id;
	//	Vec3i face;
	//	lineStream>>id>>face[0]>>face[1]>>face[2];
	//	mSurfaceMesh.mFace[id-1] = face;
	//}
}

void SSSVolMesh::LoadTets(const char* iFileName)
{
	mTets.clear();

	LoadArray<Vec4i, 4>(iFileName, mTets);
	if (!mbLoadZeroBase)
	for (int i=0; i<(int)mTets.size(); i++)
		mTets[i] = mTets[i]-1;
}

void SSSVolMesh::WriteNodes(const char* oFileName, std::vector<SSSLib::Vec3d>& iNodeList)
{
	cout<<"Num. Nodes "<<iNodeList.size()<<endl;
	ofstream outFile(oFileName);
	outFile<<iNodeList.size()<<"\t3\t0\t0\n";
	WriteArray<Vec3d, 3>(outFile, iNodeList);
}

void SSSVolMesh::LoadVtk(const char* iFileName)
{
	mVtx.clear();
	mTets.clear();
	mHexs.clear();

	ifstream inFile(iFileName);
	int n;
	do
	{
		string line;
		std::getline(inFile, line);
		stringstream lineStream(line);
		string name;
		lineStream>>name;
		std::transform(name.begin(), name.end(), name.begin(), ::tolower);
		if (name=="points")
		{
			lineStream>>n;
			break;
		}
	} while (1);
	mVtx.resize(n);
	for (int i=0; i<n; i++)
	for (int c=0; c<3; c++)
		inFile>>mVtx[i][c];

	do
	{
		string line;
		std::getline(inFile, line);
		stringstream lineStream(line);
		string name;
		lineStream>>name;
		std::transform(name.begin(), name.end(), name.begin(), ::tolower);
		if (name=="cells")
		{
			lineStream>>n;
			break;
		}
	} while (1);
	//mHexs.resize(n);
	mHexs.clear();
	mTets.clear();
	for (int i=0; i<n; i++)
	{
		int tmp;
		inFile>>tmp;
		assert(tmp==8 || tmp==4);
		if (tmp==8)
		{
			_THex hex;
			for (int c=0; c<8; c++)
				inFile>>hex[c];
			swap(hex[2], hex[3]);
			swap(hex[6], hex[7]);
			mHexs.push_back(hex);
		}
		else if (tmp==4)
		{
			Vec4i tet;
			for (int c=0; c<4; c++)
				inFile>>tet[c];
			mTets.push_back(tet);
		}
	}
}

void SSSVolMesh::LoadInp(const char* iFileName)
{
	ifstream inFile(iFileName);
	int state = -1;
	//state:
	// -2: error
	// -1: nothing
	//  1: nodes
	//  2: hex
	//  3: tet
	map<int, int> nodeID2Order;
	while (!inFile.eof())
	{
		string line;
		std::getline(inFile, line);
		transform(line.begin(), line.end(), line.begin(), ::toupper);
		if (line[0]=='*')
		{
			if (line.substr(0, 5)=="*NODE")
				state = 1;
			else if (line.substr(0, 8)=="*ELEMENT")
			{
				int place = line.find("TYPE=");
				string tmp = line.substr(place+5);
				stringstream tmpss(tmp);
				string tmp2;
				tmpss>>tmp2;
				assert(tmp2.substr(0, 4) == "C3D4" || tmp2.substr(0, 5)=="C3D8R");
				if (tmp2.substr(0, 4) == "C3D4")
					state = 3;
				else if (tmp2.substr(0, 5)=="C3D8R")
					state = 2;
				else
					state = -2;
			}
			else
				state = -1;
		}
		else if (state==1)
		{
			Vec3d node;
			int id;
			char c;
			stringstream ss(line);
			ss>>id>>c>>node[0]>>c>>node[1]>>c>>node[2];
			nodeID2Order[id] = mVtx.size();
			mVtx.push_back(node);
		}
		else if (state==2)
		{
			int id;
			_THex hex;
			char c;
			stringstream ss(line);
			ss>>id;
			for (int i=0; i<8; i++)
			{
				ss>>c>>hex[i];
				hex[i] = nodeID2Order[hex[i]];
			}
			swap(hex[2], hex[3]);
			swap(hex[6], hex[7]);
			mHexs.push_back(hex);
		}
		else if (state==3)
		{
			int id;
			Vec4i tet;
			char c;
			stringstream ss(line);
			ss>>id;
			for (int i=0; i<4; i++)
			{
				ss>>c>>tet[i];
				tet[i] = nodeID2Order[tet[i]];
			}
			mTets.push_back(tet);
		}
	}
}

void SSSVolMesh::LoadDisteneMesh(const char* iFileName)
{
	ifstream inFile(iFileName);
	char buffer[100];
	while (!inFile.eof())
	{
		string name;
		inFile>>name;
		if (name[0]=='#')
			inFile.getline(buffer, 100);
		else if (name=="Vertices")
		{
			int vtxNum;
			inFile>>vtxNum;
			mVtx.resize(vtxNum);
			for (int i=0; i<vtxNum; i++)
			{
				int tmp;
				for (int c=0; c<3; c++)
					inFile>>mVtx[i][c];
				inFile>>tmp;
			}
		}
		else if (name=="Hexahedra")
		{
			int hexNum;
			inFile>>hexNum;
			mHexs.resize(hexNum);
			for (int i=0; i<hexNum; i++)
			{
				int tmp;
				for (int c=0; c<8; c++)
					inFile>>mHexs[i][c];
				mHexs[i] = mHexs[i]-1;
				swap(mHexs[i][2], mHexs[i][3]);
				swap(mHexs[i][6], mHexs[i][7]);
				inFile>>tmp;
			}
		}
	}
	inFile.close();
	mSurfaceMesh.LoadDisteneMesh(iFileName);
}

void SSSVolMesh::WriteVtk(const char* oFileName)
{
	ofstream outFile(oFileName);
	outFile<<"# vtk DataFile Version 3.0\n"
		<<"Mesquite Mesh\n"
		<<"ASCII\n"
		<<"DATASET UNSTRUCTURED_GRID\n"
		<<"POINTS "<<mVtx.size()<<" double\n";
	for (int i=0; i<(int)mVtx.size(); i++)
		outFile<<setprecision(16)<<mVtx[i][0]<<" "<<mVtx[i][1]<<" "<<mVtx[i][2]<<endl;
	outFile<<"CELLS "<<mHexs.size()+mTets.size()<<" "<<mHexs.size()*9+mTets.size()*5<<endl;
	for (int i=0; i<(int)mHexs.size(); i++)
	{
		_THex hex = mHexs[i];
		swap(hex[2], hex[3]);
		swap(hex[6], hex[7]);
		outFile<<"8";
		for (int c=0; c<8; c++)
			outFile<<" "<<hex[c];
		outFile<<endl;
	}
	for (int i=0; i<(int)mTets.size(); i++)
	{
		Vec4i tet = mTets[i];
		outFile<<"4";
		for (int v=0; v<4; v++)
			outFile<<" "<<tet[v];
		outFile<<endl;
	}

	outFile<<"CELL_TYPES "<<mHexs.size()+mTets.size()<<endl;
	for (int i=0; i<(int)mHexs.size(); i++)
		outFile<<"12\n";
	for (int i=0; i<(int)mTets.size(); i++)
		outFile<<"10\n";
	outFile<<"POINT_DATA "<<mVtx.size()<<endl
			<<"SCALARS fixed int\n"
			<<"LOOKUP_TABLE default\n";
	for (int i=0; i<(int)mVtx.size(); i++)
		outFile<<"0\n";
}

void SSSVolMesh::WriteMesh(const char* oFileName)
{
	if (strcmp(oFileName+strlen(oFileName)-3, "vtk")==0)
		WriteVtk(oFileName);
	else if (strcmp(oFileName+strlen(oFileName)-3, "inp")==0)
		WriteAbaqusMsh(oFileName);
	else
	{
		cout<<"Unrecognized format: "<<oFileName<<endl;
	}
}

void SSSVolMesh::WriteAbaqusMsh(const char* oFileName)
{
	ofstream outFile(oFileName);
	outFile<<"*Heading\n"
			<<" "<<oFileName
			<<"\n*Node\n";
	for (int i=0; i<(int)mVtx.size(); i++)
	{
		outFile<<i+1;
		for (int c=0; c<3; c++)
			outFile<<", "<<mVtx[i][c];
		outFile<<endl;
	}

	if (mTets.size()>0)
	{
		outFile<<"*Element, type=C3D4, ELSET=PART1\n";
		for (int i=0; i<(int)mTets.size(); i++)
		{
			outFile<<i+1;
			for (int c=0; c<4; c++)
				outFile<<", "<<mTets[i][c]+1;
			outFile<<endl;
		}
	}

	if (mHexs.size()>0)
	{
		outFile<<"*Element, type=C3D8R, ELSET=PART2\n";
		for (int i=0; i<(int)mHexs.size(); i++)
		{
			outFile<<i+1;
			_THex h = mHexs[i];
			swap(h[2], h[3]);
			swap(h[6], h[7]);
			for (int c=0; c<8; c++)
				outFile<<", "<<h[c]+1;
			outFile<<endl;
		}
	}
}
