/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "SSSVolMesh.h"
#include "KDTree_wrap.h"
#include <algorithm>

using namespace std;
using namespace SSSLib;

namespace 
{
	Vec3d Pick(Vec3d v0, Vec3d v1, Vec3d v2, Vec3d v3) 
	{
		double s = randf<double>();
		double t = randf<double>();
		double u = randf<double>();
		if(s+t>1.0) { // cut'n fold the cube into a prism

		s = 1.0 - s;
		t = 1.0 - t;

		}
		if(t+u>1.0) { // cut'n fold the prism into a tetrahedron

		double tmp = u;
		u = 1.0 - s - t;
		t = 1.0 - tmp;

		} else if(s+t+u>1.0) {

		double tmp = u;
		u = s + t + u - 1.0;
		s = 1 - t - tmp;

		}
		double a=1-s-t-u; // a,s,t,u are the barycentric coordinates of the random point.
		return v0*a + v1*s + v2*t + v3*u;

	}
}

void SSSVolMesh::AddPoissonSamples(double iPr,
								   std::vector<SSSLib::Vec3d>& ioPS,
									std::vector<double>* ipSize)
{

//#define _DIS_WEIGHT
//#define _SUM_WEIGHT

#ifdef _DIS_WEIGHT
	const int checkSizeCount = 27;
	vector<Vec3d> checkDir(checkSizeCount);
	//for (int i=0; i<6; i++)
	//{
	//	checkDir[i] = Vec3d(0, 0, 0);
	//	checkDir[i][i/2] = (i%2)*2-1;
	//}
	for (int i=0; i<27; i++)
	{
		int ii=i;
		for (int c=0; c<3; c++)
		{
			checkDir[i][c] = (ii%3)-1.0;
			ii/=3;
		}
		if (i!=13)
		{
			assert(checkDir[i].Length()!=0);
			checkDir[i] = checkDir[i].Norm();
		}
	}
#endif
	double pr2 = sqr(iPr);
	vector<double> sizingFunc;
	if (!ipSize)
		sizingFunc.assign(ioPS.size(), pr2);
	else
	{
		sizingFunc.resize(ioPS.size());
		for (int i=0; i<(int)ioPS.size(); i++)
#if defined(_DIS_WEIGHT) || defined(_SUM_WEIGHT)
			sizingFunc[i] = sqr((*ipSize)[i]);
#else
			sizingFunc[i] = pr2;
#endif
	}

	bool bNewPt = true;
	vector<double> tetWeight(mTets.size());
	double tetWeightSum = 0;
	for (int i=0; i<(int)mTets.size(); i++)
	{
		double val = fabs(TetSignedVol(i));
		if (i==0)
			tetWeight[i] = val;
		else
			tetWeight[i] = val+tetWeight[i-1];
		tetWeightSum += val;
	}
	int numTrial = 1000;
	while (bNewPt)
	{
		bNewPt = false;
		for (int i=0; i<numTrial; i++)
		{
			double randTet = randf<double>()*tetWeightSum;
			int tetID = Min<int>(lower_bound(tetWeight.begin(), tetWeight.end(), randTet)
								- tetWeight.begin(),
								tetWeight.size()-1);
			Vec4i tet = mTets[tetID];
			Vec3d sample = Pick(mVtx[tet[0]], mVtx[tet[1]], mVtx[tet[2]], mVtx[tet[3]]);

			bNewPt = true;

#ifdef _DIS_WEIGHT
			vector<double> sizeW(checkSizeCount, -1);
			vector<double> sizeV(checkSizeCount, 0);
#endif
#ifdef _SUM_WEIGHT
			vector<Vec2d> checkPt;
#endif

			for (int j=0; j<(int)ioPS.size(); j++)
			{
				double disLimit = sizingFunc[j];
				//double disLimit = pr2;
				//if (ipSize && j<(int)ipSize->size())
				//	dis = sqr((*ipSize)[j]);
				Vec3d disVec = ioPS[j]-sample;
				double dis2 = disVec.SqrLength();
				if (dis2<disLimit)
				{
					bNewPt = false;
					break;
				}
#ifdef _DIS_WEIGHT
				int maxDir = 0;
				double maxDis = dot<double>(disVec, checkDir[0]);
				for (int i=1; i<checkSizeCount; i++)
				if (i!=13)
				{
					double val = dot<double>(disVec, checkDir[i]);
					if (val>maxDis)
					{
						maxDir = i;
						maxDis = val;
					}
				}

				assert(maxDis>0);

				if (sizeW[maxDir]<0 || maxDis<sizeW[maxDir])
				{
					sizeW[maxDir] = maxDis;
					sizeV[maxDir] = disLimit;
				}

#endif
				
#ifdef _SUM_WEIGHT
				if (checkPt.size()>=6)
				{
					checkPt[5] = Vec2d(dis2, disLimit);
					push_heap(checkPt.begin(), checkPt.end());
					pop_heap(checkPt.begin(), checkPt.end());
				}
				else
				{
					checkPt.push_back(Vec2d(dis2, disLimit));
					push_heap(checkPt.begin(), checkPt.end());
				}
#endif

				//for (int checkI = 0;
				//	checkI<checkSizeCount;
				//	checkI++)
				//{

				//}
			}

			if (bNewPt)
			{
				double val = pr2;
#ifdef _DIS_WEIGHT
				val = 0;
				double weight = 0;
				for (int i=0; i<checkSizeCount; i++)
				if (sizeW[i]>=0)
				{
					weight += 1.0/sizeW[i];
					val += sizeV[i]/sizeW[i];
				}
				assert(weight>0);
				val /= weight;
#endif

#ifdef _SUM_WEIGHT
				val = 0;
				for (int i=0; i<(int)checkPt.size(); i++)
					val += checkPt[i][1];
				val /= checkPt.size();
#endif

				ioPS.push_back(sample);
				sizingFunc.push_back(val);
				break;
			}
		}

		if (ioPS.size()%100==0)
			cout<<"Sampling "<<ioPS.size()<<endl;
	}
}

bool SSSVolMesh::CheckInside(Vec3d iPt)
{
	for (int t=0; t<(int)mTets.size(); t++)
	{
		Vec4i tet = mTets[t];
		bool bIn = true;
		for (int v=0; v<4 && bIn; v++)
		{
			Vec3d edges[3];
			for (int j=0; j<3; j++)
			{
				edges[j] = mVtx[tet[(j+1+v)%4]]-iPt;
			}
			if (v%2==1)
				swap(edges[0], edges[1]);
			double val = TripleProduct<double>(edges[0], edges[1], edges[2]);
			if (val<0)
				bIn = false;
		}
		if (bIn)
			return true;
	}
	return false;
}
