/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <iostream>
#include <fstream>
#include <opencv\cxcore.h>

namespace SSSLib
{
	inline void PFPfWrite(int iWidth, int iHeight, const float* iData, const char* oFileName, bool iPForPf)
	{
		FILE* oFile = fopen(oFileName, "wb");
		using namespace std;
		if (iPForPf)
			fprintf(oFile, "PF\n");
		else
			fprintf(oFile, "Pf\n");
		fprintf(oFile, "%d %d\n-1.00\n", iWidth, iHeight);
		fwrite((const char*)iData, sizeof(float)*(iPForPf?3:1), iWidth*iHeight, oFile);
		fclose(oFile);
	}

	inline void PFWrite(int iWidth, int iHeight, const float* iData, const char* oFileName)
	{
		PFPfWrite(iWidth, iHeight, iData, oFileName, true);
	}
	inline void PfWrite(int iWidth, int iHeight, const float* iData, const char* oFileName)
	{
		PFPfWrite(iWidth, iHeight, iData, oFileName, false);
	}
	inline void PfWrite(const cv::Mat& iImg, const char* oFileName)
	{
		PfWrite(iImg.cols, iImg.rows, &iImg.at<float>(0, 0), oFileName);
	}
	inline void PFWrite(const cv::Mat& iImg, const char* oFileName)
	{
		PFWrite(iImg.cols, iImg.rows, iImg.at<cv::Vec3f>(0, 0).val, oFileName);
	}

	inline cv::Mat PFPfRead(const char* iFileName)
	{
		using namespace std;
		FILE* inFile = fopen(iFileName, "rb");
		char name[100];
		int h,w;
		float f;
		fscanf(inFile, "%s\n%d %d %f\n", name, &w, &h, &f);
		cv::Mat retMat(w, h, name[1]=='F'? CV_32FC3: CV_32F);
		fread((char*)retMat.data, retMat.elemSize(), w*h, inFile);
		fclose(inFile);
		return retMat;
	}
}