/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "..\StaticFEM\Static2d.h"
#include "..\StaticADMMLib\Static2DADMM.h"
#include <gl\glew.h>
#include "..\SSSLib\SSSLibOGL\OGLFrame.h"
#include "..\SSSLib\SSSLibOGL\OGLCamModel.h"
#include "..\SSSLib\SSSLibOGL\OGLSimpleObj.h"
#include "..\SSSLib\SSSLibOGL\OGLTexture.h"
#include "..\SSSLib\SSSLibOGL\OGLSEasyProgram.h"
#include "..\SSSLib\SSSSmallVec.h"
#include "..\SSSLib\SSSColor.h"
#include <iostream>
//#include "..\FEMLib\FEMObj.h"
#include <Windows.h>
#include <GL\freeglut.h>

#ifdef _USE_MATLAB
#pragma comment(lib, "mclmcrrt.lib")
#pragma comment(lib, "FOptLib.lib")
#endif

using namespace SSSLib;
using namespace std;

OGLCamModel* g_testCam;
OGLSEasyProgram* g_pProgram;
Static2D gFEM;
ADMMLib::Static2DADMM gADMM;

int gWindowHeight = 500;
int gSteps = 0;

void TestRender(OGLScene::TStageName, int)
{
	gSteps++;
	//if (gSteps%500==0)
	//{
	//	cout<<"Restarting\n";
	//	gADMM.Restart();
	//}
	gADMM.RunOneStep();
	gADMM.UpdateResult(true, true);
//	g_testCam->SetProjMatrix();
//	g_testCam->SetViewMatrix();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(0, 0, gWindowHeight, gWindowHeight);

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glColor3f(0, 0, 1.0f);

	glBegin(GL_LINES);
		for (int i=0; i<(int)gADMM.mpXNodeF.size(); i++)
		{
			ADMMLib::XNodeF* node = gADMM.mpXNodeF[i];
			_T_FLOAT fconv = (node->mtF[0]-node->mtF0[0]).SqrLength()*node->mtRhoF[0][0]+
				(node->mtF[1]-node->mtF0[1]).SqrLength()*node->mtRhoF[1][0]+
				(node->mtF[2]-node->mtF0[2]).SqrLength()*node->mtRhoF[2][0];
			fconv = sqrt(fconv);
			_T_FLOAT vmconv =node->mtVM-node->mtVM0;

			cout<<"ID "<<i<<"\tF Conv:"<<fconv<<"\tVM Conv:"<<vmconv<<endl;

			Vec3uc clr = SSSLib::Value2RGBuc(0.4999f+float(SSSLib::clamp(node->mDispute*0.25, 0.0, 0.5)));
			glColor3ub(clr[0], clr[1], clr[2]);
			for (int j=0; j<3; j++)
			{
				glVertex2dv(node->mtX[j].val);
				glVertex2dv(node->mtX[(j+1)%3].val);
			}
		}
	glEnd();

	//glBegin(GL_LINES);
	//	for (int i=0; i<(int)gFEM.mTriangles.size(); i++)
	//	for (int j=0; j<3; j++)
	//	{
	//		glVertex2dv((double*)&(gFEM.mRest[gFEM.mTriangles[i][j]]));	//glTexCoord2dv((double*)&(gFEM.mRest[gFEM.mTriangles[i][j]]));
	//		glVertex2dv((double*)&(gFEM.mRest[gFEM.mTriangles[i][(j+1)%3]]));	//glTexCoord2dv((double*)&(gFEM.mRest[gFEM.mTriangles[i][k]]));
	//	}
	//glEnd();
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(gWindowHeight, 0, gWindowHeight, gWindowHeight);

	//glBegin(GL_LINES);
	//	for (int i=0; i<(int)gFEM.mTriangles.size(); i++)
	//	for (int j=0; j<3; j++)
	//	{
	//		glVertex2dv((double*)&(gFEM.mStatic[gFEM.mTriangles[i][j]]));	//glTexCoord2dv((double*)&(gFEM.mRest[gFEM.mTriangles[i][j]]));
	//		glVertex2dv((double*)&(gFEM.mStatic[gFEM.mTriangles[i][(j+1)%3]]));	//glTexCoord2dv((double*)&(gFEM.mRest[gFEM.mTriangles[i][k]]));
	//	}
	//glEnd();

	glBegin(GL_LINES);
		for (int i=0; i<(int)gADMM.mpXNodeF.size(); i++)
		{
			ADMMLib::XNodeF* node = gADMM.mpXNodeF[i];
			_T_FLOAT val = gADMM.mpXNodeVM[i]->m_n[0]/gADMM.mVMScale/gADMM.mMaxVM;
			Vec3uc clr = SSSLib::Value2RGBuc(0.4999f+float(SSSLib::clamp(val*0.5, 0.0, 0.5)));
			if (val>1.1)
				clr = Vec3uc(255, 0, 255);
			glColor3ub(clr[0], clr[1], clr[2]);
			for (int j=0; j<3; j++)
			{
				glVertex2dv(node->mtP[j].val);
				glVertex2dv(node->mtP[(j+1)%3].val);
			}
		}
	glEnd();

	glFlush();
	glutSwapBuffers();
	int err = glGetError();
	if (err>0)
		std::cout<<"Drawing Error:"<<err<<std::endl;
}

void Keyboard(unsigned char iKey, int x, int y)
{
	switch (iKey)
	{
	case '+':
		gADMM.mMaxVM *= 1.5;
		cout<<"New Von Mises Crit:"<<gADMM.mMaxVM<<endl;
		break;
	case '-':
		gADMM.mMaxVM /= 1.5;
		cout<<"New Von Mises Crit:"<<gADMM.mMaxVM<<endl;
		break;
	}
}

void KeyboardUp(unsigned char iKey, int x, int y)
{
	switch (iKey)
	{
	case 'r':
		gADMM.Restart();
		break;
	}
}

void main(int argc, char** argv)
{

	OGLFrame::Init(gWindowHeight*2,gWindowHeight);
	//OGLFrame::GetInstance()->SetCommonFunction();
	OGLFrame::GetInstance()->AddRenderExt(TestRender);
	OGLFrame::GetInstance()->AddKeyboard(Keyboard);
	OGLFrame::GetInstance()->AddKeyboardUp(KeyboardUp);
	OGLFrame::EnableIdleRender();
	//OGLFrame::GetInstance()->AddKey0board(
	g_testCam = new OGLCamModel();
	g_testCam->EnableCam(OGLFrame::GetInstance());

	gFEM.InitRect(2, 2, 0.9);
	gFEM.InitParam(2e3, 0.0f);
	gFEM.mForce[0] = Vec2Float(0, -10.0f);
	gFEM.mForce[1] = Vec2Float(0, -10.0f);
	gFEM.mFixedVtx[2] = 1;
	gFEM.mFixedVtx[3] = 1;

	//gFEM.InitRect(6, 6, 0.15);
	//gFEM.InitParam(1e4, 0.3f);
	////for (int i=0; i<(int)gFEM.mForce.size(); i++)
	////	gFEM.mForce[i] = Vec2Float(0, -9.8f);
	//gFEM.mForce[0] = Vec2Float(0, -10.0f);
	//gFEM.mForce[5] = Vec2Float(0, -10.0f);
	////gFEM.mForce[30] = Vec2Float(0, -10.0f);
	////gFEM.mForce[5] = Vec2Float(-10.0f, 0);
	////gFEM.mForce[33] = Vec2Float(0, 10.0f);
	////gFEM.mForce[32] = Vec2Float(0, 9.8f);
	////gFEM.mForce[35] = Vec2Float(0, -9.8f);
	//gFEM.mFixedVtx[30] = 1;
	//gFEM.mFixedVtx[31] = 1;
	//gFEM.mFixedVtx[32] = 1;
	//gFEM.mFixedVtx[33] = 1;
	//gFEM.mFixedVtx[34] = 1;
	//gFEM.mFixedVtx[35] = 1;
	
	gFEM.Compute();

	gADMM.mpFEM2D = &gFEM;
	gADMM.mRho = 1.0;
	gADMM.SetupGraph();
	gADMM.CopyInitialValue();
	//gADMM.UpdateResult(true, true);
	//gADMM.RunOneStep();

	cout<<"Length\t"<<gFEM.mRest[gFEM.mRest.size()-1]-gFEM.mRest[0]<<endl;
	//glewInit();
	//g_pProgram = new OGLSEasyProgram(NULL,"restshader.txt");
	//g_pProgram->Use();
	glutMainLoop();
}
