/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "IpoptWrap.h"
#include "PhysImpSolve3D.h"

IpoptWrap::IpoptWrap(void)
{
}

IpoptWrap::~IpoptWrap(void)
{
}

void IpoptWrap::StateIpopt2Phys(const Ipopt::Number* iX)
{
	int poffset = mpPhys->mX0.size()*3;
	if (sizeof(Ipopt::Number)==sizeof(_T_FLOAT))
	{
		memcpy(mpPhys->mX0[0].val, iX, mpPhys->mX0.size()*3*sizeof(_T_FLOAT));
		memcpy(mpPhys->mP0[0].val, iX+poffset, mpPhys->mX0.size()*3*sizeof(_T_FLOAT));
	}
	else
	{
		for (int i=0; i<(int)mpPhys->mX0.size(); i++)
		for (int c=0; c<3; c++)
		{
			mpPhys->mX0[i][c] = iX[i*3+c];
			mpPhys->mP0[i][c] = iX[i*3+c+poffset];
		}
	}
}

void IpoptWrap::StatePhys2Ipopt(Ipopt::Number* iX)
{
	int poffset = mpPhys->mX0.size()*3;
	if (sizeof(Ipopt::Number)==sizeof(_T_FLOAT))
	{
		memcpy(iX, mpPhys->mX0[0].val, mpPhys->mX0.size()*3*sizeof(_T_FLOAT));
		memcpy(iX+poffset, mpPhys->mP0[0].val, mpPhys->mX0.size()*3*sizeof(_T_FLOAT));
	}
	else
	{
		for (int i=0; i<(int)mpPhys->mX0.size(); i++)
		for (int c=0; c<3; c++)
		{
			iX[i*3+c] = mpPhys->mX0[i][c];
			iX[i*3+c+poffset] = mpPhys->mP0[i][c];
		}
	}
}

void IpoptWrap::Init()
{
	mForceIdxBegin = 0;
	mNumForceRows = mpPhys->mP0.size()*3;

	mVMIdxBegin = mForceIdxBegin+mNumForceRows;
	mNumVMRows = mpPhys->mpVolMesh->mHexs.size()*8;

	mDeformIdxBegin = mVMIdxBegin+mNumVMRows;
	mLapForceIdx.clear();
	for (int i=0; i<(int)mpPhys->mConnect.size(); i++)
	{
		if (mpPhys->mConnect[i].size()>0)
			mLapForceIdx.push_back(i);
		mNumDeformEntries += mpPhys->mConnect[i].size()+1;
	}
	mNumDeformRows = mLapForceIdx.size()*3;
}

bool IpoptWrap::get_nlp_info(Ipopt::Index& n,
							Ipopt::Index& m,
							Ipopt::Index& nnz_jac_g,
							Ipopt::Index& nnz_h_lag,
							IndexStyleEnum& index_style)
{
	n = mpPhys->mX0.size()*3*2;
	// m: force equilibrium (p*3) + VM + deform
	m = mDeformIdxBegin+mNumDeformRows;

	// jacobian sparse structure
	// force
	nnz_jac_g = mpPhys->mNumVtxSparse*9;
	// VM
	nnz_jac_g += mNumVMRows*8*3;
	// deform
	nnz_jac_g += mNumDeformEntries*3;
}

