/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "IpTNLP.hpp"
#include <vector>

class PhysImpSolve3D;

class IpoptWrap : public Ipopt::TNLP
{
public:
	IpoptWrap(void);
	~IpoptWrap(void);

	PhysImpSolve3D* mpPhys;
	void StateIpopt2Phys(const Ipopt::Number* iX);
	void StatePhys2Ipopt(Ipopt::Number* iX);

	std::vector<int> mLapForceIdx;
	int mNumForceRows;
	int mForceIdxBegin;
	int mNumVMRows;
	int mVMIdxBegin;
	int mNumDeformRows;
	int mNumDeformEntries;
	int mDeformIdxBegin;

	void Init();

	/**@name Overloaded from TNLP */
	//@{
	/** Method to return some info about the nlp */
	virtual bool get_nlp_info(Ipopt::Index& n,
							Ipopt::Index& m,
							Ipopt::Index& nnz_jac_g,
							Ipopt::Index& nnz_h_lag,
							IndexStyleEnum& index_style);

	/** Method to return the bounds for my problem */
	virtual bool get_bounds_info(Ipopt::Index n,
							Ipopt::Number* x_l,
							Ipopt::Number* x_u,
							Ipopt::Index m,
							Ipopt::Number* g_l,
							Ipopt::Number* g_u);

	/** Method to return the starting point for the algorithm */
	virtual bool get_starting_point(Ipopt::Index n,
								bool init_x,
								Ipopt::Number* x,
								bool init_z,
								Ipopt::Number* z_L,
								Ipopt::Number* z_U,
								Ipopt::Index m,
								bool init_lambda,
								Ipopt::Number* lambda);

	/** Method to return the objective value */
	virtual bool eval_f(Ipopt::Index n,
						const Ipopt::Number* x,
						bool new_x,
						Ipopt::Number& obj_value);

	/** Method to return the gradient of the objective */
	virtual bool eval_grad_f(Ipopt::Index n,
							const Ipopt::Number* x,
							bool new_x,
							Ipopt::Number* grad_f);

	/** Method to return the constraint residuals */
	virtual bool eval_g(Ipopt::Index n,
						const Ipopt::Number* x,
						bool new_x,
						Ipopt::Index m,
						Ipopt::Number* g);

	/** Method to return:
	*   1) The structure of the jacobian (if "values" is NULL)
	*   2) The values of the jacobian (if "values" is not NULL)
	*/
	virtual bool eval_jac_g(Ipopt::Index n,
							const Ipopt::Number* x,
							bool new_x,
							Ipopt::Index m,
							Ipopt::Index nele_jac,
							Ipopt::Index* iRow,
							Ipopt::Index *jCol,
							Ipopt::Number* values);

	/** Method to return:
	*   1) The structure of the hessian of the lagrangian (if "values" is NULL)
	*   2) The values of the hessian of the lagrangian (if "values" is not NULL)
	*/
	virtual bool eval_h(Ipopt::Index n,
						const Ipopt::Number* x,
						bool new_x,
						Ipopt::Number obj_factor,
						Ipopt::Index m,
						const Ipopt::Number* lambda,
						bool new_lambda,
						Ipopt::Index nele_hess,
						Ipopt::Index* iRow,
						Ipopt::Index* jCol,
						Ipopt::Number* values);

	//@}

};

