/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

//#include "engine.h"
//#include <mex.h>
#include "mclmcrrt.h"
#include "KKTEle.h"
#include <algorithm>
#include <numeric>
#include <map>

namespace MatlabUtil
{

	inline mxArray* CreateMatrix(int iRows, int iCols)
	{
		return mxCreateDoubleMatrix(iRows, iCols, mxREAL);
	}

	template <typename _TVal>
	void SetMatrix(_TVal* iData, mxArray* ioMatrix)
	{
		int rows = (int)mxGetM(ioMatrix);
		int cols = (int)mxGetN(ioMatrix);
		double* ptr = mxGetPr(ioMatrix);
		for (int r=0; r<rows; r++)
		for (int c=0; c<cols; c++)
			ptr[r+c*rows] = double(iData[r*cols+c]);
		//for (int i=0; i<iRows*iCols; i++)
		//	ptr[i] = double(iData[i]);
	}

	template <typename _TVal>
	void CreateSetMatrix(int iRows, int iCols, _TVal* iData, mxArray** oMatrix)
	{
		*oMatrix = CreateMatrix(iRows, iCols);
		SetMatrix(iData, *oMatrix);
	}

	template <typename _TVal>
	void GetMatrix(mxArray* iMatrix, _TVal* oData)
	{
		int rows = mxGetM(iMatrix);
		int cols = mxGetN(iMatrix);
		double* ptr = mxGetPr(iMatrix);
		for (int r=0; r<rows; r++)
		for (int c=0; c<cols; c++)
			oData[r*cols+c] = _TVal(ptr[r+c*rows]);
		//for (int i=0; i<iRows*iCols; i++)
		//	oData[i] = _TVal(ptr[i]);
	}

	inline mxArray* CreateSetSparseMatrix(
									int iCols,
									int iN,
									const KKTEle* iKKT
									)
	{
		using namespace std;
		using namespace SSSLib;

		typedef pair<Vec2i, _T_FLOAT> _TCRD;
		vector<_TCRD> crddata;
		for (int i=0; i<iN; i++)
		for (int j=0; j<(int)iKKT[i].mACoes.size(); j++)
		if (iKKT[i].mACoes[j].first<iCols)
			crddata.push_back(_TCRD(Vec2i(iKKT[i].mACoes[j].first, i), iKKT[i].mACoes[j].second));
		
		int kktNonZero = (int)crddata.size();
		mxArray* outMat = mxCreateSparse(iN, iCols, kktNonZero, mxREAL);
		double *data = mxGetPr(outMat);
		mwIndex *rowIndex = mxGetIr(outMat);
		mwIndex *columnPos = mxGetJc(outMat);

		sort(crddata.begin(), crddata.end());

		int prevCol = -1;
		for (int i=0; i<(int)crddata.size(); i++)
		{
			_TCRD& datapt = crddata[i];
			while (datapt.first[0]!=prevCol)
			{
				prevCol++;
				columnPos[prevCol] = i;
			}
			rowIndex[i] = datapt.first[1];
			data[i] = datapt.second;
		}
		prevCol++;
		while (prevCol<=iCols)
		{
			columnPos[prevCol] = crddata.size();
			prevCol++;
		}
		//for (int i=0; i<iN; i++)
		//{
		//	for (int j=0; j<(int)iKKT[i].mACoes.size(); j++)
		//	{
		//		rowIndex[idx] = i;
		//		columnPos[idx] = iKKT[i].mACoes[j].first;
		//		assert(iKKT[i].mACoes[j].first<iCols);
		//		data[idx] = iKKT[i].mACoes[j].second;
		//		idx++;
		//	}
		//}
		return outMat;
	}
	inline mxArray* CreateSetSparseMatrix(
										int iRows,
										int iCols,
										const std::map<SSSLib::Vec2i, _T_FLOAT>& iData)
	{
		using namespace std;
		using namespace SSSLib;

		typedef pair<Vec2i, _T_FLOAT> _TData;
		vector<_TData> datavec(iData.size());
		int idx=0;
		for (map<Vec2i, _T_FLOAT>::const_iterator ite = iData.begin();
			ite != iData.end();
			ite++)
		{
			datavec[idx].first = Vec2i(ite->first[1], ite->first[0]);
			datavec[idx].second = ite->second;
			idx++;
		}

		sort(datavec.begin(), datavec.end());

		mxArray* outMat = mxCreateSparse(iRows, iCols, iData.size(), mxREAL);
		double *data = mxGetPr(outMat);
		mwIndex *rowIndex = mxGetIr(outMat);
		mwIndex *columnPos = mxGetJc(outMat);
		int lastCol = -1;
		for (int i=0; i<(int)datavec.size(); i++)
		{
			_TData dataPt = datavec[i];
			while (dataPt.first[0] != lastCol)
			{
				lastCol++;
				columnPos[lastCol] = i;
			}
			rowIndex[i] = dataPt.first[1];
			data[i] = dataPt.second;
		}
		return outMat;
	}


	inline mxArray* CreateSetSparseMatrixTransposed(
									int iRows,
									int iN,
									const KKTEle* iKKT
									)
	{
		using namespace std;
		using namespace SSSLib;
		int kktNonZero = 0;
		for (int i=0; i<iN; i++)
		for (int j=0; j<(int)iKKT[i].mACoes.size(); j++)
		if (iKKT[i].mACoes[j].first<iRows)
			kktNonZero ++;
		mxArray* outMat = mxCreateSparse(iRows, iN, kktNonZero, mxREAL);
		double *data = mxGetPr(outMat);
		mwIndex *rowIndex = mxGetIr(outMat);
		mwIndex *columnPos = mxGetJc(outMat);

		int idx = 0;

		for (int i=0; i<iN; i++)
		{
			columnPos[i] = idx;
			for (int j=0; j<(int)iKKT[i].mACoes.size(); j++)
			{
				int rowIdx = iKKT[i].mACoes[j].first;
				if (rowIdx<iRows)
				{
					rowIndex[idx] = rowIdx;
					data[idx] = iKKT[i].mACoes[j].second;
					idx++;
				}
			}
		}
		columnPos[iN] = idx;
		return outMat;
	}

	inline mxArray* CreateSetB(
								int iN,
								const KKTEle* iKKT)
	{
		mxArray* outMat;
		std::vector<_T_FLOAT> data(iN);
		for (int i=0; i<iN; i++)
			data[i] = iKKT[i].mB;
		CreateSetMatrix<_T_FLOAT>(iN, 1, &data[0], &outMat);
		return outMat;
	}
	inline mxArray* CreateSetBTransposed(
								int iN,
								const KKTEle* iKKT)
	{
		mxArray* outMat;
		std::vector<_T_FLOAT> data(iN);
		for (int i=0; i<iN; i++)
			data[i] = iKKT[i].mB;
		CreateSetMatrix<_T_FLOAT>(1, iN, &data[0], &outMat);
		return outMat;
	}
}