/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve.h"
#include "../SSSLib/SSSFunc.h"
#include <Eigen/Sparse>
#include <Eigen/IterativeLinearSolvers>
#include <Eigen/SparseLU>
#include <Eigen/SparseQR>
#include <Eigen/IterativeLinearSolvers>
#include <iostream>
#include <set>

using namespace std;
using namespace SSSLib;

//#define _NO_X_DEFORM

MatXFloat PhysImpSolve::DetX1stOrder(SVec2Float* iX, MatXFloat* odInvDetdX)
{
	MatXFloat ddetdX = MatXFloat::Zero(6, 1);

	ddetdX(0, 0) = iX[1][1]-iX[2][1];
	ddetdX(2, 0) = iX[2][1]-iX[0][1];
	ddetdX(4, 0) = iX[0][1]-iX[1][1];

	ddetdX(1, 0) = iX[2][0]-iX[1][0];
	ddetdX(3, 0) = iX[0][0]-iX[2][0];
	ddetdX(5, 0) = iX[1][0]-iX[0][0];

	if (odInvDetdX)
	{
		_T_FLOAT det = (iX[0][0]-iX[2][0])*(iX[1][1]-iX[2][1])-(iX[0][1]-iX[2][1])*(iX[1][0]-iX[2][0]);
		*odInvDetdX = -1.0/sqr(det)*ddetdX;
	}

	return ddetdX;
}

void PhysImpSolve::U2ndOrder(SVec2Float* iX, SVec2Float* iP,
							 MatXFloat iD,  _T_FLOAT iThick,
							 MatXFloat& odUdX, MatXFloat& odUdP,
							 MatXFloat& od2UdX2, MatXFloat& od2UdXdP, MatXFloat& od2UdP2)
{
	_T_FLOAT det;
	MatXFloat B0 = Static2D::GetB0(iX[0], iX[1], iX[2], &det);
	_T_FLOAT detinv = 1.0/det;

	MatXFloat q = MatXFloat::Zero(6, 1);
	for (int i=0; i<3; i++)
	{
		q(i*2, 0) = iP[i][0];
		q(i*2+1, 0) = iP[i][1];
	}

	odUdX = MatXFloat::Zero(6, 1);
	odUdP = MatXFloat::Zero(6, 1);
	od2UdX2 = MatXFloat::Zero(6, 6);
	od2UdXdP = MatXFloat::Zero(6, 6);
	od2UdP2 = MatXFloat::Zero(6, 6);
	if (det<msDetEpsilon)
		return;

	MatXFloat B0q = B0*q;
	MatXFloat DB0q = iD*B0q;
	MatXFloat BtD = B0.transpose()*iD;

	MatXFloat ddetInvdX;
	MatXFloat ddetdX = DetX1stOrder(iX, &ddetInvdX);

	//MatXFloat ddetInvdX = -1.0*sqr(detinv)*ddetdX;

	MatXFloat d2detinvdX2 = 2*sqr(detinv)*detinv*ddetdX*ddetdX.transpose()
							-1.0*sqr(detinv)*msd2detdX2;

	// dUdP
	odUdP = 1.0/2.0*iThick*detinv*B0.transpose()*DB0q;

	// dUdX
	MatXFloat QtB0tDB0Q = B0q.transpose()*DB0q;
	assert(QtB0tDB0Q.rows()==1);
	assert(QtB0tDB0Q.cols()==1);
	
	odUdX = 1.0/4.0*iThick*QtB0tDB0Q(0, 0)*ddetInvdX;
	for (int i=0; i<6; i++)
	{
		MatXFloat tmp = detinv*DB0q.transpose()*msdBdX[i]*q;
		assert(tmp.cols() == 1);
		assert(tmp.rows() == 1);
		odUdX(i, 0) += 1.0/2.0*iThick*tmp(0, 0);
	}

	// d2UdX2
	for (int i=0; i<6; i++)
	{
		for (int j=0; j<6; j++)
		{
			MatXFloat d2udx2 = 1.0/2.0*iThick
								*DB0q.transpose()*msdBdX[j]*q*ddetInvdX(i,0);
			d2udx2 += 1.0/2.0*iThick
				*DB0q.transpose()*msdBdX[i]*q*ddetInvdX(j,0);
			d2udx2 += 1.0/4.0*iThick
				*B0q.transpose()*DB0q*d2detinvdX2(i, j);

			d2udx2 += 1.0/2.0*iThick
				*detinv*q.transpose()*msdBdX[i].transpose()*iD*msdBdX[j]*q;

			od2UdX2(i, j) += d2udx2(0, 0);
			//od2UdX2(j, i) += d2udx2(0, 0);
		}
	}

	// d2UdXdP
	//for (int i=0; i<6; i++)
	//{
	//	MatXFloat d2udxdp = BtD*msdBdX[i];
	//	MatXFloat d2udxdpt = d2udxdp.transpose();
	//	d2udxdp += d2udxdpt;
	//	d2udxdp = 0.5*iThick*detinv*d2udxdp*q;

	//	d2udxdp += 0.5*iThick*B0.transpose()*DB0q*ddetInvdX(i, 0);

	//	od2UdXdP.row(i) = d2udxdp;
	//}

	for (int i=0; i<6; i++)
	{
		MatXFloat d2udxdp = 0.5*iThick*detinv*BtD*msdBdX[i]*q;
		d2udxdp += 0.5*iThick*detinv*msdBdX[i].transpose()*DB0q;
		d2udxdp += 0.5*iThick*ddetInvdX(i, 0)*B0.transpose()*DB0q;

		od2UdXdP.row(i) = d2udxdp.transpose();
	}

	// d2Ud2P
	od2UdP2 = 0.5*iThick*detinv*BtD*B0;
}

void PhysImpSolve::PStep(_T_FLOAT iDT)
{
	// build system
	typedef Eigen::Triplet<_T_FLOAT> T;
	std::vector<T> tripleList;
	MatXFloat b = MatXFloat::Zero(mPV.size()*2, 1);

	for (int t=0; t<(int)mpStatic2D->mTriangles.size(); t++)
	{

		SVec2Float x[3];
		SVec2Float p[3];

		MatXFloat dUdX;
		MatXFloat dUdP;
		MatXFloat d2UdX2;
		MatXFloat d2UdXdP;
		MatXFloat d2UdP2;

		Vec3i tri = mpStatic2D->mTriangles[t];

		for (int i=0; i<3; i++)
		{
			x[i] = mpStatic2D->mRest[tri[i]];
			p[i] = mpStatic2D->mStatic[tri[i]];
			p[i] -= x[i];
		}

		U2ndOrder(x, p,
			mpStatic2D->mD, mpStatic2D->mThick,
			dUdX, dUdP,
			d2UdX2, d2UdXdP, d2UdP2);

		// b
		for (int i=0; i<3; i++)
		for (int j=0; j<2; j++)
			b(tri[i]*2+j, 0) -= mFDamping*iDT*dUdP(i*2+j,0);

		// A
		for (int i=0; i<6; i++)
		for (int j=0; j<6; j++)
			tripleList.push_back(T(tri[i/2]*2+(i%2), tri[j/2]*2+(j%2), 
							mFDamping*sqr(iDT)*d2UdP2(i, j)));
	}

	for (int i=0; i<(int)mpStatic2D->mRest.size(); i++)
	for (int j=0; j<2; j++)
	{
		tripleList.push_back(T(i*2+j, i*2+j, mMass[i]));
		b(i*2+j, 0) += mVDamping*mMass[i]*mPV[i][j];
		b(i*2+j, 0) += iDT*mFDamping*mExtForceP[i][j];
	}

	using namespace Eigen;
	SparseMatrix<_T_FLOAT> A(mpStatic2D->mRest.size()*2, mpStatic2D->mRest.size()*2);
	A.setFromTriplets(tripleList.begin(), tripleList.end());

	SparseQR<SparseMatrix<_T_FLOAT>, COLAMDOrdering<int> > solver;
	solver.analyzePattern(A);
	solver.factorize(A);
	if (solver.lastErrorMessage().length()>0)
		cout<<solver.lastErrorMessage();
	MatXFloat x = solver.solve(b);

	for (int i=0; i<(int)mPV.size(); i++)
	for (int j=0; j<2; j++)
	{
		mPV[i][j] = x(i*2+j,0);
		//mPV[i][j] = b(i*2+j, 0)/mMass[i];
		mpStatic2D->mStatic[i][j] += mPV[i][j]*iDT;
	}

}

void PhysImpSolve::XPStep(_T_FLOAT iDT)
{

	_T_FLOAT damping = pow(mFDamping, iDT);
	_T_FLOAT vdamping = pow(mVDamping, iDT);

	if (mbStatic)
	{
		iDT = 1.0;
		damping = 1.0;
		vdamping = 0;
	}

	_T_FLOAT dpweight = damping*iDT;
	_T_FLOAT dp2weight = damping*sqr(iDT);

	if (mbPStatic)
	{
		dpweight = 1.0;
		dp2weight = 1.0;
	}

	// build system
	typedef Eigen::Triplet<_T_FLOAT> T;
	std::vector<T> tripleList;
	MatXFloat b = MatXFloat::Zero(mXV.size()*2+mPV.size()*2, 1);

	int poffset = mXV.size()*2;

	for (int t=0; t<(int)mpStatic2D->mTriangles.size(); t++)
	{

		SVec2Float x[3];
		SVec2Float p[3];

		MatXFloat dUdX;
		MatXFloat dUdP;
		MatXFloat d2UdX2;
		MatXFloat d2UdXdP;
		MatXFloat d2UdP2;

		Vec3i tri = mpStatic2D->mTriangles[t];

		for (int i=0; i<3; i++)
		{
			x[i] = mpStatic2D->mRest[tri[i]];
			p[i] = mpStatic2D->mStatic[tri[i]];
			p[i] -= x[i];
		}

		U2ndOrder(x, p,
			mpStatic2D->mD, mpStatic2D->mThick,
			dUdX, dUdP,
			d2UdX2, d2UdXdP, d2UdP2);

#ifndef _NO_X_DEFORM

#	ifndef _ASYM_LINEAR
		for (int i=0; i<3; i++)
		for (int j=0; j<2; j++)
			b(tri[i]*2+j, 0) -= damping*iDT*(dUdX(i*2+j,0)-dUdP(i*2+j,0));
			//b(tri[i]*2+j, 0) -= mDamping*iDT*dUdX(i*2+j,0);
#	endif

#endif

		for (int i=0; i<3; i++)
		for (int j=0; j<2; j++)
			b(tri[i]*2+j+poffset, 0) -= dpweight*dUdP(i*2+j,0);

		// A
#ifndef _NO_X_DEFORM

#	ifndef _ASYM_LINEAR
		for (int i=0; i<6; i++)
		for (int j=0; j<6; j++)
			tripleList.push_back(T(tri[i/2]*2+(i%2), tri[j/2]*2+(j%2), 
							damping*sqr(iDT)*
							//d2UdX2(i, j)
							(d2UdX2(i, j)  - d2UdXdP(i, j) - d2UdXdP(j, i)
							+ d2UdP2(i, j))
							));
#	endif

		for (int i=0; i<6; i++)
		for (int j=0; j<6; j++)
		{
			_T_FLOAT val = dp2weight*
				//d2UdXdP(i, j);
				(d2UdXdP(i, j)-d2UdP2(i, j));
			//val = 0;

#	ifndef _ASYM_LINEAR
			tripleList.push_back(T(tri[i/2]*2+(i%2), tri[j/2]*2+(j%2)+poffset, 
							val));
#	endif
			if (fabs(val)>mMatrixEpsilon && !mpStatic2D->mFixedVtx[tri[j/2]])
				tripleList.push_back(T(tri[j/2]*2+(j%2)+poffset, tri[i/2]*2+(i%2),
								val));

		}
#endif
		for (int i=0; i<6; i++)
		if (!mpStatic2D->mFixedVtx[tri[i/2]])
		for (int j=0; j<6; j++)
		{
			_T_FLOAT val = dp2weight*d2UdP2(i, j);
			if (fabs(val)>mMatrixEpsilon)
				tripleList.push_back(T(tri[i/2]*2+(i%2)+poffset, tri[j/2]*2+(j%2)+poffset, 
								val));

			//tripleList.push_back(T(tri[i/2]*2+(i%2), tri[j/2]*2+(j%2), 
			//				damping*sqr(iDT)*d2UdP2(i, j)));

			//tripleList.push_back(T(tri[i/2]*2+(i%2), tri[j/2]*2+(j%2)+poffset, 
			//				-damping*sqr(iDT)*d2UdP2(i, j)));
			//tripleList.push_back(T(tri[i/2]*2+(i%2)+poffset, tri[j/2]*2+(j%2), 
			//				-damping*sqr(iDT)*d2UdP2(i, j)));
		}
	}

	for (int i=0; i<(int)mdExtForce.size(); i++)
	{
		T extXTri = mdExtForce[i];
		if (fabs(extXTri.value())<mMatrixEpsilon)
			continue;
		if (extXTri.row()<poffset || !mpStatic2D->mFixedVtx[(extXTri.row()-poffset)/2])
		{
			// a potential error in the sign here
			if (extXTri.row()<poffset || !mbPStatic)
				tripleList.push_back(T(extXTri.row(), extXTri.col(), -sqr(iDT)*damping*extXTri.value()));
			else
				tripleList.push_back(T(extXTri.row(), extXTri.col(), -dp2weight*extXTri.value()));
		}
	}

	//for (int i=0; i<(int)mdExtForceP.size(); i++)
	//{
	//	T extXTri = mdExtForceP[i];
	//	tripleList.push_back(T(extXTri.row()+poffset, extXTri.col()+poffset,
	//		-sqr(iDT)*mDamping*extXTri.value()));
	//}

	for (int i=0; i<(int)mpStatic2D->mRest.size(); i++)
	for (int j=0; j<2; j++)
	{
		_T_FLOAT xmass;
		if (mbXUniformMass)
			xmass = mXUniformMass;
		else
			xmass = mMass[i]*mXMassScale;
		if (!mbStatic)
		{
			tripleList.push_back(T(i*2+j, i*2+j, xmass));
			b(i*2+j, 0) += vdamping*xmass*mXV[i][j];
		}
		b(i*2+j, 0) += dpweight*mExtForceX[i][j];

		//b(i*2+j, 0) = 0;
		if (!mpStatic2D->mFixedVtx[i])
		{
			if (!mbStatic && !mbPStatic)
			{
				tripleList.push_back(T(i*2+j+poffset, i*2+j+poffset, mMass[i]));
				b(i*2+j+poffset, 0) += vdamping*mMass[i]*mPV[i][j];
			}
			b(i*2+j+poffset, 0) += dpweight*mExtForceP[i][j];
		}
		else
		{
			tripleList.push_back(T(i*2+j+poffset, i*2+j, 1.0));
			tripleList.push_back(T(i*2+j+poffset, i*2+j+poffset, -1.0));
			b(i*2+j+poffset, 0) = 0;
		}
	}

	// check non-referenced parameters
	{
		std::set<int> referenced;
		std::set<int> referencedC;
		for (int i=0; i<(int)tripleList.size(); i++)
		{
			if (tripleList[i].value()!=0)
			{
				referenced.insert(tripleList[i].row());
				referencedC.insert(tripleList[i].col());
			}
		}
		for (int i=0; i<(int)mpStatic2D->mRest.size()*4; i++)
		if (referenced.count(i)==0 /*|| referencedC.count(i)==0*/)
		{
			tripleList.push_back(T(i, i, 1));
			b(i) = 0;
			//cout<<"Unreference "<<i<<endl;
		}
	}

	using namespace Eigen;
	SparseMatrix<_T_FLOAT> A(mpStatic2D->mRest.size()*4, mpStatic2D->mRest.size()*4);
	A.setFromTriplets(tripleList.begin(), tripleList.end());

	SparseLU<SparseMatrix<_T_FLOAT>, COLAMDOrdering<int> > solver;
	//ConjugateGradient<SparseMatrix<_T_FLOAT> > solver;
	solver.analyzePattern(A);
	if (solver.lastErrorMessage().length()>0)
		cout<<solver.lastErrorMessage();
	solver.factorize(A);
	if (solver.lastErrorMessage().length()>0)
		cout<<solver.lastErrorMessage();
	MatXFloat x = solver.solve(b);
	//MatXFloat x0(mpStatic2D->mRest.size()*4, 1);
	//for (int i=0; i<(int)mpStatic2D->mRest.size(); i++)
	//for (int j=0; j<2; j++)
	//{
	//	x0(i*2+j, 0) = mXV[i][j];
	//	x0(i*2+j+poffset, 0) = mPV[i][j];
	//}
	//MatXFloat x = solver.solveWithGuess(b, x0);

	//assert((A*x-b).squaredNorm()<b.squaredNorm()*1e-6);

	for (int i=0; i<(int)mPV.size(); i++)
	for (int j=0; j<2; j++)
	{
		mXV[i][j] = x(i*2+j,0);
		mPV[i][j] = x(i*2+j+poffset,0);


		//if (mpStatic2D->mFixedVtx[i] && fabs(mXV[i][j]-mPV[i][j])>1e-8)
		//{
		//	cout<<"Inequal speed on fixed points: "<<i<<','<<j<<endl;
		//}
		
		//mXV[i][j] = b(i*2+j, 0)/mMass[i]/mXMassScale;

		//mPV[i][j] = b(i*2+j+poffset, 0)/mMass[i];
		if (mbStatic)
		{
			mpStatic2D->mRest[i][j] += mXV[i][j]*mXStaticStep;
		}
		else
			mpStatic2D->mRest[i][j] += mXV[i][j]*iDT;

		if (mbPStatic)
			mpStatic2D->mStatic[i][j] += mPV[i][j];
		else if (mbStatic)
			mpStatic2D->mStatic[i][j] += mPV[i][j]*mXStaticStep;
		else
			mpStatic2D->mStatic[i][j] += mPV[i][j]*iDT;

		if (mpStatic2D->mFixedVtx[i])
		{
			mPV[i][j] = mXV[i][j];
			mpStatic2D->mStatic[i][j] = mpStatic2D->mRest[i][j];
		}
	}

}