/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "../StaticFEM/Static2D.h"
#include <Eigen/Sparse>
#include <set>
#include <map>

#define _ASYM_LINEAR

class PhysImpSolve
{
public:
	PhysImpSolve(void);
	~PhysImpSolve(void);

	Static2D* mpStatic2D;

	static void Strain2ndOrder(SVec2Float* iX, SVec2Float* iP,
							MatXFloat& oS,
							MatXFloat& odSdX, MatXFloat& odSdP,
							std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >& od2SdX2,
							std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >& od2SdXdP); // first strain second X third P
							// the last term is zero
							//std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >& od2SdP2
	static MatXFloat DetX1stOrder(SVec2Float* iX, MatXFloat* odInvDetdX = NULL);
	static void U2ndOrder(SVec2Float* iX, SVec2Float* iP,
							MatXFloat iD, _T_FLOAT iThick,
							MatXFloat& odUdX, MatXFloat& odUdP,
							MatXFloat& od2UdX2, MatXFloat& od2UdXdP, MatXFloat& od2UdP2);
	static void VM2ndOrder(SVec2Float* iX, SVec2Float* iP,
							MatXFloat iD, _T_FLOAT iThick,
							_T_FLOAT& oVM,
							MatXFloat& odVMdXP,
							MatXFloat& od2VMdXP2);

	std::vector<SVec2Float> mX0;
	std::vector<SVec2Float> mPV;
	std::vector<SVec2Float> mXV;
	std::vector<SVec2Float> mExtForceP;
	//std::vector<Eigen::Triplet<_T_FLOAT> > mdExtForceP;
	std::vector<SVec2Float> mExtForceX;
	//std::vector<Eigen::Triplet<_T_FLOAT> > mdExtForceX;
	std::vector<Eigen::Triplet<_T_FLOAT> > mdExtForce;
	std::vector<_T_FLOAT> mMass;
	std::vector<std::set<int> > mVtxConnection;
	std::vector<std::set<int> > mVtxConnectionBound;
	std::vector<std::set<int> > mVtxConnectionHV[2];
	std::vector<std::set<int> > mVtxControls;
	std::vector<std::set<int> > mVtxNeighbour;
	std::vector<SVec2Float> mLaplacian;
	std::vector<SVec2Float> mBoundLaplacian;
	std::vector<_T_FLOAT> mLapWeight;
	std::vector<std::map<int, SVec2Float> > mEdges0;
	_T_FLOAT mDensity;
	_T_FLOAT mXDeformWeight;
	_T_FLOAT mLaplacianWeight;
	_T_FLOAT mVecLaplacianWeight;
	_T_FLOAT mEdgeCrossWeight;
	_T_FLOAT mGridWeight;
	_T_FLOAT mXMassScale;
	_T_FLOAT mFDamping;
	_T_FLOAT mVDamping;
	_T_FLOAT mMatrixEpsilon;
	static _T_FLOAT msDetEpsilon;

	static std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> > msdBdX;
	static MatXFloat msd2detdX2;
	static MatXFloat msV;

	bool mbFloor;
	_T_FLOAT mFloorHeight;
	_T_FLOAT mFloorForceWeight;
	bool mbGravity;
	bool mbCarryMass;
	bool mbLapMassWeight;
	bool mbXUniformMass;
	bool mbPStatic;
	_T_FLOAT mXUniformMass;
	_T_FLOAT mXSpringForce;
	_T_FLOAT mXStaticStep;

	void Init();
	void BuildVertexConnection();
	_T_FLOAT ComputeSignedArea(int idx);

	template <class _TVec>
	SVec2Float ComputeLaplacianWithNeighbour(int iVtxID, _TVec& iVtx,
		std::set<int>& iNeighbour,
		bool ibMassWeight = false, _T_FLOAT* oWeightSum = NULL)
	{
                using namespace std;
		SVec2Float retVec(0, 0);
		_T_FLOAT weight = 0;
		std::set<int>& connect = iNeighbour;
		for (set<int>::iterator v1=connect.begin();
			v1!=connect.end(); v1++)
		if (ibMassWeight)
		{
			retVec += SVec2Float(iVtx[*v1])*mMass[*v1];
			weight += mMass[*v1];
		}
		else
			retVec += SVec2Float(iVtx[*v1]);
		if (ibMassWeight)
		{
			retVec = SVec2Float(iVtx[iVtxID])-retVec/weight;
			*oWeightSum = weight;
		}
		else
			retVec = SVec2Float(iVtx[iVtxID])-retVec/connect.size();
		return retVec;
	}

	template <class _TVec>
	SVec2Float ComputeLaplacian(int iVtxID, _TVec& iVtx,
		bool ibMassWeight = false, _T_FLOAT* oWeightSum = NULL)
	{
		return ComputeLaplacianWithNeighbour(iVtxID, iVtx, mVtxConnection[iVtxID], ibMassWeight, oWeightSum);
	}

	void ComputeExtForce();
	void ComputeXForce();
	void ComputeXLapVec(int iIdx, SVec2Float iL0, std::set<int>& iConnection);
	void ComputeXLap(int iIdx, _T_FLOAT iL0, std::set<int>& iConnection);
	void ComputeXEdgeCross(int iIdx, std::set<int>& iConnection);
	void ComputeXControlForce();
	void ComputeXControlForceB();
	void ComputeXVecLaplacianForce();
	void ComputeXLaplacianForceWithNeighbour(std::vector<std::set<int> >& iNeighbourList);
	void ComputeXLaplacianForce();
	void ComputeGridForce();
	void ComputeXShapeMatching();
	void ComputePForce();
	void ComputeVMForce();
	void ComputeXSpringForce();
	
	_T_FLOAT mVMBound;
	_T_FLOAT mVMWeight;

	bool mbStatic;

	void PStep(_T_FLOAT iDT);
	void XPStep(_T_FLOAT iDT);
};
