/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "../SSSLib/SSSFunc.h"
#include "../SSSLib/SSSUtil.h"
#include "../matrix-factor-master/source/solver.h"
#ifdef _IND_SYM_SOLVE
#include <unsupported/Eigen/SparseExtra>
#endif


using namespace std;
using namespace SSSLib;

_T_FLOAT PhysImpSolve3D::XPStep(_T_FLOAT iDT, bool ibPOnly, bool ibKKT)
{
	_T_FLOAT vconverge = 0;
	_T_FLOAT vconvergeNorm = 0;

	_T_FLOAT energy = 0;

	_T_FLOAT damping = pow(mFDamping, iDT);
	_T_FLOAT vdamping = pow(mVDamping, iDT);

	if (mFDamping==0)
		damping = 0;
	if (mVDamping==0)
		vdamping = 0;

	_T_FLOAT dpweight = iDT;
	_T_FLOAT dp2weight = sqr(iDT);

	if (mbStatic)
	{
		iDT = 1.0;
		damping = 1.0;
		vdamping = 0;
		dpweight = 1.0;
		dp2weight = 1.0;
	}

	int poffset = int(mX0.size())*3;
	int kktoffset = poffset*2;
	if (mPhase==1)
		kktoffset++;

	typedef Eigen::Triplet<_T_FLOAT> T;
	std::vector<T> tripleList;
	MatXFloat b;
	//if (ibPOnly)
	if (ibKKT)
		b = MatXFloat::Zero(kktoffset+mKKT.size()+mFixedVtx.size()*3, 1);
	else
		b = MatXFloat::Zero(kktoffset+mFixedVtx.size()*3, 1);

#define _NOT_FIX_CHECK(r) ((r)<poffset || mFixedVtx.size()==0 || !mFixedVtx.count(((r)-poffset)/3))

	for (int i=0; i<(int)mdExtForce.size(); i++)
	{
		T extXTri = mdExtForce[i];
		if (!ibPOnly || extXTri.row()>=poffset)
		{
			//if (extXTri.row()>=poffset && extXTri.col()>=poffset)
			// fixed point check
			//if (_NOT_FIX_CHECK(extXTri.row()))
				tripleList.push_back(T(extXTri.row(),
						extXTri.col(),
						dp2weight*extXTri.value()));
		}
		/*if (extXTri.row()<poffset)
			tripleList.push_back(T(extXTri.row(), extXTri.col(), -sqr(iDT)*damping*extXTri.value()));
		else
			tripleList.push_back(T(extXTri.row(), extXTri.col(), -dp2weight*extXTri.value()));*/
	}

	// for debug
	for (int i=0; i<(int)mP0.size(); i++)
		energy += mMass[i]*0.5*mPV[i].SqrLength();

	//cout<<"Energy "<<energy<<endl;

	// above are physics simulation code
#ifdef _OUTPUT_LOG
	cout<<"b\n"<<b<<endl;
#endif

	for (int i=0; i<(int)mX0.size(); i++)
	for (int c=0; c<3; c++)
	{
		tripleList.push_back(T(i*3+c, i*3+c, mXUniformMass*mXMassScale));
		if (!mbStatic)
			b(i*3+c, 0) += vdamping*mXUniformMass*mXV[i][c]*mXMassScale;
		if (!ibPOnly)
			b(i*3+c, 0) += dpweight*mExtForceX[i][c];

		if (!mbStatic)
		//if (_NOT_FIX_CHECK(i*3+c+poffset))
			tripleList.push_back(T(i*3+c+poffset, i*3+c+poffset, mMass[i]));
		b(i*3+c+poffset, 0) += vdamping*mMass[i]*mPV[i][c];
		b(i*3+c+poffset, 0) += dpweight*mExtForceP[i][c];
	}

	if (mPhase==1)
	{
		// when phase==1, compute the gradient/hessian of $s - 1/t * \sum log(s-f_i(x))$
		if (!ibPOnly)
		{
			b(kktoffset-1, 0) = dpweight*(mdS-mNewtonSScale*mXMassScale);
			b(kktoffset-1, 0) += vdamping*mXUniformMass*mVS*mXMassScale;
			tripleList.push_back(T(kktoffset-1, kktoffset-1, mXUniformMass*mXMassScale));
			//b(kktoffset-1, 0) = dpweight*(mdS);
		}
		else
		{
			tripleList.push_back(T(kktoffset-1, kktoffset-1, mXUniformMass*mXMassScale));
			b(kktoffset-1, 0) = 0;
		}
	}

	// KKT
	if (ibKKT)
	for (int i=0; i<(int)mKKT.size(); i++)
	{
		const KKTEle& ele = mKKT[i];
		int nonZero = 0;
		for (int c=0; c<(int)ele.mACoes.size(); c++)
		{
			if (ele.mACoes[c].second*dp2weight!=0)
				nonZero++;
			tripleList.push_back(T(i+kktoffset, ele.mACoes[c].first, ele.mACoes[c].second*dp2weight));
			tripleList.push_back(T(ele.mACoes[c].first, i+kktoffset, ele.mACoes[c].second*dp2weight));
		}
		assert(nonZero>0);
		b(kktoffset+i, 0) = -ele.mB*dpweight;
	}
	// fixed points
	int idx = kktoffset+mKKT.size();
	for (set<int>::iterator ite = mFixedVtx.begin();
		ite != mFixedVtx.end();
		ite++)
		for (int c=0; c<3; c++)
		{
			int vtxIdx = *ite*3+c;
			b(idx, 0) = -(mX0[*ite][c]-mP0[*ite][c])*dpweight;

			tripleList.push_back(T(idx, vtxIdx, dp2weight));
			tripleList.push_back(T(vtxIdx, idx, dp2weight));

			tripleList.push_back(T(idx, vtxIdx+poffset, -dp2weight));
			tripleList.push_back(T(vtxIdx+poffset, idx, -dp2weight));
			idx++;
		}


	// solve newton
	using namespace Eigen;
	SparseMatrix<_T_FLOAT> A;

	if (ibKKT)
		A = SparseMatrix<_T_FLOAT>(kktoffset+mKKT.size()+mFixedVtx.size()*3, kktoffset+mKKT.size()+mFixedVtx.size()*3);
	else
		A = SparseMatrix<_T_FLOAT>(kktoffset+mFixedVtx.size()*3, kktoffset+mFixedVtx.size()*3);
	A.setFromTriplets(tripleList.begin(), tripleList.end());
	A.makeCompressed();


	int timebeg = _GET_TICK_COUNT();

	//SimplicialLDLT<SparseMatrix<_T_FLOAT> > solver;
	//SparseQR<SparseMatrix<_T_FLOAT>, COLAMDOrdering<int> > solver;
	SparseLU<SparseMatrix<_T_FLOAT>, COLAMDOrdering<int> > solver;
	//ConjugateGradient<SparseMatrix<_T_FLOAT> > solver;
	//BiCGSTAB<SparseMatrix<_T_FLOAT>, Eigen::IncompleteLUT<_T_FLOAT> > solver;
	solver.compute(A);

	//solver.isSymmetric(true);
	
	//solver.analyzePattern(A);
	/*if (solver.lastErrorMessage().length()>0)
	{
		cout<<solver.lastErrorMessage();
		cout<<" X len: "<<kktoffset
			<<" kkt num "<<mKKT.size()
			<<" b size "<<b.rows()
			<<endl;
		lbPhysValid = false;
		return -1;
	}*/
	//solver.factorize(A);
	/*if (solver.lastErrorMessage().length()>0)
	{
		cout<<solver.lastErrorMessage();
		cout<<" X len: "<<kktoffset
			<<" kkt num "<<mKKT.size()
			<<" b size "<<b.rows()
			<<" A size "<<A.rows()<<" "<<A.cols()
			<<endl;
		lbPhysValid = false;
		return -1;
	}*/
	MatXFloat x = solver.solve(b);

	//cout<<"Solver time "<<_GET_TICK_COUNT()-timebeg<<endl;
	//if (solver.lastErrorMessage().length()>0)
	//	cout<<solver.lastErrorMessage();
	
	//MatXFloat x;
	//SIMSolve(A, b, x);

	// for debug
	//{
	//	ofstream debugFileA[4];
	//	ofstream debugFileB("debugSolverB.txt");
	//	//int interestPts[4] = {0, 26, 78, 90};
	//	// cube
	//	int interestPts[4] = {11, 0, 21, 1};
	//	//int interestPts[4] = {4, 28, 49, 261};
	//	for (int i=0; i<4; i++)
	//	{
	//		char nameBuffer[1000];
	//		sprintf_s<1000>(nameBuffer, "debugSolverA%d.txt", i);
	//		debugFileA[i].open(nameBuffer);
	//	}

	//	//for (vector<T>::iterator ite = tripleList.begin();
	//	//	ite != tripleList.end();
	//	//	ite++)
	//	//{
	//	//	for (int i=0; i<4; i++)
	//	//	if (ite->row()%3==interestPts[i])
	//	//	{
	//	//		debugFileA[i]<<ite->value()<<endl;
	//	//		break;
	//	//	}
	//	//}


	//	for (int offset=0; offset<=1; offset++)
	//	{
	//		for (int i=0; i<4; i++)
	//		{
	//			for (int c=0; c<3; c++)
	//			{
	//				for (SparseMatrix<double>::InnerIterator ite(A, interestPts[i]*3+c+poffset*offset);
	//					ite;
	//					ite++)
	//				{
	//					debugFileA[i]<<ite.value()<<"\t";
	//				}
	//				debugFileA[i]<<endl;
	//				debugFileB<<b(interestPts[i]*3+c+poffset*offset, 0)<<"\t";
	//			}
	//			debugFileB<<endl;
	//			debugFileA[i]<<endl;
	//		}
	//		
	//		debugFileB<<endl;
	//	}
	//	debugFileB<<"\n\nP\n";
	//	for (int offset=0; offset<=1; offset++)
	//	{
	//		for (int i=0; i<4; i++)
	//		{
	//			for (int c=0; c<3; c++)
	//				debugFileB<<x(interestPts[i]*3+c+poffset*offset, 0)<<"\t";
	//			debugFileB<<endl;
	//		}
	//		debugFileB<<endl;
	//	}
	//}


	for (int i=0; i<(int)mPV.size(); i++)
	{
		for (int j=0; j<3; j++)
		{
			mXV[i][j] = x(i*3+j,0);
			mPV[i][j] = x(i*3+j+poffset,0);
			//mXV[i][j] = b(i*3+j,0)/mXUniformMass;
			//mPV[i][j] = b(i*3+j+poffset,0)/mMass[i];
			vconverge += sqr(mPV[i][j]);
			vconvergeNorm += sqr(mX0[i][j]);
		}
		if (ibPOnly)
		{
			//assert(mXV[i].Length()<mEdgeLengthEpsilon);
		}
		else
			mX0[i] += mXV[i]*iDT*mNewtonStep;
		mP0[i] += mPV[i]*iDT*mNewtonStep;
	}

	if (mPhase==-1)
	{
		// update the status
		ComputeVMForce(true, ContVM);
		UpdateVMBound(mVMBound, -1, false, true);
	}
	if (mPhase==1)
	{
		// update s (which is very tricky)

		//mS += x(kktoffset-1, 0)*iDT;
		mVS = x(kktoffset-1, 0);
		_T_FLOAT newS = mVS*iDT*mNewtonStep+mS;
		_T_FLOAT oldS = mS;

		lNewtonEnergy += mS*mXMassScale*mNewtonSScale;

		ComputeVMForce(true, ContVM);
		UpdateVMBound(mVMBound, newS, false);

		cout<<"NewS\t"<<newS
			<<"\tS\t"<<mS
			<<"\tNewtonEnergy\t"<<lNewtonEnergy
			<<"\tT\t"<<mNewtonT
			<<endl;
		bool bConv = true;
		_T_FLOAT crit = fabs(mS*1e-3);
		_T_FLOAT converge = (mS-oldS)/iDT*b(kktoffset-1, 0)/2;
		if (converge>crit)
			bConv = false;
		crit = 1e-3;

		// convergence check
		if (bConv)
		for (int i=0; i<kktoffset-1; i++)
		{
			_T_FLOAT converge = x(i, 0)*b(i, 0)/2;
			if (converge>crit)
			{
				bConv = false;
				break;
			}
		}
		//assert(converge>=0);

		// for debug
#if 1
		//if (newS<mS)
		//	mNewtonT /= mNewtonMu;
		//else
		if (bConv /*&& mS==newS*/)
		{
			mNewtonT *= mNewtonMu;
			//cout<<"New T: "<<mNewtonT<<endl;
		}
#endif
		//if (mPhase==-1)
		//	mXV.assign(mXV.size(), Vec3d(0, 0, 0));

		if (mS<=0)
		{
			cout<<"Converging\n";
			mVS = 0;
			mS = 0;
			if (mNewtonT>mNewtonTCrit)
				mPhase = -1;
		}
	}

#ifdef _OUTPUT_LOG
	cout<<"PV\n";
	for (int i=0; i<(int)mPV.size(); i++)
	{
		for (int j=0; j<3; j++)
			cout<<mPV[i][j]<<"\t";
		cout<<endl;
	}
#endif
	return sqrt(vconverge/vconvergeNorm);
}

void PhysImpSolve3D::AddInequalityConstraint(_T_FLOAT iF,
								int iN,
								int* iIdx,
								int* ibOptimize,
								const _T_FLOAT* iGrad,
								const MatXFloat& iHess)
{
	// when phase==1, compute the gradient/hessian of $s - 1/t * \sum log(s-f_i(x))$
	// when phase==-2, compute jacob only

	if (mPhase==-2)
	{
		if (mbUsePenalty)
		{
			_T_FLOAT weight = 1;
			if (mbAugmentedLag)
				weight = mMu/2;
			if (lOutputConst)
				lConst[lLapSolContNum] = Max<_T_FLOAT>(iF, 0);
			if (iF<=0)
			{
				if (mbAugmentedLag)
					lLapSolContNum++;
				return;
			}
			mFuncVal += weight*sqr(iF);

			if (mbAugmentedLag)
				mFuncVal -= mLapSolLambda[lLapSolContNum]*iF;

			for (int i=0; i<iN; i++)
			{
				_T_FLOAT* ptr;
				int idx0 = iIdx[i];

				mGrad[iIdx[i]] += 2*weight*iGrad[i]*iF;
				if (mbAugmentedLag)
					mGrad[iIdx[i]] -= mLapSolLambda[lLapSolContNum]*iGrad[i];

				if (!lFirstOrderOnly)
				for (int j=0; j<iN; j++)
				{
					int idx1 = iIdx[j];
					Vec2i idx(idx0, idx1);
					mHess[idx] = mHess[idx]+2*weight*iGrad[i]*iGrad[j]+2*weight*iF*iHess(i, j);
					if (mbAugmentedLag)
						mHess[idx] -= mLapSolLambda[lLapSolContNum]*iHess(i, j);
				}
			}

			if (mbAugmentedLag)
				lLapSolContNum++;
		}
		else
		{
			mIneqJacob.push_back(KKTEle());
			KKTEle& jacob = mIneqJacob[mIneqJacob.size()-1];
			jacob.mB = iF;
			jacob.mACoes.resize(iN);
			for (int i=0; i<iN; i++)
			{
				jacob.mACoes[i].first = iIdx[i];
				jacob.mACoes[i].second = iGrad[i];
			}
			//jacob.mIdx.assign(iIdx, iIdx+iN);
			//jacob.mGrad.assign(iGrad, iGrad+iN);

			if (mbComputeLapSolHessian)
			{
				for (int i=0; i<iN; i++)
				for (int j=0; j<iN; j++)
				{
					Vec2i idx(iIdx[i], iIdx[j]);
					mLapSolHessian[idx] = mLapSolHessian[idx]
								+iHess(i, j)*mLapSolLambda[lLapSolHessianCounter];
				}
				lLapSolHessianCounter++;
			}
		}
		return;
	}

	typedef Eigen::Triplet<_T_FLOAT> T;
	int soffset = mX0.size()*6;
	_T_FLOAT s = 0;
	_T_FLOAT invval = -1.0/iF;
	_T_FLOAT scale = 1.0;

	if (mPhase==1)
	{
		scale = 1/mNewtonT;
		_T_FLOAT dS = 1.0;
		s = mS*mXMassScale;
		lNewtonEnergy += -scale*log(s-iF);
	// dPhidS
		invval = 1.0/(s-iF);
		assert(invval>0);
		mdS += invval*scale;
		mdExtForce.push_back(T(soffset, soffset, sqr(invval*dS)*scale));
		for (int i=0; i<iN; i++)
		{
			_T_FLOAT val = -sqr(invval)*iGrad[i]*dS*scale;
			mdExtForce.push_back(T(soffset, iIdx[i], val));   
			if (ibOptimize && ibOptimize[i])
				mdExtForce.push_back(T(iIdx[i], soffset, val));
		}
	}

	for (int i=0; i<iN; i++)
	if (ibOptimize[i])
	{
		_T_FLOAT* ptr;
		int idx0 = iIdx[i];

		if (idx0<(int)mX0.size()*3)
			ptr = &mExtForceX[idx0/3][idx0%3];
		else
			ptr = &mExtForceP[idx0/3-mX0.size()][idx0%3];
		if (mPhase==-1)
			*ptr += -iGrad[i]*scale;
		else
			*ptr += -invval*iGrad[i]*scale;

		for (int j=0; j<iN; j++)
		{
			int idx1 = iIdx[j];
			_T_FLOAT val;
			if (mPhase==-1)
				val = iHess(i, j)*scale;
			else
				val = (sqr(invval)*iGrad[i]*iGrad[j]+invval*iHess(i, j))*scale;
			//val = -val;

			mdExtForce.push_back(T(idx0, idx1, val));
		}
	}
}

void PhysImpSolve3D::BuildVtxSparseStruct()
{
	mVtxSparseStruct.clear();
	mVtxSparseStruct.resize(mX0.size());
	for (int h=0; h<(int)mpVolMesh->mHexs.size(); h++)
	{
		SSSVolMesh::_THex hex = mpVolMesh->mHexs[h];
		for (int i=0; i<8; i++)
		for (int j=0; j<8; j++)
			mVtxSparseStruct[hex[i]].insert(hex[j]);
	}

	for (int t=0; t<(int)mpVolMesh->mTets.size(); t++)
	{
		Vec4i tet = mpVolMesh->mTets[t];
		for (int i=0; i<4; i++)
		for (int j=0; j<4; j++)
			mVtxSparseStruct[tet[i]].insert(tet[j]);
	}

	mNumVtxSparse = 0;
	for (int i=0; i<(int)mX0.size(); i++)
		mNumVtxSparse += mVtxSparseStruct[i].size();
}

void PhysImpSolve3D::DumpState(const char* oFileName)
{
	ofstream outFile(oFileName, ios_base::binary);
	int num = mX0.size();
	outFile.write((char*)&num, sizeof(int));
	outFile.write((char*)mX0[0].val, sizeof(SVec3Float)*mX0.size());
	outFile.write((char*)mP0[0].val, sizeof(SVec3Float)*mP0.size());
	outFile.write((char*)mXV[0].val, sizeof(SVec3Float)*mXV.size());
	outFile.write((char*)mPV[0].val, sizeof(SVec3Float)*mPV.size());
}

void PhysImpSolve3D::LoadState(const char* iFileName)
{
	ifstream inFile(iFileName, ios_base::binary);
	int num;
	inFile.read((char*)&num, sizeof(int));
	assert(num == (int)mX0.size());
	inFile.read((char*)mX0[0].val, sizeof(SVec3Float)*mX0.size());
	inFile.read((char*)mP0[0].val, sizeof(SVec3Float)*mP0.size());
	inFile.read((char*)mXV[0].val, sizeof(SVec3Float)*mXV.size());
	inFile.read((char*)mPV[0].val, sizeof(SVec3Float)*mPV.size());

}


void PhysImpSolve3D::SolveLinear(const std::vector<KKTEle>& iSystem,
							_T_FLOAT iDiagnalWeight,
							int iIdxOffset,
							_T_FLOAT& oPenalty,
							std::vector<_T_FLOAT>& oGrad)
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	using namespace Eigen;
	std::vector<T> tripleList;
	MatXFloat b;
	//if (ibPOnly)
	b = MatXFloat::Zero(iSystem.size(), 1);

	oPenalty = 0;

	for (int i=0; i<(int)iSystem.size(); i++)
	{
		const KKTEle& ele = iSystem[i];
		b(i, 0) = ele.mB;
		oPenalty += 0.5*sqr(ele.mB);
		for (int j=0; j<(int)ele.mACoes.size(); j++)
		if (ele.mACoes[j].first>=iIdxOffset)
		{
			tripleList.push_back(T(i, ele.mACoes[j].first-iIdxOffset, ele.mACoes[j].second));
		}

		if (iDiagnalWeight!=0)
			tripleList.push_back(T(i, i, iDiagnalWeight));
	}

	SparseMatrix<_T_FLOAT> A(iSystem.size(), iSystem.size());
	A.setFromTriplets(tripleList.begin(), tripleList.end());

	SparseLU<SparseMatrix<_T_FLOAT>, COLAMDOrdering<int> > solver;
	//ConjugateGradient<SparseMatrix<_T_FLOAT> > solver;
	solver.analyzePattern(A);
	if (solver.lastErrorMessage().length()>0)
		cout<<solver.lastErrorMessage();
	solver.factorize(A);
	if (solver.lastErrorMessage().length()>0)
		cout<<solver.lastErrorMessage();
	MatXFloat x = solver.solve(b);
	if (solver.lastErrorMessage().length()>0)
		cout<<solver.lastErrorMessage();

	oGrad.resize(iSystem.size());
	for (int i=0; i<(int)oGrad.size(); i++)
		oGrad[i] = x(i, 0);
}

void PhysImpSolve3D::KKTSolve(const Eigen::SparseMatrix<_T_FLOAT>& iH,
						const Eigen::SparseMatrix<_T_FLOAT>& iA,
						const MatXFloat& ig,
						const MatXFloat& ih,
						MatXFloat& ov,
						MatXFloat& ow)
{
	using namespace Eigen;
	SparseLU<SparseMatrix<_T_FLOAT>, COLAMDOrdering<int> > solver;
	solver.analyzePattern(iH);
	if (solver.lastErrorMessage().length()>0)
		cout<<solver.lastErrorMessage();
	solver.factorize(iH);
	if (solver.lastErrorMessage().length()>0)
		cout<<solver.lastErrorMessage();

	Eigen::SparseMatrix<_T_FLOAT> At = iA.transpose();
	Eigen::SparseMatrix<_T_FLOAT> HinvAt = solver.solve(At);
	Eigen::SparseMatrix<_T_FLOAT> S = -iA*HinvAt;
	MatXFloat Hinvg = solver.solve(ig);
	MatXFloat AHinvg = iA*Hinvg;
	MatXFloat AHinvg_h = AHinvg-ih;

	{
		SparseLU<SparseMatrix<_T_FLOAT>, COLAMDOrdering<int> > solver2;
		solver2.analyzePattern(S);
		if (solver2.lastErrorMessage().length()>0)
			cout<<solver2.lastErrorMessage();
		solver2.factorize(S);
		if (solver2.lastErrorMessage().length()>0)
			cout<<solver2.lastErrorMessage();
		ow = solver2.solve(AHinvg_h);
	}

	MatXFloat Atw = At*ow;
	MatXFloat _Atw_g = -Atw-ig;

	{
		SparseLU<SparseMatrix<_T_FLOAT>, COLAMDOrdering<int> > solver3;
		solver3.analyzePattern(iH);
		if (solver3.lastErrorMessage().length()>0)
			cout<<solver3.lastErrorMessage();
		solver3.factorize(iH);
		if (solver3.lastErrorMessage().length()>0)
			cout<<solver3.lastErrorMessage();
		ov = solver3.solve(_Atw_g);
	}
}

void PhysImpSolve3D::AddSpeedTerm()
{
	typedef Eigen::Triplet<_T_FLOAT> T;

	for (int i=0; i<(int)mX0.size()*3; i++)
	{
		mHess[Vec2i(i, i)] += mSpeedWeight;
		//mGrad[i] += -2*mX0[i/3][i%3]*mSpeedWeight;
	}
}

#ifdef _IND_SYM_SOLVE

void PhysImpSolve3D::SIMSolve(Eigen::SparseMatrix<_T_FLOAT>& iA,
		const MatXFloat& ib,
		MatXFloat& ox)
{
	solver<_T_FLOAT> simSolver;
	simSolver.set_reorder_scheme("amd");
	simSolver.set_equil(true);
	simSolver.set_pivot("rook");
	/*
	Eigen::saveMarket(iA, "tmp.mtx", Eigen::Symmetric);
	simSolver.load("tmp.mtx");
	int nncount = 0;
	int fillFactor = 1;
	for (int k=0; k<iA.outerSize(); k++)
	{
		int innerSize = 0;
		for (Eigen::SparseMatrix<_T_FLOAT>::InnerIterator ite(iA, k);
			ite; ++ite)
		if (ite.row()>=ite.col())
		{
			//simA.m_idx[ite.col()].push_back(ite.row());
			//simA.m_x[ite.col()].push_back(ite.value());
			nncount++;
			innerSize ++;
			//if (ite.col()!=ite.row())
			//	simA.list[ite.row()].push_back(ite.col());
		}
		fillFactor = Max(fillFactor, innerSize);
	}

	*/

	lilc_matrix<_T_FLOAT>& simA = simSolver.A;
	// convert the A matrix
	
	simA.resize(iA.rows(), iA.cols());
	std::fill(simA.first.begin(), simA.first.end(), 0);
	int nncount = 0;
	int fillFactor = 1;
	for (int k=0; k<iA.outerSize(); k++)
	{
		int innerSize = 0;
		for (Eigen::SparseMatrix<_T_FLOAT>::InnerIterator ite(iA, k);
			ite; ++ite)
		if (ite.row()>=ite.col())
		{
			simA.m_idx[ite.col()].push_back(ite.row());
			simA.m_x[ite.col()].push_back(ite.value());
			nncount++;
			innerSize ++;
			if (ite.col()!=ite.row())
				simA.list[ite.row()].push_back(ite.col());
		}
		fillFactor = Max(fillFactor, innerSize);
	}
	//simA.n_cols = iA.cols();
	//simA.n_rows = iA.rows();
	simA.nnz_count = nncount;

	vector<_T_FLOAT> b(ib.rows());
	_T_FLOAT b_l1norm = 0;
	for (int i=0; i<ib.rows(); i++)
	{
		b[i] = ib(i, 0);
		b_l1norm += fabs(b[i]);
	}

	simSolver.set_rhs(b);
	simSolver.solve((fillFactor+1.0)/(double)iA.rows(), 0, 0.5, 10000, 1e-13);
	//simSolver.solve(1, 0, 1, 1000, 1e-10, b_l1norm/ib.rows()*1e-3);
	//simSolver.solve(1, 0, 0, 10000, 1e-13, 0);

	// copy back
	
	ox = MatXFloat(ib.rows(), ib.cols());
	for (int i=0; i<ox.rows(); i++)
		ox(i, 0) = simSolver.sol_vec[i];
}

#endif
