/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "TestConfig.h"
#include "PhysImpSolve3D_util.h"
#include "../SSSLib/SSSUtil.h"

using namespace std;
using namespace SSSLib;

//#define _NO_PHYSCS
//#define _IGNORE_NEGATIVE_VOL

namespace
{
	_T_FLOAT lConvergeVal = -1;
	_T_FLOAT lNegConvergeVal = -1;
}

_T_FLOAT PhysImpSolve3D::CalScoreWithPen()
{
	lbPhysValid = true;

	if (mJacPercentBound>0)
		ComputeVMForce(true, ContVolume);
	ComputeVMForce(true, ContVM);

	if (!lbPhysValid)
		return numeric_limits<_T_FLOAT>::infinity();

	_T_FLOAT retVal = 0;


	//assert(retVal>=0);

	_T_FLOAT weight = 1;
	if (mbAugmentedLag)
		weight = mMu/2;
	_T_FLOAT jacWeight = 0;
	if (mbUsePenalty)
	{
		
		for (int h = 0; h<(int)mVMHex.size(); h++)
		{
			SmallVec<_T_FLOAT, 8>& hexVM = mVMHex[h];
			for (int v=0; v<8; v++)
			{
				if (hexVM[v]>mVMBound)
					retVal += weight*sqr((hexVM[v]-mVMBound)*mVMWeight);
				if (mJacHex[h][v]>0)
					jacWeight += weight*sqr(mJacHex[h][v]*mVolumeWeight);
			}
		}
		for (int t=0; t<(int)mVM.size(); t++)
		{
			if (mVM[t]>mVMBound)
				retVal += weight*sqr((mVM[t]-mVMBound)*mVMWeight);
			if (mJac[t]>0)
				jacWeight += weight*sqr(mJac[t]*mVolumeWeight);
		}
		retVal += jacWeight;
	}

	//cout<<"VM Val "<<retVal
	//	<<"\tJac Val "<<jacWeight<<endl;

	for (int i=0; i<(int)mX0.size()*3; i++)
	{
		_T_FLOAT val = mX0[i/3][i%3]-mXRef[i/3][i%3];
		retVal += sqr(val)*mShapeSimWeight;
	}
	//assert(retVal>=0);
#if defined(_USE_SIM_LAP) && !defined(_USE_SIM_LAP_AS_KKT)
	retVal += EvaluateLapSim();
	//assert(retVal>=0);
#endif

	retVal += ComputeHexVolumeForce(true);
	//assert(retVal>=0);

	//cout<<"Jac percentage "<<jacWeight/retVal<<endl;

	return retVal;
}

_T_FLOAT PhysImpSolve3D::AdvanceAndEvaluate(_T_FLOAT iStep)
{
	bool valid = EqKKTAdvance(iStep);
	if (!valid)
	{
		cout<<"Warning: advance faile, try tune up laplacian weight, step "<<iStep<<endl;
		return numeric_limits<_T_FLOAT>::infinity();
	}
	return CalScoreWithPen();
}

void PhysImpSolve3D::RunPhysics()
{
	mbStatic = false;
	_T_FLOAT conv;
	int iteNum=0;
	do
	{
		if (iteNum>30)
		{
			lbPhysValid = false;
			cout<<"Warning: Phys Sim Not Converge!\n";
			break;
		}

		lbPhysValid = true;
		mdExtForce.clear();
		mKKT.clear();
		ComputePhysForce();
		//OptExportLinear();
#if !defined(_USE_SIM_LAP) || defined(_EXT_SIM_LAP_ONLY)
		//ComputeLapForce();
#endif
#ifdef _USE_SIM_LAP_AS_KKT
		LapAddToKKT(1.0);
#endif

#ifdef _IGNORE_NEGATIVE_VOL
		lbPhysValid = true;
#endif

		if (lbPhysValid)
			conv = XPStep(mSimTimeStep, true, true);

		iteNum++;
	}
	while ((conv>1e-3 && lbPhysValid) || iteNum<=3);
}

bool PhysImpSolve3D::EqKKTAdvance(_T_FLOAT iStep)
{
	for (int i=0; i<(int)mX0.size(); i++)
	{
		mX0[i] += mXAdv[i]*iStep;
		mP0[i] += mPAdv[i]*iStep;
	}
	RunPhysics();

	if (!lbPhysValid)
	{
		Rollback(mStepBK);
		//mbBackupInit = false;
		return false;
	}
	return true;
}

bool PhysImpSolve3D::LineSearch(_T_FLOAT iRefT, _T_FLOAT iMinT, _T_FLOAT iMaxT, _T_FLOAT& oTotalT)
{
	_T_FLOAT valBefore = mFuncVal;
	BKStates searchBK;
	Backup(searchBK);
	oTotalT = 0;
	//assert(fabs(mFuncVal-CalScoreWithPen())<1e-3);

	_T_FLOAT tstep = iRefT;
	
	_T_FLOAT pst[3];
	_T_FLOAT val[3];
	val[0] = mFuncVal;
	val[1] = AdvanceAndEvaluate(tstep);
	if (val[0]<=val[1])
	{
		pst[0] = 0;
		pst[1] = 0;
		pst[2] = tstep;
		val[2] = val[1];
		val[1] = val[0];
		tstep = iRefT/2;
		Rollback(mStepBK);
	}
	else
	{
		pst[0] = 0;
		pst[1] = tstep;
		pst[2] = tstep*2;
		mFuncVal = val[1];
		Backup(mStepBK);
		oTotalT += tstep;
		val[2] = AdvanceAndEvaluate(tstep);
		while (val[2]<val[1])
		{
			mFuncVal = val[2];
			Backup(mStepBK);
			oTotalT += tstep;
			if (oTotalT>iMaxT)
			{
				val[0] = val[1];
				val[1] = val[2];

				pst[0] = pst[1];
				pst[1] = pst[2];
				return true;
			}
			val[0] = val[1];
			val[1] = val[2];
			tstep *= 2;
			val[2] = AdvanceAndEvaluate(tstep);
			pst[0] = pst[1];
			pst[1] = pst[2];
			pst[2] += tstep;
		}
	}

	_T_FLOAT maxStep = 0;
	do
	{
		maxStep = 0;
		for (int i=0; i<=2; i+=2)
		{
			_T_FLOAT step = (pst[i]-pst[1])*0.5;
			int i_other = 2-i;
			maxStep = Max(fabs(step), maxStep);
			if (fabs(step)>iMinT)
			{
				_T_FLOAT newVal = AdvanceAndEvaluate(step);
				if (newVal<val[1])
				{
					mFuncVal = newVal;
					Backup(mStepBK);
					oTotalT += step;
					val[i_other] = val[1];
					pst[i_other] = pst[1];
					val[1] = newVal;
					pst[1] = oTotalT;
				}
				else
				{
					Rollback(mStepBK);
					val[i] = newVal;
					pst[i] = oTotalT+step;
				}
			}
		}
	} while (maxStep>iMinT && (oTotalT==0 || maxStep/oTotalT>0.1));
	//if (oTotalT!=0 && valBefore-mFuncVal<1e-2)
	//{
	//	oTotalT = 0;
	//	Rollback(searchBK);
	//}
	if (oTotalT!=0)
		return true;
	else
		return false;
}

void PhysImpSolve3D::NormalizeKKTStep(bool ibCap)
{
	_T_FLOAT maxStep = mXAdv[0][0];
	for (int i=0; i<(int)mXAdv.size(); i++)
	{
		for (int c=0; c<3; c++)
			maxStep = Max(maxStep, fabs(mXAdv[i][c]));
	}

	_T_FLOAT normFactor = 0;
	if (maxStep!=0)
	{
		if (ibCap)
		{
			if (maxStep<0.1*mMaxDim)
				return;
			else
				normFactor = 0.1*mMaxDim/maxStep;
		}
		else
			normFactor = 1e-2*mMaxDim/maxStep;
	}
	for (int i=0; i<(int)mXAdv.size(); i++)
	{
		mXAdv[i] *= normFactor;
		mPAdv[i] *= normFactor;
	}
}

bool PhysImpSolve3D::SolveEqKKT(_T_FLOAT iStep, _T_FLOAT& iDiagWeight, _T_FLOAT iLambdaRegular, bool ibSkipHess)
{
	assert(mGrad.size()==mX0.size()*6);
	typedef Eigen::Triplet<_T_FLOAT> T;
	std::vector<T> tripleList;

	// inequality constraints treated as soft constraints
	lComputeHess = true;
	//lComputeHess = false;
	mPhase = -2;
	lbPhysValid = true;
	if (ibSkipHess)
	{
		//mGrad.assign(mX0.size()*6, 0);
		mHess.clear();
		for (int i=0; i<(int)mX0.size()*6; i++)
			mHess[Vec2i(i, i)] = iDiagWeight;
	}
	else
	{
		mKKT.clear();
		OptExportLinear();
		OptExport();
	//#if !defined(_USE_SIM_LAP) || defined(_EXT_SIM_LAP_ONLY)
	//	ComputeLapForce();
	//#endif

		//if (mSpeedWeight>0)
		//	AddSpeedTerm();

		// add distance
		for (int i=0; i<(int)mX0.size()*3; i++)
		{
			_T_FLOAT val = mX0[i/3][i%3]-mXRef[i/3][i%3];
			mFuncVal += sqr(val)*mShapeSimWeight;
			mGrad[i] += 2*val*mShapeSimWeight;
			mHess[Vec2i(i, i)] = mHess[Vec2i(i, i)]+2*mShapeSimWeight;

			if (iDiagWeight!=0)
			{
				mHess[Vec2i(i, i)] += iDiagWeight;
				//mHess[Vec2i(i+mX0.size()*3, i+mX0.size()*3)] += iDiagWeight;
			}
		
		}

		//cout<<"VM Val "<<mFuncVal<<endl;

	#if defined(_USE_SIM_LAP) && !defined(_USE_SIM_LAP_AS_KKT)
		mFuncVal += EvaluateLapSim();
	#elif defined(_USE_SIM_LAP_AS_KKT)
		LapAddToKKT(1.0);
	#endif

		cout<<"New Value "<<mFuncVal<<"\t";
		if (!lbPhysValid)
			cout<<"Warning! Invalid Physics!\t";
		if ((!mbBackupInit || mFuncVal<mStepBK.mLastFuncVal) && lbPhysValid)
		{
			if (mFuncVal>0 && mbBackupInit)
				mFuncDropSpeed = (mStepBK.mLastFuncVal-mFuncVal)/mFuncVal;
			else
				mFuncDropSpeed = 0;
			Backup(mStepBK);
			mbBackupInit = true;
		}
		else
		{
			cout<<"Rolling back\n";
			Rollback(mStepBK);
			//mbBackupInit = false;
			return false;
		}
	}
#ifdef _NO_PHYSCS
	mEqJacob.clear();
#endif

	int numKKT = mEqJacob.size()+mKKT.size();
	int kktOffset = mX0.size()*6;

#ifdef _KKT_ELIMINATION
	MatXFloat b = MatXFloat::Zero(kktOffset, 1);
#else
	MatXFloat b = MatXFloat::Zero(kktOffset+numKKT, 1);
#endif
	// soft constraints
	for (int i=0; i<(int)mGrad.size(); i++)
		b(i, 0) = -mGrad[i];
	for (map<Vec2i, _T_FLOAT>::iterator ite = mHess.begin();
		ite != mHess.end();
		ite++)
	{
		tripleList.push_back(T(ite->first[0], ite->first[1], ite->second));
		//assert(ite->second==mHess[Vec2i(ite->first[1], ite->first[0])]);
		//if (fabs(ite->second-mHess[Vec2i(ite->first[1], ite->first[0])])>1e-8)
		//{
		//	_T_FLOAT val = mHess[Vec2i(ite->first[1], ite->first[0])];
		//	cout<<"Assym Hessian!"
		//		<<ite->second<<","<<val<<endl;
		//}
	}

#if defined(_USE_SIM_LAP) && !defined(_USE_SIM_LAP_AS_KKT)
	LapAddToGradient(-1, b.data());
	if (!ibSkipHess)
		PhysImpUtil::PushTriplet(mLapSimMat, 2.0, tripleList);
#endif

	//if (!mbStatic)
	//for (int i=0; i<mX0.size(); i++)
	//for (int c=0; c<3; c++)
	//	tripleList.push_back(T(i*3+c+mX0.size()*3, i*3+c+mX0.size()*3, mMass[i]));

	Eigen::SparseMatrix<_T_FLOAT> H(mX0.size()*6, mX0.size()*6);
	H.setFromTriplets(tripleList.begin(), tripleList.end());
	// check if negative
	
	// hard constraints
#ifdef _KKT_ELIMINATION
	std::vector<T> kktTripleList;
	MatXFloat h = MatXFloat::Zero(numKKT, 1);
	int contPst = 0;
#else
	int contPst = kktOffset;
#endif
	for (int i=0; i<(int)mKKT.size(); i++)
	{
		KKTEle& ele = mKKT[i];
		_T_FLOAT val = 0;

		//for (int j=0; j<(int)ele.mACoes.size(); j++)
		//{
		//	_T_FLOAT* ptr;
		//	int idx = ele.mACoes[j].first;
		//	if (idx<mX0.size()*3)
		//		ptr = mX0[0].val+idx;
		//	else
		//		ptr = mP0[0].val+idx-mX0.size()*3;
		//	val += (*ptr)*ele.mACoes[j].second;
		//}
		//val -= ele.mB;
		val = -ele.mB;
#ifdef _KKT_ELIMINATION
		h(contPst, 0) = val;
#else
		b(contPst, 0) = val;
#endif

		for (int j=0; j<(int)ele.mACoes.size(); j++)
		{
#ifdef _KKT_ELIMINATION
			kktTripleList.push_back(T(contPst, ele.mACoes[j].first, ele.mACoes[j].second));
#else
			tripleList.push_back(T(contPst, ele.mACoes[j].first, ele.mACoes[j].second));
			tripleList.push_back(T(ele.mACoes[j].first, contPst, ele.mACoes[j].second));
#endif
		}

		contPst++;
	}

	for (int i=0; i<(int)mEqJacob.size(); i++)
	{
		KKTEle& ele = mEqJacob[i];
#ifdef _KKT_ELIMINATION
		h(contPst, 0) = ele.mB;
#else
		b(contPst, 0) = -ele.mB;
#endif

		for (int j=0; j<(int)ele.mACoes.size(); j++)
		{
#ifdef _KKT_ELIMINATION
			kktTripleList.push_back(T(contPst, ele.mACoes[j].first, ele.mACoes[j].second));
#else
			tripleList.push_back(T(contPst, ele.mACoes[j].first, ele.mACoes[j].second));
			tripleList.push_back(T(ele.mACoes[j].first, contPst, ele.mACoes[j].second));
#endif
		}

		contPst++;
	}

	//if (iDiagWeight!=0)
	//for (int i=0; i<contPst; i++)
	//	tripleList.push_back(T(i, i, iDiagWeight));

	//assert(contPst == kktOffset+numKKT);

	// solve
	int timeBeg = _GET_TICK_COUNT();
#ifdef _KKT_ELIMINATION
	using namespace Eigen;
	SparseMatrix<_T_FLOAT> H(kktOffset, kktOffset);
	H.setFromTriplets(tripleList.begin(), tripleList.end());

	SparseMatrix<_T_FLOAT> A(numKKT, kktOffset);
	A.setFromTriplets(kktTripleList.begin(), kktTripleList.end());

	MatXFloat x;
	MatXFloat w;

	KKTSolve(H, A, -b, -h, x, w);
#else

	// for debug
	//{
	//	_T_FLOAT maxval = 0;
	//	set<int> cols;
	//	for (int i=0; i<(int)tripleList.size(); i++)
	//	{
	//		_T_FLOAT val = tripleList[i].value();
	//		if (fabs(val)>1e10)
	//			cols.insert(tripleList[i].col());
	//		maxval = Max(fabs(val), maxval);
	//	}

	//	cout<<"\nMissing entries:";
	//	for (int i=0; i<contPst; i++)
	//	if (!cols.count(i))
	//		cout<<i<<"\t";
	//	cout<<endl;
	//	cout<<"Max val "<<maxval<<endl;
	//}

		// for debug
	//{
	//	ofstream debugFileA[4];
	//	ofstream debugFileB("debugB.txt");
	//	debugFileB.precision(10);

	//	for (int i=0; i<4; i++)
	//	{
	//		char nameBuffer[1000];
	//		sprintf_s<1000>(nameBuffer, "debugA%d.txt", i);
	//		debugFileA[i].open(nameBuffer);
	//	}
	//	int interestPts[4] = {0, 26, 78, 90};

	//	for (map<Vec2i, double>::iterator ite = mLapSimMat.begin();
	//		ite != mLapSimMat.end();
	//		ite++)
	//	{
	//		int idx = ite->first[0];
	//		if (idx%3==0)
	//		for (int i=0; i<4; i++)
	//			if (idx/3==interestPts[i])
	//				debugFileA[i]<<ite->second<<"\t";
	//	}
	//	for (int i=0; i<4; i++)
	//		debugFileA[i]<<endl;
	//	//for (int i=0; i<4; i++)
	//	//{
	//	//	MatXFloat col = mLapHalfSimMatEigen.col(interestPts[i]*3);
	//	//	for (int j=0; j<(int)col.rows(); j++)
	//	//	if (col(j, 0)!=0)
	//	//		debugFileA[i]<<col(j, 0)<<"\t";
	//	//	debugFileA[i]<<endl;
	//	//}

	//	for (vector<T>::iterator ite = tripleList.begin();
	//		ite != tripleList.end();
	//		ite++)
	//	{
	//		int idx = ite->row();
	//		if (idx%3==0)
	//		for (int i=0; i<4; i++)
	//			if (idx/3==interestPts[i] && ite->value()!=0)
	//				debugFileA[i]<<ite->value()<<"\n";
	//	}
	//	
	//	for (int i=0; i<4; i++)
	//	{
	//		debugFileB<<i
	//			<<"\tDefault Vol\t"<<mHexDefaultVolume[interestPts[i]]
	//			<<"\tb\t";
	//		for (int c=0; c<3; c++)
	//			debugFileB<<b(interestPts[i]*3+c, 0)<<"\t";
	//		debugFileB<<endl;
	//	}
	//	int a;
	//	a=0;
	//}

	MatXFloat x;

	if (!mbCholeskyAddedMultiple)
	{
		vector<T> newTripleList;

		//MatXFloat bsub = b.topLeftCorner(mX0.size()*6, 1);
		//MatXFloat bsubT(bsub.transpose());
		//if ((bsubT*H*bsub)(0, 0)<0)
		//{
		//	cout<<"Non-positive Hessian, going gradient direction\n";
		//	for (int i=0; i<(int)tripleList.size(); i++)
		//	if (tripleList[i].row()>=mX0.size() || tripleList[i].col()>=mX0.size())
		//		newTripleList.push_back(tripleList[i]);

		//	_T_FLOAT val = 1;
		//	if (iDiagWeight>0)
		//		val = iDiagWeight;
		//	for (int i=0; i<(int)mX0.size()*6; i++)
		//		newTripleList.push_back(T(i, i, val));

		//	for (int i=0; i<(int)mX0.size()*6; i++)
		//		b(i, 0) = 0;
		//}
		//else
		{
			newTripleList = tripleList;

			if (iDiagWeight!=0 && !ibSkipHess)
			for (int i=0; i<(int)mX0.size()*3; i++)
			{
				newTripleList.push_back(T(i, i, iDiagWeight));
				//mHess[Vec2i(i, i)] += iDiagWeight;
				//mHess[Vec2i(i+mX0.size()*3, i+mX0.size()*3)] += iDiagWeight;
			}
		}

		using namespace Eigen;
		SparseMatrix<_T_FLOAT> A(contPst, contPst);
		A.setFromTriplets(newTripleList.begin(), newTripleList.end());
		A.makeCompressed();

		//SimplicialLDLT<SparseMatrix<_T_FLOAT> > solver;

		//BiCGSTAB<SparseMatrix<_T_FLOAT>, Eigen::IncompleteLUT<_T_FLOAT> > solver;
		//ConjugateGradient<SparseMatrix<_T_FLOAT> > solver;
		//solver.compute(A);
		
		//solver.setTolerance(1e-15);
		//solver.setMaxIterations(A.rows()*2);
		//solver.compute(A);
		SparseLU<SparseMatrix<_T_FLOAT>, COLAMDOrdering<int> > solver;
		solver.isSymmetric(true);
		solver.compute(A);
		//solver.analyzePattern(A);
		//if (solver.lastErrorMessage().length()>0)
		//	cout<<solver.lastErrorMessage();
		//solver.factorize(A);
		//if (solver.lastErrorMessage().length()>0)
		//	cout<<solver.lastErrorMessage();
		//if (lXGuess.rows()==b.rows())
		//	x = solver.solveWithGuess(b, lXGuess);
		//else
			x = solver.solve(b);
		//lXGuess = x;
		//if (solver.lastErrorMessage().length()>0)
		//	cout<<solver.lastErrorMessage();
		//SIMSolve(A, b, x);
		//std::cout << "#iterations:     " << solver.iterations() << std::endl;
		//std::cout << "estimated error: " << solver.error()      << std::endl;
	#endif
		cout<<"Linear solver time "<<_GET_TICK_COUNT()-timeBeg<<"\t";
	}
	else
	{
		bool bContinue = true;
		using namespace Eigen;

		//SparseMatrix<_T_FLOAT> A(contPst, contPst);
		//A.setFromTriplets(tripleList.begin(), tripleList.end());
		//A.makeCompressed();

		while (1)
		{
			cout<<"Diag Weight:"<<iDiagWeight<<endl;
			
			vector<T> newTripleList;;
			//for (int i=0; i<contPst; i++)

			{
				newTripleList = tripleList;
				for (int i=0; i<(int)mX0.size()*6; i++)
				{
					newTripleList.push_back(T(i, i, iDiagWeight));
					//newTripleList.push_back(T(i+mX0.size()*3, i+mX0.size()*3, iDiagWeight));
					//mHess[Vec2i(i+mX0.size()*3, i+mX0.size()*3)] += iDiagWeight;
				}
			}
			SparseMatrix<_T_FLOAT> A(contPst, contPst);
			A.setFromTriplets(newTripleList.begin(), newTripleList.end());
			A.makeCompressed();
			

			//BiCGSTAB<SparseMatrix<_T_FLOAT> > solver;
			//SimplicialLLT<SparseMatrix<_T_FLOAT> > solver;
			SimplicialLDLT<SparseMatrix<_T_FLOAT> > solver;
			//solver.setShift(iDiagWeight);
			solver.analyzePattern(A);
			solver.factorize(A);
			if (solver.info()==Success)
			{
				x = solver.solve(b);
				break;
			}
			else
				iDiagWeight *= 2;
		}
	}

	lConvergeVal = 0;
	lNegConvergeVal = 0;

	for (int i=0; i<kktOffset; i++)
	{
		lConvergeVal += Max<_T_FLOAT>(x(i, 0)*b(i, 0), 0);
		lNegConvergeVal += Max<_T_FLOAT>(-x(i, 0)*b(i, 0), 0);
	}

	//if (lConvergeVal<lNegConvergeVal)
	//{
	//	vector<T> newTripleList;

	//	cout<<"Non-positive Hessian, going gradient direction\n";
	//	for (int i=0; i<(int)tripleList.size(); i++)
	//	if (tripleList[i].row()>=mX0.size() || tripleList[i].col()>=mX0.size())
	//		newTripleList.push_back(tripleList[i]);

	//	_T_FLOAT val = 1;
	//	if (iDiagWeight>0)
	//		val = iDiagWeight;
	//	for (int i=0; i<(int)mX0.size()*6; i++)
	//		newTripleList.push_back(T(i, i, val));

	//	using namespace Eigen;
	//	SparseMatrix<_T_FLOAT> A(contPst, contPst);
	//	A.setFromTriplets(newTripleList.begin(), newTripleList.end());
	//	A.makeCompressed();

	//	SparseLU<SparseMatrix<_T_FLOAT>, COLAMDOrdering<int> > solver;
	//	solver.isSymmetric(true);
	//	solver.analyzePattern(A);
	//	solver.factorize(A);
	//	x = solver.solve(b);

	//	lConvergeVal = 0;
	//	lNegConvergeVal = 0;

	//	for (int i=0; i<kktOffset; i++)
	//	{
	//		lConvergeVal += Max<_T_FLOAT>(x(i, 0)*b(i, 0), 0);
	//		lNegConvergeVal += Max<_T_FLOAT>(-x(i, 0)*b(i, 0), 0);
	//	}
	//}

	for (int i=0; i<kktOffset; i++)
	{
		//_T_FLOAT* ptr;
		_T_FLOAT* ptr1;
		if (i<mX0.size()*3)
		{
		//	ptr = ((_T_FLOAT*)&mX0[0])+i;
			ptr1 = ((_T_FLOAT*)&mXAdv[0])+i;
		}
		else
		{
		//	ptr = ((_T_FLOAT*)&mP0[0])+i-mX0.size()*3;
			ptr1 = ((_T_FLOAT*)&mPAdv[0])+i-mX0.size()*3;
		}
		//*ptr += iStep*x(i, 0);
		if (lConvergeVal>lNegConvergeVal)
			*ptr1 = x(i,0);
		else
			*ptr1 = -x(i, 0);
	}

	//for (int i=0; i<x.rows(); i++)

	// force convergence
	//mExtForceP.assign(mP0.size(), SVec3Float(0, 0, 0));
	//mbStatic = true;
	//_T_FLOAT conv;
	//do
	//{
	//	lbPhysValid = true;
	//	mdExtForce.clear();
	//	mKKT.clear();
	//	ComputePhysForce();
	//	ComputeLapForce();

	//	if (lbPhysValid)
	//		conv = XPStep(1.0, true, true);
	//}
	//while (conv>3e-3 && lbPhysValid);

	//if (!lbPhysValid)
	//{
	//	Rollback();
	//	mbBackupInit = false;
	//	return false;
	//}
	return true;
}

#define _BIN_LINE_SEARCH
#define _ADJ_DIAG_WEIGHT

void PhysImpSolve3D::OptSolveEqKKT()
{
	_T_FLOAT tInit = 0.5;
	_T_FLOAT t = tInit;
	bool bVMEverPass = false;
	OptExportMatlabInit();
	mbUsePenalty = true;
	mbBackupInit = false;
	lOutputConst = false;
	mbAugmentedLag = false;
	//mLapWeight = 1e15;
	//mVMWeight = 1e-4;
	lbNoExtraVM = true;
#ifdef _CURVATURE_WEIGHT
	//mLapSimWeight *= 1e-6;
#endif

	_T_FLOAT VMBoundBackup = mVMBound;
	//mVMBound *= 0.6;
	bool vmPass = false;

	if (mbLoadState)
	{
		cout<<"Loading contents\n";
		LoadState("tmp.data");
		// load params
		ifstream inFile("tmp_state.txt");
		inFile>>mSpeedWeight>>mVMWeight>>VMBoundBackup>>mVMBound>>vmPass;
	}

	cout<<" VM weight "<<mVMWeight
		<<" LapSim weight "<<mLapSimWeight
		<<" Sym weight "<<mSymWeight[0]
		<<" Int Lap weight "<<mLapWeight
		<<" X scale "<<mXMassScale
		<<" Num Sym Const "<<mSymConst[0].size()
		<<" Num Physics Sym Const "<<mSymConst[1].size()
		<<" Num Nodes "<<mX0.size()
		<<" SpeedWeight "<<mSpeedWeight
		<<endl;

	mXAdv.resize(mX0.size());
	mPAdv.resize(mP0.size());

	//mVMBound *= 0.99;

#ifdef _USE_SIM_LAP
	//mpVolMesh->BuildHexVtxConnect();
	//mpVolMesh->FindIntVtx();
	ComputeLapAll();
#endif
	//OptExportLinear();
	//{
	//	LoadState("tmp - converge.data");
	//	mpVolMesh->BuildSurfaceMesh(true, false);
	//	mpVolMesh->mSurfaceMesh.mColor.resize(mpVolMesh->mSurfaceMesh.mVtx.size());
	//	for (int i=0; i<(int)mX0.size(); i++)
	//	{
	//		mpVolMesh->mSurfaceMesh.mVtx[i] = mX0[i];
	//		if (mFixedDesignConst.count(i))
	//			mpVolMesh->mSurfaceMesh.mColor[i] = Vec3f(255, 0, 0);
	//		else
	//			mpVolMesh->mSurfaceMesh.mColor[i] = Vec3f(255, 255, 255);
	//	}
	//	mpVolMesh->mSurfaceMesh.WriteObj("double-converge.obj");
	//	exit(1);
	//}

	//LoadState("problematic_5000.data");

	//
	_T_FLOAT xweight0;
	_T_FLOAT xweightstop = 1;
#ifdef _TABLE_TEST
	xweight0 = 1e5;
#elif defined _2x2CUBE_TEST
	xweight0 = 1e5;
#elif defined _LOAD_MODEL
	xweight0 = 1e7;
#elif defined _LOAD_XML
	if (mSpeedWeight>0)
		xweight0 = mSpeedWeight;
	else
		xweight0 = 1;
#endif
	_T_FLOAT xweight = xweight0;
	_T_FLOAT stepScale = 1;
	mFuncDropSpeed = 0;

	bool bDump = true;

	int updateStreak = 0;
	int maxStreak = 8;
	vector<_T_FLOAT> scores(maxStreak);
	int stepCount = 0;
	bool bSearchInit = false;

	bool bLineSearchSuccess = true;

	Backup(mStepBK);
	_T_FLOAT minStep = 1e-2;
	
	int xweightFail = 0;
	bool negConv = false;

	int updateStep = 0;
	int branchStep = 0;

	bool bBranching = false;

	while (mVMWeight<mVMWeightMax)
	{

		{
			ofstream logfile("log.txt", ios::app | ios::out);
			logfile<<"running time: "<<_GET_TICK_COUNT()<<" ms"<<endl;
		}
//		if (/*mFuncDropSpeed!=0
//			&&*/
//#ifdef _BIN_LINE_SEARCH
//			!bLineSearchSuccess || 
//#else
//			t<=1e-4 ||
//#endif
//			(lConvergeVal>0 && lConvergeVal/mX0.size()/3<1e-3)
//			)
//			/*(
//			scores[stepCount]!=0
//			&& (scores[stepCount]-scores[(stepCount-1+maxStreak)%maxStreak])/scores[stepCount]<0.1))*/
//		{
//			mVMWeight *= 5;
//			t = tInit;
//			mbBackupInit = false;
//			mFuncDropSpeed = 0;
//			cout<<"Update VM weight: "<<mVMWeight<<endl;
//			scores.assign(maxStreak, 0);
//		}

#ifdef _ADJ_DIAG_WEIGHT
		if (!bLineSearchSuccess && (mNoiseSteps<=0 || !bBranching))
		//if (!bLineSearchSuccess
		//	&& lConvergeVal>lNegConvergeVal)
			//&& ((lConvergeVal>lNegConvergeVal) || (xweightFail>10)))
#else
		if (!bLineSearchSuccess)
#endif
		{
			xweightFail = 0;
		//	minStep /= 4;
			//if (!mbKKTNormStep)
			//	stepScale /= 2;
			mVMWeight *= 2;
			//mVMBound += (VMBoundBackup*0.8-mVMBound)*0.2;
			//xweight /= 10;
			//if (vmPass)
			//	xweight = xweightstop;
			//else
			//	xweight = xweight0;
			mbBackupInit = false;
			cout<<"Update VM weight: "<<mVMWeight<<endl;

			if (mNoiseSteps>0)
				branchStep = 0;
		}

		// check vm
		if (bLineSearchSuccess && (mNoiseSteps<=0 || !bBranching))
		{
			
			vmPass = true;

			CalScoreWithPen();
			for (int i=0; i<(int)mVMHex.size() && vmPass; i++)
			for (int v=0; v<8; v++)
			if (mVMHex[i][v]>VMBoundBackup)
				vmPass = false;

			for (int i=0; i<(int)mVM.size() && vmPass; i++)
			if (mVM[i]>VMBoundBackup)
				vmPass = false;

			if (vmPass)
			{
				bVMEverPass = true;
				cout<<"VM test success!\n";
				mVMBound = VMBoundBackup;
				//_T_FLOAT newW = mVMWeightMax*0.49;
				//_T_FLOAT newW = mVMWeight*10;
				
				//if (!mbKKTNormStep)
				//	stepScale /= newW/mVMWeight;
				//mVMWeight = newW;
				//minStep /= 20;
			}
		}

		//xweight = exp(mFuncDropSpeed)*1e9;
		if (mNoiseSteps>0 && branchStep==mNoiseSteps)
		{
			cout<<"Branching\n";
			bool bBranch = NoiseBranch();
			if (bBranch)
			{
				//branchStep = 0;
				cout<<"Branching:"<<mFuncVal<<endl;
				bBranching = true;
			}
			else
			{
				cout<<"Branching fail\n";
				branchStep = 0;
			}
		}
		else if (mNoiseSteps>0 && branchStep==0 && bBranching)
		{
			cout<<"Merging\n";
			bBranching = false;
			NoiseMerge();
			//branchStep = -1;
		}

		cout<<"Prev value:"<<mFuncVal<<"\t";
		mFuncVal = 0;
		mGrad.assign(mX0.size()*6, 0);
		mHess.clear();
		lLapSolContNum = 0;

		int begTime = _GET_TICK_COUNT();

		//if (mFuncVal>0)
		//	xweight = Max<_T_FLOAT>(1e3, lNegConvergeVal);
		//else
		//	xweight = 1e7;

		//_T_FLOAT searchLowBound = 

#ifdef _BIN_LINE_SEARCH
		bLineSearchSuccess = SolveEqKKT(t, xweight, 1.0, false);
		double prevVal = mFuncVal;
		if (lConvergeVal<lNegConvergeVal)
			negConv = true;
		else
			negConv = false;
		if (lConvergeVal<lNegConvergeVal && mbNegGD)
		{
			cout<<"\nNeg Determinant, perform projected gradient descent.";
			bLineSearchSuccess = SolveEqKKT(t, xweight, 1.0, true);
		}
		if (mbKKTNormStep)
			NormalizeKKTStep(false);
		else
			NormalizeKKTStep(true);
#	ifdef _ADJ_DIAG_WEIGHT
		//if (bLineSearchSuccess && lConvergeVal>lNegConvergeVal)
		if (bLineSearchSuccess)
#	else
		if (bLineSearchSuccess)
#	endif
			//bLineSearchSuccess = LineSearch(Max<_T_FLOAT>(t, 1e-3), 1e-3, 1e5, t);
			//bLineSearchSuccess = LineSearch(Max<_T_FLOAT>(t, minStep), minStep, 1e5, t);
			bLineSearchSuccess = LineSearch(stepScale, minStep*stepScale, 1e5*stepScale, t);
		else
			bLineSearchSuccess = false;
		if (!bLineSearchSuccess)
		{
			cout<<"\nWARNING: Line Search Failed!\n";
		}
		else if (!bBranching)
		{
			updateStep++;
			DumpState();
			ofstream outFile("tmp_state.txt");
			outFile<<mSpeedWeight
				<<' '<<mVMWeight
				<<' '<<VMBoundBackup
				<<' '<<mVMBound
				<<' '<<vmPass;
		}
#	ifdef _USE_SIM_LAP
		cout<<"\tLap Sim Weight Percentage:"<<EvaluateLapSim()/mFuncVal<<"\t";
#	endif
		if (mHexVolumeWeight>0)
			cout<<"\tHex Vol Weight Percentage:"<<ComputeHexVolumeForce(true)/mFuncVal<<"\t";

		mbBackupInit = false;
#else
		bool b1 = SolveEqKKT(t, xweight, 1.0);
		bool b2 = EqKKTAdvance(t);

		while (!b2 && t>1e-6)
		{
			t /= 2;
			b2 = EqKKTAdvance(t);
			updateStreak = 0;
		}

		if (b1)
		{
			updateStreak++;
			bSearchInit = true;
		}
		else
		{
			updateStreak = 0;
			t/=2;
		}
		if (updateStreak>4)
		{
			updateStreak = 0;
			t *= 2;
		}
#endif
		cout<<"Val "<<mFuncVal<<"\tstep "<<t
			<<"\tSolver Time\t"<<_GET_TICK_COUNT()-begTime
			<<"\tposConvergence\t"<<lConvergeVal
			<<"\tnegConvergence\t"<<lNegConvergeVal
			<<"\tConvergence\t"<<lConvergeVal-lNegConvergeVal
			<<endl;
#ifdef _ADJ_DIAG_WEIGHT
		if (negConv || !bLineSearchSuccess)
		{
			xweight *= 3;
			if (lConvergeVal-lNegConvergeVal<0)
				xweightFail++;
		}
		else
		{
			xweight /= 1.5;
			xweightFail = 0;
		}
#endif
		if (bSearchInit)
			scores[stepCount%maxStreak] = mFuncVal;
		//stepCount = (stepCount+1)%maxStreak;
		stepCount++;

		//if (mFuncVal<6000 && bDump)
		//{
		//	Rollback();
		//	DumpState("problematic_5000.data");
		//	bDump = false;
		//}

		if (mRecalLapStep>0 && updateStep%mRecalLapStep==0)
		{
			vector<SVec3Float> xrefBackup = mXRef;
			mXRef = mX0;
			ComputeLapAll();
			mXRef = xrefBackup;

			double score = CalScoreWithPen();
			Backup(mStepBK);
			cout<<"Update Laplacian. New Score "<<score<<endl;
		}

		if (mNoiseSteps>0)
		{
			branchStep = (branchStep+1)%(mNoiseSteps+mNoiseConvSteps);
		}

		//if (prevVal>0 && (prevVal-mFuncVal)/prevVal<1e-3)
		//{
		//	cout<<"Update too small!\n";
		//	bLineSearchSuccess = false;
		//}
	}
	cout<<"Opt Terminated\n";
	if (bVMEverPass)
		cout<<"VM Pass!\n";
	mVMBound = VMBoundBackup;
	cout<<"Max VM Percentage "<<ComputeMaxVM()/mVMBound<<endl;

	{
		mpVolMesh->BuildSurfaceMesh(true, false);
		mpVolMesh->mSurfaceMesh.mColor.resize(mpVolMesh->mSurfaceMesh.mVtx.size());
		for (int i=0; i<(int)mX0.size(); i++)
		{
			mpVolMesh->mSurfaceMesh.mVtx[i] = mX0[i];
			if (mFixedDesignConst.count(i))
				mpVolMesh->mSurfaceMesh.mColor[i] = Vec3f(255, 0, 0);
			else
				mpVolMesh->mSurfaceMesh.mColor[i] = Vec3f(255, 255, 255);
		}
		mpVolMesh->mSurfaceMesh.WriteObj("result.obj");
	}

	{
		ofstream logfile("log.txt", ios::app | ios::out);
		logfile<<"end time: "<<_GET_TICK_COUNT()<<" ms"<<endl;
		logfile<<"Max VM Percentage "<<ComputeMaxVM()/mVMBound<<endl;
	}
}
