/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "PhysImpSolve3D_util.h"
#include "../SSSLib/SSSFunc.h"

using namespace std;
using namespace SSSLib;
using namespace PhysImpUtil;

#define _USE_KKT

void PhysImpSolve3D::InitIntLap()
{
	mIntVtx.assign(mX0.size(), VTInt);
	//for (int i=0; i<(int)mpVolMesh->mSurfaceMesh.mFace.size(); i++)
	//{
	//	Vec3i face = mpVolMesh->mSurfaceMesh.mFace[i];
	//	for (int j=0; j<3; j++)
	//		mIntVtx[face[j]] = VTSurf;
	//}

	if (mpVolMesh->mTets.size()==0)
		BuildConnect(true);
	else
		BuildConnect(false);
	//BuildLap();
}

void PhysImpSolve3D::BuildConnect(bool ibIsQuad)
{
	mConnect.clear();
	mConnect.resize(mX0.size(), set<int>());

	mLap.resize(mX0.size());

	if (ibIsQuad)
	{
		if (mpHexExtEdges)
		{
			vector<Vec2i> innerEdges;
			set<Vec2i> extEdges(mpHexExtEdges->begin(), mpHexExtEdges->end());
			for (int i=0; i<(int)mpHexEdges->size(); i++)
			{
				Vec2i e = (*mpHexEdges)[i];
				Vec2i e2 = e;
				swap(e2[0], e2[1]);
				if (!extEdges.count(e) && !extEdges.count(e2))
					innerEdges.push_back(e);
			}

			BuildConnect<2>(VTInt, innerEdges);
			BuildConnect<2>(VTSurf, *mpHexExtEdges);
			BuildConnect<2>(VTEdge, *mpHexExtEdges);
		}
		else
		{
			BuildConnect<2>(VTInt, *mpHexEdges);
			BuildConnect<2>(VTSurf, *mpHexEdges);
			BuildConnect<2>(VTEdge, *mpHexEdges);
		}
	}
	else
	{
		BuildConnect<4>(VTInt, mpVolMesh->mTets);
		BuildConnect<3>(VTSurf, mpVolMesh->mSurfaceMesh.mFace);
		BuildConnect<2>(VTEdge, *mpEdges);
	}
}

//void PhysImpSolve3D::BuildConnect(VtxType iLevel, bool ibUseSurface)
//{
//	if (ibUseSurface)
//	for (int f=0; f<(int)mpVolMesh->mSurfaceMesh.mFace.size(); f++)
//	{
//		Vec3i face = mpVolMesh->mSurfaceMesh.mFace[f];
//		for (int i=0; i<3; i++)
//		for (int j=i+1; j<3; j++)
//		if (mIntVtx[face[i]] == iLevel && mIntVtx[face[j]] == iLevel)
//		{
//			mConnect[face[i]].insert(face[j]);
//			mConnect[face[j]].insert(face[i]);
//		}
//	}
//	else
//	for (int f=0; f<(int)mpVolMesh->mTets.size(); f++)
//	{
//		Vec4i tet = mpVolMesh->mTets[f];
//		for (int i=0; i<4; i++)
//		for (int j=i+1; j<4; j++)
//		if (mIntVtx[tet[i]] == iLevel && mIntVtx[tet[j]] == iLevel)
//		{
//			mConnect[tet[i]].insert(tet[j]);
//			mConnect[tet[j]].insert(tet[i]);
//		}
//	}
//
//	for (int v=0; v<(int)mX0.size(); v++)
//	if (mIntVtx[v] == iLevel)
//	{
//		mLap[v] = ComputeLaplacianWithNeighbour(v, mpVolMesh->mVtx, mConnect[v]);
//		if (mLap[v].SqrLength()>sqr(mEdgeLengthEpsilon))
//		{
//			mIntVtx[v] = (VtxType)(mIntVtx[v]+1);
//			mConnect[v].clear();
//		}
//	}
//}

void PhysImpSolve3D::ComputeLapForce()
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	for (int i=0; i<(int)mX0.size(); i++)
#ifdef _EXT_SIM_LAP_ONLY
	//if (mpVolMesh->mbIntVtx[i])
	if (mIntVtx[i] == VTInt)
#endif
	if (mConnect[i].size()>0)
	{
		set<int>& connection = mConnect[i];
		SVec3Float restLap = ComputeLaplacianWithNeighbour(i, mX0, connection);
		SVec3Float ldelta = restLap-mLap[i];

#ifdef _USE_KKT
		KKTEle newKKT[3];
		for (int c=0; c<3; c++)
		{
			newKKT[c].mB = mLapWeight*ldelta[c]*mXMassScale;
			newKKT[c].AddACoe(i*3+c, -mLapWeight*mXMassScale);
		}
#else // _USE_KKT
		mExtForceX[i] += -mLapWeight*ldelta*mXMassScale;

		for (int c=0; c<3; c++)
			mdExtForce.push_back(T(i*3+c, i*3+c, -mLapWeight*mXMassScale));
#endif // _USE_KKT

		int N = connection.size();

		for (set<int>::iterator v=connection.begin();
			v!=connection.end(); v++)
		for (int c=0; c<3; c++)
		{
#ifdef _USE_KKT
			newKKT[c].AddACoe((*v)*3+c, mLapWeight/N*mXMassScale);
#else // _USE_KKT
			mExtForceX[*v][c] += mLapWeight/N*ldelta[c]*mXMassScale;

			mdExtForce.push_back(T(i*3+c, (*v)*3+c, mLapWeight/N*mXMassScale));

			mdExtForce.push_back(T((*v)*3+c, i*3+c, mLapWeight/N*mXMassScale));

			for (set<int>::iterator v2=connection.begin();
			v2!=connection.end(); v2++)
			{
				mdExtForce.push_back(T((*v)*3+c, (*v2)*3+c, -mLapWeight/sqr(N)*mXMassScale));
			}
#endif // _USE_KKT
		}
#ifdef _USE_KKT
		for (int c=0; c<3; c++)
			mKKT.push_back(newKKT[c]);
#endif // _USE_KKT
	}
}

SVec3Float PhysImpSolve3D::Transform(SVec3Float iVec, MatXFloat iMat)
{
	Vec3Float x1(iVec.val);
	Vec4Float x1_4;
	x1_4.block(0, 0, 3, 1) = x1;
	x1_4(3) = 1;

	Vec3Float ret = (iMat*x1_4).block(0, 0, 3, 1);
	return SVec3Float(ret);
}

void PhysImpSolve3D::ComputeFixedDesignForce()
{
	_T_FLOAT symWeight = mSymWeight[0];

	for (set<int>::iterator ite = mFixedDesignConst.begin();
		ite != mFixedDesignConst.end();
		ite++)
	for (int c=0; c<3; c++)
	{
		int idx = *ite;
		_T_FLOAT val = 2*symWeight*(mX0[idx][c]-mXRef[idx][c]);
		KKTEle newKKT;
		newKKT.mB = val;
		newKKT.AddACoe(idx*3+c, 2*symWeight);

		mKKT.push_back(newKKT);
	}

	for (int c=0; c<3; c++)
	for (set<int>::iterator ite = mFixedDesignDimConst[c].begin();
		ite != mFixedDesignDimConst[c].end();
		ite++)
	{
		int idx = *ite;
		_T_FLOAT val = 2*symWeight*(mX0[idx][c]-mXRef[idx][c]);
		KKTEle newKKT;
		newKKT.mB = val;
		newKKT.AddACoe(idx*3+c, 2*symWeight);

		mKKT.push_back(newKKT);
	}
}

void PhysImpSolve3D::ComputeModelSymForce()
{
	int modelOrPhysics = 0;

	typedef Eigen::Triplet<_T_FLOAT> T;
	for (int i=0; i<(int)mSymConst[modelOrPhysics].size(); i++)
	{
		TSymConst& sym = mSymConst[modelOrPhysics][i];
		Vec3Float x1(mX0[sym.mV1].val);
		Vec3Float x2(mX0[sym.mV2].val);
		Vec4Float x1_4;
		x1_4.block(0, 0, 3, 1) = x1;
		x1_4(3) = 1;
		Mat4x4 trans;
		Mat2SMat(sym.mTransform, trans);

		_T_FLOAT symWeight = mSymWeight[modelOrPhysics];

		Vec3Float dx2 = 2*symWeight*((Vec4Float(trans*x1_4)).block(0, 0, 3, 1)-x2);
#ifdef _USE_KKT
		KKTEle newKKT[3];
		
		for (int r=0; r<3; r++)
		if (!mFixedDesignDimConst[r].count(sym.mV2))
		{
			newKKT[r].mB = dx2[r]*mXMassScale;
			
			newKKT[r].AddACoe(sym.mV2*3+r, -2*symWeight*mXMassScale);
			for (int c=0; c<3; c++)
				newKKT[r].AddACoe(sym.mV1*3+c, 2*symWeight*sym.mTransform(r, c)*mXMassScale);

			//bool trivial = true;
			//for (int i=0; i<(int)newKKT[r].mACoes.size(); i++)
			//if (newKKT[r].mACoes[i].second!=0)
			//{
			//	trivial = false;
			//	break;
			//}
			//if (!trivial)
			mKKT.push_back(newKKT[r]);
		}

#else
		mExtForceX[sym.mV2] += SVec3Float(dx2*mXMassScale);
		for (int r=0; r<3; r++)
		{
			mdExtForce.push_back(T(sym.mV2*3+r, sym.mV2*3+r, -2*mSymWeight*mXMassScale));
			for (int c=0; c<3; c++)
				mdExtForce.push_back(T(sym.mV2*3+r, sym.mV1*3+c, 2*mSymWeight*sym.mTransform(r, c)*mXMassScale));
		}
#endif
	}
}

void PhysImpSolve3D::AddSymConstraints(int iV1,
									   int iV2,
									   MatXFloat iTransform,
									   bool ibAutoInv,
									   bool ibPhysics)
{
	int modelOrPhysics = 0;
	if (ibPhysics)
		modelOrPhysics = 1;
	else
		modelOrPhysics = 0;
	TSymConst symConst;
	symConst.mV1 = iV1;
	symConst.mV2 = iV2;
	Mat2SMat(iTransform, symConst.mTransform);
	//symConst.mTransform = iTransform;
	
	//cout<<"Transform Mat\n"<<iTransform<<endl;

	mSymConst[modelOrPhysics].push_back(symConst);

	if (ibAutoInv)
	{
		symConst.mV1 = iV2;
		symConst.mV2 = iV1;
		MatXFloat transInv = iTransform.inverse();
		Mat2SMat(transInv, symConst.mTransform);
		mSymConst[modelOrPhysics].push_back(symConst);
		//cout<<"Inv Transform Mat\n"<<symConst.mTransform<<endl;
	}
}

void PhysImpSolve3D::AddRotateCenterConstraint(int iV1, int iV2)
{
	MatXFloat T = MatXFloat::Zero(4, 4);
	T(0, 0) = -1;
	T(1, 1) = -1;
	T(2, 2) = -1;
	T(3, 3) = 1;

	SVec3Float c = mpVolMesh->mVtx[iV1]+mpVolMesh->mVtx[iV2];
	for (int i=0; i<3; i++)
		T(i, 3) = c[i];

	AddSymConstraints(iV1, iV2, T, true, 0);
}

void PhysImpSolve3D::AddMirrorConstraint(int iV1, int iV2)
{
	SVec3Float v1 = mpVolMesh->mVtx[iV1];
	SVec3Float v2 = mpVolMesh->mVtx[iV2];

	if (v1==v2)
	{
		AddRotateCenterConstraint(iV1, iV2);
		return;
	}

	SVec3Float t = (v2-v1).Norm();

	MatXFloat T = MatXFloat::Identity(4, 4);
	Vec3Float et(t.val);
	MatXFloat ttT = et*et.transpose();
	_T_FLOAT b2 = dot<_T_FLOAT>(v1+v2, t);
	T.block(0, 0, 3, 3) -= 2*ttT;
	T.block(0, 3, 3, 1) = b2*et;

	AddSymConstraints(iV1, iV2, T, true, 0);
}
