/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "PhysImpSolve3D_util.h"
#include "../SSSLib/SSSFunc.h"

using namespace std;
using namespace SSSLib;
using namespace PhysImpUtil;

#define _H_TEST

void PhysImpSolve3D::ComputePhysForce()
{
	for (int i=0; i<(int)mExtForceX.size(); i++)
		mExtForceX[i] = SVec3Float(0, 0, 0);
	for (int i=0; i<(int)mExtForceP.size(); i++)
		mExtForceP[i] = SVec3Float(0, 0, 0);

	if (mbGravity)
		ComputePGravity();

	if (mbFloor)
		ComputePFloor();

	if (mVtxPresetForce.size()>0)
		ComputePresetForce();

	// for debug
	//{
	//	int interestPts[4] = {11, 0, 21, 1};
	//	cout<<endl;
	//	for (int i=0; i<4; i++)
	//		cout<<mpVolMesh->mSurfaceMesh.mVtxArea[interestPts[i]]<<endl;
	//	for (int i=0; i<4; i++)
	//		cout<<mVtxPresetForce[interestPts[i]][1]<<endl;
	//	//for (int i=0; i<4; i++)
	//	//	cout<<mExtForceP[interestPts[i]][1]<<endl;
	//}


	if (mSymWeight[1]>0)
		ComputePhysicsSimForce();

	ComputeFixedForce();

	ComputeElasticForce();

}

void PhysImpSolve3D::AddDiagonalPHess(_T_FLOAT iUniformMass)
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	for (int i=0; i<(int)mP0.size()*3; i++)
		mdExtForce.push_back(T(mX0.size()*3+i, i, iUniformMass));
}

void PhysImpSolve3D::ComputeExtForce()
{
	mdExtForce.clear();

	ComputePhysForce();

	lNewtonEnergy = 0;

	mKKT.clear();
	mdS = 0;

	//for debug
#if 0
	for (int i=0; i<(int)mExtForceX.size(); i++)
	{
		_T_FLOAT ref = mXRef[i][1]*0.5;
		mExtForceX[i][1] = mXUniformMass*mVMWeight*(ref-mX0[i][1]);
		mdExtForce.push_back(Eigen::Triplet<_T_FLOAT>(i*3+1, i*3+1, -mXUniformMass*mVMWeight));
	}
#else
	if (mJacPercentBound>0)
		ComputeVMForce(false, ContVolume);
	if (mVMWeight>0)
		ComputeVMForce(false, ContVM);
	if (mLapWeight>0)
		ComputeLapForce();
	if (mSymWeight[0]>0)
		ComputeModelSymForce();
#endif
}

void PhysImpSolve3D::ComputeElasticForce()
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	int poffset = int(mX0.size())*3;
	_T_FLOAT energy = 0;
	// elastic force in physics simulation
	for (int eleType=0; eleType<2; eleType++)
	{
		int eleNum;
		int eleSize;
		vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> > h;
		switch (eleType)
		{
		case 0:
			eleNum = mpVolMesh->mTets.size();
			eleSize = 4;
			h.resize(1);
			h[0] = HexBuildH(SVec3Float(0, 0, 0), eleSize);
			break;
		case 1:
			eleNum = mpVolMesh->mHexs.size();
			eleSize = 8;

			h.resize(8);
			for (int i=0; i<8; i++)
			{
				SVec3Float coord(i&1, (i>>1)&1, (i>>2)&1);
				coord = coord*2-1;
				coord = coord/sqrt(3);
				h[i] = HexBuildH(coord, 8);
			}
			break;
		}

#pragma omp parallel for
		for (int t=0; t<eleNum; t++)
#ifdef _H_TEST
		{
			vector<int> ele(eleSize);
			if (eleType==0)
				for (int i=0; i<eleSize; i++)
					ele[i] = mpVolMesh->mTets[t][i];
			else
				for (int i=0; i<eleSize; i++)
					ele[i] = mpVolMesh->mHexs[t][i];
			assert(eleSize<=8);
			SVec3Float x0s[8];
			Vec3Float x0[8];
			Vec3Float p0[8];
			Vec3Float x1[3];
			Vec3Float p1[3];
			for (int i=0; i<eleSize; i++)
			{
				x0s[i] = mX0[ele[i]].val;
				x0[i] = Vec3Float(mX0[ele[i]].val);
				p0[i] = Vec3Float(mP0[ele[i]].val);
			}

			bool bNoV = false;
			
			_T_FLOAT vol;
			MatXFloat dVol;
			MatXFloat d2Vol;

			bNoV = true;
			if (eleType==0 && bNoV)
			{
				Eigen::Matrix<_T_FLOAT, 12, 1> dVolTet;
				Eigen::Matrix<_T_FLOAT, 12, 12> d2VolTet;
				PhysImpUtil::TetVol(&x0s[0], &vol, &dVolTet, &d2VolTet);
				dVol = MatXFloat::Zero(eleSize*6, 1);
				dVol.block(0, 0, eleSize*3, 1) = dVolTet;
				d2Vol = MatXFloat::Zero(eleSize*6, eleSize*6);
				d2Vol.block(0, 0, eleSize*3, eleSize*3) = d2VolTet;
			}
			if (eleType==1)
			{
				Eigen::Matrix<_T_FLOAT, 24, 1> dVolHex;
				Eigen::Matrix<_T_FLOAT, 24, 24> d2VolHex;
				PhysImpUtil::HexVol(&x0s[0], &vol, &dVolHex, &d2VolHex);
				vol /= eleSize;
				dVolHex /= eleSize;
				d2VolHex /= eleSize;
				dVol = MatXFloat::Zero(eleSize*6, 1);
				dVol.block(0, 0, eleSize*3, 1) = dVolHex;
				d2Vol = MatXFloat::Zero(eleSize*6, eleSize*6);
				d2Vol.block(0, 0, eleSize*3, eleSize*3) = d2VolHex;
				bNoV = true;
			}

			for (int testPt = 0; testPt<(int)h.size(); testPt++)
			{

				HexEncode(x0, p0, h[testPt], x1, p1);

				_T_FLOAT U;
				MatXFloat dU;
				MatXFloat d2U;

				//bool bNoV;
				//if (eleType==0)
				//	bNoV = false;
				//else
				//	bNoV = true;

				if (!U2ndOrder(x1, p1, mE, bNoV, U, dU, d2U))
				{
					lbPhysValid = false;
					continue;
				}

				U /= 2;
				dU /= 2;
				d2U /= 2;

				MatXFloat dU2;
				MatXFloat d2U2;

				if (eleSize==4)
					Hex2ndOrderDecode<4>(dU, d2U, h[testPt], dU2, d2U2);
				else
					Hex2ndOrderDecode<8>(dU, d2U, h[testPt], dU2, d2U2);

				if (bNoV)
				{
					d2U2 *= vol;
					d2U2 += dU2*dVol.transpose();
					d2U2 += dVol*dU2.transpose();
					d2U2 += U*d2Vol;

					dU2 *= vol;
					dU2 += U*dVol;

					U *= vol;
				}

		//#pragma omp critical
		//		for (int v=0; v<eleSize; v++)
		//			//energy += U*mTetVtxWeight[t][v]/eleSize;
		//				energy += U/eleSize;

				// b
				//if (ibPOnly)
		#pragma omp critical
				for (int v=0; v<eleSize; v++)
				for (int c=0; c<3; c++)
				{
					_T_FLOAT val = dU2(v*3+c+eleSize*3, 0);
					
					// for debug
					//val *= 2e-3;

					//b(ele[v]*3+c+poffset, 0) -= val*mTetVtxWeight[t][v];
					//b(ele[v]*3+c+poffset, 0) -= val;

					mExtForceP[ele[v]][c] -= val;
				}
				//cout<<"b"<<endl
				//	<<b<<endl;

				// A
				//if (ibPOnly)
		#pragma omp critical
				if (!lFirstOrderOnly)
				for (int v0=eleSize; v0<eleSize*2; v0++)
				for (int c0=0; c0<3; c0++)
				for (int v1=0; v1<eleSize*2; v1++)
				// for debug
				//for (int v1=eleSize; v1<eleSize*2; v1++)
				for (int c1=0; c1<3; c1++)
				//for (int xp=1; xp<2; xp++)
				{
					_T_FLOAT val = d2U2(v0*3+c0, v1*3+c1);
					
					// for debug
					//val *= 2e-3;

					if (val==0)
						continue;

					int idx_a = ele[v0%eleSize]*3+c0+(v0/eleSize)*poffset;
					int idx_b = ele[v1%eleSize]*3+c1+(v1/eleSize)*poffset;

					//tripleList.push_back(T(idx_a, idx_b, val*mTetVtxWeight[t][v0%eleSize]));
					mdExtForce.push_back(T(idx_a, idx_b, val));

				}
			}
		}
#else
		{
			Vec4i tet = mpVolMesh->mTets[t];
			Vec3Float x[3];
			Vec3Float p[3];
			for (int i=0; i<3; i++)
			{
				x[i] = Vec3Float((mX0[tet[i+1]]-mX0[tet[0]]).val);
				p[i] = Vec3Float((mP0[tet[i+1]]-mP0[tet[0]]).val);
			}

			_T_FLOAT U;
			MatXFloat dU;
			MatXFloat d2U;

			U2ndOrder(x, p, mE, false, U, dU, d2U);

			for (int v=0; v<4; v++)
				energy += U*mTetVtxWeight[t][v]/4;

			// b
			//if (ibPOnly)
	#pragma omp critical
			for (int v=0; v<3; v++)
			for (int c=0; c<3; c++)
			{
				_T_FLOAT val = dpweight*dU(v*3+c+9, 0);
				mExtForceP[tet[v+1]][c] += val;
				b(tet[v+1]*3+c+poffset, 0) -= val*mTetVtxWeight[t][v+1];
				b(tet[0]*3+c+poffset, 0) += val*mTetVtxWeight[t][0];
			}
			//cout<<"b"<<endl
			//	<<b<<endl;

			// A
			//if (ibPOnly)
	#pragma omp critical
			for (int v0=0; v0<3; v0++)
			for (int c0=0; c0<3; c0++)
			for (int v1=0; v1<3; v1++)
			for (int c1=0; c1<3; c1++)
			for (int xp=0; xp<2; xp++)
			//for (int xp=1; xp<2; xp++)
			{
				_T_FLOAT val = dp2weight*d2U(v0*3+c0+9, v1*3+c1+xp*9);
				if (val==0)
					continue;

				int idx_a = tet[v0+1]*3+c0+poffset;
				int idx_b = tet[v1+1]*3+c1+poffset*xp;

				int idx_a0 = tet[0]*3+c0+poffset;
				int idx_b0 = tet[0]*3+c1+poffset*xp;

				tripleList.push_back(T(idx_a, idx_b, val*mTetVtxWeight[t][v0+1]));
				tripleList.push_back(T(idx_a0, idx_b, -val*mTetVtxWeight[t][0]));
				tripleList.push_back(T(idx_a, idx_b0, -val*mTetVtxWeight[t][v0+1]));
				tripleList.push_back(T(idx_a0, idx_b0, val*mTetVtxWeight[t][0]));
			}
		}
#endif
	}
}

void PhysImpSolve3D::ComputePGravity()
{
	for (int i=0; i<(int)mExtForceP.size(); i++)
		mExtForceP[i] += SVec3Float(0, -9.8*mMass[i], 0);
}

void PhysImpSolve3D::ComputePFloor()
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	for (int i=0; i<(int)mP0.size(); i++)
	{
		if (mP0[i][1]<=mFloorHeight)
		{
			mExtForceP[i][1] += mFloorForceWeight*(mFloorHeight-mP0[i][1]);
			mdExtForce.push_back(T(i*3+1+mX0.size()*3, i*3+1+mX0.size()*3, mFloorForceWeight));
		}
	}
}

void PhysImpSolve3D::ComputePresetForce()
{
	int numForce = 0;
	for (int i=0; i<(int)mP0.size(); i++)
	{
		if (mVtxPresetForce[i].Length()>1e-3)
			numForce++;
		mExtForceP[i] += mVtxPresetForce[i];
	}

	//cout<<"Num Force "<<numForce<<endl;
}

void PhysImpSolve3D::ComputeFixedForce()
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	for (map<int, _T_FLOAT>::iterator ite = mFixedForce.begin();
		ite != mFixedForce.end();
		ite++)
	{
		int idx = ite->first;
		mExtForceP[idx] += ite->second*(mX0[idx]-mP0[idx]);
		for (int c=0; c<3; c++)
		{
			mdExtForce.push_back(T(idx*3+mX0.size()*3+c, idx*3+c, -ite->second));
			mdExtForce.push_back(T(idx*3+mX0.size()*3+c, idx*3+mX0.size()*3+c, ite->second));
		}
	}

	for (int c=0; c<3; c++)
	for (map<int, _T_FLOAT>::iterator ite = mFixedForceDim[c].begin();
		ite != mFixedForceDim[c].end();
		ite++)
	{
		int idx = ite->first;
		mExtForceP[idx][c] += ite->second*(mX0[idx][c]-mP0[idx][c]);

		mdExtForce.push_back(T(idx*3+mX0.size()*3+c, idx*3+c, -ite->second));
		mdExtForce.push_back(T(idx*3+mX0.size()*3+c, idx*3+mX0.size()*3+c, ite->second));
	}
}

void PhysImpSolve3D::ComputePhysicsSimForce()
{
	int modelOrPhysics = 1;

	typedef Eigen::Triplet<_T_FLOAT> T;
	for (int i=0; i<(int)mSymConst[modelOrPhysics].size(); i++)
	{
		TSymConst& sym = mSymConst[modelOrPhysics][i];
		Vec3Float x1(mP0[sym.mV1].val);
		Vec3Float x2(mP0[sym.mV2].val);
		Vec4Float x1_4;
		x1_4.block(0, 0, 3, 1) = x1;
		x1_4(3) = 1;
		Mat4x4 trans;
		Mat2SMat(sym.mTransform, trans);

		_T_FLOAT symWeight = mSymWeight[modelOrPhysics];

		int idx = sym.mV2;

		Vec3Float dx2 = 2*symWeight*((Vec4Float(trans*x1_4)).block(0, 0, 3, 1)-x2);

		mExtForceP[idx] += SVec3Float(dx2);
		for (int r=0; r<3; r++)
		{
			mdExtForce.push_back(T(idx*3+mX0.size()*3+r, idx*3+mX0.size()*3+r, 2*symWeight));
			for (int c=0; c<3; c++)
				mdExtForce.push_back(T(idx*3+mX0.size()*3+r,
								sym.mV1*3+mX0.size()*3+c,
								-2*symWeight*sym.mTransform(r, c)));
		}

	}

}

