/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "EigenUtil.h"

using namespace std;
using namespace SSSLib;

MatXFloat PhysImpSolve3D::HexBuildH(SVec3Float iCoord, int iEleSize)
{
	if (iEleSize==8)
	{
		MatXFloat retMat(8, 3);
		for (int c=0; c<3; c++)
		{
			for (int i=0; i<8; i++)
			{
				int i3[3];
				_T_FLOAT w = 1.0;
				for (int cc=0; cc<3; cc++)
				{
					i3[cc] = (i>>cc)&1;
					int s;
					if (i3[cc]==0)
						s = -1;
					else
						s = 1;
					//if (cc==0)
					//	s = -s;
					if (cc!=c)
						w = w*(1+s*iCoord[cc]);
					else
						w = w*s;
				}
				w /= 8.0;
				retMat(i, c) = w;
			}
		}
		return retMat;
	}
	else if (iEleSize==4)
	{
		MatXFloat retMat = MatXFloat::Zero(4, 3);
		for (int c=0; c<3; c++)
		{
			retMat(c+1, c) = 1;
			retMat(0, c) = -1;
		}
		return retMat;
	}
}

void PhysImpSolve3D::HexEncode(const Vec3Float* iX,
							const MatXFloat& iH,
							Vec3Float* oX)
{
	int eleSize = iH.rows();
	for (int c=0; c<3; c++)
	{
		oX[c] = Vec3Float::Zero();
		for (int i=0; i<eleSize; i++)
			oX[c] += iX[i]*iH(i, c);
	}
	//for (int c=0; c<3; c++)
	//{
	//	oX[c] = Vec3Float::Zero();	
	//	for (int i=0; i<8; i++)
	//	{
	//		int i3[3];
	//		_T_FLOAT w = 1.0;
	//		for (int cc=0; cc<3; cc++)
	//		{
	//			i3[cc] = (i>>cc)&1;
	//			int s;
	//			if (i3[cc]==0)
	//				s = 1;
	//			else
	//				s = -1;
	//			if (cc!=c)
	//				w = w*(1+s*iCoord[cc]);
	//			else
	//				w = w*s;
	//		}
	//		w /= 8.0;
	//		oX[c] += iX[i]*w;
	//	}
	//}
}

void PhysImpSolve3D::HexEncode(const Vec3Float* iX,
							const Vec3Float* iP,
							const MatXFloat& iH,
							Vec3Float* oX,
							Vec3Float* oP)
{
	HexEncode(iX, iH, oX);
	HexEncode(iP, iH, oP);
}

//void PhysImpSolve3D::Hex2ndOrderDecode(const MatXFloat& idU,
//								const MatXFloat& id2U,
//								const MatXFloat& iH,
//								MatXFloat& odU,
//								MatXFloat& od2U)
//{
//	int eleSize = iH.rows();
//	int poffset = _ElementSize*3;
//	odU = MatXFloat::Zero(poffset*2, 1);
//	for (int xp=0; xp<2; xp++)
//	{
//		Mat3x3 dUMat = Vec2Mat3x3(idU.block(xp*9, 0, 9, 1)); // transposed by default
//		MatXFloat odUMat = iH*dUMat; // transposed
//		for (int i=0; i<_ElementSize; i++)
//			odU.block(xp*poffset+i*3, 0, 3, 1) = odUMat.row(i).transpose();
//	}
//	MatXFloat h2 = MatXFloat::Zero(_ElementSize*2, 6);
//	h2.block(0, 0, _ElementSize, 3) = iH;
//	h2.block(_ElementSize, 3, _ElementSize, 3) = iH;
//	MatXFloat h2t(h2);
//	h2t.transposeInPlace();
//
//	od2U = MatXFloat::Zero(poffset*2, poffset*2);
//	
//	for (int c1=0; c1<3; c1++)
//	for (int c2=0; c2<3; c2++)
//		Eigen::Map<Eigen::Matrix<_T_FLOAT, 2*_ElementSize, 2*_ElementSize>, 0, Eigen::Stride<3*3*_ElementSize*2, 3> >(od2U.data()+c1+c2*od2U.rows())
//			= h2*
//				Eigen::Map<const Eigen::Matrix<_T_FLOAT, 2*3, 2*3>, 0, Eigen::Stride<3*3*6, 3> >(id2U.data()+c1+c2*id2U.rows())
//				*h2t;
//}
