/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "PhysImpSolve3D_util.h"

using namespace std;
using namespace SSSLib;
using namespace PhysImpUtil;

_T_FLOAT PhysImpSolve3D::ComputeHexVolumeForce(bool ibRetTmpVal)
{
	_T_FLOAT sumVal = 0;
	{
		vector<SVec3Float> vtx(8);
		Eigen::Matrix<_T_FLOAT, 24, 1> grad;
		Eigen::Matrix<_T_FLOAT, 24, 24> hess;
		for (int h=0; h<(int)mpVolMesh->mHexs.size(); h++)
		{
			SSSLib::SSSVolMesh::_THex hex = mpVolMesh->mHexs[h];
		
			for (int i=0; i<8; i++)
				vtx[i] = mX0[hex[i]];

			_T_FLOAT vol;
			_T_FLOAT vol0 = mHexDefaultVolume[h];
			if (mbMinVol)
				vol0 = 0;

			if (ibRetTmpVal)
			{
				HexVol(&vtx[0], &vol, NULL, NULL);
				sumVal += sqr(vol-vol0)*mHexVolumeWeight;
				continue;
			}
		
			HexVol(&vtx[0], &vol, &grad, &hess);

			// for debug
			//{
			//	int interestHex[4] = {270, 0, 5, 269};
			//	int idx = -1;

			//	for (int i=0; i<4; i++)
			//		if (interestHex[i]==h)
			//			idx = i;
			//	
			//	if (idx>=0)
			//	{
			//		ofstream debugFileA;
			//		char nameBuffer[1000];
			//		sprintf_s<1000>(nameBuffer, "debugHexVol%d.txt", idx);
			//		debugFileA.open(nameBuffer);
			//		for (int r=0; r<hess.rows(); r++)
			//		for (int c=0; c<hess.cols(); c++)
			//		if (fabs(hess(r, c))<1e-5)
			//			hess(r, c) = 0;
			//		debugFileA<<vol<<endl
			//			<<grad<<endl
			//			<<hess<<endl;
			//	}
			//}

			sumVal += sqr(vol-vol0)*mHexVolumeWeight;

			for (int i=0; i<24; i++)
				mGrad[hex[i/3]*3+(i%3)] += 2*mHexVolumeWeight*(vol-vol0)
										*grad(i, 0);

			for (int r=0; r<24; r++)
			for (int c=0; c<24; c++)
			{
				Vec2i idx(hex[r/3]*3+(r%3), hex[c/3]*3+(c%3));
				mHess[idx] = mHess[idx]+2*mHexVolumeWeight*grad(r, 0)*grad(c, 0)
									+2*mHexVolumeWeight*(vol-vol0)*hess(r, c);
			}
		}
	}

	{
		vector<SVec3Float> vtx(4);
		Eigen::Matrix<_T_FLOAT, 12, 1> grad;
		Eigen::Matrix<_T_FLOAT, 12, 12> hess;
		for (int t=0; t<(int)mpVolMesh->mTets.size(); t++)
		{
			Vec4i tet = mpVolMesh->mTets[t];
		
			for (int i=0; i<4; i++)
				vtx[i] = mX0[tet[i]];

			_T_FLOAT vol;
			_T_FLOAT vol0 = mTetDefaultVolume[t];
			if (mbMinVol)
				vol0 = 0;

			if (ibRetTmpVal)
			{
				TetVol(&vtx[0], &vol, NULL, NULL);
				sumVal += sqr(vol-vol0)*mHexVolumeWeight;
				continue;
			}
		
			TetVol(&vtx[0], &vol, &grad, &hess);

			sumVal += sqr(vol-vol0)*mHexVolumeWeight;

			for (int i=0; i<12; i++)
				mGrad[tet[i/3]*3+(i%3)] += 2*mHexVolumeWeight*(vol-vol0)
										*grad(i, 0);

			for (int r=0; r<12; r++)
			for (int c=0; c<12; c++)
			{
				Vec2i idx(tet[r/3]*3+(r%3), tet[c/3]*3+(c%3));
				mHess[idx] = mHess[idx]+2*mHexVolumeWeight*grad(r, 0)*grad(c, 0)
									+2*mHexVolumeWeight*(vol-vol0)*hess(r, c);
			}
		}
	}

	if (!ibRetTmpVal)
		mFuncVal += sumVal;

	return sumVal;
}

