/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "../SSSLib/SSSFunc.h"
//#include "MatlabUtil.h"
//#include "engine.h"
//#include "mclmcrrt.h"
//#include "..\matlab_fmincon\fmincon_export.h"
//#include "..\matlab_fmincon\fminunc_export.h"

//#define _INTERNAL_P

using namespace std;
using namespace SSSLib;

PhysImpSolve3D* PhysImpSolve3D::spPhysImp = NULL;

vector<double> gRef;

/*
void SimpleF(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
{
	vector<double> val(gRef.size());
	double* pr = mxGetPr(prhs[0]);
	assert(mxGetM(prhs[0])*mxGetN(prhs[0])==gRef.size());
	double diff = 0;
	for (int i=0; i<(int)gRef.size(); i++)
		diff += (pr[i]-gRef[i])*(pr[i]-gRef[i]);
	plhs[0] = mxCreateDoubleScalar(diff);
}

void SimpleCon(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
{
	plhs[0] = mxCreateDoubleMatrix(0, 0, mxREAL);
	plhs[1] = mxCreateDoubleMatrix(0, 0, mxREAL);
}

void SimpleHess(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
{
	plhs[0] = mxCreateDoubleMatrix(0, 0, mxREAL);
}

void PhysImpSolve3D::MatlabSimpleTest()
{
	mclInitializeApplication(NULL,0);
	fmincon_exportInitialize();

	gRef.resize(2);
	gRef[0] = 1.0;
	gRef[1] = 2.0;

	mxArray* x0 = mxCreateDoubleMatrix(2, 1, mxREAL);
	double* pr = mxGetPr(x0);
	pr[0] = 0.0;
	pr[1] = 1.0;

	mxArray* x_out = mxCreateDoubleMatrix(2, 1, mxREAL);
	mxArray* fval = mxCreateDoubleScalar(0);

	mxArray *Aeq = mxCreateDoubleMatrix(0, 0, mxREAL);
	mxArray *Beq = mxCreateDoubleMatrix(0, 0, mxREAL);

	mxArray* f = mclCreateSimpleFunctionHandle(SimpleF);
	mxArray* nonlcon = mclCreateSimpleFunctionHandle(SimpleCon);

	mxArray* hess = mclCreateSimpleFunctionHandle(SimpleHess);

	mlfFmincon_export(2, &x_out, &fval, f, x0, Aeq, Beq, nonlcon, hess);

	pr = mxGetPr(x_out);
	cout<<"x_out: "<<pr[0]<<","<<pr[1]<<endl;
	cout<<"f:"<<mxGetScalar(fval)<<endl;
	char c;
	cin>>c;
}*/

void PhysImpSolve3D::OptExportLinear()
{
#ifdef _EXT_SIM_LAP_ONLY
	if (mLapWeight>0)
		ComputeLapForce();
#endif
	if (mSymWeight[0]>0)
	{
		ComputeFixedDesignForce();
		ComputeModelSymForce();
	}
}

void PhysImpSolve3D::OptExport()
{
	for (int i=0; i<(int)mExtForceX.size(); i++)
		mExtForceX[i] = SVec3Float(0, 0, 0);
	for (int i=0; i<(int)mExtForceP.size(); i++)
		mExtForceP[i] = SVec3Float(0, 0, 0);

	int poffset = int(mX0.size())*3;
	typedef Eigen::Triplet<_T_FLOAT> T;
	mIneqJacob.clear();
	mEqJacob.clear();

	mPhase = -2;
	mdExtForce.clear();
#ifdef _INTERNAL_P
	_T_FLOAT vconverge = 0;
	int numLoop = 0;
	do
	{
		for (int i=0; i<(int)mExtForceP.size(); i++)
			mExtForceP[i] = SVec3Float(0, 0, 0);
		mdExtForce.clear();
		ComputePhysForce();
		vconverge = XPStep(0.5, true);
		numLoop++;
		//cout<<"Phys Loop: "<<numLoop<<endl;
	}
	while (vconverge>1e-5 && numLoop<50);
#else
	lFirstOrderOnly = false;
	ComputePhysForce();

	map<Vec2i, _T_FLOAT> eqJacob;
	for (int i=0; i<(int)mdExtForce.size(); i++)
	{
		T& tri = mdExtForce[i];
		assert(tri.row()>=poffset && tri.row()<poffset*2);
		//if (tri.row()<poffset || tri.row()>=poffset*2)
		//	continue;
		// force is the negative gradient

		// check fixed
		if (mFixedVtx.size()==0 || !mFixedVtx.count((tri.row()-poffset)/3))
			eqJacob[Vec2i(tri.row(), tri.col())] = eqJacob[Vec2i(tri.row(), tri.col())]-tri.value();
	}

	mEqJacob.resize(poffset);
	for (map<Vec2i, _T_FLOAT>::iterator ite = eqJacob.begin();
		ite != eqJacob.end();
		ite++)
	{
		mEqJacob[ite->first[0]-poffset].AddACoe(ite->first[1], ite->second*mForceWeight);
	}
	for (int i=0; i<poffset; i++)
	{
		// add mass?
		if (!mbStatic)
			mEqJacob[i].AddACoe(i+poffset, mMass[i/3]*mForceWeight);
		mEqJacob[i].mB = mExtForceP[i/3][i%3]*mForceWeight;
		
	}

	// check fixed vertices
	for (set<int>::iterator ite = mFixedVtx.begin();
		ite != mFixedVtx.end();
		ite++)
		for (int c=0; c<3; c++)
		{
			int idx = (*ite)*3+c;
			mEqJacob[idx].mB = (mX0[*ite][c]-mP0[*ite][c])*mForceWeight;
			mEqJacob[idx].mACoes.clear();
			mEqJacob[idx].AddACoe(idx, mForceWeight);
			mEqJacob[idx].AddACoe(idx+poffset, -mForceWeight);
		}

#endif
	if (mVMWeight>0)
	{
		if (!lComputeHess)
			lFirstOrderOnly = true;
		ComputeVMForce(false, ContVM);
//#ifndef _USE_NEG_VOL
		if (mJacPercentBound>0)
			ComputeVMForce(false, ContVolume);
//#endif
	}

	if (mHexVolumeWeight>0)
		ComputeHexVolumeForce(false);

	lLapSolContNum = mIneqJacob.size();
	// deformation constraints are linear
}

//#define _SOLVE_FORCE_GRAD
/*
void PhysImpSolve3D::OptExportMatlabCallbackHead(mxArray* iX)
{
	spPhysImp->CopybackState(iX);
	spPhysImp->mbComputeLapSolHessian = false;
	if (spPhysImp->lNeedUpdate)
	{
		spPhysImp->OptExport();
		spPhysImp->lNeedUpdate = false;
	}
}
*/

void PhysImpSolve3D::OptExportPen()
{
	int xsize = mX0.size()*6;
	mFuncVal = 0;
	mGrad.assign(xsize, 0);
	mHess.clear();
	lLapSolContNum = 0;

	OptExport();

#ifdef _SOLVE_FORCE_GRAD
	{
		vector<_T_FLOAT> pgrad;
		_T_FLOAT funVal;
		SolveLinear(mEqJacob, 0, mX0.size()*3, funVal, pgrad);
		mFuncVal += funVal;
		for (int i=0; i<(int)pgrad.size(); i++)
			mGrad[i+mX0.size()*3] += pgrad[i];
	}
#else

	for (int i=0; i<(int)mEqJacob.size(); i++)
	{
		KKTEle& ele = mEqJacob[i];
		_T_FLOAT val = ele.mB;
		_T_FLOAT weight = 1.0;
		if (mbAugmentedLag)
		{
			weight = mMu/2;
			mFuncVal -= mLapSolLambda[lLapSolContNum]*val;
		}
		if (lOutputConst)
			lConst[lLapSolContNum] = val;
		mFuncVal += weight*sqr(val);
		for (int j=0; j<(int)ele.mACoes.size(); j++)
		{
			mGrad[ele.mACoes[j].first] +=
									2*weight*val*
									ele.mACoes[j].second;
			if (mbAugmentedLag)
				mGrad[ele.mACoes[j].first] -=
						mLapSolLambda[lLapSolContNum]*ele.mACoes[j].second;
		}
		if (mbAugmentedLag)
			lLapSolContNum++;
	}
#endif

	// deal with the linear constraints in the kkt
	for (int i=0; i<(int)mKKT.size(); i++)
	{
		KKTEle& ele = mKKT[i];
		_T_FLOAT val = 0;

		for (int j=0; j<(int)ele.mACoes.size(); j++)
		{
			_T_FLOAT* ptr;
			int idx = ele.mACoes[j].first;
			if (idx<mX0.size()*3)
				ptr = mX0[0].val+idx;
			else
				ptr = mP0[0].val+idx-mX0.size()*3;
			val += (*ptr)*ele.mACoes[j].second;
		}
		val -= ele.mB;

		if (lOutputConst)
			lConst[lLapSolContNum] = val;

		_T_FLOAT weight = 1;
		if (mbAugmentedLag)
		{
			mFuncVal -= mLapSolLambda[lLapSolContNum]*val;
			weight = mMu/2;
		}
		mFuncVal += weight*sqr(val);
		for (int j=0; j<(int)ele.mACoes.size(); j++)
		{
			mGrad[ele.mACoes[j].first] += 2*weight*val*ele.mACoes[j].second;
			if (mbAugmentedLag)
				mGrad[ele.mACoes[j].first] -= mLapSolLambda[lLapSolContNum]*ele.mACoes[j].second;
		}
		if (mbAugmentedLag)
			lLapSolContNum++;
	}	

	for (int i=0; i<(int)mX0.size(); i++)
	{
		mFuncVal += (mX0[i]-mXRef[i]).SqrLength();
		for (int c=0; c<3; c++)
		{
			mGrad[i*3+c] += 2*(mX0[i][c]-mXRef[i][c]);
		}
	}

}

/*
void PhysImpSolve3D::OptExportMatlabCallbackCon(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
{
	//cout<<"Calling\n";
#ifdef _INTERNAL_P
	int vecLen = (int)spPhysImp->mX0.size()*3;
#else
	int vecLen = spPhysImp->mX0.size()*6;
#endif
	OptExportMatlabCallbackHead(prhs[0]);
	if (spPhysImp->mIneqJacob.size()>0)
		plhs[0] = MatlabUtil::CreateSetB(spPhysImp->mIneqJacob.size(), &spPhysImp->mIneqJacob[0]);
	else
		plhs[0] = MatlabUtil::CreateMatrix(0, 0);
	//plhs[1] = MatlabUtil::CreateMatrix(0, 0);
	if (spPhysImp->mEqJacob.size()>0)
		plhs[1] = MatlabUtil::CreateSetB((int)spPhysImp->mEqJacob.size(), &spPhysImp->mEqJacob[0]);
	else
		plhs[1] = MatlabUtil::CreateMatrix(0, 0);
	if (nlhs>=3)
	{
		if (spPhysImp->mIneqJacob.size()>0)
			plhs[2] = MatlabUtil::CreateSetSparseMatrixTransposed(
							vecLen,
							(int)spPhysImp->mIneqJacob.size(),
							&spPhysImp->mIneqJacob[0]);
		else
			plhs[2] = MatlabUtil::CreateMatrix(0, 0);
		//plhs[3] = MatlabUtil::CreateMatrix(0, 0);
		if (spPhysImp->mEqJacob.size()>0)
			plhs[3] = MatlabUtil::CreateSetSparseMatrixTransposed(
							vecLen,
							(int)spPhysImp->mEqJacob.size(),
							&spPhysImp->mEqJacob[0]);
		else
			plhs[3] = MatlabUtil::CreateMatrix(0, 0);
	}

	//if (RenderCallback)
	//	RenderCallback();
}

void PhysImpSolve3D::OptExportMatlabCallbackHess(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
{
#ifdef _INTERNAL_P
	int vecLen = (int)spPhysImp->mX0.size()*3;
#else
	int vecLen = spPhysImp->mX0.size()*6;
#endif
	bool bSame = spPhysImp->CopybackState(prhs[0]);
	spPhysImp->CopyLapLambda(prhs[1]);
	spPhysImp->mbComputeLapSolHessian = true;
	spPhysImp->mLapSolHessian.clear();
	spPhysImp->lLapSolHessianCounter = 0;
	spPhysImp->OptExport();
	// hessian of f
	for (int i=0; i<(int)spPhysImp->mX0.size()*3; i++)
		spPhysImp->mLapSolHessian[Vec2i(i, i)] = spPhysImp->mLapSolHessian[Vec2i(i, i)]+2;

	plhs[0] = MatlabUtil::CreateSetSparseMatrix(vecLen, vecLen, spPhysImp->mLapSolHessian);
}
*/

void PhysImpSolve3D::OptExportMatlabInit()
{
	mVDamping = 0;
	//mLapWeight = 1e4;
	//mVMWeight = 1e-4;
	//mSymWeight = 0;
	mXMassScale = 1.0;
	mForceWeight = 1;
	mVolumeWeight = 1e6;
	lNeedUpdate  =true;
	mFWeight = 1;
	//mbAugmentedLag = false;
}

/*
void PhysImpSolve3D::OptExportMatlab()
{
	OptExportMatlabInit();

	int poffset = int(mX0.size())*3;

	mclInitializeApplication(NULL,0);
	fmincon_exportInitialize();

	//if (!(mMatlabEngine = engOpen("")))
	//{
	//	cout<<"\nCan't start MATLAB engine\n";
	//}
	mxArray* test = mxCreateDoubleMatrix(10, 10, mxREAL);

	OptExportLinear();

	//int kktNonZero = 0;
	//for (int i=0; i<(int)mKKT.size(); i++)
	//	kktNonZero += (int)mKKT[i].mACoes.size();

	mxArray *Aeq;
	mxArray *Beq;
	mxArray *x0;

	if (mKKT.size()!=0)
	{
#ifdef _INTERNAL_P
		Aeq = MatlabUtil::CreateSetSparseMatrix((int)poffset, (int)mKKT.size(), &mKKT[0]);
#else
		Aeq = MatlabUtil::CreateSetSparseMatrix((int)poffset*2, (int)mKKT.size(), &mKKT[0]);
#endif
		//Beq = MatlabUtil::CreateSetB((int)mKKT.size(), &mKKT[0]);
		vector<double> zero(mKKT.size(), 0);
		MatlabUtil::CreateSetMatrix(mKKT.size(), 1, &zero[0], &Beq);
	}
	else
	{
		Aeq = mxCreateDoubleMatrix(0, 0, mxREAL);
		Beq = mxCreateDoubleMatrix(0, 0, mxREAL);
	}

	//engPutVariable(mMatlabEngine, "Aeq", Aeq);
	//engPutVariable(mMatlabEngine, "Beq", Beq);

	x0 = EncodeState();
	//engPutVariable(mMatlabEngine, "x0", x0);
	
	mxArray* f = mclCreateSimpleFunctionHandle(OptExportMatlabCallbackF);
	mxArray* con = mclCreateSimpleFunctionHandle(OptExportMatlabCallbackCon);

	mxArray* x_out = MatlabUtil::CreateMatrix((int)mX0.size()*6, 1);
	mxArray* fval = mxCreateDoubleScalar(0);

	mxArray* hessianfcn = mclCreateSimpleFunctionHandle(OptExportMatlabCallbackHess);
	mlfFmincon_export(2, &x_out, &fval, f, x0, Aeq, Beq, con, hessianfcn);

	cout<<"fmincon terminated\n";
	DumpState();
	//CopybackState(x_out);

	//mexEvalString("options = optimoptions('fmincon','GradObj','on',"
	//			" 'GradConstr','on', 'DerivativeCheck', 'on')");

	//mxArray* options;
	//mxArray* optstr[7];
	//optstr[0] = mxCreateString("fmincon");
	//optstr[1] = mxCreateString("GradObj");
	//optstr[2] = mxCreateString("on");
	//optstr[3] = mxCreateString("GradConstr");
	//optstr[4] = mxCreateString("on");
	//optstr[5] = mxCreateString("DerivativeCheck");
	//optstr[6] = mxCreateString("on");

	//engPutVariable(mMatlabEngine, "f", f);
	//engPutVariable(mMatlabEngine, "con", con);
	//assert(0==engEvalString(mMatlabEngine, "options = optimoptions('fmincon','GradObj','on',"
	//										" 'GradConstr','on', 'DerivativeCheck', 'on')"));

	mxDestroyArray(Aeq);
	mxDestroyArray(Beq);
	mxDestroyArray(x0);
	mxDestroyArray(f);
	mxDestroyArray(con);
	mxDestroyArray(x_out);
	mxDestroyArray(fval);

	fmincon_exportTerminate();
}
*/

void PhysImpSolve3D::OptExportPenInit()
{
	//mVDamping = 0;
	////mLapWeight = 1e5;
	////mLapWeight = 1;
	//mLapWeight = 0;
	//mVMWeight = 1e-4;
	////mVMWeight = 0;
	//mSymWeight = 0;
	//mXMassScale = 1.0;
	//mForceWeight = 1e2;
	//mVolumeWeight = 1e6;
	//mbUsePenalty = true;
	//lNeedUpdate = true;
	OptExportMatlabInit();
	//mForceWeight = 10;
	mbUsePenalty = true;

	OptExportLinear();

	if (mbAugmentedLag)
	{
		int numCon = 0;
		numCon = (int)mpVolMesh->mHexs.size()*16;
		numCon += mKKT.size();
		numCon += mP0.size()*3;
		mLapSolLambda.assign(numCon, 0.0);
		lConst.resize(numCon);
	}
}

/*
void PhysImpSolve3D::OptExportMatlabPen()
{
	OptExportPenInit();

	int poffset = int(mX0.size())*3;

	mclInitializeApplication(NULL,0);
	fminunc_exportInitialize();

	mxArray *x0 = EncodeState();
	
	mxArray* f = mclCreateSimpleFunctionHandle(OptExportMatlabCallbackFPen);

	mxArray* x_out = MatlabUtil::CreateMatrix((int)mX0.size()*6, 1);
	mxArray* fval = mxCreateDoubleScalar(0);

	// for debug
	//{
	//	mxArray* argLeft[2];
	//	mxArray* argRight[1];
	//	argRight[0] = x0;
	//	OptExportMatlabCallbackFPen(2, argLeft, 1, argRight);
	//	cout<<"Init Evaluation "<<mxGetScalar(argLeft[0])<<endl;
	//}

	for (int i=0; i<30; i++)
	{
		lOutputConst = false;
		mlfFminunc_export(2, &x_out, &fval, f, x0);
		CopybackState(x_out);
		if (mbAugmentedLag)
		{
			lOutputConst = true;
			OptExportPen();
			for (int i=0; i<(int)mLapSolLambda.size(); i++)
			{
				mLapSolLambda[i] -= mMu*lConst[i];
			}
		}
		//mMu *= 10;
		cout<<"fminunc ite "<<i<<"\n";
		DumpState();
	}
	
	mxDestroyArray(x0);
	mxDestroyArray(f);
	mxDestroyArray(x_out);
	mxDestroyArray(fval);

	fminunc_exportTerminate();
}

void PhysImpSolve3D::OptExportMatlabCallbackFPen(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
{
	int poffset = spPhysImp->mX0.size()*3;
	int xsize = poffset*2;
	spPhysImp->CopybackState(prhs[0]);

	spPhysImp->OptExportPen();

	plhs[0] = mxCreateDoubleScalar(spPhysImp->mFuncVal);

	cout<<"Func val "<<spPhysImp->mFuncVal<<endl;

	if (nlhs>1)
	{
		plhs[1] = MatlabUtil::CreateMatrix((int)spPhysImp->mGrad.size(), 1);
		memcpy(mxGetPr(plhs[1]),
			(double*)&spPhysImp->mGrad[0],
			sizeof(double)*spPhysImp->mGrad.size());
	}
}

void PhysImpSolve3D::OptExportMatlabCallbackF(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
{
	int poffset = spPhysImp->mX0.size()*3;
	spPhysImp->CopybackState(prhs[0]);
	//cout<<"Calling F\n";
#ifdef _INTERNAL_P
	vector<double> grad(spPhysImp->mX0.size()*3, 0);
#else
	vector<double> grad(spPhysImp->mX0.size()*6, 0);
#endif

	//plhs[0] = mxCreateDoubleScalar(0);
	//MatlabUtil::CreateSetMatrix((int)grad.size(), 1, &grad[0], &plhs[1]);
	//return;

	_T_FLOAT diff = 0;
	for (int i=0; i<(int)spPhysImp->mX0.size(); i++)
	{
		diff += (spPhysImp->mX0[i]-spPhysImp->mXRef[i]).SqrLength()*spPhysImp->mFWeight;
//#ifndef _INTERNAL_P
//		diff += (spPhysImp->mP0[i]-spPhysImp->mXRef[i]).SqrLength();
//#endif
		for (int c=0; c<3; c++)
		{
			grad[i*3+c] = 2*(spPhysImp->mX0[i][c]-spPhysImp->mXRef[i][c])*spPhysImp->mFWeight;
//#ifndef _INTERNAL_P
//			grad[i*3+c+poffset] = 2*(spPhysImp->mP0[i][c]-spPhysImp->mXRef[i][c]);
//#endif
		}
	}
	plhs[0] = mxCreateDoubleScalar(diff);
	if (nlhs>1)
	{
		plhs[1] = MatlabUtil::CreateMatrix((int)grad.size(), 1);
		memcpy(mxGetPr(plhs[1]), (double*)&grad[0], sizeof(double)*grad.size());
	}
}

mxArray* PhysImpSolve3D::EncodeState()
{
#ifdef _INTERNAL_P
	mxArray* x0 = MatlabUtil::CreateMatrix((int)mX0.size()*3, 1);
#else
	mxArray* x0 = MatlabUtil::CreateMatrix((int)mX0.size()*6, 1);
#endif
	{
		double* ptr = mxGetPr(x0);
		if (sizeof(mX0[0][0])==sizeof(double))
		{
			memcpy(ptr, mX0[0].val, mX0.size()*3*sizeof(double));
#ifndef _INTERNAL_P
			memcpy(ptr+mX0.size()*3, mP0[0].val, mX0.size()*3*sizeof(double));
#endif
		}
		else
		for (int i=0; i<(int)mX0.size(); i++)
		for (int c=0; c<3; c++)
		{
			ptr[i*3+c] = mX0[i][c];
#ifndef _INTERNAL_P
			ptr[i*3+c+mX0.size()*3] = mP0[i][c];
#endif
		}
	}
	return x0;
}

bool PhysImpSolve3D::CopybackState(const double* iXP)
{
	const double* inPtr = iXP;
//	if (sizeof(mX0[0][0])==sizeof(double))
//	{
//		memcpy(mX0[0].val, inPtr, mX0.size()*3*sizeof(double));
//#ifndef _INTERNAL_P
//		memcpy(mP0[0].val, inPtr+mX0.size()*3, mX0.size()*3*sizeof(double));
//#endif
//	}
//	else
	for (int i=0; i<(int)mX0.size(); i++)
	for (int c=0; c<3; c++)
	{
		if (mX0[i][c]!=inPtr[i*3+c])
		{
			mX0[i][c] = inPtr[i*3+c];
			lNeedUpdate = true;
		}
#ifndef _INTERNAL_P
		if (mP0[i][c]!=inPtr[i*3+c+mX0.size()*3])
		{
			mP0[i][c] = inPtr[i*3+c+mX0.size()*3];
			lNeedUpdate = true;
		}
#endif
	}
	return false;
}

bool PhysImpSolve3D::CopybackState(mxArray* iXP)
{
	double* inPtr = mxGetPr(iXP);
	return CopybackState(inPtr);
}

void PhysImpSolve3D::CopyLapLambda(mxArray* iLambda)
{
	double* inPtr = mxGetPr(iLambda);
	mLapSolLambda.resize(lLapSolContNum);
	for (int i=0; i<(int)mLapSolLambda.size(); i++)
		mLapSolLambda[i] = mxGetScalar(mxGetField(iLambda, i, "ineqnonlin"));
}
*/

void PhysImpSolve3D::RunPhysics(double iConv, int iIteNum)
{
	_T_FLOAT converge;
	int iteNum = 0;
	do
	{
		mdExtForce.clear();
		ComputePhysForce();
		//converge = spPhysImp->XPStep(0.45, true);
		converge = XPStep(mSimTimeStep, true);
		//cout<<"Vtx0 "<<g_PhysImp.mP0[0][0]<<","<<g_PhysImp.mP0[0][1]
		//	<<","<<g_PhysImp.mP0[0][2]<<endl;
		iteNum ++;
		cout<<"Conv "<<converge<<endl;
	}
	//while (0);
	while (converge>iConv || iteNum<iIteNum);
}

void PhysImpSolve3D::MatlabOptThread(void* argv)
{
	//assert(0==engEvalString(spPhysImp->mMatlabEngine,
	//	"[x_out]=fmincon(f, x0, [], [], Aeq, Beq, [], [], con, options)"));
	//mxArray* x = engGetVariable(spPhysImp->mMatlabEngine, "x_out");

	//spPhysImp->CopybackState(x);
	//mxDestroyArray(x);
	//spPhysImp->OptExportMatlab();
	//spPhysImp->OptExportMatlabSnopt();
	//spPhysImp->mbAugmentedLag = true;
	//spPhysImp->OptExportMatlabPen();
	//spPhysImp->mbStatic = true;
	spPhysImp->mbStatic = false;
	//_T_FLOAT converge;
	//int iteNum = 0;
	//do
	//{
	//	spPhysImp->mdExtForce.clear();
	//	spPhysImp->ComputePhysForce();
	//	converge = spPhysImp->XPStep(spPhysImp->mSimTimeStep, true);
	//	iteNum ++;
	//}
	//while (converge>2e-3 || iteNum<((int*)argv)[0] || ((int*)argv)[0]<0);
	// for debug
	spPhysImp->RunPhysics(2e-3, ((int*)argv)[0]);

	spPhysImp->OptSolveEqKKT();
}
