/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "../SSSLib/SSSFunc.h"
#include "MatlabUtil.h"
//#include "engine.h"
#include "mclmcrrt.h"
#include "..\matlab_fmincon\TOMLAB_export.h"

using namespace SSSLib;
using namespace std;

void PhysImpSolve3D::OptExportMatlabCallbackCOnly(int nlhs,
												  mxArray *plhs[],
												  int nrhs,
												  mxArray *prhs[])
{
	OptExportMatlabCallbackHead(prhs[0]);
	plhs[0] = mxCreateDoubleMatrix(
		spPhysImp->mEqJacob.size()+spPhysImp->mIneqJacob.size(), 1, mxREAL);
	double* pr = mxGetPr(plhs[0]);
	for (int i=0; i<(int)spPhysImp->mEqJacob.size(); i++)
	{
		*pr = spPhysImp->mEqJacob[i].mB;
		pr++;
	}
	for (int i=0; i<(int)spPhysImp->mIneqJacob.size(); i++)
	{
		*pr = spPhysImp->mIneqJacob[i].mB;
		pr++;
	}
}

void PhysImpSolve3D::OptExportMatlabCallbackdC(int nlhs,
											   mxArray *plhs[],
											   int nrhs,
											   mxArray *prhs[])
{
	int vecLen = (int)spPhysImp->mX0.size()*6;
	OptExportMatlabCallbackHead(prhs[0]);
	vector<KKTEle> kktVec = spPhysImp->mEqJacob;
	kktVec.insert(kktVec.end(), spPhysImp->mIneqJacob.begin(), spPhysImp->mIneqJacob.end());

	plhs[0] = MatlabUtil::CreateSetSparseMatrix(vecLen, (int)kktVec.size(), &kktVec[0]);
}

mxArray* PhysImpSolve3D::OptExportMatlabComputeConsPattern()
{
	OptExport();
	int vecLen = (int)spPhysImp->mX0.size()*6;
	vector<KKTEle> kktVec = spPhysImp->mEqJacob;
	kktVec.insert(kktVec.end(), spPhysImp->mIneqJacob.begin(), spPhysImp->mIneqJacob.end());
	for (int i=0; i<(int)kktVec.size(); i++)
	{
		KKTEle& ele = kktVec[i];
		set<int> idx;
		for (int j=0; j<(int)ele.mACoes.size(); j++)
			idx.insert(ele.mACoes[j].first);
		ele.mACoes.clear();
		for (set<int>::iterator ite = idx.begin();
			ite != idx.end();
			ite++)
			ele.mACoes.push_back(KKTEle::KKTCoe(*ite, 1.0));
	}

	return MatlabUtil::CreateSetSparseMatrix(vecLen, (int)kktVec.size(), &kktVec[0]);
}

void PhysImpSolve3D::OptExportMatlabCallbackFOnly(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
{
	int poffset = spPhysImp->mX0.size()*3;
	spPhysImp->CopybackState(prhs[0]);

	_T_FLOAT diff = 0;
	for (int i=0; i<(int)spPhysImp->mX0.size(); i++)
	{
		diff += (spPhysImp->mX0[i]-spPhysImp->mXRef[i]).SqrLength();
	}
	plhs[0] = mxCreateDoubleScalar(diff*spPhysImp->mFWeight);
}

void PhysImpSolve3D::OptExportMatlabCallbackdF(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
{
	spPhysImp->CopybackState(prhs[0]);
	vector<double> grad(spPhysImp->mX0.size()*6, 0);
	for (int i=0; i<(int)spPhysImp->mX0.size(); i++)
		for (int c=0; c<3; c++)
			grad[i*3+c] = 2*(spPhysImp->mX0[i][c]-spPhysImp->mXRef[i][c])*spPhysImp->mFWeight;
	
	plhs[0] = MatlabUtil::CreateMatrix((int)grad.size(), 1);
	memcpy(mxGetPr(plhs[0]), (double*)&grad[0], sizeof(double)*grad.size());
}

void PhysImpSolve3D::OptExportMatlabCallbackd2F(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
{
	int vecLen = (int)spPhysImp->mX0.size()*6;
	plhs[0] = mxCreateSparse(vecLen, vecLen, vecLen/2, mxREAL);
	double *data = mxGetPr(plhs[0]);
	mwIndex *rowIndex = mxGetIr(plhs[0]);
	mwIndex *columnPos = mxGetJc(plhs[0]);
	for (int i=0; i<vecLen/2; i++)
	{
		rowIndex[i] = i;
		columnPos[i] = i;
		data[i] = 2*spPhysImp->mFWeight;
	}
}

mxArray* PhysImpSolve3D::OptExportMatlabComputeHessPattern()
{
	//int vecLen = (int)spPhysImp->mX0.size()*6;
	//mxArray* ret = mxCreateSparse(vecLen, vecLen, vecLen/2, mxREAL);
	//double *data = mxGetPr(ret);
	//mwIndex *rowIndex = mxGetIr(ret);
	//mwIndex *columnPos = mxGetJc(ret);
	//for (int i=0; i<vecLen/2; i++)
	//{
	//	rowIndex[i] = i;
	//	columnPos[i] = i;
	//	data[i] = 1;
	//}
	//return ret;

	int vecLen = (int)spPhysImp->mX0.size()*6;
	int poffset = (int)spPhysImp->mX0.size()*3;
	vector<set<int> > sparsePattern(vecLen);
	for (int h=0; h<(int)mpVolMesh->mHexs.size(); h++)
	{
		SSSVolMesh::_THex& hex = mpVolMesh->mHexs[h];
		for (int v1=0; v1<8; v1++)
		for (int c1=0; c1<3; c1++)
		for (int v2=0; v2<8; v2++)
		for (int c2=0; c2<3; c2++)
		{
			sparsePattern[hex[v1]*3+c1].insert(hex[v2]*3+c2);
			sparsePattern[hex[v1]*3+c1].insert(hex[v2]*3+c2+poffset);
			sparsePattern[hex[v1]*3+c1+poffset].insert(hex[v2]*3+c2+poffset);
			sparsePattern[hex[v1]*3+c1+poffset].insert(hex[v2]*3+c2);
		}
	}

	int entries = 0;
	for (int i=0; i<(int)sparsePattern.size(); i++)
		entries += sparsePattern[i].size();

	mxArray* ret = mxCreateSparse(vecLen, vecLen, entries, mxREAL);
	double *data = mxGetPr(ret);
	mwIndex *rowIndex = mxGetIr(ret);
	mwIndex *columnPos = mxGetJc(ret);
	int idx = 0;
	for (int i=0; i<(int)sparsePattern.size(); i++)
	{
		columnPos[i] = idx;
		for (set<int>::iterator ite = sparsePattern[i].begin();
			ite != sparsePattern[i].end();
			ite++)
		{
			rowIndex[idx] = *ite;
			data[idx] = 1;
			idx++;
		}
	}
	return ret;
}

void PhysImpSolve3D::OptExportMatlabSnopt()
{
	OptExportMatlabInit();

	int poffset = int(mX0.size())*3;

	mclInitializeApplication(NULL,0);
	TOMLAB_exportInitialize();

	OptExportLinear();

	mxArray *Aeq;
	mxArray *Beq;
	mxArray *x0;

	if (mKKT.size()!=0)
	{
#ifdef _INTERNAL_P
		Aeq = MatlabUtil::CreateSetSparseMatrix((int)poffset, (int)mKKT.size(), &mKKT[0]);
#else
		Aeq = MatlabUtil::CreateSetSparseMatrix((int)poffset*2, (int)mKKT.size(), &mKKT[0]);
#endif
		//Beq = MatlabUtil::CreateSetB((int)mKKT.size(), &mKKT[0]);
		vector<double> zero(mKKT.size(), 0);
		MatlabUtil::CreateSetMatrix(mKKT.size(), 1, &zero[0], &Beq);
	}
	else
	{
		Aeq = mxCreateDoubleMatrix(0, 0, mxREAL);
		Beq = mxCreateDoubleMatrix(0, 0, mxREAL);
	}

	x0 = EncodeState();

	mxArray* f = mclCreateSimpleFunctionHandle(OptExportMatlabCallbackFOnly);
	mxArray* df = mclCreateSimpleFunctionHandle(OptExportMatlabCallbackdF);
	mxArray* d2f = mclCreateSimpleFunctionHandle(OptExportMatlabCallbackd2F);
	mxArray* con = mclCreateSimpleFunctionHandle(OptExportMatlabCallbackCOnly);
	mxArray* dcon = mclCreateSimpleFunctionHandle(OptExportMatlabCallbackdC);
	//mxArray* dcon = mxCreateDoubleMatrix(0, 0, mxREAL);
	//mxArray* df = mxCreateDoubleMatrix(0, 0, mxREAL);
	//mxArray* d2f = mxCreateDoubleMatrix(0, 0, mxREAL);
	//mxArray* dcon = mxCreateDoubleMatrix(0, 0, mxREAL);

	mxArray* x_out = MatlabUtil::CreateMatrix((int)mX0.size()*6, 1);
	mxArray* fval = mxCreateDoubleScalar(0);

	//mxArray* hessPattern = mxCreateDoubleMatrix(0, 0, mxREAL);
	//mxArray* nonlconPattern = mxCreateDoubleMatrix(0, 0, mxREAL);
	mxArray* hessPattern = OptExportMatlabComputeHessPattern();
	mxArray* nonlconPattern = OptExportMatlabComputeConsPattern();

	mxArray* eqConNum = mxCreateDoubleScalar(mEqJacob.size());
	mxArray* neqConNum = mxCreateDoubleScalar(mIneqJacob.size());

	mxArray* empty = mxCreateDoubleMatrix(0, 0, mxREAL);

	mlfTOMLAB_export(2, &x_out, &fval,
						f, df, d2f, hessPattern, x0,
						Aeq, Beq,
						nonlconPattern, con, dcon, 
						eqConNum, neqConNum);

	cout<<"Tomlab opt terminated\n";
	OptExportMatlabCallbackHead(x_out);
	DumpState();

	mxDestroyArray(Aeq);
	mxDestroyArray(Beq);
	mxDestroyArray(x0);
	mxDestroyArray(df);
	mxDestroyArray(d2f);
	mxDestroyArray(con);
	mxDestroyArray(dcon);
	mxDestroyArray(x_out);
	mxDestroyArray(fval);
	mxDestroyArray(hessPattern);
	mxDestroyArray(nonlconPattern);
	mxDestroyArray(eqConNum);
	mxDestroyArray(neqConNum);

	TOMLAB_exportTerminate();
}
