/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "PhysImpSolve3D_util.h"
#include "TestConfig.h"
#include <tinyxml2.h>
#include <boost/pending/disjoint_sets.hpp>

using namespace std;
using namespace SSSLib;

bool PhysImpSolve3D::lFirstOrderOnly = false;
bool PhysImpSolve3D::mbCorotation = false;

PhysImpSolve3D::PhysImpSolve3D(void)
	: mFDamping(1.0)

	, mbLoadState(false)

	, mbInvLapWeight(false)

	, mRecalLapStep(0)

	, mbNegGD(false)

	, mJacPercentBound(0)
	
	, mLapSimNumRing(1)

	, mbCholeskyAddedMultiple(false)

	, mbMassLapWeight(false)
	, mSpeedWeight(1.0)

	, mbUseEleSim(false)

	, mbStatic(true)
	, mSimTimeStep(0.4)
	
	, mbKKTNormStep(false)

	, mbMinVol(false)
	, mHexVolumeWeight(0)

	, mVMWeightMax(1)

	//, mLapSimWeight(1e3)
	//, mLapSimWeight(1e10)
	//, mLapSimWeight(0)
	, mShapeSimWeight(0)
	
	, mDensity(1e1)
	, mbGravity(true)
	, mbFloor(true)
	, mFloorForceWeight(1e7)
	, mNewtonTCrit(1e8)
	, mVolumeWeight(1e2)
	, mForceWeight(1)
	, mbComputeLapSolHessian(false)
	, mbUsePenalty(false)
	, mFWeight(1)
	, mMu(1.0)
	, mbAugmentedLag(false)
	, lComputeHess(false)
	, lbNoExtraVM(false)
#ifdef _2x2CUBE_TEST
	, mNewtonSScale(1.0)
	, mXUniformMass(1e4)
	, mVMWeight(1e-8)
	//, mVMBound(1e7)
	, mVMBound(3e3)
	, mVDamping(0.1)
	, mFloorHeight(-0.5)
	, mNewtonStep(1.0)
	, mNewtonMu(1.1)
	, mXMassScale(1e3)
	, mNewtonTInitScale(1)
#elif defined _SINGLE_STAND_TEST
	, mNewtonSScale(1.0)
	, mXUniformMass(1e4)
	, mVMWeight(1e-5)
	//, mVMBound(1e7)
	, mVMBound(2e6)
	, mVDamping(0.0)
	, mFloorHeight(-0.5)
	, mNewtonStep(1.0)
	, mNewtonMu(1.1)
	, mXMassScale(1e3)
	, mNewtonTInitScale(1e-8)
#elif defined _TWO_STAND_TEST
	, mNewtonSScale(1.0)
	, mXUniformMass(1)
	, mVMWeight(1e-8)
	//, mVMBound(1e7)
#	ifdef _TET
	, mVMBound(1e3)
#	else
	, mVMBound(3e4)
#	endif
	, mVDamping(0.0)
	, mFloorHeight(-0.5)
	, mNewtonStep(1.0)
	, mNewtonMu(1.1)
	, mXMassScale(1e6)
	, mNewtonTInitScale(10)
#elif defined _TABLE_TEST
	, mLapSimWeight(1e6)
	//, mXUniformMass(1e5)
	, mXUniformMass(1)
	//, mVMWeight(1)
	, mVMWeight(1e-5)
	//, mVMBound(1e7)
#	ifdef _TET
	, mVMBound(2e3)
#	else
	, mVMBound(1e5)
#	endif
	, mVDamping(0.0)
	, mFloorHeight(-0.5)
	, mNewtonStep(1.0)
	, mNewtonMu(2.0)
	, mXMassScale(1e6)
	, mNewtonTInitScale(10)
	, mNewtonSScale(1)
#else
	, mLapSimWeight(1e9)
	, mLapFixedWeight(0)
	, mLapScaleWeight(0)
#	ifdef _2ND_BOUND
	, mVMWeight(1e-20)
#	else
	, mVMWeight(1e-4)
	//, mVMWeight(0.001)
#	endif
	, mXUniformMass(3e5)
	//, mXUniformMass(1e-3)
#	ifdef _CHAIR_B_RELAXED_B
	, mVMBound(3e5)
#	elif defined (_BEETHOVEN)
	, mVMBound(3e5)
#	elif defined (_ELLIPSOID)
	, mVMBound(3e5)
#	endif
	, mVDamping(0)
	, mFloorHeight(-0.7)
	, mNewtonStep(1.0)
	, mNewtonMu(1)
	, mNewtonTInitScale(10)
	, mNewtonSScale(1)
	//, mXMassScale(1e10)
	, mXMassScale(1)
	, mForcePenWeight(1e3)
#endif
	//, mVMBound(1e5)
	, mLapWeight(1e17)
	//, mLapWeight(0)
	, mEdgeLengthEpsilon(2e-3)
	//, mSymWeight[0](1e15)
	//, mSymWeight[1](1e5)
	//, mSymWeight(0)
	
	, mPhase(-1)
	
	, mpHexExtEdges(NULL)

	, mNoiseSteps(0)
	, mNoiseConvSteps(5)
	, mNoiseProportion(0.1)
{
	mSymWeight[0] = 1e15;
	mSymWeight[1] = 1e15;

	mSymConst[0].clear();
	mSymConst[1].clear();
}

PhysImpSolve3D::~PhysImpSolve3D(void)
{
}

void PhysImpSolve3D::Init(SSSLib::SSSVolMesh* ipVolMesh)
{
	mpVolMesh = ipVolMesh;
	mX0.resize(mpVolMesh->mVtx.size());
	mXV.clear();
	mXV.resize(mX0.size(), SVec3Float(0, 0, 0));

	for (int i=0; i<(int)mX0.size(); i++)
		mX0[i] = mpVolMesh->mVtx[i];

	mXRef = mX0;
	mP0 = mX0;
	mPV = mXV;

	//mpVolMesh->ComputeMass((float)mDensity);
	mMass.resize(mpVolMesh->mMass.size());
	for (int i=0; i<(int)mMass.size(); i++)
		mMass[i] = _T_FLOAT(mpVolMesh->mMass[i]);

	mExtForceX.resize(mX0.size());
	mExtForceP.resize(mP0.size());

	mTetVtxWeight.resize(mpVolMesh->mTets.size());
	for (int t=0; t<(int)mTetVtxWeight.size(); t++)
	{
		//mTetVtxWeight[t] = mpVolMesh->mSolidAngles[t];
		mTetVtxWeight[t] = SVec4Float(1.0, 1.0, 1.0, 1.0);
	}

	mVM.resize(ipVolMesh->mTets.size(), 0);
	mVMHex.resize(ipVolMesh->mHexs.size());
	InitIntLap();

	//for (int i=0; i<(int)mX0.size(); i++)
	//if (mX0[i][0]>0)
	//	mX0[i][0] *= 1.5;

	mHexDefaultVolume.resize(mpVolMesh->mHexs.size());
	for (int h=0; h<(int)mHexDefaultVolume.size(); h++)
	{
		SVec3Float vtx[8];
		SSSVolMesh::_THex hex = mpVolMesh->mHexs[h];
		for (int i=0; i<8; i++)
			vtx[i] = mX0[hex[i]];
		PhysImpUtil::HexVol(vtx, &mHexDefaultVolume[h], NULL, NULL);
	}

	mTetDefaultVolume.resize(mpVolMesh->mTets.size());
	for (int t=0; t<(int)mTetDefaultVolume.size(); t++)
	{
		SVec3Float vtx[4];
		Vec4i tet = mpVolMesh->mTets[t];
		for (int i=0; i<4; i++)
			vtx[i] = mX0[tet[i]];
		PhysImpUtil::TetVol(vtx, &mTetDefaultVolume[t], NULL, NULL);
	}
	//mSymConst.clear();
	ComputeVMForce(true, ContVM);
	mJacRef.resize(mpVolMesh->mTets.size());
	mJacRefHex.resize(mpVolMesh->mHexs.size());
	mJac.resize(mpVolMesh->mTets.size());
	mJacHex.resize(mpVolMesh->mHexs.size());
	ComputeVMForce(true, InitJac);
	UpdateVMBound(mVMBound, 0, true, true);
	mPhase = 1;

	//mIgnoreLap.clear();
	//mIgnoreLapInput.clear();

	//int numMaxThread = omp_get_max_threads();
	//omp_set_num_threads(numMaxThread-1);
	//cout<<"Setting "<<numMaxThread-1<<" threads.\n";

	/*mSIMSolver.set_reorder_scheme("amd");
	mSIMSolver.set_equil(true);
	mSIMSolver.set_pivot("rook");*/
}

void PhysImpSolve3D::SetMaterialParam(_T_FLOAT iYoungs, _T_FLOAT iPoissionRatio)
{
	// for debug
	cout<<"Mat Size "<<sizeof(mE)<<endl;

	mE = MatXFloat::Zero(6, 6);

	for (int i=0; i<3; i++)
	{
		for (int j=0; j<3; j++)
		if (i!=j)
			mE(i, j) = iPoissionRatio;
		mE(i, i) = 1-iPoissionRatio;
		mE(i+3, i+3) = 1-2*iPoissionRatio;
	}

	mE *= iYoungs/((1+iPoissionRatio)*(1-2*iPoissionRatio));
	mYoungs = iYoungs;
	mPR = iPoissionRatio;
}

namespace
{
	template<class _TOut>
	void String2Array(const char* iString, int iN, _TOut* oArray)
	{
		stringstream ss(iString);
		for (int i=0; i<iN; i++)
			ss>>oArray[i];
	}
}

void PhysImpSolve3D::ParseConfigFile(const char* iFileName,
						SSSLib::SSSVolMesh& oVolMesh)
{
	using namespace tinyxml2;
	XMLDocument doc;
	doc.LoadFile(iFileName);
	bool bAreaWeight = false;

	XMLElement* pPhysImpNode = doc.FirstChildElement("PhysImp");

	// settings
	XMLElement* pSettings = pPhysImpNode->FirstChildElement("Settings");
	double modelScaling = 1;
	if (pSettings)
	{
		XMLElement* node;
		if (node=pSettings->FirstChildElement("RelativeEpsilon"))
			node->QueryDoubleText(&mEdgeLengthEpsilon);

		if (node=pSettings->FirstChildElement("bCorotation"))
			node->QueryBoolText(&mbCorotation);

		if (node=pSettings->FirstChildElement("bNegGD"))
			node->QueryBoolText(&mbNegGD);

		if (node=pSettings->FirstChildElement("bFloor"))
			node->QueryBoolText(&mbFloor);
		if (node=pSettings->FirstChildElement("FloorHeight"))
			node->QueryDoubleText(&mFloorHeight);
		if (node=pSettings->FirstChildElement("FloorForce"))
			node->QueryDoubleText(&mFloorForceWeight);

		if (node=pSettings->FirstChildElement("bGravity"))
			node->QueryBoolText(&mbGravity);
		if (node=pSettings->FirstChildElement("VMBound"))
		{
			node->QueryDoubleText(&mVMBound);
			mVMBound = sqr(mVMBound);
		}
		if (node=pSettings->FirstChildElement("Youngs"))
			node->QueryDoubleText(&mYoungs);
		if (node=pSettings->FirstChildElement("PoissonRatio"))
			node->QueryDoubleText(&mPR);
		if (node=pSettings->FirstChildElement("Density"))
			node->QueryDoubleText(&mDensity);

		if (node=pSettings->FirstChildElement("bUseEleSim"))
			node->QueryBoolText(&mbUseEleSim);

		if (node=pSettings->FirstChildElement("bMinVol"))
			node->QueryBoolText(&mbMinVol);

		if (node=pSettings->FirstChildElement("bMassLapWeight"))
			node->QueryBoolText(&mbMassLapWeight);

		if (node=pSettings->FirstChildElement("bInvLapWeight"))
			node->QueryBoolText(&mbInvLapWeight);

		if (node=pSettings->FirstChildElement("bAreaWeight"))
			node->QueryBoolText(&bAreaWeight);

		if (node=pSettings->FirstChildElement("bCholeskyAddedMultiple"))
			node->QueryBoolText(&mbCholeskyAddedMultiple);

		if (node=pSettings->FirstChildElement("LapSimNumRing"))
			node->QueryIntText(&mLapSimNumRing);

		if (node=pSettings->FirstChildElement("JacPercentBound"))
			node->QueryDoubleText(&mJacPercentBound);

		if (node=pSettings->FirstChildElement("RecalLapStep"))
			node->QueryIntText(&mRecalLapStep);
		
		if (node=pSettings->FirstChildElement("Scaling"))
			node->QueryDoubleText(&modelScaling);

		if (node=pSettings->FirstChildElement("NoiseSteps"))
			node->QueryIntText(&mNoiseSteps);
		
		if (node=pSettings->FirstChildElement("NoiseConvSteps"))
			node->QueryIntText(&mNoiseConvSteps);

		if (node=pSettings->FirstChildElement("NoiseProportion"))
			node->QueryDoubleText(&mNoiseProportion);

		if (node=pSettings->FirstChildElement("bStatic"))
			node->QueryBoolText(&mbStatic);

		if (node=pSettings->FirstChildElement("SimTimeStep"))
			node->QueryDoubleText(&mSimTimeStep);
	}

	// weights
	XMLElement* pWeights = pPhysImpNode->FirstChildElement("Weights");
	if (pWeights)
	{
		XMLElement* node;
		if (node=pWeights->FirstChildElement("VMWeight"))
			node->QueryDoubleText(&mVMWeight);
		if (node=pWeights->FirstChildElement("VMWeightMax"))
			node->QueryDoubleText(&mVMWeightMax);
		if (node=pWeights->FirstChildElement("LapSimWeight"))
			node->QueryDoubleText(&mLapSimWeight);
		if (node=pWeights->FirstChildElement("LapScaleWeight"))
			node->QueryDoubleText(&mLapScaleWeight);
		if (node=pWeights->FirstChildElement("LapFixedWeight"))
			node->QueryDoubleText(&mLapFixedWeight);
		if (node=pWeights->FirstChildElement("ShapeSimWeight"))
			node->QueryDoubleText(&mShapeSimWeight);
		if (node=pWeights->FirstChildElement("HexVolumeWeight"))
			node->QueryDoubleText(&mHexVolumeWeight);
		if (node=pWeights->FirstChildElement("SpeedWeight"))
			node->QueryDoubleText(&mSpeedWeight);
		if (node=pWeights->FirstChildElement("SmallJacWeight"))
			node->QueryDoubleText(&mVolumeWeight);
	}

	SetMaterialParam(mYoungs, mPR);

	string meshFileName = pPhysImpNode->FirstChildElement("VolumeMesh")->GetText();
	
	if (meshFileName.substr(meshFileName.length()-3, 3)=="inp")
		oVolMesh.LoadInp(meshFileName.c_str());
	else if (meshFileName.substr(meshFileName.length()-3, 3)=="vtk")
		oVolMesh.LoadVtk(meshFileName.c_str());
	else
		oVolMesh.LoadVolMesh(meshFileName.c_str());

	oVolMesh.Scale(modelScaling);

	oVolMesh.RemoveIsolatedVtx();

	oVolMesh.FindIntVtx();
	oVolMesh.BuildSurfaceMesh(true, true);

	oVolMesh.ComputeMass(mDensity);

	oVolMesh.BuildHexVtxConnect(true);
	oVolMesh.BuildVtxConnect(false);

	if (mNoiseProportion>0)
		mNoiseScale = oVolMesh.AvgEdgeLength()*mNoiseProportion;
	else
		mNoiseScale = 0;

	set<int> surfaceNodes;

	if (oVolMesh.mTets.size()>0 && oVolMesh.mHexs.size()>0)
		cout<<"Error: mixed tet and hex elements, not supported"<<endl;
	if (oVolMesh.mHexs.size()>0)
	{
		oVolMesh.BuildSurfaceMesh(false, true);
		for (int i=0; i<(int)oVolMesh.mSurfaceMesh.mQuad.size(); i++)
		{
			Vec4i quad = oVolMesh.mSurfaceMesh.mQuad[i];
			for (int v=0; v<4; v++)
				surfaceNodes.insert(quad[v]);
		}
	}

	if (oVolMesh.mTets.size()>0)
	{
		oVolMesh.BuildSurfaceMesh(true, false);
		for (int i=0; i<(int)oVolMesh.mSurfaceMesh.mFace.size(); i++)
		{
			Vec3i tri = oVolMesh.mSurfaceMesh.mFace[i];
			for (int v=0; v<3; v++)
				surfaceNodes.insert(tri[v]);
		}
	}

	{
		Vec3d high;
		Vec3d low;
		oVolMesh.mSurfaceMesh.ComputeBB(low, high);
		mMaxDim = Max(Max(high[0]-low[0], high[1]-low[1]), high[2]-low[2]);
		mEdgeLengthEpsilon *= mMaxDim;
		
	}

	oVolMesh.mSurfaceMesh.ComputeVtxAreaWeight();
	oVolMesh.mSurfaceMesh.BuildVtxConnect();

	KDTree3d<int> meshNodes;
	vector<int> ordering(oVolMesh.mVtx.size());
	for (int i=0; i<(int)oVolMesh.mVtx.size(); i++)
		ordering[i] = i;
	meshNodes.BuildTree(oVolMesh.mVtx.size(), &oVolMesh.mVtx[0], &ordering[0]);

	mVtxPresetForce.clear();
	mVtxPresetForce.resize(oVolMesh.mVtx.size(), SVec3Float(0, 0, 0));
	mFixedVtx.clear();
	mFixedForce.clear();
	mSymConst[0].clear();
	mSymConst[1].clear();
	mFixedDesignConst.clear();

	XMLElement* constraintNodes = pPhysImpNode->FirstChildElement("PointSet");
	while (constraintNodes!=NULL)
	{
		string constType = constraintNodes->Attribute("type");
		XMLElement* tmpNode;


		vector<Vec3d> nodeList;
		XMLElement* vtxFile = constraintNodes->FirstChildElement("File");
		while (vtxFile!=NULL)
		{
			SSSMesh mesh;
			mesh.LoadObj(vtxFile->GetText());
			mesh.Scale(Vec3d(modelScaling, modelScaling, modelScaling));
			nodeList.insert(nodeList.end(), mesh.mVtx.begin(), mesh.mVtx.end());

			vtxFile = vtxFile->NextSiblingElement("File");
		}

		KDTree3d<Vec3d> nodeTree;
		for (int i=0; i<(int)nodeList.size(); i++)
			nodeTree.InsertPt(nodeList[i], nodeList[i]);
		nodeTree.InitTree();

		vector<int> vtxList;
		vector<_T_FLOAT> vtxWeight;
		_T_FLOAT weightSum = 0;
		
		for (int i=0; i<(int)oVolMesh.mVtx.size(); i++)
		{
			Vec3d vtx = oVolMesh.mVtx[i];
			Vec3d closePt;
			Vec3d data;
			nodeTree.QueryClosestPt(vtx, &data, &closePt);
			//assert(data==closePt);
			//if ((vtx-closePt).Length()<1e-3 || (vtx-data).Length()<1e-3)
			if ((vtx-data).Length()<mEdgeLengthEpsilon)
			{
				vtxList.push_back(i);

				_T_FLOAT weight = 1.0;
				/*
				if (mbMassLapWeight)
					weight = oVolMesh.mMass[i];*/
				if (bAreaWeight)
					weight = oVolMesh.mSurfaceMesh.mVtxArea[i];

				weightSum += weight;
				vtxWeight.push_back(weight);
			}
		}

		cout<<"Loading Num Nodes "<<nodeList.size()<<endl
			<<" Num Vertices "<<vtxList.size()<<endl;

		if (constType=="fixed")
		{
			for (int i=0; i<(int)vtxList.size(); i++)
				mFixedVtx.insert(vtxList[i]);
		}
		else if (constType=="fixed_force")
		{
			_T_FLOAT youngs;
			constraintNodes->FirstChildElement("Youngs")->QueryDoubleText(&youngs);
			for (int i=0; i<(int)vtxList.size(); i++)
				mFixedForce[vtxList[i]] = youngs;
		}
		else if (constType=="fixed_force_dim")
		{
			_T_FLOAT youngs;
			constraintNodes->FirstChildElement("Youngs")->QueryDoubleText(&youngs);
			if (constraintNodes->FirstChildElement("DimX"))
				for (int i=0; i<(int)vtxList.size(); i++)
					mFixedForceDim[0][vtxList[i]] = youngs;
			if (constraintNodes->FirstChildElement("DimY"))
				for (int i=0; i<(int)vtxList.size(); i++)
					mFixedForceDim[1][vtxList[i]] = youngs;
			if (constraintNodes->FirstChildElement("DimZ"))
				for (int i=0; i<(int)vtxList.size(); i++)
					mFixedForceDim[2][vtxList[i]] = youngs;
			
		}
		else if (constType=="force")
		{
			Vec3d force(0, 0, 0);
			XMLElement* forceSum = constraintNodes->FirstChildElement("ForceSum");
			forceSum->FirstChildElement("x")->QueryDoubleText(force.val);
			forceSum->FirstChildElement("y")->QueryDoubleText(force.val+1);
			forceSum->FirstChildElement("z")->QueryDoubleText(force.val+2);

			//force /= vtxList.size();
			for (int i=0; i<(int)vtxList.size(); i++)
				mVtxPresetForce[vtxList[i]] += force*vtxWeight[i]/weightSum;
		}
		else if (constType=="fixed_design")
		{
			mFixedDesignConst.insert(vtxList.begin(), vtxList.end());
			mFixedDesignAny.insert(vtxList.begin(), vtxList.end());
		}
		else if (constType=="fixed_design_dim")
		{
			if (constraintNodes->FirstChildElement("DimX"))
				mFixedDesignDimConst[0].insert(vtxList.begin(), vtxList.end());
			if (constraintNodes->FirstChildElement("DimY"))
				mFixedDesignDimConst[1].insert(vtxList.begin(), vtxList.end());
			if (constraintNodes->FirstChildElement("DimZ"))
				mFixedDesignDimConst[2].insert(vtxList.begin(), vtxList.end());
			mFixedDesignAny.insert(vtxList.begin(), vtxList.end());
		}
		else if (constType=="Ignore_Lap")
		{
			mIgnoreLapInput.insert(vtxList.begin(), vtxList.end());
		}
		else
			cout<<"Unknown constraint type "<<constType<<endl;

		constraintNodes = constraintNodes->NextSiblingElement("PointSet");
	}

	vector<int> id(oVolMesh.mVtx.size());
	for (int i=0; i<(int)id.size(); i++)
		id[i] = i;
	vector<int> parents = id;
	boost::disjoint_sets<int*, int*> symSets(&id[0], &parents[0]);

	for (int i=0; i<2; i++)
	{

		XMLElement* symmetryNodes;
		if (i==0)
			symmetryNodes = pPhysImpNode->FirstChildElement("Symmetry");
		else
			symmetryNodes = pPhysImpNode->FirstChildElement("PhysicsSymmetry");
		while (symmetryNodes!=NULL)
		{
			vector<_T_FLOAT> transformData(16); 
			const char* matrixText = symmetryNodes->FirstChildElement("TransformMatrix")->GetText();
			String2Array(matrixText, 16, (_T_FLOAT*)&transformData[0]);
			MatXFloat transformMat(4, 4);
		
			for (int r=0; r<4; r++)
			for (int c=0; c<4; c++)
				transformMat(r, c) = transformData[r*4+c];

			_T_FLOAT symEpsilon = mEdgeLengthEpsilon;

			XMLElement* epsilonNode = symmetryNodes->FirstChildElement("Epsilon");
			if (epsilonNode)
				epsilonNode->QueryDoubleText(&symEpsilon);

			if (symmetryNodes->FirstChildElement("Global"))
			{
				//for (int i=0; i<(int)oVolMesh.mVtx.size(); i++)
				set<Vec2i> donePair;
				for (set<int>::iterator ite = surfaceNodes.begin();
					ite != surfaceNodes.end();
					ite++)
				{
					Vec3d vtx = oVolMesh.mVtx[*ite];
					Vec3d symVtx;
					Vec3d closePt;
					int idx;
					symVtx = Transform(vtx, transformMat);
					meshNodes.QueryClosestPt(symVtx, &idx, &closePt);
					closePt = oVolMesh.mVtx[idx];

					if ((closePt-symVtx).Length()<symEpsilon
							&& idx!=*ite
							&& (i==1 ||
									(
									!donePair.count(Vec2i(idx, *ite))
									&& symSets.find_set(idx)!=symSets.find_set(*ite)
									)
								)
						)
					{
						if (i==1 || (!mFixedDesignAny.count(idx) || !mFixedDesignAny.count(*ite)))
							AddSymConstraints(*ite, idx, transformMat, false, i);

						if (i==0)
						{
							donePair.insert(Vec2i(*ite, idx));

							symSets.link(idx, *ite);
						}
					}
				}

				int aa=0;
			}
			else
				cout<<"Unrecognize symmetry syntax!\n";

			if (i==0)
				symmetryNodes = symmetryNodes->NextSiblingElement("Symmetry");
			else
				symmetryNodes = symmetryNodes->NextSiblingElement("PhysicsSymmetry");
		}
	}
	// regularize symmtry pairs
	//for (int i=0; i<(int)mSymConst.size(); i++)
	//{

	//}
	// cleanup constraints

	for (int d=0; d<3; d++)
	{
		set<int> rmSet;
		set_intersection(mFixedDesignConst.begin(), mFixedDesignConst.end(),
			mFixedDesignDimConst[d].begin(), mFixedDesignDimConst[d].end(),
			std::inserter(rmSet, rmSet.begin()));
		for (set<int>::iterator ite = rmSet.begin();
			ite != rmSet.end();
			ite++)
			mFixedDesignDimConst[d].erase(*ite);
	}
}
