/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D.h"
#include "..\SSSLib\SSSFunc.h"

using namespace std;
using namespace SSSLib;

lbfgsfloatval_t PhysImpSolve3D::OptExportLBFGSCallback(
								void *instance,
								const lbfgsfloatval_t *iX,
								lbfgsfloatval_t *oG,
								const int iN,
								const lbfgsfloatval_t iStep)
{
	assert(sizeof(lbfgsfloatval_t)==sizeof(double));
	spPhysImp->CopybackState(iX);
	spPhysImp->OptExportPen();
	assert(iN == spPhysImp->mGrad.size());
	memcpy(oG, &spPhysImp->mGrad[0], iN);
	cout<<"Evaluating Result "<<spPhysImp->mFuncVal<<endl;
	return spPhysImp->mFuncVal;
}

int PhysImpSolve3D::OptExportLBFGSProgress(
				void *instance,
				const lbfgsfloatval_t *x,
				const lbfgsfloatval_t *g,
				const lbfgsfloatval_t fx,
				const lbfgsfloatval_t xnorm,
				const lbfgsfloatval_t gnorm,
				const lbfgsfloatval_t step,
				int n,
				int k,
				int ls)
{
	//if (n%100==0)
		cout<<"Num. Step: "<<n<<" Val "<<fx<<endl;
	return 0;
}

void PhysImpSolve3D::LBFGSOptThread(void* argv)
{
	spPhysImp->OptExportPenInit();

	int N = spPhysImp->mX0.size()*6;
	lbfgsfloatval_t *x = lbfgs_malloc(N*sizeof(_T_FLOAT));
    lbfgs_parameter_t param;

	assert(sizeof(lbfgsfloatval_t)==sizeof(_T_FLOAT));

	memcpy(x, spPhysImp->mX0[0].val, spPhysImp->mX0.size()*3*sizeof(_T_FLOAT));
	memcpy(x+spPhysImp->mX0.size()*3, spPhysImp->mP0[0].val, spPhysImp->mX0.size()*3*sizeof(_T_FLOAT));

	lbfgs_parameter_init(&param);

	vector<_T_FLOAT> grad(N);
	lbfgsfloatval_t val = OptExportLBFGSCallback(NULL, x, &grad[0], N, 1.0);
	cout<<"Init Val "<<val<<endl;

	lbfgs(N, x, NULL, OptExportLBFGSCallback, OptExportLBFGSProgress, NULL, &param);

	cout<<"l-BFGS finished."<<endl;

	lbfgs_free(x);
}
