/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve3D_util.h"
#include <fstream>

using namespace SSSLib;
using namespace std;
using namespace PhysImpUtil;

void PhysImpUtil::HexVol(const SVec3Float* iVtx,
		_T_FLOAT* oVol,
		Eigen::Matrix<_T_FLOAT, 24, 1>* oGrad,
		Eigen::Matrix<_T_FLOAT, 24, 24>* oHess)
{
	_T_FLOAT vol = 0;
	SVec3Float ctr(0, 0, 0);
	for (int j=0; j<8; j++)
		ctr += iVtx[j];
	ctr /= 8.0;

	Eigen::Matrix<_T_FLOAT, 24, 1> dVoldX;
	dVoldX.setZero();
	Eigen::Matrix<_T_FLOAT, 24, 24> d2VoldX2;
	d2VoldX2.setZero();

	//SVec3Float faceVtx0[4];
	SVec3Float faceVtx[3];
	Vec4i idx0;
	Vec2i idx;

	// for test
	//ofstream testOutput("test.txt");

	for (int fdir=0; fdir<3; fdir++)
	for (int fs=0; fs<2; fs++)
	for (int t=0; t<4; t++)
	{
		SVec3Float faceCtr(0, 0, 0);
		for (int v=0; v<4; v++)
		{
			idx0[v] = (fs<<fdir)
						+(((v&1)^fs)<<((fdir+1)%3))
						+((((v>>1)&1))<<((fdir+2)%3));
			faceCtr += iVtx[idx0[v]];
			//faceVtx0[v] = iVtx[idx0[v]];
		}
		faceCtr /= 4;

		switch (t)
		{
		case 0:
			idx = Vec2i(idx0[0], idx0[1]);
			break;
		case 1:
			idx = Vec2i(idx0[1], idx0[3]);
			break;
		case 2:
			idx = Vec2i(idx0[3], idx0[2]);
			break;
		case 3:
			idx = Vec2i(idx0[2], idx0[0]);
			break;
		default:
			break;
		}
		for (int v=0; v<2; v++)
			faceVtx[v] = iVtx[idx[v]];
		faceVtx[2] = faceCtr;

		SVec3Float left = ctr-faceVtx[2];
		SVec3Float right = cross3(faceVtx[0]-faceVtx[2], faceVtx[1]-faceVtx[2]);
		SVec3Float r1 = faceVtx[0]-faceVtx[2];
		SVec3Float r2 = faceVtx[1]-faceVtx[2];

		_T_FLOAT incvol = dot<_T_FLOAT>(left, right)/6.0;
		vol += incvol;
		//assert(incvol>0);

		if (oGrad==NULL && oHess==NULL)
			continue;
		
		Eigen::Matrix<_T_FLOAT, 24, 3> dLeft;
		dLeft.setZero();
		for (int c=0; c<3; c++)
		{
			for (int rr=0; rr<8; rr++)
				dLeft(rr*3+c, c) += 1.0/8.0;
			for (int rr=0; rr<4; rr++)
				dLeft(idx0[rr]*3+c, c) -= 1.0/4.0;
		}

		Eigen::Matrix<_T_FLOAT, 24, 3> dr1;
		dr1.setZero();
		Eigen::Matrix<_T_FLOAT, 24, 3> dr2;
		dr2.setZero();
		for (int c=0; c<3; c++)
		{
			dr1(idx[0]*3+c, c) += 1.0;
			dr2(idx[1]*3+c, c) += 1.0;

			for (int rr=0; rr<4; rr++)
			{
				dr1(idx0[rr]*3+c, c) -= 1.0/4.0;
				dr2(idx0[rr]*3+c, c) -= 1.0/4.0;
			}
		}

		Eigen::Matrix<_T_FLOAT, 24, 3> dRight;
		dRight.setZero();
		Eigen::Matrix<_T_FLOAT, 24, 24> d2Right[3];
		for (int c=0; c<3; c++)
		{
			dRight.col(c) += (dr1.col((c+1)%3))*r2[(c+2)%3]
							+r1[(c+1)%3]*(dr2.col((c+2)%3))
							-(dr1.col((c+2)%3))*r2[(c+1)%3]
							-r1[(c+2)%3]*(dr2.col((c+1)%3));

			if (oHess)
				d2Right[c] = MatXFloat(dr1.col((c+1)%3))*MatXFloat(dr2.col((c+2)%3)).transpose()
					-MatXFloat(dr1.col((c+2)%3))*MatXFloat(dr2.col((c+1)%3)).transpose();
		}

		for (int c=0; c<3; c++)
		{
			dVoldX += left[c]*dRight.col(c)/6.0
					+right[c]*dLeft.col(c)/6.0;
			if (oHess)
				d2VoldX2 += left[c]/6.0*d2Right[c];
				//d2VoldX2 += d2Right[c];
		}

		//testOutput<<dVoldX.transpose()<<endl;

		if (oHess)
			d2VoldX2 += dLeft/3.0*MatXFloat(dRight).transpose();
		
	}

	if (oVol)
		*oVol = vol;
	if (oGrad)
		*oGrad = dVoldX;

	if (oHess)
	{
		d2VoldX2 = MatXFloat(d2VoldX2)+MatXFloat(d2VoldX2).transpose();
		*oHess = d2VoldX2;
	}
}

void PhysImpUtil::TetVol(const SVec3Float* iVtx,
		_T_FLOAT* oVol,
		Eigen::Matrix<_T_FLOAT, 12, 1>* oGrad,
		Eigen::Matrix<_T_FLOAT, 12, 12>* oHess)
{
	SVec3Float edges[3];
	for (int i=0; i<3; i++)
		edges[i] = iVtx[i+1]-iVtx[0];

	SVec3Float left = edges[0];
	SVec3Float right = cross3(edges[1], edges[2]);

	_T_FLOAT vol = dot<_T_FLOAT>(left, right)/6;
	//assert(vol>0);

	Eigen::Matrix<_T_FLOAT, 12, 1> dVoldX;
	Eigen::Matrix<_T_FLOAT, 12, 12> d2VoldX2;

	if (oGrad!=NULL || oHess!=NULL)
	{
		
		dVoldX.setZero();
		d2VoldX2.setZero();

		Eigen::Matrix<_T_FLOAT, 12, 3> dLeft;
		dLeft.setZero();
		Eigen::Matrix<_T_FLOAT, 12, 3> dRight;
		dRight.setZero();

		Eigen::Matrix<_T_FLOAT, 12, 3> dr1;
		dr1.setZero();
		Eigen::Matrix<_T_FLOAT, 12, 3> dr2;
		dr2.setZero();

		Eigen::Matrix<_T_FLOAT, 12, 12> d2Right[3];

		SVec3Float r1 = edges[1];
		SVec3Float r2 = edges[2];

		for (int c=0; c<3; c++)
		{
			dLeft(c, c) = -1;
			dLeft(c+3, c) = 1;

			dr1(c, c) = -1;
			dr1(c+6, c) = 1;

			dr2(c, c) = -1;
			dr2(c+9, c) = 1;
		}

		// cross product
		for (int c=0; c<3; c++)
		{
			dRight.col(c) = dr1.col((c+1)%3)*r2[(c+2)%3]
							+r1[(c+1)%3]*dr2.col((c+2)%3)
							-dr1.col((c+2)%3)*r2[(c+1)%3]
							-r1[(c+2)%3]*dr2.col((c+1)%3);

			if (oHess)
				d2Right[c] = dr1.col((c+1)%3)*MatXFloat(dr2.col((c+2)%3)).transpose()
					-dr1.col((c+2)%3)*MatXFloat(dr2.col((c+1)%3)).transpose();
		}

		for (int c=0; c<3; c++)
		{
			if (oGrad)
				dVoldX += left[c]*dRight.col(c)/6.0
					+right[c]*dLeft.col(c)/6.0;
			if (oHess)
				d2VoldX2 += left[c]*d2Right[c]/6.0;
		}

		if (oHess)
			d2VoldX2 += dLeft/6.0*MatXFloat(dRight).transpose();
	}

	if (oVol)
		*oVol = vol;
	if (oGrad)
		*oGrad = dVoldX;

	if (oHess)
	{
		*oHess = d2VoldX2;
		*oHess = *oHess+MatXFloat(d2VoldX2).transpose();
	}
}