/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "../StaticFEM/Config.h"
#include <Eigen/Sparse>
#include <set>
#include <map>

namespace PhysImpUtil
{
	inline void PushTriplet(const std::map<SSSLib::Vec2i, _T_FLOAT>& iEles,
					_T_FLOAT iScale,
					std::vector<Eigen::Triplet<_T_FLOAT> >& ioTripleList)
	{
		using namespace std;
		using namespace SSSLib;

		for (map<Vec2i, _T_FLOAT>::const_iterator ite = iEles.begin();
			ite != iEles.end();
			ite++)
		if (ite->second!=0)
			ioTripleList.push_back(Eigen::Triplet<_T_FLOAT>(
						ite->first[0], ite->first[1], ite->second*iScale));
	}

	template <class _TMatIn, class _TMatOut>
	inline void Mat2SMat(const _TMatIn& iMat, _TMatOut& oMat)
	{
		for (int r=0; r<iMat.rows(); r++)
		for (int c=0; c<iMat.cols(); c++)
			oMat(r, c) = iMat(r, c);
	}

	extern void HexVol(const SVec3Float* iVtx,
		_T_FLOAT* oVol,
		Eigen::Matrix<_T_FLOAT, 24, 1>* oGrad,
		Eigen::Matrix<_T_FLOAT, 24, 24>* oHess);

	extern void TetVol(const SVec3Float* iVtx,
		_T_FLOAT* oVol,
		Eigen::Matrix<_T_FLOAT, 12, 1>* oGrad,
		Eigen::Matrix<_T_FLOAT, 12, 12>* oHess);

}

