/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve.h"
#include "..\SSSLib\SSSFunc.h"
#include <eigen/sparse>
#include <eigen/iterativeLinearSolvers>
#include <eigen/sparselu>
#include <iostream>

using namespace std;
using namespace SSSLib;

void PhysImpSolve::Strain2ndOrder(SVec2Float* iX, SVec2Float* iP,
							MatXFloat& oS,
							MatXFloat& odSdX, MatXFloat& odSdP,
							std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >& od2SdX2,
							std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >& od2SdXdP)
							//std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >& od2SdP2)
{
	MatXFloat dinvdetdX;
	MatXFloat ddetdX = DetX1stOrder(iX, &dinvdetdX);
	_T_FLOAT det;
	MatXFloat B0 = Static2D::GetB0(iX[0], iX[1], iX[2], &det);

	if (det<msDetEpsilon)
	{
		oS = MatXFloat::Zero(3, 1);
		odSdX = MatXFloat::Zero(3, 6);
		odSdP = MatXFloat::Zero(3, 6);
		od2SdX2 = vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >(3, MatXFloat::Zero(6, 6));
		od2SdXdP = vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >(3, MatXFloat::Zero(6, 6));
		return;
	}

	_T_FLOAT invdet = 1.0/det;
	MatXFloat q = MatXFloat::Zero(6, 1);
	for (int i=0; i<3; i++)
	{
		q(i*2, 0) = iP[i][0];
		q(i*2+1, 0) = iP[i][1];
	}
	MatXFloat B0q = B0*q;
	oS = invdet*B0q;
	odSdX = B0q*dinvdetdX.transpose();
	for (int i=0; i<6; i++)
		odSdX.col(i) += invdet*msdBdX[i]*q;

	odSdP = invdet*B0;

	// d2SdX2
	od2SdX2 = vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> >(3, MatXFloat::Zero(6, 6));
	vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> > d2SdX2_XSX(6);
	for (int i=0; i<6; i++)
	{
		d2SdX2_XSX[i] = msdBdX[i]*q*dinvdetdX.transpose();
		for (int j=0; j<3; j++)
		{
			od2SdX2[j].row(i) += d2SdX2_XSX[i].row(j);
		}
	}

	MatXFloat d2InvDetdX2 = 2*sqr(invdet)*invdet*ddetdX*ddetdX.transpose()
							-sqr(invdet)*msd2detdX2;

	for (int i=0; i<3; i++)
	{
		MatXFloat d2SdX2Trans = od2SdX2[i].transpose();
		od2SdX2[i] = od2SdX2[i]+d2SdX2Trans;
		od2SdX2[i] += B0q(i,0)*d2InvDetdX2;
	}

	// d2SdXdP
	//od2SdXdP.resize(3, MatXFloat::Zero(6, 6));
	od2SdXdP.resize(3);
	for (int i=0; i<3; i++)
	{
		od2SdXdP[i] = dinvdetdX*B0.row(i);
		for (int j=0; j<6; j++)
			od2SdXdP[i].row(j) += invdet*msdBdX[j].row(i);
	}

}

void PhysImpSolve::VM2ndOrder(SVec2Float* iX, SVec2Float* iP,
							MatXFloat iD, _T_FLOAT iThick,
							_T_FLOAT& oVM,
							MatXFloat& odVMdXP,
							MatXFloat& od2VMdXP2)
{
	MatXFloat dSdX;
	MatXFloat dSdP;
	std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> > d2SdX2;
	std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> > d2SdXdP;
	MatXFloat strain;

	Strain2ndOrder(iX, iP, strain, dSdX, dSdP, d2SdX2, d2SdXdP);

	MatXFloat dSdXP(3, 12);
	dSdXP<<dSdX,dSdP;

	std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> > d2SdXP(3, MatXFloat(12, 12));

	for (int i=0; i<3; i++)
		d2SdXP[i]<<d2SdX2[i], d2SdXdP[i], d2SdXdP[i].transpose(), MatXFloat::Zero(6, 6);

	MatXFloat dtvd = iD.transpose()*msV*iD;
	MatXFloat stdtvd = strain.transpose()*dtvd;

	oVM = (stdtvd*strain)(0, 0);
	
	odVMdXP = 2*(stdtvd*dSdXP).transpose();
	
	od2VMdXP2 = MatXFloat::Zero(12, 12);
	for (int i=0; i<3; i++)
	{
		od2VMdXP2 += 2*stdtvd(0,i)*d2SdXP[i];
		for (int j=0; j<3; j++)
		{
			/*MatXFloat val = dSdXP.col(i).transpose()*dtvd(i, j)*dSdXP.col(j);
			assert(val.rows()==1 && val.cols()==1);
			od2VMdXP2(i, j) += 2*val(0, 0);*/
			MatXFloat val = dSdXP.row(i).transpose()*dtvd(i, j)*dSdXP.row(j);
			//assert(val.rows()==1 && val.cols()==1);
			od2VMdXP2 += 2*val;
		}
	}
}

void PhysImpSolve::ComputeVMForce()
{
	for (int t=0; t<(int)mpStatic2D->mTriangles.size(); t++)
	{
		Vec3i tri = mpStatic2D->mTriangles[t];

		vector<SVec2Float> x(3);
		vector<SVec2Float> p(3);

		for (int v=0; v<3; v++)
		{
			x[v] = mpStatic2D->mRest[tri[v]];
			p[v] = mpStatic2D->mStatic[tri[v]];
			p[v] -= x[v];
		}

		SVec3Float stress;
		//mpStatic2D->ComputeStress(&x[0], &p[0], mpStatic2D->mD, stress);
		//_T_FLOAT vm = mpStatic2D->VonMisesSqr(stress);
		_T_FLOAT vm = mpStatic2D->ComputeVM(t);

		if (mVMWeight>0 && vm>mVMBound)
		{		
			MatXFloat dVMdXP;
			MatXFloat d2VMdXP2;

			_T_FLOAT tmpVM;
			
			VM2ndOrder(&x[0], &p[0],
				mpStatic2D->mD, mpStatic2D->mThick,
				tmpVM,
				dVMdXP, d2VMdXP2);

			//assert(fabs(tmpVM-vm)<vm*1e-7);

			_T_FLOAT scale = -mVMWeight*(vm-mVMBound);
			//_T_FLOAT scale = mVMWeight;
			for (int v=0; v<3; v++)
			for (int c=0; c<2; c++)
			{ 
				//_T_FLOAT tmpMonitor = dVMdXP(v*2+c, 0);
				//_T_FLOAT tmpMonitor2 = scale*tmpMonitor;
				mExtForceX[tri[v]][c] += scale*(dVMdXP(v*2+c, 0)-dVMdXP(v*2+c+6, 0));
#ifndef _ASYM_LINEAR
				mExtForceP[tri[v]][c] += scale*dVMdXP(v*2+c+6, 0);
#endif

				// for debug
#if 1
				{
					_T_FLOAT sigma = 1e-2;
					vector<SVec2Float> tmpx = x;
					vector<SVec2Float> tmpp = p;
					for (int i=0; i<3; i++)
						tmpp[i] = x[i]+p[i];

					SVec3Float stress;
					_T_FLOAT vmp;
					_T_FLOAT vmm;
					
					tmpx[v][c] = x[v][c]+sigma;
					mpStatic2D->ComputeStress(&tmpx[0], &tmpp[0], mpStatic2D->mD, stress);
					vmp = mpStatic2D->VonMisesSqr(stress);

					tmpx[v][c] = x[v][c]-sigma;
					mpStatic2D->ComputeStress(&tmpx[0], &tmpp[0], mpStatic2D->mD, stress);
					vmm = mpStatic2D->VonMisesSqr(stress);

					_T_FLOAT grad = (vmp-vmm)/sigma/2.0;
					_T_FLOAT compgrad = dVMdXP(v*2+c, 0)-dVMdXP(v*2+c+6, 0);

					_T_FLOAT delta = grad-compgrad;

				}
#endif
			}
			//MatXFloat d2fdX2 = scale*d2VMdXP2;
			MatXFloat d2fdX2 = -mVMWeight*dVMdXP*dVMdXP.transpose()
								+scale*d2VMdXP2;
#ifdef _ASYM_LINEAR
			for (int i=0; i<6; i++)
#else
			for (int i=0; i<12; i++)
#endif
			for (int j=0; j<12; j++)
			//if (d2fdX2(i, j)!=0)
			{
				int idx0 = tri[(i%6)/2]*2+i%2;
				int idx1 = tri[(j%6)/2]*2+j%2;
				if (i>=6)
					idx0 += (int)mXV.size()*2;
				if (j>=6)
					idx1 += (int)mXV.size()*2;
				_T_FLOAT val = d2fdX2(i, j);
				if (i<6 && j<6)
				{
					val += -d2fdX2(i, j+6) - d2fdX2(i+6, j) + d2fdX2(i+6, j+6);
				}
				else if (i<6)
					val += -d2fdX2(i+6, j);
				else if (j<6)
					val += -d2fdX2(i, j+6);
				if (val!=0)
					mdExtForce.push_back(Eigen::Triplet<_T_FLOAT>(idx0, idx1, val));
			}
		}
	}
}