/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PhysImpSolve.h"
#include "..\SSSLib\SSSFunc.h"
#include <eigen/sparse>
#include <eigen/iterativeLinearSolvers>
#include <eigen/sparselu>
#include <iostream>

using namespace std;
using namespace SSSLib;

std::vector<MatXFloat, Eigen::aligned_allocator<MatXFloat> > PhysImpSolve::msdBdX;
MatXFloat PhysImpSolve::msd2detdX2;
MatXFloat PhysImpSolve::msV;
//_T_FLOAT PhysImpSolve::msDetEpsilon = 1e-2;
_T_FLOAT PhysImpSolve::msDetEpsilon = 0;

PhysImpSolve::PhysImpSolve(void)
	: mbCarryMass(true)
	, mbGravity(false)
	//, mbFloor(false)
	, mbFloor(true)
	, mFloorHeight(-0.6)
	, mFloorForceWeight(1e7)
	, mFDamping(1.0)
	, mVDamping(0.3)
	//, mDamping(1)
	, mXDeformWeight(3e4)
	, mLaplacianWeight(1e12)
	//, mLaplacianWeight(1e5)
	//, mLaplacianWeight(3e7)
	//, mVecLaplacianWeight(1e7)
	//, mVecLaplacianWeight(1e23)
	//, mGridWeight(1e7)
	, mVecLaplacianWeight(1e15)
	, mEdgeCrossWeight(1e15)
	, mVMWeight(2e-22)
	//, mVMWeight(0)
	, mVMBound(5e12)
	, mXMassScale(1000.0)
	, mbStatic(false)
	, mbLapMassWeight(false)
	, mbXUniformMass(true)
	, mXUniformMass(1e4)
	, mXSpringForce(1e6)
	, mbPStatic(false)
	, mXStaticStep(1e-2)
	, mMatrixEpsilon(1e-10)
	//, mXStaticStep(1e-5)
{
}

PhysImpSolve::~PhysImpSolve(void)
{
}

void PhysImpSolve::Init()
{
	//mX0 = mpStatic2D->mRest;

	mPV.resize(mpStatic2D->mRest.size(), SVec2Float(0, 0));
	mXV.resize(mpStatic2D->mRest.size(), SVec2Float(0, 0));
	mX0.resize(mpStatic2D->mRest.size(), SVec2Float(0, 0));

	for (int i=0; i<(int)mpStatic2D->mRest.size(); i++)
		mX0[i] = mpStatic2D->mRest[i];

	mExtForceP.resize((int)mpStatic2D->mForce.size());
	for (int i=0; i<(int)mExtForceP.size(); i++)
		mExtForceP[i] = mpStatic2D->mForce[i];
	//mdExtForceP = vector<SMat2x2>(mExtForceP.size(), SMat2x2::Zeros());

	mExtForceX = vector<SVec2Float>(mpStatic2D->mRest.size(), SVec2Float(0, 0));
	//mdExtForceX = vector<SMat2x2>(mExtForceX.size(), SMat2x2::Zeros());

	mMass.resize(mpStatic2D->mRest.size(), 0);
	for (int t=0; t<(int)mpStatic2D->mTriangles.size(); t++)
	{
		Vec3i tri = mpStatic2D->mTriangles[t];
		_T_FLOAT area = ComputeSignedArea(t);
		for (int i=0; i<3; i++)
		{
			_T_FLOAT length[2];
			SVec2Float edge[2];
			for (int j=0; j<2; j++)
			{
				edge[j] = mpStatic2D->mRest[tri[(i+j+1)%3]]-mpStatic2D->mRest[tri[i]];
				length[j] = edge[j].Length();
			}
			_T_FLOAT angle = asin(area*2/length[0]/length[1]);
			if (dot<_T_FLOAT>(edge[0], edge[1])<0)
				angle = Pi-angle;
			mMass[tri[i]] += area/Pi*angle*mpStatic2D->mThick*mDensity;
		}
	}

	mpStatic2D->mStatic = mpStatic2D->mRest;
	//for (int i=0; i<(int)mpStatic2D->mStatic.size(); i++)
	//	mpStatic2D->mStatic[i] *= 1.5;

	msdBdX.clear();
	msdBdX.resize(6, MatXFloat::Zero(3, 6));

	
	msdBdX[0](1, 3) = 1;
	msdBdX[0](1, 5) = -1;
	msdBdX[0](2, 2) = 1;
	msdBdX[0](2, 4) = -1;

	msdBdX[2](1, 5) = 1;
	msdBdX[2](1, 1) = -1;
	msdBdX[2](2, 4) = 1;
	msdBdX[2](2, 0) = -1;

	msdBdX[4](1, 1) = 1;
	msdBdX[4](1, 3) = -1;
	msdBdX[4](2, 0) = 1;
	msdBdX[4](2, 2) = -1;

	msdBdX[1](0, 4) = 1;
	msdBdX[1](0, 2) = -1;
	msdBdX[1](2, 5) = 1;
	msdBdX[1](2, 3) = -1;

	msdBdX[3](0, 0) = 1;
	msdBdX[3](0, 4) = -1;
	msdBdX[3](2, 1) = 1;
	msdBdX[3](2, 5) = -1;

	msdBdX[5](0, 2) = 1;
	msdBdX[5](0, 0) = -1;
	msdBdX[5](2, 3) = 1;
	msdBdX[5](2, 1) = -1;

	msd2detdX2 = MatXFloat::Zero(6, 6);

	msd2detdX2(0, 3) = msd2detdX2(3, 0) = 1;
	msd2detdX2(0, 5) = msd2detdX2(5, 0) = -1;
	msd2detdX2(2, 5) = msd2detdX2(5, 2) = 1;
	msd2detdX2(2, 1) = msd2detdX2(1, 2) = -1;
	msd2detdX2(4, 1) = msd2detdX2(1, 4) = 1;
	msd2detdX2(4, 3) = msd2detdX2(3, 4) = -1;

	msV = MatXFloat::Zero(3, 3);
	msV(0, 0) = 1;
	msV(0, 1) = -0.5;
	msV(1, 0) = -0.5;
	msV(1, 1) = 1;
	msV(2, 2) = 3;

	mpStatic2D->ComputeBoundaryVtx();

	BuildVertexConnection();
}

void PhysImpSolve::BuildVertexConnection()
{
	mVtxConnection.clear();
	mVtxConnection.resize(mpStatic2D->mRest.size(), set<int>());

	mVtxConnectionHV[0].clear();
	mVtxConnectionHV[1].clear();
	mVtxConnectionHV[0].resize(mpStatic2D->mRest.size(), set<int>());
	mVtxConnectionHV[1].resize(mpStatic2D->mRest.size(), set<int>());

	mEdges0.clear();
	mEdges0.resize(mpStatic2D->mRest.size());

	for (int f=0; f<(int)mpStatic2D->mTriangles.size(); f++)
	{
		Vec3i tri = mpStatic2D->mTriangles[f];
		for (int v=0; v<3; v++)
		for (int v2=1; v2<=2; v2++)
		{
			int v2idx = tri[(v+v2)%3];
			mVtxConnection[tri[v]].insert(v2idx);
			mEdges0[tri[v]][v2idx] = mX0[tri[v]]-mX0[v2idx];

			if (fabs(mEdges0[tri[v]][v2idx][0])<1e-6)
				mVtxConnectionHV[1][tri[v]].insert(v2idx);
			else if (fabs(mEdges0[tri[v]][v2idx][1])<1e-6)
				mVtxConnectionHV[0][tri[v]].insert(v2idx);
		}
	}

	mVtxControls.clear();
	mVtxControls.resize(mpStatic2D->mRest.size());
	for (int i=0; i<(int)mVtxControls.size(); i++)
	{
		for (int c=0; c<2; c++)
		if (mVtxConnectionHV[c][i].size()==2)
		{
			set<int>::iterator ite = mVtxConnectionHV[c][i].begin();
			int v[2];
			v[0] = *ite;
			ite++;
			v[1] = *ite;
			if (!mpStatic2D->mBoundaryVtx[i] || 
				(mpStatic2D->mBoundaryVtx[v[0]] && mpStatic2D->mBoundaryVtx[v[1]]))
				mVtxControls[i].insert(mVtxConnectionHV[c][i].begin(), mVtxConnectionHV[c][i].end());
		}
	}
	
	mVtxConnectionBound.clear();
	mVtxConnectionBound.resize(mpStatic2D->mRest.size());

	for (int f=0; f<(int)mpStatic2D->mTriangles.size(); f++)
	{
		Vec3i tri = mpStatic2D->mTriangles[f];
		for (int v=0; v<3; v++)
		for (int v2=1; v2<=2; v2++)
		{
			int vidx = tri[v];
			int v2idx = tri[(v+v2)%3];
			if (mpStatic2D->mBoundaryVtx[vidx] && mpStatic2D->mBoundaryVtx[v2idx]
				&& (mVtxConnectionHV[0][vidx].count(v2idx)>0
						|| mVtxConnectionHV[1][vidx].count(v2idx)>0))
				mVtxConnectionBound[vidx].insert(v2idx);
		}
	}

	mBoundLaplacian.clear();
	mBoundLaplacian.resize(mpStatic2D->mRest.size(), SVec2Float(0, 0));
	for (int i=0; i<(int)mBoundLaplacian.size(); i++)
	if (mVtxConnectionBound[i].size()>0)
	{
		mBoundLaplacian[i] = ComputeLaplacianWithNeighbour(i, mX0, mVtxConnectionBound[i]);
	}

	mLaplacian.clear();
	mLaplacian.resize(mX0.size(), SVec2Float(0, 0));
	mLapWeight.resize(mX0.size(), 0);
	for (int v=0; v<(int)mLaplacian.size(); v++)
		mLaplacian[v] = ComputeLaplacian(v, mX0, mbLapMassWeight, &mLapWeight[v]);
}

_T_FLOAT PhysImpSolve::ComputeSignedArea(int idx)
{
	Vec3i tri = mpStatic2D->mTriangles[idx];
	SVec2Float vtx[3];
	for (int i=0; i<3; i++)
		vtx[i] = mpStatic2D->mRest[tri[i]];
	return 0.5*cross2<_T_FLOAT>(vtx[1]-vtx[0], vtx[2]-vtx[0]);
}

void PhysImpSolve::ComputeExtForce()
{
	mdExtForce.clear();
	for (int i=0; i<(int)mExtForceX.size(); i++)
		mExtForceX[i] = SVec2Float(0, 0);
	for (int i=0; i<(int)mExtForceP.size(); i++)
		mExtForceP[i] = SVec2Float(0, 0);
	ComputeXForce();
	ComputePForce();
	ComputeVMForce();
}

void PhysImpSolve::ComputeXForce()
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	//for (int i=0; i<(int)mX0.size(); i++)
	//{
	//	mExtForceX[i] += mXDeformWeight*(mX0[i]-SVec2Float(mpStatic2D->mRest[i]));
	//	//mdExtForceX[i] = -mXDeformWeight*SMat2x2::Ones();
	//	mdExtForce.push_back(T(i*2, i*2, -mXDeformWeight));
	//	mdExtForce.push_back(T(i*2+1, i*2+1, -mXDeformWeight));
	//}

	//ComputeXVecLaplacianForce();
	ComputeXControlForce();
	//ComputeXControlForceB();
	//ComputeGridForce();
	//ComputeXSpringForce();
	//ComputeXLaplacianForce();
}

void PhysImpSolve::ComputeXLapVec(int iIdx, SVec2Float iL0, std::set<int>& iConnection)
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	int i = iIdx;
	set<int>& connection = iConnection;

	SVec2Float restLap = ComputeLaplacianWithNeighbour(i, mpStatic2D->mRest, connection);
	SVec2Float ldelta = restLap-iL0;

	mExtForceX[i] += -mVecLaplacianWeight*ldelta;
		
	for (int c=0; c<2; c++)
		mdExtForce.push_back(T(i*2+c, i*2+c, -mVecLaplacianWeight));

	int N = connection.size();

	for (set<int>::iterator v=connection.begin();
		v!=connection.end(); v++)
	for (int c=0; c<2; c++)
	{

		mExtForceX[*v][c] += mVecLaplacianWeight/N*ldelta[c];

		mdExtForce.push_back(T(i*2+c, (*v)*2+c, mVecLaplacianWeight/N));

		mdExtForce.push_back(T((*v)*2+c, i*2+c, mVecLaplacianWeight/N));

		for (set<int>::iterator v2=connection.begin();
		v2!=connection.end(); v2++)
		{
			mdExtForce.push_back(T((*v)*2+c, (*v2)*2+c, -mVecLaplacianWeight/sqr(N)));
		}
	}
}

void PhysImpSolve::ComputeXLap(int iIdx, _T_FLOAT iL0, std::set<int>& iConnection)
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	set<int>& connection = iConnection;
	int i = iIdx;

	SVec2Float restLap = ComputeLaplacianWithNeighbour(i, mpStatic2D->mRest, connection);
	_T_FLOAT l = restLap.SqrLength();
	_T_FLOAT l0 = iL0;

	Vec2Float el;
	el(0) = restLap[0];
	el(1) = restLap[1];

	SVec2Float dEldX = mLaplacianWeight*2*(l-l0)*restLap;
	MatXFloat d2EldX2 = mLaplacianWeight*4*el*el.transpose()
							+mLaplacianWeight*2*(l-l0)*MatXFloat::Identity(2, 2);

	mExtForceX[i] += -dEldX;
	for (int r=0; r<2; r++)
	for (int c=0; c<2; c++)
		mdExtForce.push_back(T(i*2+r, i*2+c, -d2EldX2(r, c)));

	int N = connection.size();

	for (set<int>::iterator v=connection.begin();
		v!=connection.end(); v++)
	{
		mExtForceX[*v] += 1.0/N*dEldX;

		for (int r=0; r<2; r++)
		for (int c=0; c<2; c++)
		{
			mdExtForce.push_back(T(i*2+r, *v+c, 1.0/N*d2EldX2(r, c)));
			mdExtForce.push_back(T(*v+r, i*2+c, 1.0/N*d2EldX2(r, c)));
		}

		for (set<int>::iterator v2=connection.begin();
		v2!=connection.end(); v2++)
		{
			for (int r=0; r<2; r++)
			for (int c=0; c<2; c++)
				mdExtForce.push_back(T(*v+r, *v2+c, -1.0/N/N*d2EldX2(r, c)));
		}
	}

}

void PhysImpSolve::ComputeXEdgeCross(int iIdx, std::set<int>& iConnection)
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	for (set<int>::iterator ite = iConnection.begin();
		ite != iConnection.end();
		ite++)
	{
		SVec2Float e0 = mX0[iIdx]-mX0[*ite];
		SVec2Float e = mpStatic2D->mRest[iIdx]-mpStatic2D->mRest[*ite];
		_T_FLOAT E = cross2<_T_FLOAT>(e, e0);
		SVec2Float dEdx(E*e0[1], -E*e0[0]);
		dEdx *= mEdgeCrossWeight;
		Vec2Float crossVec;
		crossVec(0) = e0[1];
		crossVec(1) = -e0[0];
		MatXFloat d2Edx2 = mEdgeCrossWeight*crossVec*crossVec.transpose();

		mExtForceX[iIdx] += -dEdx;
		mExtForceX[*ite] += dEdx;

		for (int r=0; r<2; r++)
		for (int c=0; c<2; c++)
		{
			mdExtForce.push_back(T(iIdx*2+r, iIdx*2+c, -d2Edx2(r, c)));
			mdExtForce.push_back(T(iIdx*2+r, *ite*2+c, d2Edx2(r, c)));
			mdExtForce.push_back(T(*ite*2+r, iIdx*2+c, d2Edx2(r, c)));
			mdExtForce.push_back(T(*ite*2+r, *ite*2+c, -d2Edx2(r, c)));
		}

	}
}

void PhysImpSolve::ComputeXControlForce()
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	for (int i=0; i<(int)mX0.size(); i++)
	if (mVtxControls[i].size()>0)
	{
		ComputeXLapVec(i, SVec2Float(0, 0), mVtxControls[i]);
	}
	else
		ComputeXEdgeCross(i, mVtxConnectionBound[i]);
}

void PhysImpSolve::ComputeXControlForceB()
{
	for (int i=0; i<(int)mX0.size(); i++)
	if (mpStatic2D->mBoundaryVtx[i] && mVtxConnectionBound[i].size()>0)
		ComputeXLap(i, mBoundLaplacian[i].SqrLength(), mVtxConnectionBound[i]);
	else if (!mpStatic2D->mBoundaryVtx[i] && mVtxControls[i].size()>0)
	{
		ComputeXLapVec(i, SVec2Float(0, 0), mVtxControls[i]);
	}
}

void PhysImpSolve::ComputeXVecLaplacianForce()
{
	ComputeXLaplacianForceWithNeighbour(mVtxConnection);
}

void PhysImpSolve::ComputeXLaplacianForceWithNeighbour(std::vector<std::set<int> >& iNeighbourList)
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	for (int i=0; i<(int)mX0.size(); i++)
	//if (!mpStatic2D->mBoundaryVtx[i])
	{
		ComputeXLapVec(i, mLaplacian[i], iNeighbourList[i]);
	}
}

void PhysImpSolve::ComputeXLaplacianForce()
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	for (int i=0; i<(int)mX0.size(); i++)
	if (mpStatic2D->mBoundaryVtx[i])
	{
		ComputeXLap(i, mLaplacian[i].SqrLength(), mVtxConnection[i]);
	}
}

void PhysImpSolve::ComputeGridForce()
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	for (int i=0; i<(int)mX0.size(); i++)
	for (int c=0; c<2; c++)
	if (mVtxConnectionHV[c][i].size()==2)
	{
		set<int>& connection = mVtxConnectionHV[c][i];
		set<int>::iterator ite = connection.begin();
		int v[2];
		v[0] = *ite;
		ite++;
		v[1] = *ite;

		SVec2Float restLap = mpStatic2D->mRest[i]
							-0.5*(mpStatic2D->mRest[v[0]]+mpStatic2D->mRest[v[1]]);
		SVec2Float ldelta = restLap;

		mExtForceX[i][c] += -mVecLaplacianWeight*ldelta[c];
		mdExtForce.push_back(T(i*2+c, i*2+c, -mVecLaplacianWeight));

		int N = 2;

		for (set<int>::iterator v=connection.begin();
			v!=connection.end(); v++)
		{

			mExtForceX[*v][c] += mVecLaplacianWeight/N*ldelta[c];

			mdExtForce.push_back(T(i*2+c, (*v)*2+c, mVecLaplacianWeight/N));

			mdExtForce.push_back(T((*v)*2+c, i*2+c, mVecLaplacianWeight/N));

			for (set<int>::iterator v2=connection.begin();
			v2!=connection.end(); v2++)
			{
				mdExtForce.push_back(T((*v)*2+c, (*v2)*2+c, -mVecLaplacianWeight/sqr(N)));
			}
		}
	}
}

void PhysImpSolve::ComputeXSpringForce()
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	for (int v0=0; v0<(int)mEdges0.size(); v0++)
	{
		for (map<int, SVec2Float>::iterator v1 = mEdges0[v0].begin();
			v1 != mEdges0[v0].end();
			v1++)
		{
			SVec2Float edge = mpStatic2D->mRest[v0]-mpStatic2D->mRest[v1->first];
			_T_FLOAT l = edge.Length();
			_T_FLOAT l0 = v1->second.Length();
			SVec2Float dUdX = -mXSpringForce*edge.Norm()*(l-l0);
			mExtForceX[v0] += dUdX;

			Vec2Float eigenEdge;
			eigenEdge(0) = edge[0];
			eigenEdge(1) = edge[1];
			MatXFloat d2UdX2;
			d2UdX2 = mXSpringForce*(-MatXFloat::Identity(2, 2)
								+l0/l
									*(MatXFloat::Identity(2, 2)-eigenEdge*eigenEdge.transpose()/sqr(l)));

			for (int r=0; r<2; r++)
			for (int c=0; c<2; c++)
			if (d2UdX2(r, c)!=0)
			{
				mdExtForce.push_back(T(v0, v0, d2UdX2(r, c)));
				mdExtForce.push_back(T(v0, v1->first, -d2UdX2(r, c)));
			}

		}
	}
}

void PhysImpSolve::ComputePForce()
{
	typedef Eigen::Triplet<_T_FLOAT> T;
	for (int i=0; i<(int)mX0.size(); i++)
	{
		mExtForceP[i] += SVec2Float(mpStatic2D->mForce[i]);
		if (mbGravity)
			mExtForceP[i][1] -= 9.8*mMass[i];
		if (mbFloor)
		if (mpStatic2D->mStatic[i][1]<mFloorHeight)
		{
			mExtForceP[i][1] += mFloorForceWeight*(mFloorHeight-mpStatic2D->mStatic[i][1]);
			mdExtForce.push_back(T(i*2+1+mXV.size()*2, i*2+1+mXV.size()*2, -mFloorForceWeight));
		}
	}
}
