/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Static2DADMM.h"
#include "XNodeF.h"
#include "XNodeFixedX.h"
#include "XNodeXDeform.h"
#include "XNodeFSum.h"
#include "XNodeVM.h"

#ifdef _USE_MATLAB
#include "..\matlab\FOptLib.h"
#endif

using namespace ADMMLib;
using namespace std;

Static2DADMM::Static2DADMM(void)
	: mMaxVM(600)
	, mForceScale(0.5)
	, mVMScale(0.05)
{
#ifdef _USE_MATLAB
	if (!mclInitializeApplication(NULL, 0))
	{
		cout<<"Matlab Init Failed!\n";
	}
	if (!FOptLibInitialize())
	{
		mclTerminateApplication();
		cout<<"FOptLib Init Failed!\n";
	}
#endif
}

Static2DADMM::~Static2DADMM(void)
{
#ifdef _USE_MATLAB
	FOptLibTerminate();
#endif
}

void Static2DADMM::SetupGraph()
{
	mpZNodes.clear();
	mpXNodes.clear();
	// setup z-nodes
	// z-node list
	mZRest.clear();
	mZStatic.clear();
	mZTriangleForce.clear();
	mZVM.clear();
	//mXForce.clear();
	mpXNodeF.clear();
	mpXNodeVM.clear();

	mZRest.resize(mpFEM2D->mRest.size()*2);
	mZStatic.resize(mpFEM2D->mStatic.size()*2);
	mZTriangleForce.resize(mpFEM2D->mTriangles.size()*6);
	mZVM.resize(mpFEM2D->mTriangles.size());
	mpXNodeVM.resize(mpFEM2D->mTriangles.size());
	// setup vertices
	for (int i=0; i<(int)mpFEM2D->mRest.size(); i++)
	{
		mpZNodes.push_back(unique_ptr<ZNode>(new ZNode(this, mRho)));
		mpZNodes.push_back(unique_ptr<ZNode>(new ZNode(this, mRho)));
		mZRest[i*2] = mpZNodes[mpZNodes.size()-2].get();
		mZRest[i*2+1] = mpZNodes[mpZNodes.size()-1].get();
	}
	for (int i=0; i<(int)mpFEM2D->mStatic.size(); i++)
	if (!mpFEM2D->mFixedVtx[i])
	{
		mpZNodes.push_back(unique_ptr<ZNode>(new ZNode(this, mRho)));
		mpZNodes.push_back(unique_ptr<ZNode>(new ZNode(this, mRho)));
		mZStatic[i*2] = mpZNodes[mpZNodes.size()-2].get();
		mZStatic[i*2+1] = mpZNodes[mpZNodes.size()-1].get();
	}
	else // combine the u-nodes and x-nodes
	{
		mZStatic[i*2] = mZRest[i*2];
		mZStatic[i*2+1] = mZRest[i*2+1];
	}

	// setup triangle forces
	for (int i=0; i<(int)mpFEM2D->mTriangles.size(); i++)
	for (int j=0; j<6; j++)
	{
		mZTriangleForce[i*6+j] = new ZNode(this, mRho);
		mpZNodes.push_back(unique_ptr<ZNode>(mZTriangleForce[i*6+j]));
	}

	// setup VM nodes
	for (int i=0; i<(int)mpFEM2D->mTriangles.size(); i++)
	{
		mZVM[i] = new ZNode(this, mRho);
		mpZNodes.push_back(unique_ptr<ZNode>(mZVM[i]));
	}

	// setup x-nodes
	// setup vertices
	std::vector<XNodeFSum*> XFSum(mpFEM2D->mRest.size(), NULL);
	for (int i=0; i<(int)mpFEM2D->mRest.size(); i++)
	//if (mpFEM2D->mFixedVtx[i])
	//{
	//	XNodeFixedX* pxfixed = new XNodeFixedX;
	//	pxfixed->mFixedX = mpFEM2D->mRest[i];
	//	pxfixed->mp_z.resize(2);
	//	pxfixed->mp_z[0] = mZRest[i*2];
	//	pxfixed->mp_z[1] = mZRest[i*2+1];
	//	mpXNodes.push_back(unique_ptr<XNode>(pxfixed));

	//	XNodeFixedX* pxfixed = new XNodeFixedX;
	//	pxfixed->mFixedX = mpFEM2D->mRest[i];
	//	pxfixed->mp_z.resize(2);
	//	pxfixed->mp_z[0] = mZStatic[i*2];
	//	pxfixed->mp_z[1] = mZStatic[i*2+1];
	//	mpXNodes.push_back(unique_ptr<XNode>(pxfixed));
	//}
	//else
	{
		XNodeXDeform* pxdeform = new XNodeXDeform;
		pxdeform->mpADMM = this;
		pxdeform->mX0 = mpFEM2D->mRest[i];
		pxdeform->mW = 0.01;
		pxdeform->mp_z[0] = mZRest[i*2];
		pxdeform->mp_z[1] = mZRest[i*2+1];
		mpXNodes.push_back(unique_ptr<XNode>(pxdeform));

		//XNodeFixedX* pxfixed = new XNodeFixedX;
		//pxfixed->mFixedX = mpFEM2D->mRest[i];
		//pxfixed->mp_z[0] = mZRest[i*2];
		//pxfixed->mp_z[1] = mZRest[i*2+1];
		//mpXNodes.push_back(unique_ptr<XNode>(pxfixed));

		if (!mpFEM2D->mFixedVtx[i])
		{
			XNodeFSum* pfsum = new XNodeFSum;
			pfsum->mpADMM = this;
			pfsum->mp_z.clear();
			pfsum->mExtF = mpFEM2D->mForce[i]*mForceScale;
			XFSum[i] = pfsum;
			mpXNodes.push_back(unique_ptr<XNode>(pfsum));
		}
	}
	// setup f-sum
	for (int i=0; i<(int)mpFEM2D->mTriangles.size(); i++)
	for (int v=0; v<3; v++)
	{
		XNodeFSum* pfsum = XFSum[mpFEM2D->mTriangles[i][v]];
		if (pfsum)
		{
			pfsum->mp_z.push_back(mZTriangleForce[i*6+v*2]);
			pfsum->mp_z.push_back(mZTriangleForce[i*6+v*2+1]);
		}
	}
	// setup vm
	for (int i=0; i<(int)mpFEM2D->mTriangles.size(); i++)
	{
		XNodeVM* pvm = new XNodeVM;
		pvm->mpADMM = this;
		pvm->mpADMMStatic2D = this;
		pvm->mp_z[0] = mZVM[i];
		mpXNodeVM[i] = pvm;
		mpXNodes.push_back(unique_ptr<XNode>(pvm));
	}
	// setup f
	for (int i=0; i<(int)mpFEM2D->mTriangles.size(); i++)
	{
		XNodeF* pf = new XNodeF;
		pf->mpADMM = this;
		pf->mpADMMStatic2D = this;
		pf->mD = mpFEM2D->mD;
		pf->mThick = mpFEM2D->mThick;
		pf->mVMScale = mVMScale;
		pf->mForceScale = mForceScale;
		//pf->InitSearchStep();
		mpXNodes.push_back(unique_ptr<XNode>(pf));
		mpXNodeF.push_back(pf);
		//mXForce.push_back(pf);
		for (int v=0; v<3; v++)
		{
			int vid = mpFEM2D->mTriangles[i][v];
			pf->mp_z[v*2] = mZRest[vid*2];
			pf->mp_z[v*2+1] = mZRest[vid*2+1];

			pf->mp_z[6+v*2] = mZStatic[vid*2];
			pf->mp_z[6+v*2+1] = mZStatic[vid*2+1];

			pf->mp_z[12+v*2] = mZTriangleForce[i*6+v*2];
			pf->mp_z[12+v*2+1] = mZTriangleForce[i*6+v*2+1];
		}
		pf->mp_z[18] = mZVM[i];
	}

	InitXNodes();
}

void Static2DADMM::CopyInitialValue()
{

	for (int i=0; i<(int)mpFEM2D->mStatic.size(); i++)
	{
		mZStatic[i*2]->m_z = mpFEM2D->mStatic[i](0);
		mZStatic[i*2+1]->m_z = mpFEM2D->mStatic[i](1);
	}
	//for (int i=0; i<(int)mpFEM2D->mStatic.size(); i++)
	//{
	//	mZStatic[i*2]->m_z = mpFEM2D->mRest[i](0);
	//	mZStatic[i*2+1]->m_z = mpFEM2D->mRest[i](1);
	//}

	for (int i=0; i<(int)mpFEM2D->mRest.size(); i++)
	{
		mZRest[i*2]->m_z = mpFEM2D->mRest[i](0);
		mZRest[i*2+1]->m_z = mpFEM2D->mRest[i](1);
	}
	
	vector<SVec2Float> forceTest(mpFEM2D->mRest.size(), SVec2Float(0, 0));

	//for (int i=0; i<(int)mpFEM2D->mTriangles.size(); i++)
	//	for (int v=0; v<3; v++)
	//		for (int c=0; c<2; c++)
	//			mZTriangleForce[i*6+v*2+c]->m_z = 0;

	for (int i=0; i<(int)mpFEM2D->mTriangles.size(); i++)
	{
		SVec2Float x[3];
		SVec2Float p[3];
		SVec2Float f[3];
		SSSLib::Vec3i triangle = mpFEM2D->mTriangles[i];
		for (int v=0; v<3; v++)
		{
			x[v] = mpFEM2D->mRest[triangle[v]];
			p[v] = mpFEM2D->mStatic[triangle[v]];
		}
		mpFEM2D->ComputeF(x, p, mpFEM2D->mThick, mpFEM2D->mD, f);
		for (int v=0; v<3; v++)
		{
			for (int c=0; c<2; c++)
				mZTriangleForce[i*6+v*2+c]->m_z = f[v][c]*mForceScale;
			forceTest[mpFEM2D->mTriangles[i][v]] += f[v];
		}
		mZVM[i]->m_z = 0;
		mZVM[i]->m_rhoOut = 0;
	}
	for (int i=0; i<(int)mpFEM2D->mRest.size(); i++)
	if (!mpFEM2D->mFixedVtx[i])
		assert(fabs((forceTest[i]-SVec2Float(mpFEM2D->mForce[i])).Length())<1e-4);

	for (int i=0; i<(int)mpXNodeF.size(); i++)
		mpXNodeF[i]->InitSearchStep();
	InitXNodeVals();
}

void Static2DADMM::UpdateResult(bool ibRest, bool ibStatic)
{
	if (ibRest)
	for (int i=0; i<(int)mpFEM2D->mRest.size(); i++)
	{
		mpFEM2D->mRest[i](0) = mZRest[i*2]->m_z;
		mpFEM2D->mRest[i](1) = mZRest[i*2+1]->m_z;
	}

	if (ibStatic)
	for (int i=0; i<(int)mpFEM2D->mStatic.size(); i++)
	{
		mpFEM2D->mStatic[i](0) = mZStatic[i*2]->m_z;
		mpFEM2D->mStatic[i](1) = mZStatic[i*2+1]->m_z;
	}
}

void Static2DADMM::Restart()
{
	UpdateResult(true, false);
	mpFEM2D->Compute();
	CopyInitialValue();
}

