/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "XNodeF.h"
#include "..\StaticFEM\Static2D.h"
//#include "..\..\SSSLib\SSSFunc.h"
#include "ADMMUtil.h"
#include "..\ADMMLib\ADMM.h"
#include "Static2DADMM.h"

#ifdef _USE_MATLAB
#include "MatlabUtil.h"
using namespace MatlabUtil;
#endif

using namespace std;
using namespace ADMMLib;

XNodeF::XNodeF(void)
	: XNode(19)
	, mMinStep(1e-6)
	//, mMaxVM(0)
{
#ifdef _USE_MATLAB
	InitMatlabDS();
#endif
}


XNodeF::~XNodeF(void)
{
#ifdef _USE_MATLAB
	ReleaseMatlabDS();
#endif
}

void XNodeF::InitSearchStep(_T_FLOAT iVal)
{
	mDefaultStep = iVal;
	for (int i=0; i<12; i++)
	{
		mtStep[i] = iVal;
		mtOpsStep[i] = -iVal;
	}
}

void XNodeF::InitVal()
{
	XNode::InitVal();
	for (int i=0; i<6; i++)
	{
		mtX[i/2][i%2] = mp_z[i]->m_z;
		mtP[i/2][i%2] = mp_z[i+6]->m_z;
		mtF[i/2][i%2] = mp_z[i+12]->m_z;
	}
}

void XNodeF::SolveX()
{
	Decode();
#ifdef _USE_MATLAB
	MatlabOptimize();
#else
	NumericalOptimize(1e-5);
#endif
	Encode();
	for (int i=0; i<19; i++)
		m_rhoOut[i] = mpADMM->mRho;
}

void XNodeF::Encode()
{
	ADMMLib::Encode(mtX, 3, &m_x[0]);
	ADMMLib::Encode(mtP, 3, &m_x[6]);
	ADMMLib::Encode(mtF, 3, &m_x[12]);
	m_x[18] = mtVM;
}

void XNodeF::Decode()
{
	ADMMLib::Decode(&m_n[0], 3, mtX0);
	ADMMLib::Decode(&m_n[6], 3, mtP0);
	ADMMLib::Decode(&m_n[12], 3, mtF0);
	mtVM0 = m_n[18];

	ADMMLib::Decode(&m_rhoIn[0], 3, mtRhoX, mtFixedX);
	ADMMLib::Decode(&m_rhoIn[6], 3, mtRhoP, mtFixedP);
	ADMMLib::Decode(&m_rhoIn[12], 3, mtRhoF, mtFixedF);
	mtRhoVM = m_rhoIn[18];
}

bool XNodeF::ComputeF(SVec2Float* iX, SVec2Float* iP, SVec2Float* oF)
{
	//if (mpADMMStatic2D->mMaxVM>0)
	//{
	//	SVec3Float stress;
	//	Static2D::ComputeStress(iX, iP, mD, stress);
	//	if (Static2D::VonMises(stress)>mpADMMStatic2D->mMaxVM)
	//		return false;
	//}

	Static2D::ComputeF(iX, iP, mThick, mD, oF);
	return true;
}

_T_FLOAT XNodeF::EvaluateTarget(SVec2Float* iX, SVec2Float* iP, SVec2Float* iF, _T_FLOAT iVM)
{
	using namespace SSSLib;
	Vec2Float v;
	_T_FLOAT retVal = 0;
	for (int i=0; i<3; i++)
		retVal += dot<_T_FLOAT>(mtRhoX[i], (iX[i]-mtX0[i]).unaryExpr(SSSLib::sqr<_T_FLOAT>))
					+dot<_T_FLOAT>(mtRhoP[i], (iP[i]-mtP0[i]).unaryExpr(SSSLib::sqr<_T_FLOAT>))
					+dot<_T_FLOAT>(mtRhoF[i], (iF[i]*mForceScale-mtF0[i]).unaryExpr(SSSLib::sqr<_T_FLOAT>));

	retVal += mtRhoVM*sqr(iVM*mVMScale-mtVM0);

	return retVal;
}

void XNodeF::NumericalOptimize(_T_FLOAT iThres)
{
	InitSearchStep(mDefaultStep);

	int evaluateIteNum = 4;
	_T_FLOAT stepScale = 2.0;

	SVec2Float tX[3];
	SVec2Float tP[3];
	SVec2Float tF[3];
	_T_FLOAT tVM;
	SVec3Float stress;

	for (int i=0; i<3; i++)
	{
		tX[i] = mtX[i];
		tP[i] = mtP[i];
		tF[i] = mtF[i];
	}
	bool isvalid = ComputeF(tX, tP, tF);
	while (!isvalid)
	{
		for (int i=0; i<3; i++)
		{
			SVec2Float delta = tP[i]-tX[i];
			if (!mtFixedP[i][0])
				tP[i] = delta*0.9+tX[i];
			if (!mtFixedX[i][0])
				tX[i] = tP[i]-0.9*delta;
		}
		isvalid = ComputeF(tX, tP, tF);
	}
	Static2D::ComputeStress(tX, tP, mD, stress);
	tVM = sqrt(Static2D::VonMisesSqr(stress));

	vector<_T_FLOAT> lastVal(evaluateIteNum, EvaluateTarget(tX, tP, tF, tVM));
	_T_FLOAT newVal = lastVal[0];
	
	int ite = 0;
	bool searchStepContinue = false;
	do {
		searchStepContinue = false;
		for (int i=0; i<12; i++)
		{
			SVec2Float* changeVal;
			SSSLib::Vec2b* checkFix;
			SVec2Float* ref;
			SVec2Float* rho;
			int changeI;
			int changeJ;
			bool isRegularize = true;
			if (i/6==1)
			{
				changeVal = tX;
				checkFix = mtFixedX;
				ref = mtX0;
				rho = mtRhoX;
			}
			else
			{
				changeVal = tP;
				checkFix = mtFixedP;
				ref = mtP0;
				rho = mtRhoP;
			}

			for (int j=0; j<3; j++)
			if (checkFix[j][0] || checkFix[j][1])
				isRegularize = false;

			changeI = (i%6)/2;
			changeJ = (i%6)%2;

			if (checkFix[changeI][changeJ])
				continue;



			//mtStep[i] = mDefaultStep;
			//mtOpsStep[i] = -mDefaultStep;

			//while (fabs(mtStep[i])>mMinStep && fabs(mtOpsStep[i])>mMinStep)
			{
				_T_FLOAT bk = changeVal[changeI][changeJ];
				SVec2Float bkVec[3];

				if (isRegularize)
					memcpy(bkVec, changeVal, sizeof(SVec2Float)*3);

				SVec2Float bkF[3];
				_T_FLOAT bkVM = tVM;
				for (int j=0; j<3; j++) bkF[j] = tF[j];
				changeVal[changeI][changeJ] += mtStep[i];

				if (isRegularize)
					Centerize(ref, rho, changeVal);

				isvalid = ComputeF(tX, tP, tF);
				Static2D::ComputeStress(tX, tP, mD, stress);
				tVM = sqrt(Static2D::VonMisesSqr(stress));

				_T_FLOAT cost;
				if (isvalid)
					cost = EvaluateTarget(tX, tP, tF, tVM);
				if (isvalid && cost<newVal)
				{
					newVal = cost;
					mtOpsStep[i] = -mtStep[i]*0.5;
					mtStep[i] *= stepScale;
				}
				else
				{
					if (isRegularize)
					{
						memcpy(changeVal, bkVec, sizeof(SVec2Float)*3);
						changeVal[changeI][changeJ] += mtOpsStep[i];
						Centerize(ref, rho, changeVal);
					}
					else
						changeVal[changeI][changeJ] += mtOpsStep[i]-mtStep[i];
					
					isvalid = ComputeF(tX, tP, tF);
					Static2D::ComputeStress(tX, tP, mD, stress);
					tVM = sqrt(Static2D::VonMisesSqr(stress));

					if (isvalid)
						cost = EvaluateTarget(tX, tP, tF, tVM);
					if (isvalid && cost<newVal)
					{
						newVal = cost;
						mtStep[i] = mtOpsStep[i]*0.5;
						mtOpsStep[i] = -mtOpsStep[i]*0.5;
					}
					else if (!isvalid || cost>newVal) // rollback
					{
						mtStep[i] *= 0.5;
						mtOpsStep[i] *= 0.5;

						if (isRegularize)
							memcpy(changeVal, bkVec, sizeof(SVec2Float)*3);
						else
							changeVal[changeI][changeJ] = bk;

						for (int j=0; j<3; j++) tF[j] = bkF[j];
						tVM = bkVM;
					}
					else
					{
						mtStep[i] *= stepScale;
						mtOpsStep[i] *= stepScale;
					}
				}
				if (mtStep[i]>mMinStep || mtOpsStep[i]>mMinStep)
					searchStepContinue = true;
			}
		}
		lastVal[ite%evaluateIteNum] = newVal;
		ite++;
	} while (ite<evaluateIteNum || lastVal[ite%evaluateIteNum]-newVal>iThres || searchStepContinue);

	mDispute = newVal;

	for (int i=0; i<3; i++)
	{
		mtX[i] = tX[i];
		mtP[i] = tP[i];
		mtF[i] = tF[i]*mForceScale;
	}
	mtVM = tVM*mVMScale;
}

void XNodeF::Centerize(SVec2Float* iRef, SVec2Float* iRho, SVec2Float* ioChange)
{
	SVec2Float ctrref(0, 0);
	SVec2Float ctrchange(0, 0);
	_T_FLOAT rhoSum = 0;
	for (int i=0; i<3; i++)
	{
		ctrref += iRef[i]*iRho[i][0];
		ctrchange += ioChange[i]*iRho[i][0];
		rhoSum += iRho[i][0];
	}

	if (rhoSum==0)
		return;

	ctrref/=rhoSum;
	ctrchange/=rhoSum;

	for (int i=0; i<3; i++)
		ioChange[i] += ctrref-ctrchange;
}

#ifdef _USE_MATLAB

void XNodeF::InitMatlabDS()
{
	mpMLXin = MatlabUtil::CreateMatrix(3,2);
	mpMLPin = MatlabUtil::CreateMatrix(3,2);
	mpMLThick = MatlabUtil::CreateMatrix(1,1);
	mpMLD = MatlabUtil::CreateMatrix(3,3);
	mpMLX0 = MatlabUtil::CreateMatrix(3,2);
	mpMLP0 = MatlabUtil::CreateMatrix(3,2);
	mpMLF0 = MatlabUtil::CreateMatrix(3,2);
	mpMLXFixed = MatlabUtil::CreateMatrix(3,2);
	mpMLPFixed = MatlabUtil::CreateMatrix(3,2);
	mpMLRhoX = MatlabUtil::CreateMatrix(3,2);
	mpMLRhoP = MatlabUtil::CreateMatrix(3,2);
	mpMLRhoF = MatlabUtil::CreateMatrix(3,2);
	mpStepTol = MatlabUtil::CreateMatrix(1,1);
	mpValTol = MatlabUtil::CreateMatrix(1,1);
	mpMaxVM = MatlabUtil::CreateMatrix(1,1);

	_T_FLOAT stepTol = 1e-8;
	_T_FLOAT valTol = 1e-8;
	SetMatrix(&stepTol, mpStepTol);
	SetMatrix(&valTol, mpValTol);
}

void XNodeF::ReleaseMatlabDS()
{
	mxDestroyArray(mpMLXin);
	mxDestroyArray(mpMLPin);
	mxDestroyArray(mpMLThick);
	mxDestroyArray(mpMLD);
	mxDestroyArray(mpMLX0);
	mxDestroyArray(mpMLP0);
	mxDestroyArray(mpMLF0);
	mxDestroyArray(mpMLRhoX);
	mxDestroyArray(mpMLRhoP);
	mxDestroyArray(mpMLRhoF);
	mxDestroyArray(mpMLXFixed);
	mxDestroyArray(mpMLPFixed);
	mxDestroyArray(mpStepTol);
	mxDestroyArray(mpValTol);
	mxDestroyArray(mpMaxVM);
}

void XNodeF::MatlabOptimize()
{
	cout<<"MatlabOptimize begin\n";
	SetMatrix(mtX[0].val, mpMLXin);
	SetMatrix(mtP[0].val, mpMLPin);
	SetMatrix(&mThick, mpMLThick);
	SetMatrix(mD.data(), mpMLD);
	SetMatrix(mtX0[0].val, mpMLX0);
	SetMatrix(mtP0[0].val, mpMLP0);
	SetMatrix(mtF0[0].val, mpMLF0);
	SetMatrix(mtRhoX[0].val, mpMLRhoX);
	SetMatrix(mtRhoP[0].val, mpMLRhoP);
	SetMatrix(mtRhoF[0].val, mpMLRhoF);
	SetMatrix(&mpADMMStatic2D->mMaxVM, mpMaxVM);
	
	double fixed[6];
	for (int r=0; r<3; r++)
	for (int c=0; c<2; c++)
		fixed[r*2+c] = double(mtFixedX[r][c]);
	SetMatrix(fixed, mpMLXFixed);

	for (int r=0; r<3; r++)
	for (int c=0; c<2; c++)
		fixed[r*2+c] = double(mtFixedP[r][c]);
	SetMatrix(fixed, mpMLPFixed);

	mxArray* ox = CreateMatrix(3, 2);
	mxArray* op = CreateMatrix(3, 2);
	mxArray* of = CreateMatrix(3, 2);
	mxArray* oval;

	bool isSuccess = mlfOptimizeXPF(3, &ox, &op, &of, &oval,
			mpMLXin, mpMLPin,
			mpMLThick, mpMLD,
			mpMLX0, mpMLP0, mpMLF0,
			mpMLXFixed, mpMLPFixed,
			mpMLRhoX, mpMLRhoP, mpMLRhoF,
			mpStepTol, mpValTol,
			mpMaxVM);

	GetMatrix(ox, mtX[0].val);
	GetMatrix(op, mtP[0].val);
	GetMatrix(of, mtF[0].val);

	mxDestroyArray(ox);
	mxDestroyArray(op);
	mxDestroyArray(of);
	//mxDestroyArray(oval);
	
	cout<<"MatlabOptimize end\n";
}

#endif