/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "..\ADMMLib\XNode.h"
#include "..\SSSLib\SSSSmallVec.h"
//#include "..\matlab\FOptLib.h"
#include <memory>

namespace ADMMLib
{

class Static2DADMM;

// calculate the f for three vertice on a triangle
class XNodeF : public XNode
{
public:
	XNodeF(void);
	~XNodeF(void);

	void InitSearchStep(_T_FLOAT iVal = 1e-3);
	virtual void InitVal();
	virtual void SolveX();
	void Encode();
	void Decode();

	MatXFloat mD;
	_T_FLOAT mThick;

	SVec2Float mtX0[3];
	SVec2Float mtP0[3];
	SVec2Float mtF0[3];
	_T_FLOAT mtVM0;

	SVec2Float mtRhoX[3];
	SVec2Float mtRhoP[3];
	SVec2Float mtRhoF[3];
	_T_FLOAT mtRhoVM;

	SSSLib::Vec2b mtFixedX[3];
	SSSLib::Vec2b mtFixedP[3];
	SSSLib::Vec2b mtFixedF[3];

	SVec2Float mtX[3];
	SVec2Float mtP[3];
	SVec2Float mtF[3];
	_T_FLOAT mtVM;
	
	_T_FLOAT mtStep[12];
	_T_FLOAT mtOpsStep[12];
	_T_FLOAT mDefaultStep;
	_T_FLOAT mMinStep;

	//_T_FLOAT mMaxVM;
	Static2DADMM* mpADMMStatic2D;

	bool ComputeF(SVec2Float* iX, SVec2Float* iP, SVec2Float* oF);
	_T_FLOAT EvaluateTarget(SVec2Float* iX, SVec2Float* iP, SVec2Float* iF, _T_FLOAT iVM);
	void NumericalOptimize(_T_FLOAT iThres);
	void Centerize(SVec2Float* iRef, SVec2Float* iRho, SVec2Float* ioChange);

	_T_FLOAT mDispute;
	_T_FLOAT mVMScale;
	_T_FLOAT mForceScale;
#ifdef _USE_MATLAB
	void InitMatlabDS();
	void ReleaseMatlabDS();
	void MatlabOptimize();

	// matlab data structure
	mxArray* mpMLXin;
	mxArray* mpMLPin;
	mxArray* mpMLThick;
	mxArray* mpMLD;
	mxArray* mpMLX0;
	mxArray* mpMLP0;
	mxArray* mpMLF0;
	mxArray* mpMLXFixed;
	mxArray* mpMLPFixed;
	mxArray* mpMLRhoX;
	mxArray* mpMLRhoP;
	mxArray* mpMLRhoF;
	mxArray* mpStepTol;
	mxArray* mpValTol;
	mxArray* mpMaxVM;
	
#endif
};

}