/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include "Config.h"
#include "../SSSLib/SSSSmallVec.h"
#include "../SSSLib/SSSSmallMat.h"
#include "../SSSLib/SSSFunc.h"

class Static2D
{
public:
	Static2D(void);
	~Static2D(void);

	std::vector<SSSLib::Vec3i> mTriangles;
	std::vector<Vec2Float, Eigen::aligned_allocator<Vec2Float> > mRest;
	std::vector<Vec2Float, Eigen::aligned_allocator<Vec2Float> > mStatic;
	std::vector<Vec2Float, Eigen::aligned_allocator<Vec2Float> > mForce;
	std::vector<bool> mFixedVtx;

	std::vector<bool> mBoundaryVtx;
	void ComputeBoundaryVtx();

	void SetGravity(float iDensity);
	void SetGravityWithMassVec(const std::vector<_T_FLOAT>& iMass);

	static MatXFloat GetB0(SVec2Float& ix1, SVec2Float& ix2, SVec2Float& ix3, _T_FLOAT* oDet = NULL)
	{
		_T_FLOAT x1 = ix1[0];
		_T_FLOAT x2 = ix2[0];
		_T_FLOAT x3 = ix3[0];

		_T_FLOAT y1 = ix1[1];
		_T_FLOAT y2 = ix2[1];
		_T_FLOAT y3 = ix3[1];

		if (oDet)
			*oDet = (x1-x3)*(y2-y3)-(y1-y3)*(x2-x3);

		MatXFloat B = MatXFloat::Zero(3, 6);

		B(0, 0) = y2-y3;
		B(0, 2) = y3-y1;
		B(0, 4) = y1-y2;

		B(1, 1) = x3-x2;
		B(1, 3) = x1-x3;
		B(1, 5) = x2-x1;

		B(2, 0) = x3-x2;
		B(2, 1) = y2-y3;
		B(2, 2) = x1-x3;
		B(2, 3) = y3-y1;
		B(2, 4) = x2-x1;
		B(2, 5) = y1-y2;

		return B;
	}

	void Compute();
	void InitRect(int iWidth, int iHeight, _T_FLOAT iScale);
	void InitParam(_T_FLOAT iYM, _T_FLOAT iPR);

	MatXFloat mD;
	_T_FLOAT mThick;

	std::vector<Mat2x2, Eigen::aligned_allocator<Mat2x2> > mTriangleStrain;
	std::vector<Mat2x2, Eigen::aligned_allocator<Mat2x2> > mTriangleStress;
	std::vector<Vec2Float, Eigen::aligned_allocator<Vec2Float> > mTrianglePS;
	std::vector<_T_FLOAT> mTriangleNormalizedPS;

	std::vector<Mat2x2, Eigen::aligned_allocator<Mat2x2> > mVtxStrain;
	std::vector<Mat2x2, Eigen::aligned_allocator<Mat2x2> > mVtxStress;
	std::vector<Vec2Float, Eigen::aligned_allocator<Vec2Float> > mVtxPS;
	std::vector<_T_FLOAT> mVtxNormalizedPS;
	_T_FLOAT mMinPS;
	_T_FLOAT mMaxPS;

	void ComputeStrainStress();
	_T_FLOAT ComputeVM(int iTri)
	{
		SVec2Float x[3];
		SVec2Float p[3];
		SVec3Float stress;
		SSSLib::Vec3i tri = mTriangles[iTri];
		for (int i=0; i<3; i++)
		{
			x[i] = mRest[tri[i]];
			p[i] = mStatic[tri[i]];
		}

		ComputeStress(x, p, mD, stress);
		return VonMisesSqr(stress);
	}
	static void ComputeF(SVec2Float iX[3], SVec2Float iP[3], _T_FLOAT iScale, MatXFloat iD, SVec2Float* oF);
	static void ComputeStress(SVec2Float iX[3], SVec2Float iP[3], MatXFloat iD, SVec3Float& oStress);
	static _T_FLOAT VonMisesSqr(SVec3Float iStress)
	{
		using namespace SSSLib;
		return sqr(iStress[0])-iStress[0]*iStress[1]+sqr(iStress[1])+3*sqr(iStress[2]);
	}
	Vec2Float SamplePS(const Vec2Float& iPst, int iTriID);
};
