/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "StaticTable.h"
#include <opencv2\opencv.hpp>
#include "..\SSSLib\SSSFunc.h"
#include "..\SSSLib\SSSColor.h"

using namespace std;
using namespace SSSLib;

StaticTable::StaticTable(void)
{
}

StaticTable::~StaticTable(void)
{
}

void StaticTable::BuildRect(int iTop, int iLeft, int iBottom, int iRight)
{
	for (int x=iLeft; x<iRight; x++)
	for (int y=iBottom; y<iTop; y++)
	{
		Vec2i pix[4];
		pix[0] = Vec2i(x, y);
		pix[1] = Vec2i(x+1, y);
		pix[2] = Vec2i(x, y+1);
		pix[3] = Vec2i(x+1, y+1);
		int vid[4];
		for (int i=0; i<4; i++)
		{
			if (mPt2Idx.count(pix[i])==0)
				mPt2Idx[pix[i]] = (int)mPt2Idx.size();
			vid[i] = mPt2Idx[pix[i]];
		}
		mTriangles.push_back(Vec3i(vid[0], vid[1], vid[2]));
		mTriangles.push_back(Vec3i(vid[1], vid[3], vid[2]));
	}
}

bool StaticTable::BuildTable()
{
	
	mPt2Idx.clear();
	mTriangles.clear();
	// left leg
	assert((mSW-mLD)%2==0);
	assert(mBT==0 || mBH+mBT<mLL);
	int left = (mSW-mLD)/2-mLW;
	int right = left+mLW;
	int bottom = 0;
	int top = mLL;
	BuildRect(top, left, bottom, right);
	// right leg
	left = right+mLD;
	right = left+mLW;
	BuildRect(top, left, bottom, right);
	// top
	left = 0;
	right = mSW;
	bottom = mLL;
	top = bottom+mST;
	BuildRect(top, left, bottom, right);
	// the bar
	if (mBT!=0)
	{
		left = (mSW-mLD)/2;
		right = left+mLD;
		bottom = mBH;
		top = bottom+mBT;
		BuildRect(top, left, bottom, right);
	}

	mOrigin[0] = mSW*0.5f;
	mOrigin[1] = (mLL+mST)*0.5f;

	BuildFloatVtx();
	// fix the legs
	//for (map<Vec2i, int>::iterator ite = mPt2Idx.begin();
	//	ite != mPt2Idx.end();
	//	ite++)
	//if (ite->first[1]==0)
	//{
	//	mFixedVtx[ite->second] = 1;
	//}

	//SetGravity(100.0f);
	return true;
}

void StaticTable::BuildFloatVtx()
{
	mRest.clear();
	mRest.resize(mPt2Idx.size());
	for (map<Vec2i, int>::iterator ite = mPt2Idx.begin();
		ite != mPt2Idx.end();
		ite++)
	{
		/*mRest[ite->second] = Vec2Float((ite->first[0]-mOrigin[0])*mGridScale,
									(ite->first[1]-mOrigin[1])*mGridScale);*/
		mRest[ite->second] = Grid2TrianCoord(Vec2Float(ite->first[0], ite->first[1]));
	}

	mForce.resize(mRest.size());
	mForce.assign(mRest.size(), Vec2Float(0, 0));
	mFixedVtx.resize(mRest.size());
	mFixedVtx.assign(mFixedVtx.size(), 0);
}

Vec2Float StaticTable::SamplePS(const Vec2Float& iWorldPst, bool* obOnMesh)
{
	Vec2Float gridPst = iWorldPst/mGridScale+mOrigin;
	Vec2i gridPt0(int(floor(gridPst[0])), int(floor(gridPst[1])));
	Vec2i gridPt1 = gridPt0+1;
	int vid[4];
	for (int x=0; x<2; x++)
	for (int y=0; y<2; y++)
	if (mPt2Idx.count(gridPt0+Vec2i(x, y))==0)
	{
		if (obOnMesh)
			*obOnMesh = false;
		return Vec2Float(0, 0);
	}
	else
		vid[x+y*2] = mPt2Idx[gridPt0+Vec2i(x, y)];
	Vec3i tri;
	if (gridPst[0]-gridPt0[0]+gridPst[1]-gridPt0[1]>1.0)
		tri = Vec3i(vid[0], vid[1], vid[2]);
	else
		tri = Vec3i(vid[1], vid[3], vid[2]);
	int tid = 0;
	for (tid=0; tid<(int)mTriangles.size() && mTriangles[tid]!=tri; tid++);
	assert(tid<(int)mTriangles.size());
	if (obOnMesh)
		*obOnMesh = true;
	return Static2D::SamplePS(iWorldPst, tid);
}

void StaticTable::SampleTable(float iLeft,
										float iBottom,
										float iRight,
										float iTop,
										int iSampleReso,
										int iDownsampleReso,
										std::vector<SSSLib::Vec2f>& oPS,
										std::vector<float>& oDensity)
{
	vector<Vec2f>& retVec = oPS;
	retVec.clear();
	retVec.resize(iDownsampleReso*iDownsampleReso, Vec2f(0, 0));

	oDensity.clear();
	oDensity.resize(retVec.size(), 0);

	int sampleX = iSampleReso/iDownsampleReso;
	bool bOnMesh;
	for (int r=0; r<iSampleReso; r++)
	for (int c=0; c<iSampleReso; c++)
	{
		int rr = r/sampleX;
		int cc = c/sampleX;
		Vec2Float pst((c+0.5)/iSampleReso*(iRight-iLeft)+iLeft,
						(r+0.5f)/iSampleReso*(iTop-iBottom)+iBottom);
		int idx = rr*iDownsampleReso+cc;
		Vec2Float val = SamplePS(pst, &bOnMesh);
		Vec2f sval((float)val[0], (float)val[1]);
		retVec[idx] = retVec[idx]+sval/float(sampleX*sampleX);
		if (bOnMesh)
			oDensity[idx] += 1.0f/float(sampleX*sampleX);
	}

	cv::Mat testMat(iDownsampleReso, iDownsampleReso, CV_8UC3);
	cv::Mat densityMat(iDownsampleReso, iDownsampleReso, CV_8U);

	//_T_FLOAT mmax = retVec[0][0];
	//_T_FLOAT mmin = retVec[0][0];
	//for (int i=0; i<(int)retVec.size(); i++)
	//{
	//	if (mmax<retVec[i][0])
	//		mmax = retVec[i][0];
	//	if (mmin>retVec[i][0])
	//		mmin = retVec[i][0];
	//}

	for (int rr=0; rr<iDownsampleReso; rr++)
	for (int cc=0; cc<iDownsampleReso; cc++)
	{
		_T_FLOAT val = retVec[rr*iDownsampleReso+cc][0];
		Vec3f clr = Value2HSVf(float((val-mMinPS)/(mMaxPS-mMinPS)*0.5));
		if (val==0)
			clr = Vec3f(0.0f, 0.0f, 0.0f);
		testMat.at<cv::Vec3b>(iDownsampleReso-1-rr, cc) = cv::Vec3b(int(clr[2]*255),
																	int(clr[1]*255),
																	int(clr[0]*255));

		_T_FLOAT density = oDensity[rr*iDownsampleReso+cc];

		densityMat.at<unsigned char>(iDownsampleReso-1-rr, cc) = int(density*255);
	}
	cv::imwrite("test.png", testMat);
	cv::imwrite("testdensity.png", densityMat);
}
