/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once
#include "static2d.h"
#include <map>
#include <vector>
#include <fstream>

class StaticTable :
	public Static2D
{
public:
	StaticTable(void);
	~StaticTable(void);

	void ParseTable();

	int mSW;
	int mST;
	int mLL;
	int mLD;
	int mLW;
	int mBH;
	int mBT;

	void SetParameters(int iSW, int iST, int iLL, int iLD, int iLW, int iBH, int iBT, float iGridScale)
	{
		mSW = iSW;
		mST = iST;
		mLL = iLL;
		mLD = iLD;
		mLW = iLW;
		mBH = iBH;
		mBT = iBT;
		mGridScale = iGridScale;
	}
	bool BuildTable();
	void BuildRect(int iTop, int iLeft, int iBottom, int iRight);

	float mGridScale;
	Vec2Float mOrigin;
	std::map<SSSLib::Vec2i, int> mPt2Idx;
	Vec2Float Grid2TrianCoord(const Vec2Float& iGridPst)
	{return (iGridPst-mOrigin)*mGridScale;}
	void BuildFloatVtx();

	Vec2Float SamplePS(const Vec2Float& iWorldPst, bool* obOnMesh = NULL);

	 void SampleTable(float iLeft,
										float iBottom,
										float iRight,
										float iTop,
										int iSampleReso,
										int iDownsampleReso,
										std::vector<SSSLib::Vec2f>& oPS,
										std::vector<float>& oDensity);

	template<class _TDATA>
	static void SaveVec(const char* iFileName, const std::vector<_TDATA>& iVec)
	{
		ofstream inFile(iFileName, ios_base::binary);
		int num = iVec.size();
		inFile.write((char*)&num, 4);
		inFile.write((char*)&iVec[0], sizeof(_TDATA)*num);
	}
};

