/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "..\SSSLib\SSSLibOGL\SSSVolMesh.h"
#include <string>
#include <Windows.h>
#include <time.h>

using namespace std;
using namespace SSSLib;

int main(int argc, char* argv[])
{
	srand(1);

	if (argc<=3)
	{
		cout<<"Usage: [input tet mesh] [input surface sample] [poisson disk ratio] [output mesh]\n";
		return 1;
	}

	string inFileName(argv[1]);

	SSSVolMesh volMesh;
	if (inFileName.substr(inFileName.length()-3, 3)=="inp")
		volMesh.LoadInp(argv[1]);
	else if (inFileName.substr(inFileName.length()-3, 3)=="vtk")
		volMesh.LoadVtk(argv[1]);
	else
		volMesh.LoadVolMesh(argv[1]);

	SSSMesh samples;
	samples.LoadObj(argv[2]);

	double pr;
	sscanf_s(argv[3], "%lf", &pr);

	vector<double> sizingFunc;
	if (argc>5)
	{
		int vecSize;
		ifstream sizingFile(argv[5], ios_base::binary);
		sizingFile.read((char*)&vecSize, sizeof(int));
		assert(vecSize==samples.mVtx.size());
		sizingFunc.resize(vecSize);
		sizingFile.read((char*)&sizingFunc[0], sizeof(double)*vecSize);
		for (int i=0; i<(int)sizingFunc.size(); i++)
			sizingFunc[i] /= 2.5;
			//sizingFunc[i] = pr;
	}

	vector<Vec3d> nodes = samples.mVtx;

	volMesh.AddPoissonSamples(pr, nodes, &sizingFunc);

	SSSVolMesh::WriteNodes("tmp.node", nodes);

	SSSVolMesh resampleMesh;

	cout<<"Num. Surface Nodes "<<samples.mVtx.size()<<endl
		<<"Num. Total Nodes "<<nodes.size()<<endl;

	bool bContinue = true;

	{
		SSSMesh sampleMesh;
		sampleMesh.mVtx = nodes;
		sampleMesh.WriteObj("InitSampleMesh.obj");
	}

	// check external points
	int iteNum = 0;
	//while (bContinue || iteNum<3)
	while (bContinue)
	//for (int iteNum=0; iteNum<3; iteNum++)
	{
		cout<<"Ite: "<<iteNum<<endl;
		iteNum++;
		system("tetgen.exe -k tmp.node");

		resampleMesh.LoadVtk("tmp.1.vtk");
		// for debug
		//{
		//	for (int i=0; i<(int)nodes.size(); i++)
		//	{
		//		if ((nodes[i]-resampleMesh.mVtx[i]).Length()>1e-3)
		//		{
		//			printf("Error node %d: (%f, %f, %f) <-> (%f, %f, %f)\n", 
		//				i, nodes[i][0], nodes[i][1], nodes[i][2],
		//				resampleMesh.mVtx[i][0], resampleMesh.mVtx[i][1], resampleMesh.mVtx[i][2]);
		//		}
		//	}
		//}


		resampleMesh.BuildSurfaceMesh();
		resampleMesh.mSurfaceMesh.WriteObj("tmp_surface.obj");

		vector<int> eraseList;

		eraseList.clear();

		#pragma omp parallel for
		for (int t=(int)resampleMesh.mTets.size()-1;
			t>=0;
			t--)
		{
			Vec4i tet = resampleMesh.mTets[t];
			Vec3d ctrPt(0, 0, 0);
			Vec3d pts[4];
			bool onSurface = true;
			for (int i=0; i<4; i++)
			{
				pts[i] = resampleMesh.mVtx[tet[i]];
				if (tet[i]>=samples.mVtx.size())
					onSurface = false;
				ctrPt += pts[i];
			}
			if (!onSurface)
				continue;
			else
			{
				ctrPt/=4;
				bool inout = volMesh.CheckInside(ctrPt);
				int inoutVote = 0;
				if (!inout)
					inoutVote++;
				//if (0)
				for (int i=0; i<4; i++)
				{
					Vec3d newPt = ctrPt+(pts[i]-ctrPt)/4;
					bool newInout = volMesh.CheckInside(newPt);
					if (!newInout)
						inoutVote++;
				}

				if (inoutVote>=1)
				{
					#pragma omp critical
					{
						eraseList.push_back(t);
					}
					//resampleMesh.mTets.erase(resampleMesh.mTets.begin()+t);
					//cout<<"Erase "<<t<<endl;
				}
			}
			//if (t%100==0)
			//	cout<<"Num "<<t<<endl;
		}

		cout<<"Erase ext tet:"<<eraseList.size()<<endl;

		sort(eraseList.begin(), eraseList.end());

		for (int i=(int)eraseList.size()-1;
			i>=0;
			i--)
			resampleMesh.mTets.erase(resampleMesh.mTets.begin()+eraseList[i]);

		eraseList.clear();

		bool bUpdate = true;

		resampleMesh.mbIntVtx.resize(resampleMesh.mVtx.size());
		for (int i=0; i<(int)resampleMesh.mVtx.size(); i++)
		{
			if (i<nodes.size())
				resampleMesh.mbIntVtx[i] = false;
			else
				resampleMesh.mbIntVtx[i] = true;
		}

		while (bUpdate)
		{

			//resampleMesh.VtxRelax();
			resampleMesh.BuildVtxConnect();
			assert(resampleMesh.mVtxConnect.size()>0);
			//SSSLib::ZeroLapRelax(resampleMesh.mVtxConnect, resampleMesh.mbIntVtx, resampleMesh.mVtx);

			resampleMesh.BuildSurfaceMesh();
			// remove low dihedral angles
			#pragma omp parallel for
			for (int t=0; t<(int)resampleMesh.mTets.size(); t++)
			if (resampleMesh.mbSurfaceTet[t])
			{
				Vec4i tet = resampleMesh.mTets[t];
				Vec3d vtx[4];
				bool onSurface = true;
				for (int i=0; i<4; i++)
				{
					vtx[i] = resampleMesh.mVtx[tet[i]];
					if (tet[i]>=samples.mVtx.size())
						onSurface = false;
				}
				if (!onSurface)
					continue;

				bool eraseFlag = false;

				for (int i=0; i<4; i++)
				{
					Vec3d normal[2];
					for (int j=0; j<2; j++)
						normal[j] = cross3(vtx[i]-vtx[(i+2+j)%4], vtx[(i+1)%4]-vtx[(i+2+j)%4]).Norm();

					double val = dot<double>(normal[0], normal[1]);

					if (fabs(val)>cos(Pi/180*10))
					{
						eraseFlag = true;
						break;
					}
				}

				Vec3d ctr = (vtx[0]+vtx[1]+vtx[2]+vtx[3])/4;

				if (eraseFlag)
				{
					eraseFlag = false;
					for (int f=0; f<resampleMesh.mSurfaceMesh.mFace.size(); f++)
					{
						double dis = resampleMesh.mSurfaceMesh.DisPt2Triangle(ctr, f);
						if (dis<pr*0.01)
						{
							eraseFlag = true;
							break;
						}
					}
					if (eraseFlag)
					{
						#pragma omp critical
						eraseList.push_back(t);
					}
				}
			}
			cout<<"Erase small dihedral angle:"<<eraseList.size()<<endl;

			if (eraseList.size()!=0)
				bUpdate = true;
			else
				bUpdate = false;

			sort(eraseList.begin(), eraseList.end());

			for (int i=(int)eraseList.size()-1;
				i>=0;
				i--)
				resampleMesh.mTets.erase(resampleMesh.mTets.begin()+eraseList[i]);

			eraseList.clear();
		}		

		resampleMesh.FindIntVtx();
		for (int i=0; i<(int)resampleMesh.mVtx.size(); i++)
		{
			bool bIn;
			if (i<samples.mVtx.size())
				bIn = false;
			else
				bIn = true;
			if (bIn != resampleMesh.mbIntVtx[i])
			{
				cout<<"Int Vtx Calculation Error!"<<endl;
			}
		}

		bContinue = false;

		vector<Vec3d> tmpVtx(resampleMesh.mVtx);
		//SSSLib::ZeroLapRelax(resampleMesh.mVtxConnect, resampleMesh.mbIntVtx, resampleMesh.mVtx);

		int negTet = CountTet(resampleMesh, NULL, 0);
		if (negTet>0)
		{
			bContinue = true;
			for (int i=0; i<(int)resampleMesh.mVtx.size(); i++)
			{
				resampleMesh.mVtx[i] += (tmpVtx[i]-resampleMesh.mVtx[i])*randf<double>();
			}
		}

		vector<Vec3d> vtx = resampleMesh.mVtx;
		bool bEraseNode = false;
		int eraseNodes = 0;
		for (int i=(int)resampleMesh.mVtx.size()-1;
			i>=samples.mVtx.size();
			i--)
		if (!volMesh.CheckInside(vtx[i]))
		{
			vtx.erase(vtx.begin()+i);
			bEraseNode = true;
			bContinue = true;
			eraseNodes++;
		}

		if (bEraseNode)
			cout<<"Erased node:"<<eraseNodes<<endl;
		else
			cout<<"Reserve all nodes!\n";

		{
			int newPt = 0;
			#pragma omp parallel for
			for (int t=0; t<(int)resampleMesh.mTets.size(); t++)
			{
				Vec4i tet = resampleMesh.mTets[t];
				Vec3d vtx_tmp[4];
				bool onSurface = true;
				bool eraseFlag = false;

				for (int i=0; i<4; i++)
				{
					Vec3d normal[2];
					for (int j=0; j<2; j++)
						normal[j] = cross3(vtx_tmp[i]-vtx_tmp[(i+2+j)%4], vtx_tmp[(i+1)%4]-vtx_tmp[(i+2+j)%4]).Norm();

					double val = dot<double>(normal[0], normal[1]);

					if (fabs(val)>cos(Pi/180*10))
					{
						eraseFlag = true;
						break;
					}
				}

				Vec3d ctr = (vtx_tmp[0]+vtx_tmp[1]+vtx_tmp[2]+vtx_tmp[3])/4;

				if (eraseFlag)
				{
					#pragma omp critical
					{
						vtx.push_back(ctr);
						newPt++;
					}
					bContinue = true;
				}
			}
			if (newPt>0)
				cout<<"Insert points "<<newPt<<endl;
		}

		SSSVolMesh::WriteNodes("tmp.node", vtx);
	}

	resampleMesh.RemoveIsolatedVtx();
	resampleMesh.WriteVtk(argv[4]);

	resampleMesh.BuildSurfaceMesh();
	resampleMesh.mSurfaceMesh.WriteObj("finished_surface.obj");

	cout<<"Press any key to continue\n";
	char c;
	cin>>c;

	return 0;
}
