/*Copyright (C) 2016 Yahan Zhou

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "../SSSLib/SSSLibOGL/SSSVolMesh.h"
#include "../SSSLib/SSSLibOGL/SSSMeshUtil.h"
#include <iostream>
#include <fstream>
#include <string.h>

using namespace SSSLib;
using namespace std;

vector<bool> gWriteMask;

Vec3d gModelCenter(0, 0, 0);
double gModelScale = 1.0f;

//#define _PREVIEW

//int CountTet(SSSVolMesh& imesh, set<int>* oNegVtx, double iVolBound = 0.0f)
//{
//	int pos = 0;
//	int neg = 0;
//	int zero = 0;
//	for (int i=0; i<(int)imesh.mTets.size(); i++)
//	{
//		double vol = imesh.TetSignedVol(i);
//		if (vol<iVolBound)
//		{
//			cout<<i<<"\t"<<vol<<endl;
//			neg++;
//			if (gWriteMask.size()>0)
//			{
//				for (int j=0; j<4; j++)
//					gWriteMask[imesh.mTets[i][j]] = false;
//			}
//			if (oNegVtx)
//				for (int j=0; j<4; j++)
//					oNegVtx->insert(imesh.mTets[i][j]);
//		}
//		else
//			pos++;
//	}
//
//	cout<<"Positive Tet: "<<pos<<endl
//		<<"Negative Tet: "<<neg<<endl
//		<<"Zero Tet: "<<zero<<endl;
//	return neg;
//}

void FilterOneRing(double iDis, SSSVolMesh& iMesh)
{
	for (int i=0; i<(int)iMesh.mVtxConnect.size(); i++)
	{
		double dis = 0;
		for (set<int>::iterator ite = iMesh.mVtxConnect[i].begin();
			ite != iMesh.mVtxConnect[i].end();
			ite++)
			dis += (iMesh.mVtx[i]-iMesh.mVtx[*ite]).Length();
		dis /= iMesh.mVtxConnect[i].size();

		if (dis<iDis)
			gWriteMask[i] = false;
	}
}

int main(int argc, char* argv[])
{
	double ringDis = 0.3;
	double volBound = 0;
	//if (argc>2)
	//	ringDis = atof(argv[2]);
	//if (argc>3)
	//	volBound = atof(argv[3]);

	char buffer[10000];

	SSSVolMesh mesh;
	//ofstream logFile("log.txt");
	//cout.rdbuf(logFile.rdbuf());
	//sprintf(buffer, "%s.node", argv[1]);
	//mesh.LoadVolMesh(argv[1]);

	if (strcmp(argv[1]+strlen(argv[1])-3, "vtk")==0)
		mesh.LoadVtk(argv[1]);
	else if (strcmp(argv[1]+strlen(argv[1])-3, "inp")==0)
		mesh.LoadInp(argv[1]);
	else if (strcmp(argv[1]+strlen(argv[1])-4, "mesh")==0)
		mesh.LoadDisteneMesh(argv[1]);
	else
	{
		cout<<"Unrecognized format: "<<argv[1]<<endl;
		exit(1);
	}
	
	mesh.BuildSurfaceMesh();

	Vec3d low, high;
	mesh.mSurfaceMesh.ComputeBB(low, high);
	gModelCenter = (low+high)*0.5f;
	Vec3d delta = high-low;
	gModelScale = 1.0f/Max(Max(delta[0], delta[1]), delta[2]);

	mesh.Translate(-gModelCenter);
	//mesh.Scale(25.4); // inch to mm
	//mesh.Scale(gModelScale);
	//mesh.Rotate(Vec3d(1.0, 0, 0), -Pi/2);
	//mesh.Rotate(Vec3d(0, 1.0, 0), Pi);
	//mesh.Scale(Vec3d(0.1, 0.1, 0.1));

	mesh.mSurfaceMesh.mVtx = mesh.mVtx;

#ifndef _PREVIEW
	cout<<"Before\n";
	CountTet(mesh, NULL, volBound);
	//int count = mesh.CheckCrossTet();
	//cout<<"Crossed Tet: "<<count<<endl;

	mesh.FindIntVtx();
	//mesh.VtxRelax();


	gWriteMask.assign(mesh.mVtx.size(), true);

	cout<<"After\n";
	
	//FilterOneRing(ringDis, mesh);

	set<int> remVtx;
	int negNum = CountTet(mesh, &remVtx, volBound);

	vector<Vec3d> nodes;
	for (int i=0; i<(int)gWriteMask.size(); i++)
	if (gWriteMask[i])
		nodes.push_back(mesh.mVtx[i]);
	//count = mesh.CheckCrossTet();
	//cout<<"Crossed Tet: "<<count<<endl;

	/*sprintf(buffer, "%s.a.node", argv[1]);
	mesh.WriteNodes(buffer);

	//sprintf(buffer, "%s.vtk", argv[1]);
	//mesh.WriteVtk(buffer);

	sprintf(buffer, "%s_d.a.node", argv[1]);
	SSSVolMesh::WriteNodes(buffer, nodes);*/

#endif

	if (argc>2)
		mesh.WriteMesh(argv[2]);

	if (argc>3)
	{
		//mesh.mSurfaceMesh.WriteStl(argv[3]);
		mesh.mSurfaceMesh.WriteObj(argv[3]);
	}
	//	sprintf(buffer, "%s_surface.stl", argv[1]);
	//	mesh.mSurfaceMesh.WriteStl(buffer);

	return 0;
}